// Copyright 2019 Canonical Ltd.

package main

import (
	"context"
	"io/ioutil"
	"net/http/httptest"
	"os"
	"testing"

	"github.com/rogpeppe/go-internal/testscript"
	errgo "gopkg.in/errgo.v1"

	contracts "github.com/CanonicalLtd/ua-contracts"
	"github.com/CanonicalLtd/ua-contracts/internal/memstore"
	"github.com/CanonicalLtd/ua-contracts/internal/saauth"
	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
)

func TestMain(m *testing.M) {
	os.Exit(testscript.RunMain(m, map[string]func() int{
		"contract": runMain,
	}))
}

func Test(t *testing.T) {
	testscript.Run(t, testscript.Params{
		Dir:   "testdata",
		Setup: setup,
		Cmds: map[string]func(ts *testscript.TestScript, neg bool, args []string){
			"check-password":  checkPassword,
			"cpenv":           cpEnv,
			"create-contract": createContract,
		},
	})
}

type storeKey struct{}

// setup initialises the environment for running contracts-admin tests.
func setup(env *testscript.Env) error {
	ctx := context.Background()
	st := memstore.New()
	env.Values[storeKey{}] = st

	err := saauth.SetPassword(ctx, st, "admin", "admin-password")
	if err != nil {
		return errgo.Mask(err)
	}
	env.Vars = append(env.Vars, "CONTRACTS_ADMIN_CREDS=admin:admin-password")
	hnd, err := contracts.NewServer(ctx, contracts.Params{
		Store: st,
	})
	if err != nil {
		return errgo.Mask(err)
	}
	srv := httptest.NewServer(hnd)
	env.Defer(srv.Close)
	env.Vars = append(env.Vars, "CONTRACTS_URL="+srv.URL)
	return nil
}

// checkPassword is a command that checks that the given service-account
// has the given password.
func checkPassword(ts *testscript.TestScript, neg bool, args []string) {
	ctx := context.Background()

	if len(args) != 2 {
		ts.Fatalf("incorrect arguments for check-password, expected <username> <password>")
	}
	id, err := saauth.Validate(ctx, getStore(ts), args[0], args[1])
	ts.Check(err)
	ok, err := id.Allow(ctx, args[:1])
	ts.Check(err)
	if neg == ok {
		ts.Fatalf("password does not have expected value")
	}
}

// createContract creates a contract in the store, associated with
// an arbitrary account. It sets the following environment variables:
//   CONTRACTID - the id of the contract.
//   ACCOUNTID - the id of the account that the contract is associated with.
func createContract(ts *testscript.TestScript, neg bool, args []string) {
	if neg {
		ts.Fatalf("neg not supported")
	}
	st := getStore(ts)
	ctx := context.Background()
	accID := token.MustNewAccountID()
	err := st.AddAccount(ctx, store.Account{
		ID:   accID,
		Name: "test account",
	}, "userid")
	ts.Check(err)
	err = st.AddProduct(ctx, store.Product{
		ID:   "productID",
		Name: "some product name",
		Resources: []store.ProductResource{{
			CanOverride:  true,
			ResourceType: "fips",
			Data: `{
			"architectures": ["amd64"],
			"series": ["xenial"]
			}`,
		}},
	})
	ts.Check(err)
	contractID, err := st.AddContract(ctx, store.Contract{
		Name:       "contract-1",
		AccountID:  accID,
		ProductIDs: []string{"productID"},
	})
	ts.Check(err)
	ts.Setenv("CONTRACTID", contractID.String())
	ts.Setenv("ACCOUNTID", accID.String())
}

// cpEnv copies one file to another, expanding environment variables
// as it does so.
func cpEnv(ts *testscript.TestScript, neg bool, args []string) {
	if neg {
		ts.Fatalf("neg not supported")
	}
	if len(args) != 2 {
		ts.Fatalf("usage: cpenv src dst")
	}
	src, dst := args[0], args[1]
	data, err := ioutil.ReadFile(ts.MkAbs(src))
	ts.Check(err)
	err = ioutil.WriteFile(ts.MkAbs(dst), []byte(os.Expand(string(data), ts.Getenv)), 0666)
	ts.Check(err)
}

// getStore returns the store added to env.Values by setup.
func getStore(ts *testscript.TestScript) store.Store {
	return ts.Value(storeKey{}).(store.Store)
}
