// Copyright 2019 Canonical Ltd.

package main

import (
	"context"

	"github.com/juju/cmd"
	"github.com/juju/gnuflag"
	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/params"
)

type showContractCommand struct {
	cmd.CommandBase

	contractID string
	output     cmd.Output
}

const showContractDoc = `
This command shows all the entitlements for a particular
contract, also taking into account contract-specific configuration.

	contract show-contract <contract-id>
`

// Info implements cmd.Command.Info.
func (c *showContractCommand) Info() *cmd.Info {
	return &cmd.Info{
		Name:           "show-contract",
		Purpose:        "Show details of contract",
		Args:           "<contract-id>",
		Doc:            showContractDoc[1:],
		ShowSuperFlags: commonFlags,
	}
}

// SetFlags implements cmd.Command.SetFlags.
func (c *showContractCommand) SetFlags(f *gnuflag.FlagSet) {
	formatters := map[string]cmd.Formatter{
		"json": cmd.FormatJson,
		"yaml": formatYAML,
	}
	c.output.AddFlags(f, "yaml", formatters)
}

// Run implements cmd.Command.Run.
func (c *showContractCommand) Init(args []string) error {
	if len(args) < 1 {
		return errgo.New("contract ID not specified")
	}
	c.contractID = args[0]
	if len(args) > 1 {
		return errgo.Newf("too many args")
	}
	return nil
}

// Run implements cmd.Command.Run.
func (c *showContractCommand) Run(ctxt *cmd.Context) error {
	ctx := context.Background()

	info, err := client().GetContract(ctx, &params.GetContractRequest{
		Contract: c.contractID,
	})
	if err != nil {
		switch errorCode(err) {
		case params.ErrContractNotFound:
			return errgo.Newf("contract %q not found", c.contractID)
		case params.ErrForbidden:
			return errgo.New("permission denied")
		}
		return errgo.Mask(err)
	}
	return errgo.Mask(c.output.Write(ctxt, info))
}
