// Copyright 2019 Canonical Ltd.

package main

import (
	"context"

	"github.com/juju/cmd"
	"github.com/juju/gnuflag"
	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/params"
)

// showProductCommand implements the show-product command.
type showProductCommand struct {
	cmd.CommandBase

	product string
	output  cmd.Output
}

const showProductDoc = `
Show the specified product.

    contract show-product free
`

// Info implements cmd.Command.Info.
func (c *showProductCommand) Info() *cmd.Info {
	return &cmd.Info{
		Name:           "show-product",
		Purpose:        "Show a product.",
		Args:           "<product id>",
		Doc:            showProductDoc[1:],
		ShowSuperFlags: commonFlags,
	}
}

// SetFlags implements cmd.Command.SetFlags.
func (c *showProductCommand) SetFlags(f *gnuflag.FlagSet) {
	formatters := map[string]cmd.Formatter{
		"json": cmd.FormatJson,
		"yaml": formatYAML,
	}
	c.output.AddFlags(f, "yaml", formatters)
}

// Init implements cmd.Command.Init.
func (c *showProductCommand) Init(args []string) error {
	if len(args) < 1 {
		return errgo.New("product not specified")
	}
	c.product = args[0]
	if len(args) > 1 {
		return errgo.Newf("too many args")
	}
	return nil
}

// Run implements cmd.Command.Run.
func (c *showProductCommand) Run(ctxt *cmd.Context) error {
	ctx := context.Background()

	resp, err := client().GetProduct(ctx, &params.GetProductRequest{
		ProductID: c.product,
	})
	if err != nil {
		switch errorCode(err) {
		case params.ErrNotFound:
			return errgo.Newf("product %q not found", c.product)
		case params.ErrForbidden:
			return errgo.New("permission denied")
		default:
			return errgo.Mask(err)
		}
	}
	return errgo.Mask(c.output.Write(ctxt, resp))
}
