// Copyright 2018 Canonical Ltd.

package main

import (
	"context"
	"database/sql"
	"net/http"
	"time"

	"github.com/CanonicalLtd/salesforce"
	aclstore "github.com/juju/aclstore/v2"
	"github.com/juju/simplekv/memsimplekv"
	"github.com/juju/simplekv/sqlsimplekv"
	"github.com/juju/zaputil"
	"github.com/juju/zaputil/zapctx"
	_ "github.com/lib/pq"
	"go.uber.org/zap"
	errgo "gopkg.in/errgo.v1"
	"gopkg.in/macaroon-bakery.v2/bakery"
	"gopkg.in/macaroon-bakery.v2/bakery/postgresrootkeystore"

	contracts "github.com/CanonicalLtd/ua-contracts"
	"github.com/CanonicalLtd/ua-contracts/config"
	"github.com/CanonicalLtd/ua-contracts/internal/kafkaauditor"
	"github.com/CanonicalLtd/ua-contracts/internal/memstore"
	"github.com/CanonicalLtd/ua-contracts/internal/postgresstore"
	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/storeutil"
)

func main() {
	zapctx.LogLevel.SetLevel(config.LogLevel())

	ctx := context.Background()

	hnd, err := newHandler(ctx)
	if err != nil {
		zapctx.Error(ctx, "error starting server", zaputil.Error(err))
		return
	}

	addr := config.ListenAddr()
	zapctx.Info(ctx, "listening", zap.String("address", addr))
	err = http.ListenAndServe(addr, hnd)
	zapctx.Error(ctx, "server stopped", zaputil.Error(err))
}

func newHandler(ctx context.Context) (http.Handler, error) {
	st, acls, rks, err := newStore(ctx)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	err = storeutil.EnsureDefaultProducts(ctx, st)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	if pw := config.BootstrapAdminPassword(); pw != "" {
		err = storeutil.EnsureAdminServiceAccount(ctx, st, pw)
		if err != nil {
			return nil, errgo.Mask(err)
		}
	}
	pkey, err := config.SSOPublicKey()
	if err != nil {
		return nil, errgo.Mask(err)
	}
	tokenSecret, err := config.TokenSecret()
	if err != nil {
		return nil, errgo.Mask(err)
	}
	kafkaConfig, err := kafkaauditor.KafkaConfig()
	if err != nil {
		return nil, errgo.Mask(err)
	}
	if kafkaConfig != nil {
		wrappedStore, err := kafkaConfig.Store(st)
		if err != nil {
			return nil, errgo.Mask(err)
		}
		st = wrappedStore
	}

	return contracts.NewServer(ctx, contracts.Params{
		ACLStore:           acls,
		AWSCertificates:    config.AWSCertificates(),
		BakeryRootKeyStore: rks,
		SSOLocation:        config.SSOLocation(),
		SSOPublicKey:       pkey,
		Store:              st,
		SalesforceConfig:   salesforceConfig(),
		TokenSecret:        tokenSecret,
	})
}

func newStore(ctx context.Context) (store.Store, aclstore.ACLStore, bakery.RootKeyStore, error) {
	if config.UseMemStore() {
		return memstore.New(), aclstore.NewACLStore(memsimplekv.NewStore()), bakery.NewMemRootKeyStore(), nil
	}
	fail := func(err error) (store.Store, aclstore.ACLStore, bakery.RootKeyStore, error) {
		return nil, nil, nil, err
	}
	db, err := sql.Open("postgres", "")
	if err != nil {
		return fail(errgo.Mask(err))
	}
	// TODO Think about how to do this going forward. May be removed entirely.
	err = postgresstore.Init(ctx, db)
	if err != nil {
		return fail(errgo.Mask(err))
	}
	st, err := postgresstore.New(db)
	if err != nil {
		return fail(errgo.Mask(err))
	}

	kvs, err := sqlsimplekv.NewStore("postgres", db, "acls")
	if err != nil {
		return fail(errgo.Mask(err))
	}

	rks := postgresrootkeystore.NewRootKeys(db, "bakery_rootkeys", 10).NewStore(postgresrootkeystore.Policy{
		ExpiryDuration: 365 * 24 * time.Hour,
	})

	return st, aclstore.NewACLStore(kvs), rks, nil
}

func salesforceConfig() *salesforce.Config {
	if config.SalesforceDisabled() {
		return nil
	}
	return &salesforce.Config{
		ConsumerID:         config.SalesforceConsumerID(),
		ConsumerSecret:     config.SalesforceConsumerSecret(),
		User:               config.SalesforceUser(),
		Password:           config.SalesforcePassword(),
		SecurityToken:      config.SalesforceSecurityToken(),
		URL:                config.SalesforceURL(),
		AuthSessionTimeout: config.SalesforceSessionTimeout(),
		RequestTimeout:     config.SalesforceRequestTimeout(),
		TenantURL:          config.SalesforceTenantURL(),
		ProductNames:       map[string][]string{},
	}
}
