package main

import (
	"context"
	"crypto/hmac"
	"crypto/rand"
	"crypto/sha256"
	"encoding/base64"
	"fmt"
	"net/http"
	"net/url"
	"os"

	"github.com/juju/usso"
	"github.com/juju/usso/openid"
	errgo "gopkg.in/errgo.v1"
	macaroon "gopkg.in/macaroon.v2"

	"github.com/CanonicalLtd/ua-contracts/contractclient"
)

// serviceSecret is used to create the HMAC of login cookies.
var serviceSecret = func() []byte {
	var buf []byte
	if s := os.Getenv("SERVICE_SECRET"); s != "" {
		buf, _ = macaroon.Base64Decode([]byte(s))
	}
	if buf == nil {
		buf = make([]byte, 20)
		_, err := rand.Read(buf)
		if err != nil {
			panic(err)
		}
	}
	return buf
}()

var ssoClient = openid.NewClient(usso.ProductionUbuntuSSOServer, nil, nil)

// login handles the /login endpoint, it both initiates the openID
// request and completes it.
func login(w http.ResponseWriter, req *http.Request) {
	ctx := req.Context()
	req.ParseForm()
	if req.Form.Get("openid.mode") == "" {
		startLogin(ctx, w, req)
		return
	}
	completeLogin(ctx, w, req)
}

// startLogin initiates the login process. This performs the following
// steps:
//     1. Retrieve a login macaroon from $CONTRACT_URL.
//     2. Store the macaroon, along with a random nonce in a cookie.
//     3. Extract the third-party caveat from the macaroon.
//     4. Calculate an OpenID URL on login.canonical.com to redirect the
//        browser to.
//     5. Redirect the browser to that URL.
func startLogin(ctx context.Context, w http.ResponseWriter, req *http.Request) {
	m, err := contractclient.GetCanonicalSSOMacaroon(ctx, contractURL)
	if err != nil {
		panic(err)
	}
	state, err := writeLoginCookie(w, m)
	if err != nil {
		panic(err)
	}
	next := "/"
	if n := req.Form.Get("next"); n != "" {
		next = n
	}
	var caveatID []byte
	for _, cav := range m.Caveats() {
		if len(cav.VerificationId) == 0 {
			continue
		}
		caveatID = cav.Id
		break
	}
	oreq := openid.Request{
		ReturnTo: fmt.Sprintf("%s/login?next=%s&state=%s", serviceURL, next, state),
		Realm:    serviceURL + "/login",
		CaveatID: string(caveatID),
	}
	http.Redirect(w, req, ssoClient.RedirectURL(&oreq), http.StatusTemporaryRedirect)
}

// completeLogin finishes the login process. This will be called when the
// browser is redirected back from login.ubuntu.com. This performs the
// following steps:
//     1. Validate the response from login.ubuntu.com.
//     2. Validate the response is for the login initiated in this
//        session.
//     3. Bind the discharge macaroon to the original received from the
//        contract service.
//     4. Store the bound macaroons as an authentication cookie.
//     5. Redirect the browser back to the page the login was initiated
//        from.
func completeLogin(_ context.Context, w http.ResponseWriter, req *http.Request) {
	reqURL, err := url.Parse(serviceURL)
	if err != nil {
		panic(err)
	}
	reqURL.Path = req.URL.Path
	reqURL.RawQuery = req.URL.RawQuery
	ssoResp, err := ssoClient.Verify(reqURL.String())
	if err != nil {
		// TODO deal with login errors
		panic(err)
	}
	m, err := readLoginCookie(req, req.Form.Get("state"))
	if err != nil {
		// TODO deal with login errors
		panic(err)
	}
	ssoResp.Discharge.Bind(m.Signature())
	ms := macaroon.Slice{m, ssoResp.Discharge}
	buf, err := ms.MarshalBinary()
	if err != nil {
		panic(err)
	}
	http.SetCookie(w, &http.Cookie{
		Name:   "contracts-auth",
		Value:  base64.RawURLEncoding.EncodeToString(buf),
		MaxAge: 86400,
	})
	http.Redirect(w, req, req.Form.Get("next"), http.StatusTemporaryRedirect)
}

// writeLoginCookie creates a login cookie. This cookie is a session
// cookie that stores the macaroon received from the contracts API. The
// format of the cookie is as follows:
//      1 byte  - version (0)
//     16 bytes - random nonce
//      x bytes - serialized macaroon
// The HMAC value of the cookie is returned, this should be round tripped
// through the OpenID server and will be used to ensure the OpenID
// callback corresponds to request initiated in this session. This is to
// prevent possible CSRF attacks.
func writeLoginCookie(w http.ResponseWriter, m *macaroon.Macaroon) (string, error) {
	nonce := make([]byte, 16)
	_, err := rand.Read(nonce)
	if err != nil {
		return "", errgo.Mask(err)
	}
	data := []byte{0}
	data = append(data, nonce...)
	buf, err := m.MarshalBinary()
	if err != nil {
		return "", errgo.Mask(err)
	}
	data = append(data, buf...)
	http.SetCookie(w, &http.Cookie{
		Name:  "contracts-login",
		Value: base64.RawURLEncoding.EncodeToString(data),
	})
	h := hmac.New(sha256.New, serviceSecret)
	h.Write(data)
	return base64.RawURLEncoding.EncodeToString(h.Sum(nil)), nil
}

// readLoginCookie validates the login cookie and returns the macaroon
// stored within it. The login cookie is validated by calculating the
// HMAC of the stored cookie and ensuring it corresponds with state. The
// state value should have been round-tripped through the OpenID server
// and thus ensures that the OpenID request was initiated by this browser
// session.
func readLoginCookie(req *http.Request, state string) (*macaroon.Macaroon, error) {
	c, err := req.Cookie("contracts-login")
	if err != nil {
		return nil, errgo.Mask(err)
	}
	st, err := base64.RawURLEncoding.DecodeString(state)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	value, err := base64.RawURLEncoding.DecodeString(c.Value)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	h := hmac.New(sha256.New, serviceSecret)
	h.Write(value)
	if !hmac.Equal(st, h.Sum(nil)) {
		return nil, errgo.New("invalid login state")
	}
	var m macaroon.Macaroon
	if value[0] != 0 {
		return nil, errgo.New("unsupported login cookie")
	}
	if err := m.UnmarshalBinary(value[17:]); err != nil {
		return nil, errgo.Mask(err)
	}
	return &m, nil
}
