package main

import (
	"context"
	"fmt"
	"html/template"
	"net/http"
	"strings"

	"github.com/CanonicalLtd/ua-contracts/contractclient"
	"github.com/CanonicalLtd/ua-contracts/params"
)

// serveResourceToken handles the /get-resource-token endpoint which uses the
// contract token provided as a query to first register a machine and then
// obtain a resource token.
func serveResourceToken(w http.ResponseWriter, req *http.Request) {
	ctx := req.Context()
	// Check the request.
	q := req.URL.Query()
	contractToken := q.Get("token")
	resName := q.Get("resource")
	var found bool
	for _, choice := range resourceChoices {
		if choice == resName {
			found = true
			break
		}
	}
	if contractToken == "" || !found {
		w.WriteHeader(http.StatusBadRequest)
		mustRender(w, badRequestTemplate, fmt.Sprintf("contract token: %q, resource name: %q", contractToken, resName))
		return
	}
	machineID := q.Get("machine-id")
	if machineID == "" {
		machineID = "3682052cd10d4cdfa87d11fb9df14d7e"
	}
	// Obtain a machine token.
	machineToken, err := registerMachine(ctx, contractToken, machineID)
	if err != nil {
		panic(err)
	}

	// Obtain a resource token.
	resourceToken, entitlement, err := getResourceToken(ctx, machineToken, machineID, resName)
	resp := resourceResponse{
		Name:        resName,
		Token:       resourceToken,
		Affordances: make(map[string]string),
		Directives:  make(map[string]string),
		Obligations: make(map[string]string),
	}
	if entitlement == nil {
		panic(fmt.Sprintf("no entitlement found for resource %q", resName))
	}
	switch e := entitlement.(type) {
	case params.EntitlementAptRepository:
		resp.Affordances["architectures"] = strings.Join(e.Affordances.Architectures, ", ")
		resp.Affordances["series"] = strings.Join(e.Affordances.Series, ", ")
		resp.Directives["apt key"] = e.Directives.AptKey
		resp.Directives["apt URL"] = e.Directives.AptURL
		resp.Directives["suites"] = strings.Join(e.Directives.Suites, ", ")
		resp.Obligations["enabled by default"] = fmt.Sprint(e.Obligations.EnableByDefault)
	case params.EntitlementLivepatch:
		resp.Affordances["tier"] = e.Affordances.Tier
		resp.Affordances["architectures"] = strings.Join(e.Affordances.Architectures, ", ")
		resp.Affordances["kernel flavors"] = strings.Join(e.Affordances.KernelFlavors, ", ")
		resp.Affordances["min kernel version"] = e.Affordances.MinKernelVersion
		resp.Affordances["series"] = strings.Join(e.Affordances.Series, ", ")
		resp.Directives["CA cert"] = e.Directives.CaCerts
		resp.Directives["remote server"] = e.Directives.RemoteServer
		resp.Obligations["enabled by default"] = fmt.Sprint(e.Obligations.EnableByDefault)
	default:
		// This should never happen.
		panic(fmt.Sprintf("invalid entitlement of type %T", e))
	}
	mustRender(w, resourceTemplate, resp)
}

func registerMachine(ctx context.Context, token, machineID string) (machineToken string, err error) {
	// Create a client authorizing with the contracts token.
	client := contractclient.New(contractclient.NewParams{
		BaseURL:    contractURL,
		Authorizer: contractclient.BearerAuthorizer(token),
	})
	// Register a machine.
	resp, err := client.AddContractMachine(ctx, &params.AddContractMachineRequest{
		Body: params.AddContractMachineBody{
			MachineId:    machineID,
			Architecture: "amd64",
			Os: params.OS{
				Distribution: "Ubuntu",
				Kernel:       "42.47",
				Release:      "42.10",
				Series:       "Cosmic",
				Type:         "Linux",
			},
		},
	})
	if err != nil {
		return "", err
	}
	return resp.MachineToken, nil
}

func getResourceToken(ctx context.Context, token, machineID, resName string) (resourceToken string, entitlement params.Entitlement, err error) {
	// Create a client authorizing with the machine token.
	client := contractclient.New(contractclient.NewParams{
		BaseURL:    contractURL,
		Authorizer: contractclient.BearerAuthorizer(token),
	})
	// Obtain a resource token.
	resp, err := client.GetResourceMachineAccess(ctx, &params.GetResourceMachineAccessRequest{
		Resource: resName,
		Machine:  machineID,
	})
	if err != nil {
		return "", nil, err
	}
	return resp.ResourceToken, resp.Entitlement, nil
}

type resourceResponse struct {
	Name        string
	Token       string
	Affordances map[string]string
	Directives  map[string]string
	Obligations map[string]string
}

var badRequestTemplate = template.Must(template.New("badRequest").Parse(`
<!doctype html>
<html>
<head><title>Contracts - Bad Request</title></head>
<body>
  <p>Bad Request: {{.}}</p>
  <p><a href="/">Home...</a></p>
</body>
</html>
`[1:]))

var resourceTemplate = template.Must(template.New("resource").Parse(`
<!doctype html>
<html>
<head><title>{{.Name}} Resource</title></head>
<body>
<h1>Resource {{.Name}}</h1>
<p>Token: <strong>{{.Token}}</strong></p>
<h2>Affordances</h2>
<table>
{{ range $key, $value := .Affordances }}
  <tr><th>{{ $key }}</th><td>{{ $value }}</td></tr>
{{ end }}
</table>
<h2>Directives</h2>
<table>
{{ range $key, $value := .Directives }}
  <tr><th>{{ $key }}</th><td>{{ $value }}</td></tr>
{{ end }}
</table>
<h2>Obligations</h2>
<table>
{{ range $key, $value := .Obligations }}
  <tr><th>{{ $key }}</th><td>{{ $value }}</td></tr>
{{ end }}
</table>
</body>
</html>
`[1:]))
