package main

import (
	"context"
	"encoding/json"
	"flag"
	"fmt"
	"io"
	"log"
	"os"
	"time"

	randomdata "github.com/Pallinder/go-randomdata"
	"github.com/Shopify/sarama"
	"github.com/juju/clock/testclock"

	"github.com/CanonicalLtd/ua-contracts/internal/kafkaauditor"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
)

var (
	nAccounts            = 100
	nContractsPerAccount = 1
	nMachinesPerContract = 10

	startTimeArg string
	intervalArg  string
)

type generator struct {
	kafkaauditor.Auditor
	serviceProducts []string
	supportProducts []string

	clock    *testclock.Clock
	interval time.Duration
}

func init() {
	flag.IntVar(&nAccounts, "accounts", 100, "number of accounts")
	flag.IntVar(&nContractsPerAccount, "contracts", 1, "number of contracts per account")
	flag.IntVar(&nMachinesPerContract, "machines", 10, "number of machines per contract")
	flag.StringVar(&startTimeArg, "start", "", "starting timestamp for generated events")
	flag.StringVar(&intervalArg, "interval", "", "time interval between generated events")

	sarama.Logger = log.New(os.Stderr, "", log.LstdFlags)
}

func main() {
	flag.Parse()

	now := time.Now().UTC()
	startTime := time.Date(now.Year(), now.Month(), 1, 0, 0, 0, 0, time.UTC)
	interval := time.Second
	var err error
	if startTimeArg != "" {
		startTime, err = time.Parse(time.RFC3339, startTimeArg)
		if err != nil {
			panic(err)
		}
	}
	if intervalArg != "" {
		interval, err = time.ParseDuration(intervalArg)
		if err != nil {
			panic(err)
		}
	}
	clk := testclock.NewClock(startTime)

	cfg, err := kafkaauditor.KafkaConfig()
	if err != nil {
		log.Fatalf("failed to configure a kafka auditor: %v ", err)
	}
	auditor, err := cfg.Auditor()
	if err != nil {
		log.Fatalf("failed to create new auditor: %v", err)
	}
	defer func() {
		err := auditor.(io.Closer).Close()
		if err != nil {
			log.Fatalf("failed to close auditor: %v", err)
		}
	}()

	g := generator{Auditor: auditor, clock: clk, interval: interval}
	g.generate()
	fmt.Println(g.clock.Now().Format(time.RFC3339), interval)
}

func (g *generator) generate() {
	g.serviceProducts = []string{"livepatch", "esm", "fips"}
	for _, id := range g.serviceProducts {
		p := &kafkaauditor.AuditAddProduct{
			ProductID:   id,
			ProductName: id,
			Resources: []kafkaauditor.ProductResource{{
				Type: id,
			}},
		}
		g.Audit(context.Background(), g.clock.Now(), p)
	}
	g.supportProducts = nil
	for _, level := range []string{"essential", "standard", "advanced"} {
		id := "support-" + level
		data := json.RawMessage([]byte(`{"level":"` + level + `"}`))
		p := &kafkaauditor.AuditAddProduct{
			ProductID:   id,
			ProductName: id,
			Resources: []kafkaauditor.ProductResource{{
				Type: "support",
				Data: &data,
			}},
		}
		g.Audit(context.Background(), g.clock.Now(), p)
		g.supportProducts = append(g.supportProducts, id)
	}
	g.generateAccounts()
}

func (g *generator) generateAccounts() {
	for i := 0; i < nAccounts; i++ {
		account := &kafkaauditor.AuditAddAccount{
			AccountID:      token.MustNewAccountID(),
			AccountName:    randomdata.Email(),
			AccountCreated: g.clock.Now(),
			UserID:         randomdata.Letters(12),
		}
		g.Audit(context.Background(), g.clock.Now(), account)
		g.clock.Advance(g.interval)
		g.generateContracts(account)
	}
}

func (g *generator) generateContracts(account *kafkaauditor.AuditAddAccount) {
	for i := 0; i < nContractsPerAccount; i++ {
		now := g.clock.Now()
		contract := &kafkaauditor.AuditAddContract{
			ContractID:    token.MustNewContractID(),
			Name:          randomdata.SillyName(),
			AccountID:     account.AccountID,
			ProductIDs:    append(g.serviceProducts, randomdata.StringSample(g.supportProducts...)),
			Created:       now,
			CreatedBy:     account.AccountName,
			EffectiveFrom: &now,
		}
		g.Audit(context.Background(), g.clock.Now(), contract)
		g.clock.Advance(g.interval)
		g.generateContractMachines(contract)
	}
}

func (g *generator) generateContractMachines(contract *kafkaauditor.AuditAddContract) {
	for i := 0; i < nMachinesPerContract; i++ {
		contractMachine := &kafkaauditor.AuditAddContractMachine{
			ContractID: contract.ContractID,
			MachineID:  randomdata.Letters(32),
		}
		g.Audit(context.Background(), g.clock.Now(), contractMachine)
		g.clock.Advance(g.interval)
		for _, resource := range g.serviceProducts {
			resourceToken := &kafkaauditor.AuditResourceToken{
				ContractID: contract.ContractID,
				MachineID:  contractMachine.MachineID,
				Resource:   resource,
			}
			g.Audit(context.Background(), g.clock.Now(), resourceToken)
			g.clock.Advance(g.interval)
		}
	}
}
