// Copyright 2019 Canonical Ltd.

// Package config supports configuration options for the contract server.
package config

import (
	"crypto/rsa"
	"crypto/tls"
	"crypto/x509"
	"encoding/base64"
	"encoding/pem"
	"fmt"
	"os"
	"strconv"
	"strings"
	"time"

	"go.uber.org/zap/zapcore"
	"gopkg.in/errgo.v1"
)

// ListenAddr returns the address the server should listen on.
func ListenAddr() string {
	return fmt.Sprintf(":%s", os.Getenv("PORT"))
}

// LogLevel returns the level of logging to perform. If the
// environment variable is not set, the level will be the default
// INFO level.
func LogLevel() zapcore.Level {
	var level zapcore.Level
	level.UnmarshalText([]byte(os.Getenv("LOGLEVEL")))
	return level
}

// TokenSecret returns the token secret set in an environment variable.
// It attempts to decode the environment variable as base64, and returns
// an error if unable to do so.
func TokenSecret() ([]byte, error) {
	secretStr := os.Getenv("TOKEN_SECRET")
	if secretStr == "" {
		return nil, nil
	}
	data, err := base64.StdEncoding.DecodeString(secretStr)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return data, nil
}

// UseMemStore returns whether the USE_MEMSTORE environment variable is set or not.
func UseMemStore() bool {
	return os.Getenv("USE_MEMSTORE") != ""
}

// SSOLocation returns the location of the SSO server.
func SSOLocation() string {
	return os.Getenv("SSO_LOCATION")
}

// SSOPublicKey returns the configured public key for the SSO server.
func SSOPublicKey() (*rsa.PublicKey, error) {
	ssoPublicKey := []byte(os.Getenv("SSO_PUBLICKEY"))
	if len(ssoPublicKey) == 0 {
		return nil, nil
	}
	block, rest := pem.Decode(ssoPublicKey)
	if len(rest) == len(ssoPublicKey) {
		return nil, errgo.New("failed to read PEM block from SSO_PUBLICKEY")
	}
	pk, err := x509.ParsePKIXPublicKey(block.Bytes)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	pkey, ok := pk.(*rsa.PublicKey)
	if !ok {
		return nil, errgo.Newf("unsupported key type %T", pk)
	}
	return pkey, nil
}

func SalesforceConsumerID() string {
	return os.Getenv("SF_CONSUMER_ID")
}

func SalesforceConsumerSecret() string {
	return os.Getenv("SF_CONSUMER_SECRET")
}

func SalesforceUser() string {
	return os.Getenv("SF_USER")
}

func SalesforcePassword() string {
	return os.Getenv("SF_PASSWORD")
}

func SalesforceSecurityToken() string {
	return os.Getenv("SF_SECURITY_TOKEN")
}

func SalesforceURL() string {
	return os.Getenv("SF_URL")
}

func SalesforceSessionTimeout() time.Duration {
	sessionTimeout, err := time.ParseDuration(os.Getenv("SF_SESSION_TIMEOUT"))
	if err != nil {
		return 0
	}
	return sessionTimeout
}

func SalesforceRequestTimeout() time.Duration {
	requestTimeout, err := time.ParseDuration(os.Getenv("SF_REQUEST_TIMEOUT"))
	if err != nil {
		return 0
	}
	return requestTimeout
}

func SalesforceTenantURL() string {
	return os.Getenv("SF_TENANT_URL")
}

func SalesforceDisabled() bool {
	val := os.Getenv("SALESFORCE_DISABLED")
	v, err := strconv.ParseBool(val)
	if err != nil {
		return false
	}
	return v
}

// KafkaBrokerURLs returns the KAFKA_BROKERS environment variable
func KafkaBrokerURLs() []string {
	brokers := os.Getenv("KAFKA_BROKERS")
	if brokers == "" {
		return nil
	}
	return strings.Split(brokers, ",")
}

// TLSConfig contains values a client needs to connect to a server via tls.
type TLSConfig struct {
	// Certificate holds the client certificate.
	Certificate *tls.Certificate
	// CACertificate holds the CA certificate of the authority
	// that created the client certificate.
	CACertificate *x509.Certificate
}

// Config loads specified certificates and returns tls.Config.
func (c *TLSConfig) Config() (*tls.Config, error) {
	if c.Certificate != nil {
		config := &tls.Config{
			Certificates: []tls.Certificate{*c.Certificate},
		}
		if c.CACertificate != nil {
			caCertPool := x509.NewCertPool()
			caCertPool.AddCert(c.CACertificate)
			config.RootCAs = caCertPool
		}
		return config, nil
	}
	return nil, errgo.New("certificate not specified")
}

// KafkaTLS fetches KAFKA_CLIENT_CERT, KAFKA_CLIENT_KEY and KAFKA_CA_CERT
// environment variables and returns a tls config structure.
func KafkaTLS() (*TLSConfig, error) {
	clientCertString := os.Getenv("KAFKA_CLIENT_CERT")
	clientKeyString := os.Getenv("KAFKA_CLIENT_KEY")
	caCertString := os.Getenv("KAFKA_CA_CERT")

	if clientCertString == "" && clientKeyString == "" {
		return nil, nil
	}
	cert, err := tls.X509KeyPair([]byte(clientCertString), []byte(clientKeyString))
	if err != nil {
		return nil, errgo.Mask(err)
	}
	config := TLSConfig{
		Certificate: &cert,
	}
	if caCertString != "" {
		pemData, _ := pem.Decode([]byte(caCertString))
		if pemData == nil {
			return nil, errgo.New("failed to decode CA certificate")
		}
		caCert, err := x509.ParseCertificate(pemData.Bytes)
		if err != nil {
			return nil, errgo.NoteMask(err, "invalid CA certificate")
		}
		config.CACertificate = caCert
	}
	return &config, nil
}

// BootstrapAdminPassword returns the password to use for the admin
// service-account, if the service account does not exist.
func BootstrapAdminPassword() string {
	return os.Getenv("BOOTSTRAP_ADMIN_PASSWORD")
}

// AWSCertificates loads AWS certificates. These are the certificates
// used to verify AWS machine information. Any failures loading
// certificates are ignored.
func AWSCertificates() []*x509.Certificate {
	var certs []*x509.Certificate
	rest := []byte(os.Getenv("AWS_CERTIFICATES"))
	var block *pem.Block
	for len(rest) > 0 {
		block, rest = pem.Decode(rest)
		if block == nil {
			break
		}
		if block.Type != "CERTIFICATE" {
			continue
		}
		cert, err := x509.ParseCertificate(block.Bytes)
		if err != nil {
			continue
		}
		certs = append(certs, cert)
	}
	return certs
}
