// Copyright 2019 Canonical Ltd.

package config_test

import (
	"encoding/base64"
	"os"
	"testing"

	qt "github.com/frankban/quicktest"
	"go.uber.org/zap/zapcore"

	"github.com/CanonicalLtd/ua-contracts/config"
)

func TestListenAddr(t *testing.T) {
	c := qt.New(t)

	os.Setenv("PORT", "1234")
	c.Assert(config.ListenAddr(), qt.Equals, ":1234")
}

func TestLogLevel(t *testing.T) {
	c := qt.New(t)

	os.Setenv("LOGLEVEL", "DEBUG")
	c.Assert(config.LogLevel(), qt.Equals, zapcore.DebugLevel)

	os.Setenv("LOGLEVEL", "ERROR")
	c.Assert(config.LogLevel(), qt.Equals, zapcore.ErrorLevel)

	os.Unsetenv("LOGLEVEL")
	c.Assert(config.LogLevel(), qt.Equals, zapcore.InfoLevel)
}

func TestSSOLocation(t *testing.T) {
	c := qt.New(t)

	c.Assert(config.SSOLocation(), qt.Equals, "")
	os.Setenv("SSO_LOCATION", "login.example.com")
	c.Assert(config.SSOLocation(), qt.Equals, "login.example.com")
}

func TestTokenSecret(t *testing.T) {
	c := qt.New(t)

	token := []byte("7BR#%fjkHt-S8TCK")
	b64 := base64.StdEncoding.EncodeToString(token)
	c.Setenv("TOKEN_SECRET", b64)
	res, err := config.TokenSecret()
	c.Assert(err, qt.Equals, nil)
	c.Assert(res, qt.DeepEquals, token)

	c.Setenv("TOKEN_SECRET", "notb64!!")
	res, err = config.TokenSecret()
	c.Assert(err, qt.ErrorMatches, "illegal base64 data at input byte 6")

	c.Setenv("TOKEN_SECRET", "")
	res, err = config.TokenSecret()
	c.Assert(err, qt.Equals, nil)
	c.Assert(res, qt.IsNil)
}

var ssoPublicKeyTests = []struct {
	name        string
	env         string
	expectError string
}{{
	name: "default",
}, {
	name: "invalid-block",
	env: `MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAtO9bUDClVY5VKD8cLRbo
bw/ruGTTepdvhDc3oewHrnog/i7ymQpB7fCbn3iryNs704p9Yg1UaQM/iFJ75ojB
4gsGHF0yZWj4vf0Wt8vjpY7wdy/7MHerizED132ncoSuc676ulydVzyOPUGKQbPe
/EkYJhfoxyJHrBPvIHj445EK//uScrrKmwXteu0sOytDSLU+EPJgTQQ38sKak9FM
+MAW4R4wPvGZibUsm4JWfGGwf68FPXwZ0a/BIJXL8hBvuGaQlFy9jrafXlJKqY+I
XAQrHqnBTYxhyhhXCQrwBJOxcVTZe/zowGC3W/YO2yIWpqgkgkg4KkmyzayvhpQ2
4wIDAQAB
-----END PUBLIC KEY-----
`,
	expectError: `failed to read PEM block from SSO_PUBLICKEY`,
}, {
	name: "invalid-key",
	env: `-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAtO9bUDClVY5VKD8cLRbo
bw/ruGTTepdvhDc3oewHrnog/i7ymQpB7fCbn3iryNs704p9Yg1UaQM/iFJ75ojB
4gsGHF0yZWj4vf0Wt8vjpY7wdy/7MHerizED132ncoSuc676ulydVzyOPUGKQbPe
/EkYJhfoxyJHrBPvIHj445EK//uScrrKmwXteu0sOytDSLU+EPJgTQQ38sKak9FM
+MAW4R4wPvGZibUsm4JWfGGwf68FPXwZ0a/BIJXL8hBvuGaQlFy9jrafXlJKqY+I
XAQrHqnBTYxhyhhXCQrwBJOxcVTZe/zowGC3W/YO2yIWpqgkgkg4KkmyzayvhpQ2
4wIDAQABAA==
-----END PUBLIC KEY-----
`,
	expectError: `x509: trailing data after ASN.1 of public-key`,
}, {
	name: "invalid-key-type",
	env: `-----BEGIN PUBLIC KEY-----
MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAEpS0TN4JLay9g9sjSqIdf4AhlIpsi
nMfnogcF/val8eELGiavmOj9lfvoUHS76vlfAmXniWg5/nslChsq6hYlVQ==
-----END PUBLIC KEY-----
`,
	expectError: `unsupported key type \*ecdsa.PublicKey`,
}}

func TestSSOPublicKey(t *testing.T) {
	c := qt.New(t)

	for _, test := range ssoPublicKeyTests {
		c.Run(test.name, func(c *qt.C) {
			os.Setenv("SSO_PUBLICKEY", test.env)
			pkey, err := config.SSOPublicKey()
			if test.expectError != "" {
				c.Assert(err, qt.ErrorMatches, test.expectError)
				return
			}
			c.Assert(err, qt.Equals, nil)
			c.Assert(pkey, qt.IsNil)
		})
	}
}

func TestUseMemstore(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	c.Setenv("USE_MEMSTORE", "")
	c.Assert(config.UseMemStore(), qt.Equals, false)
	c.Setenv("USE_MEMSTORE", "x")
	c.Assert(config.UseMemStore(), qt.Equals, true)
}

func TestKafkaBrokerURLs(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	os.Unsetenv("KAFKA_BROKERS")
	c.Assert(config.KafkaBrokerURLs(), qt.IsNil)
	os.Setenv("KAFKA_BROKERS", "localhost:1234")
	c.Assert(config.KafkaBrokerURLs(), qt.DeepEquals, []string{"localhost:1234"})
}

func TestKafkaTLS(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	cfg, err := config.KafkaTLS()
	c.Assert(err, qt.IsNil)
	c.Assert(cfg, qt.IsNil)

	certPem := []byte(`-----BEGIN CERTIFICATE-----
MIIBhTCCASugAwIBAgIQIRi6zePL6mKjOipn+dNuaTAKBggqhkjOPQQDAjASMRAw
DgYDVQQKEwdBY21lIENvMB4XDTE3MTAyMDE5NDMwNloXDTE4MTAyMDE5NDMwNlow
EjEQMA4GA1UEChMHQWNtZSBDbzBZMBMGByqGSM49AgEGCCqGSM49AwEHA0IABD0d
7VNhbWvZLWPuj/RtHFjvtJBEwOkhbN/BnnE8rnZR8+sbwnc/KhCk3FhnpHZnQz7B
5aETbbIgmuvewdjvSBSjYzBhMA4GA1UdDwEB/wQEAwICpDATBgNVHSUEDDAKBggr
BgEFBQcDATAPBgNVHRMBAf8EBTADAQH/MCkGA1UdEQQiMCCCDmxvY2FsaG9zdDo1
NDUzgg4xMjcuMC4wLjE6NTQ1MzAKBggqhkjOPQQDAgNIADBFAiEA2zpJEPQyz6/l
Wf86aX6PepsntZv2GYlA5UpabfT2EZICICpJ5h/iI+i341gBmLiAFQOyTDT+/wQc
6MF9+Yw1Yy0t
-----END CERTIFICATE-----`)
	keyPem := []byte(`-----BEGIN EC PRIVATE KEY-----
MHcCAQEEIIrYSSNQFaA2Hwf1duRSxKtLYX5CB04fSeQ6tF1aY/PuoAoGCCqGSM49
AwEHoUQDQgAEPR3tU2Fta9ktY+6P9G0cWO+0kETA6SFs38GecTyudlHz6xvCdz8q
EKTcWGekdmdDPsHloRNtsiCa697B2O9IFA==
-----END EC PRIVATE KEY-----`)
	caCertPem := []byte(`-----BEGIN CERTIFICATE-----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-----END CERTIFICATE-----`)

	os.Setenv("KAFKA_CLIENT_CERT", string(certPem))
	os.Setenv("KAFKA_CLIENT_KEY", string(keyPem))
	os.Setenv("KAFKA_CA_CERT", string(caCertPem))
	cfg, err = config.KafkaTLS()
	c.Assert(err, qt.IsNil)
	c.Assert(cfg.Certificate, qt.Not(qt.IsNil))
	c.Assert(cfg.CACertificate, qt.Not(qt.IsNil))
}

func TestBootstrapAdminPassword(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	c.Setenv("BOOTSTRAP_ADMIN_PASSWORD", "")
	c.Assert(config.BootstrapAdminPassword(), qt.Equals, "")
	c.Setenv("BOOTSTRAP_ADMIN_PASSWORD", "x")
	c.Assert(config.BootstrapAdminPassword(), qt.Equals, "x")
}
