package contractclient

import (
	"context"
	"encoding/base64"
	"net/http"

	errgo "gopkg.in/errgo.v1"
	httprequest "gopkg.in/httprequest.v1"
	macaroon "gopkg.in/macaroon.v2"

	"github.com/CanonicalLtd/ua-contracts/params"
)

//go:generate httprequest-generate-client github.com/CanonicalLtd/ua-contracts/internal/v1 handler client

// Client represents the client of an contracts server.
type Client struct {
	client
}

// NewParams holds the parameters for creating a new client.
type NewParams struct {
	// BaseURL holds the URL of the identity manager.
	BaseURL string

	// Authorizer, if set, is used to authorize requests made by this
	// client.
	Authorizer Authorizer
}

// New returns a new client.
func New(p NewParams) *Client {
	var c Client
	c.Client.BaseURL = p.BaseURL
	c.Client.UnmarshalError = httprequest.ErrorUnmarshaler(new(params.Error))
	if p.Authorizer != nil {
		c.Client.Doer = authDoer(p.Authorizer)
	}
	return &c
}

// An Authorizer modifies and returns the given request to add
// authentication information.
type Authorizer func(*http.Request)

// BasicAuthorizer returns an Authorizer that uses http basic
// authentication, with the given credentials, to authorize requests.
func BasicAuthorizer(username, password string) Authorizer {
	return func(req *http.Request) {
		req.SetBasicAuth(username, password)
	}
}

// BearerAuthorizer returns an Authorizer that uses http bearer
// authentication, with the given credential, to authorize requests.
func BearerAuthorizer(token string) Authorizer {
	auth := "Bearer " + token
	return func(req *http.Request) {
		req.Header.Set("Authorization", auth)
	}
}

// MacaroonAuthorizer returns an Authorizer that uses macaroon
// authentication, with the given credential, to authorize requests.
func MacaroonAuthorizer(ms macaroon.Slice) (Authorizer, error) {
	buf, err := ms.MarshalBinary()
	if err != nil {
		return nil, errgo.Mask(err)
	}
	auth := "Macaroon " + base64.RawURLEncoding.EncodeToString(buf)
	return func(req *http.Request) {
		req.Header.Set("Authorization", auth)
	}, nil
}

type authDoer Authorizer

func (a authDoer) Do(req *http.Request) (*http.Response, error) {
	Authorizer(a)(req)
	return http.DefaultClient.Do(req)
}

func GetCanonicalSSOMacaroon(ctx context.Context, baseURL string) (*macaroon.Macaroon, error) {
	c := New(NewParams{
		BaseURL: baseURL,
	})
	resp, err := c.GetCanonicalSSOMacaroon(ctx, &params.GetCanonicalSSOMacaroonRequest{})
	if err != nil {
		return nil, errgo.Mask(err, errgo.Any)
	}
	buf, err := macaroon.Base64Decode([]byte(resp.Macaroon))
	if err != nil {
		return nil, errgo.Mask(err)
	}
	var m macaroon.Macaroon
	if err := m.UnmarshalBinary(buf); err != nil {
		return nil, errgo.Mask(err)
	}
	return &m, nil
}
