// Copyright 2019 Canonical Ltd.

// Package awsauth provides authentication for AWS PKCs#7 signed
// documents.
package awsauth

import (
	"crypto/x509"
	"encoding/base64"
	"encoding/json"
	"time"

	"github.com/fullsailor/pkcs7"
	errgo "gopkg.in/errgo.v1"
)

// A Decoder is used to verify and decode an AWS Instance Identity
// Documents. See
// https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instance-identity-documents.html.
type Decoder struct {
	Certs []*x509.Certificate
}

// Decoder decodes the IID and verifies the signature.
func (d Decoder) Decode(msg string) (*InstanceInformationDocument, error) {
	buf, err := base64.StdEncoding.DecodeString(msg)
	if err != nil {
		return nil, errgo.Notef(err, "cannot decode PKCS#7 document")
	}
	data, err := pkcs7.Parse(buf)
	if err != nil {
		return nil, errgo.Notef(err, "cannot parse PKCS#7 document")
	}

	data.Certificates = d.Certs
	if err := data.Verify(); err != nil {
		return nil, errgo.Notef(err, "cannot verify PKCS#7 document")
	}
	var doc InstanceInformationDocument
	if err := json.Unmarshal(data.Content, &doc); err != nil {
		return nil, errgo.Notef(err, "invalid instance information document")
	}
	return &doc, nil
}

// An instanceInformationDocument holds information about an AWS
// instance.
type InstanceInformationDocument struct {
	DevpayProductCodes      []string  `json:"devpayProductCodes"`
	MarketplaceProductCodes []string  `json:"marketplaceProductCodes"`
	ImageId                 string    `json:"imageId"`
	InstanceType            string    `json:"instanceType"`
	Version                 string    `json:"version"`
	BillingProducts         []string  `json:"billingProducts"`
	InstanceId              string    `json:"instanceId"`
	PendingTime             time.Time `json:"pendingTime"`
	AccountId               string    `json:"accountId"`
	AvailabilityZone        string    `json:"availabilityZone"`
	KernelId                *string   `json:"kernelId"`
	RamdiskId               *string   `json:"ramdiskId"`
	Architecture            string    `json:"architecture"`
	PrivateIp               string    `json:"privateIp"`
	Region                  string    `json:"region"`
}
