// Copyright 2019 Canonical Ltd.

package kafkaauditor

import (
	"encoding/json"
	"time"

	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
	"github.com/CanonicalLtd/ua-contracts/params"
)

var _ AuditEntry = (*AuditResourceToken)(nil)

// AuditResourceToken represents a call to issue a resource token for
// a machine.
type AuditResourceToken struct {
	ContractID token.ContractID `json:"contractID"`
	MachineID  string           `json:"machineID"`
	Resource   string           `json:"resource"`
}

// KafkaInfo implements the AuditEntry interface.
func (a *AuditResourceToken) KafkaInfo() KafkaInfo {
	return KafkaInfo{
		Topic: "ua-resource-token",
		Key:   a.ContractID.String() + "-" + a.MachineID,
	}
}

var _ AuditEntry = (*AuditAddContract)(nil)

// AuditAddContract represents a call to the AddContract method.
type AuditAddContract struct {
	ContractID    token.ContractID `json:"contractID"`
	Name          string           `json:"name"`
	AccountID     token.AccountID  `json:"accountID"`
	ProductIDs    []string         `json:"productID"`
	Created       time.Time        `json:"created"`
	CreatedBy     string           `json:"createdBy"`
	EffectiveFrom *time.Time       `json:"effectiveFrom,omitempty"`
	EffectiveTo   *time.Time       `json:"effectiveTo,omitempty"`
}

// KafkaInfo implements the AuditEntry interface.
func (a *AuditAddContract) KafkaInfo() KafkaInfo {
	return KafkaInfo{
		Topic: "ua-contract",
		Key:   a.ContractID.String(),
	}
}

var _ AuditEntry = (*AuditAddContract)(nil)

// AuditAddContractMachine represents a call to the AddContractMachine method..
type AuditAddContractMachine struct {
	ContractID token.ContractID `json:"contractID"`
	MachineID  string           `json:"machineID"`
}

// KafkaInfo implements the AuditEntry interface.
func (a *AuditAddContractMachine) KafkaInfo() KafkaInfo {
	return KafkaInfo{
		Topic: "ua-contract-machine",
		Key:   a.ContractID.String() + "-" + a.MachineID,
	}
}

var _ AuditEntry = (*AuditAddContractToken)(nil)

// AuditAddContractToken represents a call to the AddContractToken method.
type AuditAddContractToken struct {
	ContractToken token.ContractToken `json:"contractToken"`
	ContractID    token.ContractID    `json:"contractID"`
	Expires       time.Time           `json:"expires"`
}

// KafkaInfo implements the AuditEntry interface.
func (a *AuditAddContractToken) KafkaInfo() KafkaInfo {
	return KafkaInfo{
		Topic: "ua-contract-token",
		Key:   a.ContractID.String(),
	}
}

var _ AuditEntry = (*AuditAddAccount)(nil)

// AuditAddAccount represents a call to the AddAccount method.
type AuditAddAccount struct {
	AccountID      token.AccountID `json:"accountID"`
	AccountName    string          `json:"accountName"`
	AccountCreated time.Time       `json:"accountCreated"`
	UserID         string          `json:"userID"`
}

// KafkaInfo implements the AuditEntry interface.
func (a *AuditAddAccount) KafkaInfo() KafkaInfo {
	return KafkaInfo{
		Topic: "ua-account",
		Key:   a.AccountID.String(),
	}
}

var _ AuditEntry = (*AuditEnsureMachine)(nil)

// AuditEnsureMachine represents a call to the EnsureMachine method..
type AuditEnsureMachine struct {
	MachineID    string    `json:"machineID"`
	Architecture string    `json:"architecture"`
	OS           params.OS `json:"os"`
}

// KafkaInfo implements the AuditEntry interface.
func (a *AuditEnsureMachine) KafkaInfo() KafkaInfo {
	return KafkaInfo{
		Topic: "ua-machine",
		Key:   a.MachineID,
	}
}

var _ AuditEntry = (*AuditEnsureUser)(nil)

// AuditEnsure user represents a call to the EnsureUser method.
type AuditEnsureUser struct {
	ID          string `json:"id"`
	Username    string `json:"username"`
	DisplayName string `json:"displayName"`
	Email       string `json:"email"`
	Verified    bool   `json:"verified"`
}

// KafkaInfo implements the AuditEntry interface.
func (a *AuditEnsureUser) KafkaInfo() KafkaInfo {
	return KafkaInfo{
		Topic: "ua-user",
		Key:   a.ID,
	}
}

var _ AuditEntry = (*AuditRemoveContractMachine)(nil)

// AuditRemoveContractMachine represents a call to the RemoveContractMachine method.
type AuditRemoveContractMachine struct {
	ContractID token.ContractID `json:"contractID"`
	MachineID  string           `json:"machineID"`
}

// KafkaInfo implements the AuditEntry interface.
func (a *AuditRemoveContractMachine) KafkaInfo() KafkaInfo {
	return KafkaInfo{
		Topic: "ua-contract-machine-remove",
		Key:   a.ContractID.String() + "-" + a.MachineID,
	}
}

var _ AuditEntry = (*AuditSetAccountAccess)(nil)

// AuditSetAccountAccess represents a call to the SetAccountAccess method.
type AuditSetAccountAccess struct {
	AccountID     token.AccountID     `json:"accountID"`
	UserID        string              `json:"userID"`
	AccountAccess store.AccountAccess `json:"accountAccess"`
}

// KafkaInfo implements the AuditEntry interface.
func (a *AuditSetAccountAccess) KafkaInfo() KafkaInfo {
	return KafkaInfo{
		Topic: "ua-account-access",
		Key:   a.AccountID.String(),
	}
}

var _ AuditEntry = (*AuditAddProduct)(nil)

// AuditAddProduct represents a call to the AddProduct method.
type AuditAddProduct struct {
	ProductID   string            `json:"productID'`
	ProductName string            `json:"productName"`
	Resources   []ProductResource `json:"resources"`
}

// KafkaInfo implements the AuditEntry interface.
func (a *AuditAddProduct) KafkaInfo() KafkaInfo {
	return KafkaInfo{
		Topic: "ua-product",
		Key:   a.ProductID,
	}
}

var _ AuditEntry = (*AuditConfigureContract)(nil)

// AuditConfigureContract represents a call to the ConfigureContract method.
type AuditConfigureContract struct {
	ContractID token.ContractID  `json:"contractID"`
	Resources  []ProductResource `json:"resources"`
}

// KafkaInfo implements the AuditEntry interface.
func (a *AuditConfigureContract) KafkaInfo() KafkaInfo {
	return KafkaInfo{
		Topic: "ua-contract-configure",
		Key:   a.ContractID.String(),
	}
}

// ProductResource represents a resource associated with a product.
type ProductResource struct {
	Type string           `json:"type"`
	Data *json.RawMessage `json:"data,omitempty"`
}

// AuditUpdateProduct represents a call to the UpdateProduct method.
type AuditUpdateProduct struct {
	AuditAddProduct
}
