// Copyright 2019 Canonical Ltd.

// Package kafkaauditor contains a store.Auditor implementation
// that sends audit messages needed to implement KPIs
// to kafka.
package kafkaauditor

import (
	"bytes"
	"context"
	"encoding/json"
	"sync"
	"time"

	"github.com/Shopify/sarama"
	"github.com/juju/zaputil"
	"github.com/juju/zaputil/zapctx"
)

var (
	messageSkipped = (func())(nil)
)

func newClient(producer sarama.AsyncProducer) *client {
	c := &client{producer: producer}
	c.w.Add(2)
	go c.readSuccesses()
	go c.readErrors()
	return c
}

type client struct {
	producer sarama.AsyncProducer
	w        sync.WaitGroup
}

// Close flushes all buffered messages and closes the client.
func (c *client) Close() error {
	// AsyncClose triggers a producer shutdown. We must continue
	// reading from errors and successes channels until they are closed.
	// The readSuccesses and readErrors methods will exit once the channels
	// are closed.
	c.producer.AsyncClose()
	c.w.Wait()
	return nil
}

// Audit implements the store.Auditor interface.
// It sends audit logs to kafka using the async producer.
func (c *client) Audit(ctx context.Context, ts time.Time, entry AuditEntry) {
	ts = ts.UTC().Truncate(time.Second)
	data, err := json.Marshal(entry)
	if err != nil {
		zapctx.Error(ctx, "failed to marshal the audit entry", zaputil.Error(err))
		return
	}
	data = bytes.TrimPrefix(data, []byte("{"))
	payloadData := append(ts.AppendFormat([]byte(`{ "timestamp": "`), time.RFC3339), []byte(`", `)...)
	payloadData = append(payloadData, data...)

	info := entry.KafkaInfo()

	msg := &sarama.ProducerMessage{
		Topic:     info.Topic,
		Key:       sarama.ByteEncoder(info.Key),
		Value:     sarama.ByteEncoder(payloadData),
		Timestamp: ts,
	}
	if msg.Topic == "" {
		if messageSkipped != nil {
			messageSkipped()
		}
		return
	}
	select {
	case c.producer.Input() <- msg:
	case <-time.After(time.Second):
		zapctx.Error(ctx, "failed to send audit message", zaputil.Error(err))
	}
}

func (c *client) readSuccesses() error {
	for {
		msg := <-c.producer.Successes()
		if msg == nil {
			c.w.Done()
			return nil
		}
	}
}

func (c *client) readErrors() error {
	for {
		// TODO (alesstimec) add monitoring of errors
		err := <-c.producer.Errors()
		if err == nil {
			c.w.Done()
			return nil
		}
		zapctx.Error(context.Background(), "failed to send audit message", zaputil.Error(err))
	}
}
