// Copyright 2019 Canonical Ltd.

package kafkaauditor

import (
	"context"
	"encoding/json"
	"time"

	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
)

var _ store.Store = (*auditStore)(nil)

// auditStore logs auditable events performed on the Store.
type auditStore struct {
	Auditor
	store store.Store
}

func (s *auditStore) audit(err *error, ctx context.Context, entry AuditEntry) {
	if *err != nil {
		return
	}
	s.Auditor.Audit(ctx, time.Now().UTC(), entry)
}

// AddContract implements store.Store.AddContract.
func (s *auditStore) AddContract(ctx context.Context, c store.Contract) (_ token.ContractID, err error) {
	defer s.audit(&err, ctx, &AuditAddContract{
		ContractID:    c.ID,
		Name:          c.Name,
		AccountID:     c.AccountID,
		ProductIDs:    c.ProductIDs,
		Created:       c.Created,
		CreatedBy:     c.CreatedBy,
		EffectiveFrom: c.EffectiveFrom,
		EffectiveTo:   c.EffectiveTo,
	})
	return s.store.AddContract(ctx, c)
}

// ConfigureContract implements store.Store.ConfigureContract.
func (s *auditStore) ConfigureContract(ctx context.Context, contractID token.ContractID, resources []store.ProductResource) (err error) {
	defer s.audit(&err, ctx, &AuditConfigureContract{
		ContractID: contractID,
		Resources:  auditProductResources(resources),
	})
	return s.store.ConfigureContract(ctx, contractID, resources)
}

// AddContractMachine implements store.Store.AddContractMachine
func (s *auditStore) AddContractMachine(ctx context.Context, cid token.ContractID, machineID string) (err error) {
	defer s.audit(&err, ctx, &AuditAddContractMachine{
		ContractID: cid,
		MachineID:  machineID,
	})
	return s.store.AddContractMachine(ctx, cid, machineID)
}

// AddContractToken implements store.Store.AddContractToken.
func (s *auditStore) AddContractToken(ctx context.Context, info store.ContractTokenInfo) (err error) {
	defer s.audit(&err, ctx, &AuditAddContractToken{
		ContractToken: info.ContractToken,
		ContractID:    info.ContractID,
		Expires:       info.Expires,
	})
	return s.store.AddContractToken(ctx, info)
}

// AddAccount implements store.Store.CreateAccount.
func (s *auditStore) AddAccount(ctx context.Context, a store.Account, userID string) (err error) {
	defer s.audit(&err, ctx, &AuditAddAccount{
		AccountID:      a.ID,
		AccountName:    a.Name,
		AccountCreated: a.Created,
		UserID:         userID,
	})
	return s.store.AddAccount(ctx, a, userID)
}

// EnsureMachine implements store.Store.EnsureMachine.
func (s *auditStore) EnsureMachine(ctx context.Context, info store.MachineInfo) (err error) {
	defer s.audit(&err, ctx, &AuditEnsureMachine{
		MachineID:    info.MachineID,
		Architecture: info.Architecture,
		OS:           info.OS,
	})
	return s.store.EnsureMachine(ctx, info)
}

// EnsureUser implements store.Store.EnsureUser.
func (s *auditStore) EnsureUser(ctx context.Context, user store.User) (err error) {
	defer s.audit(&err, ctx, &AuditEnsureUser{
		ID:          user.ID,
		Username:    user.Username,
		DisplayName: user.DisplayName,
		Email:       user.Email,
		Verified:    user.Verified,
	})
	return s.store.EnsureUser(ctx, user)
}

// RemoveContractMachine implements store.Store.RemoveContractMachine.
func (s *auditStore) RemoveContractMachine(ctx context.Context, cid token.ContractID, machineID string) (err error) {
	defer s.audit(&err, ctx, &AuditRemoveContractMachine{
		ContractID: cid,
		MachineID:  machineID,
	})
	return s.store.RemoveContractMachine(ctx, cid, machineID)
}

// SetAccountAccess gives the given user the given access level
// on the given account.
func (s *auditStore) SetAccountAccess(ctx context.Context, accountID token.AccountID, userID string, a store.AccountAccess) (err error) {
	defer s.audit(&err, ctx, &AuditSetAccountAccess{
		AccountID:     accountID,
		UserID:        userID,
		AccountAccess: a,
	})
	return s.store.SetAccountAccess(ctx, accountID, userID, a)
}

func (s *auditStore) AddProduct(ctx context.Context, p store.Product) (err error) {
	defer s.audit(&err, ctx, auditAddProduct(p))
	return s.store.AddProduct(ctx, p)
}

func (s *auditStore) UpdateProduct(ctx context.Context, p store.Product) (err error) {
	entry := AuditUpdateProduct{*auditAddProduct(p)}
	defer s.audit(&err, ctx, &entry)
	return s.store.UpdateProduct(ctx, p)
}

func (s *auditStore) UpdateContractProducts(ctx context.Context, contractID token.ContractID, productIDs []string) error {
	// TODO(fabricematrat) audit update contract's products.
	return s.store.UpdateContractProducts(ctx, contractID, productIDs)
}

func (s *auditStore) EnsureServiceAccount(ctx context.Context, account store.ServiceAccount) error {
	// TODO(mhilton) audit service accounts.
	return s.store.EnsureServiceAccount(ctx, account)
}

func (s *auditStore) Account(ctx context.Context, id token.AccountID) (store.Account, error) {
	return s.store.Account(ctx, id)
}

func (s *auditStore) AccountForSalesforceID(ctx context.Context, id string) (store.Account, error) {
	return s.store.AccountForSalesforceID(ctx, id)
}

func (s *auditStore) AccountAccess(ctx context.Context, accountID token.AccountID, userID string) (store.AccountAccess, error) {
	return s.store.AccountAccess(ctx, accountID, userID)
}

func (s *auditStore) AccountContracts(ctx context.Context, accountID token.AccountID) ([]store.FullContract, error) {
	return s.store.AccountContracts(ctx, accountID)
}

func (s *auditStore) AllProducts(ctx context.Context) ([]store.Product, error) {
	return s.store.AllProducts(ctx)
}

func (s *auditStore) Contract(ctx context.Context, id token.ContractID) (store.FullContract, error) {
	return s.store.Contract(ctx, id)
}

func (s *auditStore) ContractToken(ctx context.Context, ct token.ContractToken) (store.ContractTokenInfo, error) {
	return s.store.ContractToken(ctx, ct)
}

func (s *auditStore) ContractTokens(ctx context.Context, cid token.ContractID) ([]store.ContractTokenInfo, error) {
	return s.store.ContractTokens(ctx, cid)
}

func (s *auditStore) MachineHasContract(ctx context.Context, cid token.ContractID, machineID string) (bool, error) {
	return s.store.MachineHasContract(ctx, cid, machineID)
}

func (s *auditStore) MachineInfo(ctx context.Context, machineID string) (store.MachineInfo, error) {
	return s.store.MachineInfo(ctx, machineID)
}

func (s *auditStore) User(ctx context.Context, id string) (store.User, error) {
	return s.store.User(ctx, id)
}

func (s *auditStore) UserAccounts(ctx context.Context, id string) ([]store.AccountWithAccess, error) {
	return s.store.UserAccounts(ctx, id)
}

func (s *auditStore) Product(ctx context.Context, productID string) (store.Product, error) {
	return s.store.Product(ctx, productID)
}

// ServiceAccount implements store.Store.ServiceAccount.
func (s *auditStore) ServiceAccount(ctx context.Context, name string) (store.ServiceAccount, error) {
	return s.store.ServiceAccount(ctx, name)
}

func auditAddProduct(p store.Product) *AuditAddProduct {
	return &AuditAddProduct{
		ProductID:   p.ID,
		ProductName: p.Name,
		Resources:   auditProductResources(p.Resources),
	}
}

func auditProductResources(resources []store.ProductResource) []ProductResource {
	prs := make([]ProductResource, len(resources))
	for i, pr := range resources {
		data := json.RawMessage(pr.Data)
		prs[i] = ProductResource{
			Type: pr.ResourceType,
			Data: &data,
		}
	}
	return prs
}
