// Copyright 2019 Canonical Ltd.

package kafkaauditor_test

import (
	"context"
	"encoding/json"
	"testing"
	"time"

	qt "github.com/frankban/quicktest"
	"github.com/juju/zaputil/zapctx"
	"go.uber.org/zap/zaptest"

	"github.com/CanonicalLtd/ua-contracts/internal/kafkaauditor"
	"github.com/CanonicalLtd/ua-contracts/internal/memstore"
	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/storetest"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
	"github.com/CanonicalLtd/ua-contracts/params"
)

func TestAuditStore(t *testing.T) {
	storetest.TestStore(qt.New(t), func(c *qt.C) store.Store {
		return newFixture(memstore.New()).store
	})
}

func TestAddContract(t *testing.T) {
	c := qt.New(t)
	defer c.Done()
	ctx := testContext(c)

	st := memstore.New()
	f := newFixture(st)

	now := time.Now().UTC()

	account := store.Account{
		ID:      token.MustNewAccountID(),
		Name:    "test-account",
		Created: now,
	}
	err := f.store.AddAccount(ctx, account, "test-user")
	c.Assert(err, qt.Equals, nil)
	c.Assert(f.auditor.entries, qt.DeepEquals, []kafkaauditor.AuditEntry{
		&kafkaauditor.AuditAddAccount{
			AccountID:      account.ID,
			AccountName:    account.Name,
			AccountCreated: now,
			UserID:         "test-user",
		},
	})

	f = newFixture(st)
	product := store.Product{
		ID:   "test-product-id",
		Name: "namey name name",
		Resources: []store.ProductResource{{
			ResourceType: "fips",
			Data:         `{"a": 5}`,
		}},
	}
	err = f.store.AddProduct(ctx, product)
	c.Assert(err, qt.Equals, nil)
	data := json.RawMessage(`{"a": 5}`)
	c.Assert(f.auditor.entries, qt.DeepEquals, []kafkaauditor.AuditEntry{
		&kafkaauditor.AuditAddProduct{
			ProductID:   product.ID,
			ProductName: product.Name,
			Resources: []kafkaauditor.ProductResource{{
				Type: "fips",
				Data: &data,
			}},
		},
	})

	f = newFixture(st)
	contract := store.Contract{
		ID:            token.MustNewContractID(),
		Name:          "test-contract",
		AccountID:     account.ID,
		ProductIDs:    []string{"test-product-id"},
		Created:       now,
		CreatedBy:     "test-user",
		EffectiveFrom: &now,
		EffectiveTo:   &now,
	}
	_, err = f.store.AddContract(ctx, contract)
	c.Assert(err, qt.Equals, nil)

	c.Assert(f.auditor.entries, qt.DeepEquals, []kafkaauditor.AuditEntry{
		&kafkaauditor.AuditAddContract{
			ContractID:    contract.ID,
			Name:          contract.Name,
			AccountID:     contract.AccountID,
			ProductIDs:    contract.ProductIDs,
			Created:       contract.Created,
			CreatedBy:     contract.CreatedBy,
			EffectiveFrom: contract.EffectiveFrom,
			EffectiveTo:   contract.EffectiveTo,
		},
	})
}

func TestEnsureMachine(t *testing.T) {
	c := qt.New(t)
	defer c.Done()
	ctx := testContext(c)

	st := memstore.New()
	f := newFixture(st)
	m := store.MachineInfo{
		MachineID:    "m1",
		Architecture: "amd64",
		OS: params.OS{
			Distribution: "Ubuntu 16.04.3 LTS",
			Kernel:       "4.4.0-142-generic",
			Release:      "16.04",
			Series:       "xenial",
			Type:         "linux",
		},
	}
	err := f.store.EnsureMachine(ctx, m)
	c.Assert(err, qt.Equals, nil)
	c.Assert(f.auditor.entries, qt.DeepEquals, []kafkaauditor.AuditEntry{
		&kafkaauditor.AuditEnsureMachine{
			MachineID:    "m1",
			Architecture: "amd64",
			OS: params.OS{
				Distribution: "Ubuntu 16.04.3 LTS",
				Kernel:       "4.4.0-142-generic",
				Release:      "16.04",
				Series:       "xenial",
				Type:         "linux",
			},
		},
	})
	m1, err := f.store.MachineInfo(ctx, "m1")
	c.Assert(err, qt.Equals, nil)
	c.Assert(m1, qt.DeepEquals, m)
}

type fixture struct {
	store   store.Store
	auditor *testAuditor
}

// newFixture creates a fixture that holds the original permanent store, while
// creating a new store from the given parameter.
func newFixture(store store.Store) *fixture {
	auditor := &testAuditor{}

	return &fixture{
		store: kafkaauditor.New(
			store,
			auditor,
		),
		auditor: auditor,
	}
}

func testContext(c *qt.C) context.Context {
	return zapctx.WithLogger(context.Background(), zaptest.NewLogger(c))
}

type testAuditor struct {
	entries []kafkaauditor.AuditEntry
}

func (t *testAuditor) Audit(ctx context.Context, _ time.Time, entry kafkaauditor.AuditEntry) {
	t.entries = append(t.entries, entry)
}

func (t *testAuditor) clear() {
	t.entries = []kafkaauditor.AuditEntry{}
}
