package memstore

import (
	"context"
	"encoding/json"
	"fmt"
	"sort"
	"sync"
	"time"

	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
)

type auditEntry struct {
	ts   time.Time
	data string
}

type memStore struct {
	// mu protects the individual fields of the store.
	mu                 sync.Mutex
	maxContractID      int
	accountAccess      map[string]map[token.AccountID]store.AccountAccess
	accounts           map[token.AccountID]store.Account
	contractMachines   map[token.ContractID][]string
	contracts          map[token.ContractID]store.Contract
	contractTokens     map[token.ContractToken]store.ContractTokenInfo
	machines           map[string]*store.MachineInfo
	contractConfig     map[token.ContractID][]store.ProductResource
	products           map[string]store.Product
	salesforceAccounts map[string]store.Account
	serviceAccounts    map[string]store.ServiceAccount
	users              map[string]store.User
}

// New returns a new in-memory store instance.
func New() store.Store {
	return &memStore{
		accountAccess:      make(map[string]map[token.AccountID]store.AccountAccess),
		accounts:           make(map[token.AccountID]store.Account),
		contractMachines:   make(map[token.ContractID][]string),
		contracts:          make(map[token.ContractID]store.Contract),
		contractTokens:     make(map[token.ContractToken]store.ContractTokenInfo),
		machines:           make(map[string]*store.MachineInfo),
		contractConfig:     make(map[token.ContractID][]store.ProductResource),
		products:           make(map[string]store.Product),
		salesforceAccounts: make(map[string]store.Account),
		serviceAccounts:    make(map[string]store.ServiceAccount),
		users:              make(map[string]store.User),
	}
}

// AddProduct implements store.Store.AddProduct.
func (s *memStore) AddProduct(ctx context.Context, p store.Product) error {
	s.mu.Lock()
	defer s.mu.Unlock()
	if _, ok := s.products[p.ID]; ok {
		return errgo.WithCausef(nil, store.ErrDuplicateProduct, "")
	}
	return s.addProduct(ctx, p)
}

func (s *memStore) UpdateProduct(ctx context.Context, p store.Product) error {
	s.mu.Lock()
	defer s.mu.Unlock()
	if _, ok := s.products[p.ID]; !ok {
		return errgo.WithCausef(nil, store.ErrProductNotFound, "")
	}
	return s.addProduct(ctx, p)
}

func (s *memStore) addProduct(ctx context.Context, p store.Product) error {
	if len(p.Resources) == 0 {
		return errgo.Newf("product %q added with no resources", p.ID)
	}
	if err := checkProductResources(p.Resources); err != nil {
		return errgo.Mask(err, errgo.Is(store.ErrProductConflict))
	}
	p.Resources = append([]store.ProductResource(nil), p.Resources...)
	s.products[p.ID] = p
	return nil
}

func (s *memStore) ConfigureContract(ctx context.Context, contractID token.ContractID, resources []store.ProductResource) error {
	s.mu.Lock()
	defer s.mu.Unlock()
	if err := checkProductResources(resources); err != nil {
		return errgo.Mask(err, errgo.Is(store.ErrProductConflict))
	}
	_, ok := s.contracts[contractID]
	if !ok {
		return errgo.WithCausef(nil, store.ErrContractNotFound, "")
	}
	if len(resources) == 0 {
		delete(s.contractConfig, contractID)
	} else {
		s.contractConfig[contractID] = append([]store.ProductResource(nil), resources...)
	}
	return nil
}

// Product implements store.Store.Product.
func (s *memStore) Product(ctx context.Context, productID string) (store.Product, error) {
	s.mu.Lock()
	defer s.mu.Unlock()
	p, ok := s.products[productID]
	if !ok {
		return store.Product{}, errgo.WithCausef(nil, store.ErrProductNotFound, "")
	}
	return copyProduct(p), nil
}

// AllProducts implements store.Store.AllProducts.
func (s *memStore) AllProducts(ctx context.Context) ([]store.Product, error) {
	s.mu.Lock()
	defer s.mu.Unlock()
	prods := make([]store.Product, 0, len(s.products))
	for _, prod := range s.products {
		prods = append(prods, copyProduct(prod))
	}
	sort.Slice(prods, func(i, j int) bool {
		return prods[i].ID < prods[j].ID
	})
	return prods, nil
}

// AddContract implements store.Store.AddContract.
func (s *memStore) AddContract(ctx context.Context, c store.Contract) (token.ContractID, error) {
	s.mu.Lock()
	defer s.mu.Unlock()
	c.ID = token.MustNewContractID()
	if _, ok := s.accounts[c.AccountID]; !ok {
		return token.ContractID{}, errgo.Newf("account for contract does not exist")
	}
	for _, p := range c.ProductIDs {
		if _, ok := s.products[p]; !ok {
			return token.ContractID{}, errgo.Newf("contract has invalid product %q", p)
		}
	}
	for _, sc := range s.contracts {
		if sc.EffectivelyEqual(c) {
			return token.ContractID{}, errgo.WithCausef(nil, store.ErrContractExists, "")
		}
	}
	s.contracts[c.ID] = c
	return c.ID, nil
}

// AddContractToken implements store.Store.AddContractToken.
func (s *memStore) AddContractToken(ctx context.Context, info store.ContractTokenInfo) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if _, ok := s.contractTokens[info.ContractToken]; ok {
		panic(fmt.Errorf("duplicated contract token %q (borked random number generation?)", info.ContractToken))
	}
	if info.Expires.IsZero() {
		panic("zero expiry time for contract token")
	}
	s.contractTokens[info.ContractToken] = info
	return nil
}

// Contract implements store.Store.Contract.
func (s *memStore) Contract(ctx context.Context, id token.ContractID) (store.FullContract, error) {
	s.mu.Lock()
	defer s.mu.Unlock()

	c, ok := s.contracts[id]
	if !ok {
		return store.FullContract{}, errgo.WithCausef(nil, store.ErrContractNotFound, "")
	}
	return store.FullContract{
		Contract: c,
		Account:  s.accounts[c.AccountID],
		Config:   s.contractConfig[id],
	}, nil
}

// ContractToken implements store.Store.ContractToken.
func (s *memStore) ContractToken(ctx context.Context, ct token.ContractToken) (store.ContractTokenInfo, error) {
	s.mu.Lock()
	defer s.mu.Unlock()

	info, ok := s.contractTokens[ct]
	if !ok {
		return store.ContractTokenInfo{}, errgo.WithCausef(nil, store.ErrTokenNotFound, "")
	}
	return info, nil
}

// ContractTokens implements store.Store.ContractTokens.
func (s *memStore) ContractTokens(ctx context.Context, id token.ContractID) ([]store.ContractTokenInfo, error) {
	s.mu.Lock()
	defer s.mu.Unlock()

	var tokens []store.ContractTokenInfo
	for _, v := range s.contractTokens {
		if v.ContractID == id {
			tokens = append(tokens, v)
		}
	}

	return tokens, nil
}

// EnsureMachine implements store.Store.EnsureMachine.
func (s *memStore) EnsureMachine(ctx context.Context, info store.MachineInfo) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	m, ok := s.machines[info.MachineID]
	if ok {
		*m = info
	} else {
		s.machines[info.MachineID] = &info
	}
	return nil
}

// MachineInfo implements store.Store.MachineInfo.
func (s *memStore) MachineInfo(ctx context.Context, machineID string) (store.MachineInfo, error) {
	s.mu.Lock()
	defer s.mu.Unlock()
	m, ok := s.machines[machineID]
	if !ok {
		return store.MachineInfo{}, errgo.WithCausef(nil, store.ErrMachineNotFound, "")
	}
	return *m, nil
}

// AddContractMachine implements store.Store.AddContractMachine.
func (s *memStore) AddContractMachine(ctx context.Context, cid token.ContractID, machineID string) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	var found bool
	for _, m := range s.contractMachines[cid] {
		if m == machineID {
			found = true
			break
		}
	}
	if !found {
		s.contractMachines[cid] = append(s.contractMachines[cid], machineID)
	}
	return nil
}

// RemoveContractMachine implements store.Store.RemoveContractMachine.
func (s *memStore) RemoveContractMachine(ctx context.Context, cid token.ContractID, machineID string) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	var cm []string
	for _, m := range s.contractMachines[cid] {
		if m == machineID {
			continue
		}
		cm = append(cm, m)
	}
	s.contractMachines[cid] = cm
	return nil
}

// MachineHasContract implements store.Store.MachineHasContract.
func (s *memStore) MachineHasContract(ctx context.Context, cid token.ContractID, machineID string) (bool, error) {
	s.mu.Lock()
	defer s.mu.Unlock()

	var found bool
	for _, m := range s.contractMachines[cid] {
		if m == machineID {
			found = true
			break
		}
	}
	return found, nil
}

// EnsureUser implements store.Store.EnsureUser.
func (s *memStore) EnsureUser(ctx context.Context, u store.User) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	u1, ok := s.users[u.ID]
	if ok && u == u1 {
		return nil
	}
	s.users[u.ID] = u
	return nil
}

// User implements store.Store.User.
func (s *memStore) User(ctx context.Context, id string) (store.User, error) {
	s.mu.Lock()
	defer s.mu.Unlock()
	u, ok := s.users[id]
	if !ok {
		return store.User{}, errgo.WithCausef(nil, store.ErrUserNotFound, "cannot find user %q", id)
	}
	return u, nil
}

// Account implements store.Store.Account.
func (s *memStore) Account(ctx context.Context, id token.AccountID) (store.Account, error) {
	s.mu.Lock()
	defer s.mu.Unlock()

	a, ok := s.accounts[id]
	if !ok {
		return a, errgo.WithCausef(nil, store.ErrAccountNotFound, "account %q not found", id)
	}
	return a, nil
}

// AccountForSalesforceID implements store.Store.AccountForSalesforceID.
func (s *memStore) AccountForSalesforceID(ctx context.Context, id string) (store.Account, error) {
	s.mu.Lock()
	defer s.mu.Unlock()

	a, ok := s.salesforceAccounts[id]
	if !ok {
		return a, errgo.WithCausef(nil, store.ErrAccountNotFound, "account for %q not found", id)
	}
	return a, nil
}

// UserAccounts implements store.Store.UserAccounts.
func (s *memStore) UserAccounts(ctx context.Context, id string) ([]store.AccountWithAccess, error) {
	s.mu.Lock()
	defer s.mu.Unlock()

	var accounts []store.AccountWithAccess
	for k, v := range s.accountAccess[id] {
		accounts = append(accounts, store.AccountWithAccess{
			Account: s.accounts[k],
			Access:  v,
		})
	}
	return accounts, nil
}

// AddAccount implements store.Store.AddAccount.
func (s *memStore) AddAccount(ctx context.Context, a store.Account, userID string) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if _, ok := s.accounts[a.ID]; ok {
		return errgo.WithCausef(nil, store.ErrAccountExists, "account %q already exists", a.ID)
	}
	if a.SalesforceAccountID != "" {
		if _, ok := s.salesforceAccounts[a.SalesforceAccountID]; ok {
			return errgo.WithCausef(nil, store.ErrDuplicateSalesforceAccount, "account for %q already exists", a.SalesforceAccountID)
		}
		s.salesforceAccounts[a.SalesforceAccountID] = a
	}
	s.accounts[a.ID] = a
	userAccess, ok := s.accountAccess[userID]
	if !ok {
		s.accountAccess[userID] = make(map[token.AccountID]store.AccountAccess)
		userAccess = s.accountAccess[userID]
	}
	userAccess[a.ID] = store.AdminAccountAccess
	return nil
}

// AccountAccess implements store.Store.AccountAccess.
func (s *memStore) AccountAccess(ctx context.Context, accountID token.AccountID, userID string) (store.AccountAccess, error) {
	s.mu.Lock()
	defer s.mu.Unlock()

	a, _ := s.accountAccess[userID][accountID]
	return a, nil
}

// SetAccountAccess implements store.Store.SetAccountAccess.
func (s *memStore) SetAccountAccess(ctx context.Context, accountID token.AccountID, userID string, a store.AccountAccess) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	userAccess, ok := s.accountAccess[userID]
	if !ok {
		s.accountAccess[userID] = make(map[token.AccountID]store.AccountAccess)
		userAccess = s.accountAccess[userID]
	}
	userAccess[accountID] = a
	return nil
}

// AccountContracts implements store.Store.AccountContracts.
func (s *memStore) AccountContracts(_ context.Context, accountID token.AccountID) ([]store.FullContract, error) {
	s.mu.Lock()
	defer s.mu.Unlock()

	var contracts []store.FullContract
	for _, contract := range s.contracts {
		if contract.AccountID != accountID {
			continue
		}
		contracts = append(contracts, store.FullContract{
			Contract: contract,
			Account:  s.accounts[contract.AccountID],
		})
	}

	return contracts, nil
}

func isValidResourceType(rt string) bool {
	for _, t := range store.ResourceTypes {
		if t == rt {
			return true
		}
	}
	return false
}

// EnsureServiceAccount implements store.Store.EnsureServiceAccount.
func (s *memStore) EnsureServiceAccount(ctx context.Context, account store.ServiceAccount) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.serviceAccounts[account.Name] = account
	return nil
}

// ServiceAccount implements store.Store.ServiceAccount.
func (s *memStore) ServiceAccount(ctx context.Context, name string) (store.ServiceAccount, error) {
	s.mu.Lock()
	defer s.mu.Unlock()

	account, ok := s.serviceAccounts[name]
	if !ok {
		return account, store.ErrServiceAccountNotFound
	}
	return account, nil
}

// UpdateContractProducts implements store.Store.UpdateContractProducts.
func (s *memStore) UpdateContractProducts(ctx context.Context, contractID token.ContractID, productIDs []string) error {
	s.mu.Lock()
	defer s.mu.Unlock()
	contract, ok := s.contracts[contractID]
	if !ok {
		return errgo.WithCausef(nil, store.ErrContractNotFound, "")
	}
	contract.ProductIDs = productIDs
	s.contracts[contractID] = contract
	return nil
}

func copyProduct(p store.Product) store.Product {
	p1 := p
	p1.Resources = append([]store.ProductResource(nil), p.Resources...)
	return p1
}

func checkProductResources(rs []store.ProductResource) error {
	found := make(map[string]bool)
	for _, r := range rs {
		if !isValidResourceType(r.ResourceType) {
			return errgo.Newf("invalid resource type %q", r.ResourceType)
		}
		if found[r.ResourceType] {
			return errgo.WithCausef(nil, store.ErrProductConflict, "")
		}
		found[r.ResourceType] = true
		var x map[string]interface{}
		if err := json.Unmarshal([]byte(r.Data), &x); err != nil {
			return errgo.Notef(err, "bad resource data %q", r.Data)
		}
	}
	return nil
}
