// Copyright 2019 Canonical Ltd.

package monitoring

import (
	"net/http"

	"github.com/julienschmidt/httprouter"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	"gopkg.in/httprequest.v1"
)

var (
	RequestDuration = prometheus.NewSummaryVec(prometheus.SummaryOpts{
		Namespace: "http",
		Subsystem: "request",
		Name:      "duration_seconds",
		Help:      "The duration of API HTTP requests.",
	}, []string{"method", "handler", "code"})

	SalesforceErrorCount = prometheus.NewCounter(prometheus.CounterOpts{
		Namespace: "salesforce",
		Name:      "error_count",
		Help:      "A counter for errors when interacting with Salesforce.",
	})
)

func init() {
	prometheus.MustRegister(
		RequestDuration,
		SalesforceErrorCount,
	)
}

// AddHandlers adds all the handlers in the given slice to the given *httprouter.Router,
// wrapping the handlers in Prometheus metrics middleware to track request details,
// durations, and status.
func AddHandlers(r *httprouter.Router, hs []httprequest.Handler) {
	for _, h := range hs {
		AddHTTPHandler(r, h.Method, h.Path, httprequest.ToHTTP(h.Handle))
	}
}

// AddHTTPHandler adds the handler with the given method and path to the
// given *httprouter.Router, wrapping the handler in a Prometheus metrics
// middleware to track request details, durations, and status.
func AddHTTPHandler(r *httprouter.Router, method, path string, hnd http.Handler) {
	h := promhttp.InstrumentHandlerDuration(
		RequestDuration.MustCurryWith(
			prometheus.Labels{"handler": path},
		),
		hnd,
	)
	r.HandlerFunc(method, path, h)
}

// AddMetricsHandler adds a prometheus metrics handler to the given router on the given
// path (usually /metrics).
func AddMetricsHandler(r *httprouter.Router, path string) {
	r.Handler("GET", path, promhttp.Handler())
}
