package postgresstore

import (
	"context"
	"database/sql"

	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
)

// Note: it is expected that there should be exactly two methods defined
// in this file per prepared SQL statement; the prepare method should
// be adjacent to the method that uses the statement so that it's
// easy to see the correspondence between positional parameters
// and their occurrence in the actual SQL statement.

//go:generate stringer -trimprefix stmt -type stmtID

// stmtID identifies one of the prepared SQL statements.
type stmtID int

const (
	_ stmtID = iota - 1
	stmtAddAccount
	stmtAddAccountAccess
	stmtAddContract
	stmtAddContractConfig
	stmtAddContractMachine
	stmtAddContractProduct
	stmtAddContractToken
	stmtAddProduct
	stmtAddProductResource
	stmtDeleteProductResources
	stmtEnsureServiceAccount
	stmtEnsureUser
	stmtGetAccount
	stmtGetAccountAccess
	stmtGetAccountContracts
	stmtGetAccountWithSalesforceID
	stmtGetAllProducts
	stmtGetContract
	stmtGetContractConfig
	stmtGetContractProducts
	stmtGetContractToken
	stmtGetContractTokens
	stmtGetMachine
	stmtGetMachineHasContract
	stmtGetMachineType
	stmtGetProduct
	stmtGetProductResources
	stmtGetServiceAccount
	stmtGetUser
	stmtGetUserAccounts
	stmtInsertMachineType
	stmtRemoveContractMachine
	stmtRemoveContractConfig
	stmtRemoveContractProducts
	stmtSetAccountAccess
	stmtUpdateProduct
	stmtUpsertMachine
	numStmts
)

var prepareMethods = []func(*pgStore) error{
	(*pgStore).prepareAddAccount,
	(*pgStore).prepareAddAccountAccess,
	(*pgStore).prepareAddContract,
	(*pgStore).prepareAddContractConfig,
	(*pgStore).prepareAddContractMachine,
	(*pgStore).prepareAddContractProduct,
	(*pgStore).prepareAddContractToken,
	(*pgStore).prepareAddProduct,
	(*pgStore).prepareAddProductResource,
	(*pgStore).prepareDeleteProductResources,
	(*pgStore).prepareEnsureServiceAccount,
	(*pgStore).prepareEnsureUser,
	(*pgStore).prepareGetAccount,
	(*pgStore).prepareGetAccountAccess,
	(*pgStore).prepareGetAccountContracts,
	(*pgStore).prepareGetAccountWithSalesforceID,
	(*pgStore).prepareGetAllProducts,
	(*pgStore).prepareGetContract,
	(*pgStore).prepareGetContractConfig,
	(*pgStore).prepareGetContractProducts,
	(*pgStore).prepareGetContractToken,
	(*pgStore).prepareGetContractTokens,
	(*pgStore).prepareGetMachine,
	(*pgStore).prepareGetMachineHasContract,
	(*pgStore).prepareGetMachineType,
	(*pgStore).prepareGetProduct,
	(*pgStore).prepareGetProductResources,
	(*pgStore).prepareGetServiceAccount,
	(*pgStore).prepareGetUser,
	(*pgStore).prepareGetUserAccounts,
	(*pgStore).prepareInsertMachineType,
	(*pgStore).prepareRemoveContractMachine,
	(*pgStore).prepareRemoveContractConfig,
	(*pgStore).prepareRemoveContractProducts,
	(*pgStore).prepareSetAccountAccess,
	(*pgStore).prepareUpdateProduct,
	(*pgStore).prepareUpsertMachine,
}

func (st *pgStore) prepareAddAccount() error {
	return st.prepare(stmtAddAccount, `
		INSERT INTO accounts (id, name, created, salesforce_account_id) VALUES ($1, $2, $3, $4)
`)
}

func (st *pgStore) addAccount(ctx context.Context, a store.Account) error {
	sfid := sql.NullString{
		String: a.SalesforceAccountID,
		Valid:  a.SalesforceAccountID != "",
	}
	_, err := st.stmt(stmtAddAccount).ExecContext(ctx, a.ID[:], a.Name, a.Created, sfid)
	return errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareGetAccount() error {
	return st.prepare(stmtGetAccount, `
		SELECT id, name, created, salesforce_account_id FROM accounts
		WHERE id = $1
`)
}

func (st *pgStore) getAccount(ctx context.Context, id token.AccountID) (store.Account, error) {
	var a store.Account
	var sfid sql.NullString
	err := st.stmt(stmtGetAccount).QueryRowContext(ctx, id[:]).Scan(
		&a.ID,
		&a.Name,
		&a.Created,
		&sfid,
	)
	a.SalesforceAccountID = sfid.String
	return a, errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareGetAccountWithSalesforceID() error {
	return st.prepare(stmtGetAccountWithSalesforceID, `
		SELECT id, name, created, salesforce_account_id FROM accounts
		WHERE salesforce_account_id = $1
`)
}

func (st *pgStore) getAccountWithSalesforceID(ctx context.Context, id string) (store.Account, error) {
	var a store.Account
	var sfid sql.NullString
	err := st.stmt(stmtGetAccountWithSalesforceID).QueryRowContext(ctx, id[:]).Scan(
		&a.ID,
		&a.Name,
		&a.Created,
		&sfid,
	)
	a.SalesforceAccountID = sfid.String
	return a, errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareAddAccountAccess() error {
	return st.prepare(stmtAddAccountAccess, `
	INSERT INTO account_access (
		account_id,
		user_id,
		access_level
	) VALUES ($1, $2, $3)
`)
}

func (st *pgStore) addAccountAccess(ctx context.Context, accountID token.AccountID, userID string, accessLevel store.AccountAccess) error {
	_, err := st.stmt(stmtAddAccountAccess).ExecContext(ctx, accountID[:], userID, accessLevel)
	return errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareGetAccountAccess() error {
	return st.prepare(stmtGetAccountAccess, `
	SELECT access_level FROM account_access 
	WHERE account_id = $1 AND user_id = $2
`)
}

func (st *pgStore) getAccountAccess(ctx context.Context, accountID token.AccountID, userID string) (store.AccountAccess, error) {
	var accessLevel int
	err := st.stmt(stmtGetAccountAccess).QueryRowContext(ctx, accountID[:], userID).Scan(
		&accessLevel,
	)
	return store.AccountAccess(accessLevel), errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareGetAccountContracts() error {
	return st.prepare(stmtGetAccountContracts, `
	SELECT
		contracts.id,
		contracts.name,
		contracts.account_id,
		contracts.created,
		contracts.created_by,
		contracts.effective_from,
		contracts.effective_to,
		accounts.name,
		accounts.created
	FROM contracts
	INNER JOIN accounts ON
		contracts.account_id=accounts.id
	WHERE contracts.account_id = $1
`)
}

func (st *pgStore) getAccountContracts(ctx context.Context, accountID token.AccountID) ([]store.FullContract, error) {
	rows, err := st.stmt(stmtGetAccountContracts).QueryContext(ctx, accountID[:])
	if err != nil {
		return nil, errgo.Mask(err, isDBError)
	}
	defer rows.Close()
	var contracts []store.FullContract
	for rows.Next() {
		var c store.FullContract
		if err := rows.Scan(
			&c.Contract.ID,
			&c.Contract.Name,
			&c.Contract.AccountID,
			&c.Contract.Created,
			&c.Contract.CreatedBy,
			&c.Contract.EffectiveFrom,
			&c.Contract.EffectiveTo,
			&c.Account.Name,
			&c.Account.Created,
		); err != nil {
			return nil, errgo.Mask(err)
		}
		c.Account.ID = c.Contract.AccountID
		contracts = append(contracts, c)
	}
	if err := rows.Err(); err != nil {
		return nil, errgo.Mask(err, isDBError)
	}
	return contracts, nil
}

func (st *pgStore) prepareSetAccountAccess() error {
	return st.prepare(stmtSetAccountAccess, `
	INSERT INTO account_access (account_id, user_id, access_level) VALUES ($1, $2, $3)
	ON CONFLICT (account_id, user_id)
	DO UPDATE SET access_level = $3
`)
}

func (st *pgStore) setAccountAccess(ctx context.Context, accountID token.AccountID, userID string, level store.AccountAccess) error {
	_, err := st.stmt(stmtSetAccountAccess).ExecContext(ctx, accountID[:], userID, level)
	return errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareEnsureUser() error {
	return st.prepare(stmtEnsureUser, `
	INSERT INTO users (id, name, display_name, email, verified)
	VALUEs ($1, $2, $3, $4, $5)
	ON CONFLICT (id)
	DO UPDATE SET
		name=$2,
		display_name=$3,
		email=$4,
		verified=$5
`)
}

func (st *pgStore) ensureUser(ctx context.Context, u store.User) error {
	_, err := st.stmt(stmtEnsureUser).ExecContext(ctx, u.ID, u.Username, u.DisplayName, u.Email, u.Verified)
	return errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareGetUser() error {
	return st.prepare(stmtGetUser, `
	SELECT id, name, display_name, email, verified
	FROM users
	WHERE id=$1
`)
}

func (st *pgStore) getUser(ctx context.Context, id string) (store.User, error) {
	var u store.User
	err := st.stmt(stmtGetUser).QueryRowContext(ctx, id).Scan(
		&u.ID,
		&u.Username,
		&u.DisplayName,
		&u.Email,
		&u.Verified,
	)
	if err != nil {
		return store.User{}, errgo.Mask(err, isDBError)
	}
	return u, nil
}

func (st *pgStore) prepareGetUserAccounts() error {
	return st.prepare(stmtGetUserAccounts, `
	SELECT accounts.id, accounts.name, accounts.created, account_access.access_level
	FROM account_access
	LEFT JOIN accounts
	ON account_access.account_id = accounts.id
	WHERE account_access.user_id = $1
`)
}

func (st *pgStore) getUserAccounts(ctx context.Context, userID string) ([]store.AccountWithAccess, error) {
	rows, err := st.stmt(stmtGetUserAccounts).QueryContext(ctx, userID)
	if err != nil {
		return nil, errgo.Mask(err, isDBError)
	}
	defer rows.Close()
	var accounts []store.AccountWithAccess
	for rows.Next() {
		var c store.AccountWithAccess
		if err := rows.Scan(
			&c.ID,
			&c.Name,
			&c.Created,
			&c.Access,
		); err != nil {
			return nil, errgo.Mask(err)
		}
		accounts = append(accounts, c)
	}
	if err := rows.Err(); err != nil {
		return nil, errgo.Mask(err, isDBError)
	}
	return accounts, nil
}

func (st *pgStore) prepareAddContract() error {
	return st.prepare(stmtAddContract, `
	INSERT INTO contracts (
		id,
		name,
		account_id,
		created,
		created_by,
		effective_from,
		effective_to
	) VALUES ($1, $2, $3, $4, $5, $6, $7)
`)
}

func (st *pgStore) addContract(ctx context.Context, c store.Contract) error {
	_, err := st.stmt(stmtAddContract).ExecContext(ctx,
		c.ID[:],
		c.Name,
		c.AccountID[:],
		c.Created,
		c.CreatedBy,
		c.EffectiveFrom,
		c.EffectiveTo,
	)
	return errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareAddContractProduct() error {
	return st.prepare(stmtAddContractProduct, `
	INSERT INTO contract_products (
		contract_id,
		product_id
	) VALUES ($1, $2)
`)
}

func (st *pgStore) addContractProduct(ctx context.Context, cid token.ContractID, pid string) error {
	_, err := st.stmt(stmtAddContractProduct).ExecContext(ctx,
		cid[:],
		pid,
	)
	return errgo.Mask(err, isDBError)
}

// addContractProducts adds multiple products to a contract.
// The addProducts function is called to add the products;
// it is passed a function that can add an individual product.
func (st *pgStore) addContractProducts(ctx context.Context, addProducts func(addProduct func(contractID token.ContractID, productID string) error) error) error {
	addProducts1 := func(addRow func(...interface{}) error) error {
		return addProducts(func(contractID token.ContractID, productID string) error {
			return errgo.Mask(addRow(contractID[:], productID), isDBError)
		})
	}
	err := st.bulkInsert(ctx, addProducts1, "contract_products",
		"contract_id",
		"product_id",
	)
	return errgo.Mask(err, errgo.Any)
}

func (st *pgStore) prepareAddContractConfig() error {
	return st.prepare(stmtAddContractConfig, `
		INSERT INTO contract_config (contract_id, resource_type, data)
		VALUES ($1, $2, $3)
`)
}

func (st *pgStore) addContractConfig(ctx context.Context, cid token.ContractID, r store.ProductResource) error {
	_, err := st.stmt(stmtAddContractConfig).ExecContext(ctx,
		cid[:],
		r.ResourceType,
		r.Data,
	)
	return errgo.Mask(err, isDBError)
}

// addContractConfigResources adds multiple resources to a contract configuration.
// The addResources function is called to add the products;
// it is passed a function that can add an individual resource.
func (st *pgStore) addContractConfigResources(ctx context.Context, addResources func(addResource func(cid token.ContractID, r store.ProductResource) error) error) error {
	addResources1 := func(addRow func(...interface{}) error) error {
		return addResources(func(cid token.ContractID, r store.ProductResource) error {
			return errgo.Mask(addRow(cid[:], r.ResourceType, r.Data), isDBError)
		})
	}
	err := st.bulkInsert(ctx, addResources1, "contract_config",
		"contract_id",
		"resource_type",
		"data",
	)
	return errgo.Mask(err, errgo.Any)
}

func (st *pgStore) prepareRemoveContractConfig() error {
	return st.prepare(stmtRemoveContractConfig, `
	DELETE FROM contract_machines
	WHERE contract_id = $1
`)
}

func (st *pgStore) removeContractConfig(ctx context.Context, cid token.ContractID) error {
	_, err := st.stmt(stmtRemoveContractConfig).ExecContext(ctx, cid[:])
	return errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareGetContractConfig() error {
	return st.prepare(stmtGetContractConfig, `
		SELECT
			resource_type,
			data
		FROM contract_config
		WHERE contract_id = $1
	`)
}

func (st *pgStore) getContractConfig(ctx context.Context, contractID token.ContractID) ([]store.ProductResource, error) {
	rows, err := st.stmt(stmtGetContractConfig).QueryContext(ctx, contractID[:])
	if err != nil {
		return nil, errgo.Mask(err, isDBError)
	}
	defer rows.Close()
	var resources []store.ProductResource
	for rows.Next() {
		var r store.ProductResource
		if err := rows.Scan(
			&r.ResourceType,
			&r.Data,
		); err != nil {
			return nil, errgo.Mask(err)
		}
		resources = append(resources, r)
	}
	if err := rows.Err(); err != nil {
		return nil, errgo.Mask(err, isDBError)
	}
	return resources, nil
}

func (st *pgStore) prepareGetAllProducts() error {
	return st.prepare(stmtGetAllProducts, `
	SELECT
		products.id,
		products.name,
		product_resources.resource_type,
		product_resources.data
	FROM products
	LEFT JOIN product_resources ON
		products.id = product_resources.product_id
	ORDER BY
		products.id
`)
}

func (st *pgStore) getAllProducts(ctx context.Context) ([]store.Product, error) {
	rows, err := st.stmt(stmtGetAllProducts).QueryContext(ctx)
	if err != nil {
		return nil, errgo.Mask(err, isDBError)
	}
	var products []store.Product
	for rows.Next() {
		var p store.Product
		var r store.ProductResource
		if err := rows.Scan(
			&p.ID,
			&p.Name,
			&r.ResourceType,
			&r.Data,
		); err != nil {
			return nil, errgo.Mask(err, isDBError)
		}
		if len(products) > 0 {
			prevProd := &products[len(products)-1]
			if p.ID == prevProd.ID {
				// The product id matches the previous one, so
				// append to the resources rather adding a new
				// product.
				prevProd.Resources = append(prevProd.Resources, r)
				continue
			}
		}
		p.Resources = []store.ProductResource{r}
		products = append(products, p)
	}
	if err := rows.Err(); err != nil {
		return nil, errgo.Mask(err, isDBError)
	}
	return products, nil
}

func (st *pgStore) prepareGetContract() error {
	return st.prepare(stmtGetContract, `
	SELECT
		accounts.id,
		accounts.name,
		accounts.created,
		contracts.id,
		contracts.name,
		contracts.account_id,
		contracts.created,
		contracts.created_by,
		contracts.effective_from,
		contracts.effective_to
	FROM contracts
	LEFT JOIN accounts ON
		contracts.account_id=accounts.id
	WHERE contracts.id=$1
`)
}

func (st *pgStore) getContract(ctx context.Context, id token.ContractID) (store.FullContract, error) {
	var c store.FullContract
	err := st.stmt(stmtGetContract).QueryRowContext(ctx, id[:]).Scan(
		&c.Account.ID,
		&c.Account.Name,
		&c.Account.Created,
		&c.Contract.ID,
		&c.Contract.Name,
		&c.Contract.AccountID,
		&c.Contract.Created,
		&c.Contract.CreatedBy,
		&c.Contract.EffectiveFrom,
		&c.Contract.EffectiveTo,
	)
	if err != nil {
		return c, errgo.Mask(err, isDBError)
	}
	c.Contract.ProductIDs, err = st.getContractProducts(ctx, id)
	if err != nil {
		return c, errgo.Mask(err, isDBError)
	}
	c.Config, err = st.getContractConfig(ctx, id)
	if err != nil {
		return c, errgo.Mask(err, isDBError)
	}
	return c, nil
}

func (st *pgStore) prepareGetContractProducts() error {
	return st.prepare(stmtGetContractProducts, `
	SELECT
		product_id
	FROM contract_products
	WHERE contract_id=$1
`)
}

func (st *pgStore) getContractProducts(ctx context.Context, id token.ContractID) ([]string, error) {
	var products []string
	rows, err := st.stmt(stmtGetContractProducts).QueryContext(ctx, id[:])
	if err != nil {
		return nil, errgo.Mask(err, isDBError)
	}
	for rows.Next() {
		var productID string
		err = rows.Scan(
			&productID,
		)
		if err != nil {
			return nil, errgo.Mask(err, isDBError)
		}
		products = append(products, productID)
	}
	if err := rows.Err(); err != nil {
		return nil, errgo.Mask(err, isDBError)
	}
	return products, nil
}

func (st *pgStore) prepareAddContractMachine() error {
	return st.prepare(stmtAddContractMachine, `
	INSERT INTO contract_machines (
		contract_id,
		machine_id
	)
	VALUES ($1, $2)
	ON CONFLICT DO NOTHING
`)
}

func (st *pgStore) addContractMachine(ctx context.Context, contractID token.ContractID, machineID string) error {
	_, err := st.stmt(stmtAddContractMachine).ExecContext(ctx, contractID[:], machineID)
	return errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareAddContractToken() error {
	return st.prepare(stmtAddContractToken, `
	INSERT INTO contract_tokens (
		contract_token,
		contract_id,
		expires
	) VALUES ($1, $2, $3)
`)
}

func (st *pgStore) addContractToken(ctx context.Context, info store.ContractTokenInfo) error {
	_, err := st.stmt(stmtAddContractToken).ExecContext(ctx,
		info.ContractToken[:],
		info.ContractID[:],
		info.Expires,
	)
	return errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareGetContractToken() error {
	return st.prepare(stmtGetContractToken, `
	SELECT contract_token, contract_id, expires
	FROM contract_tokens 
	WHERE contract_token = $1
`)
}

func (st *pgStore) getContractToken(ctx context.Context, t token.ContractToken) (store.ContractTokenInfo, error) {
	var info store.ContractTokenInfo
	err := st.stmt(stmtGetContractToken).QueryRowContext(ctx, t[:]).Scan(
		&info.ContractToken,
		&info.ContractID,
		&info.Expires,
	)
	return info, errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareGetContractTokens() error {
	return st.prepare(stmtGetContractTokens, `
	SELECT contract_token, contract_id, expires
	FROM contract_tokens 
	WHERE contract_id = $1
`)
}

func (st *pgStore) getContractTokens(ctx context.Context, contractID token.ContractID) ([]store.ContractTokenInfo, error) {
	var tokens []store.ContractTokenInfo
	rows, err := st.stmt(stmtGetContractTokens).QueryContext(ctx, contractID[:])
	if err != nil {
		return nil, errgo.Mask(err, isDBError)
	}
	defer rows.Close()
	for rows.Next() {
		var info store.ContractTokenInfo
		if err := rows.Scan(&info.ContractToken, &info.ContractID, &info.Expires); err != nil {
			return nil, errgo.Mask(err, isDBError)
		}
		tokens = append(tokens, info)
	}
	if err := rows.Err(); err != nil {
		return nil, errgo.Mask(err, isDBError)
	}
	return tokens, nil
}

func (st *pgStore) prepareGetMachineHasContract() error {
	return st.prepare(stmtGetMachineHasContract, `
	SELECT COUNT(*) FROM contract_machines
	WHERE contract_id = $1
	AND machine_id = $2
`)
}

func (st *pgStore) getMachineHasContract(ctx context.Context, contractID token.ContractID, machineID string) (bool, error) {
	var count int
	err := st.stmt(stmtGetMachineHasContract).QueryRowContext(ctx, contractID[:], machineID).Scan(&count)
	return count > 0, errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareRemoveContractMachine() error {
	return st.prepare(stmtRemoveContractMachine, `
	DELETE FROM contract_machines
	WHERE contract_id = $1
	AND machine_id = $2
`)
}

func (st *pgStore) removeContractMachine(ctx context.Context, cid token.ContractID, machineID string) error {
	_, err := st.stmt(stmtRemoveContractMachine).ExecContext(ctx, cid[:], machineID)
	return errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareGetMachine() error {
	// Use left join so that we're sure that even if the
	// machine type isn't available, we'll still see the record.
	return st.prepare(stmtGetMachine, `
SELECT
	machines.id,
	machine_types.architecture,
	machine_types.os_distribution,
	machine_types.os_kernel,
	machine_types.os_release,
	machine_types.os_series,
	machine_types.os_type
FROM machines
LEFT JOIN machine_types ON
	machines.machine_type=machine_types.machine_type
WHERE
	machines.id = $1
`)
}

func (st *pgStore) getMachine(ctx context.Context, id string) (store.MachineInfo, error) {
	var m store.MachineInfo
	err := st.stmt(stmtGetMachine).QueryRowContext(ctx, id).Scan(
		&m.MachineID,
		&m.Architecture,
		&m.OS.Distribution,
		&m.OS.Kernel,
		&m.OS.Release,
		&m.OS.Series,
		&m.OS.Type,
	)
	return m, errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareGetMachineType() error {
	return st.prepare(stmtGetMachineType, `
SELECT machine_type
FROM machine_types
WHERE
	architecture = $1 AND
	os_distribution = $2 AND
	os_kernel = $3 AND
	os_release = $4 AND
	os_series = $5 AND
	os_type = $6
`)
}

func (st *pgStore) getMachineType(ctx context.Context, m store.MachineInfo) (int, error) {
	var machineType int
	err := st.stmt(stmtGetMachineType).
		QueryRowContext(ctx,
			m.Architecture,
			m.OS.Distribution,
			m.OS.Kernel,
			m.OS.Release,
			m.OS.Series,
			m.OS.Type,
		).Scan(&machineType)
	return machineType, errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareInsertMachineType() error {
	return st.prepare(stmtInsertMachineType, `
INSERT into machine_types (
	architecture,
	os_distribution,
	os_kernel,
	os_release,
	os_series,
	os_type
) VALUES (
	$1, $2, $3, $4, $5, $6
) ON CONFLICT DO NOTHING
RETURNING machine_type
`)
}

// insertMachineType inserts a new machine type; it will
// fail if the type already exists.
func (st *pgStore) insertMachineType(ctx context.Context, m store.MachineInfo) (int, error) {
	var machineType int
	err := st.stmt(stmtInsertMachineType).
		QueryRowContext(ctx,
			m.Architecture,
			m.OS.Distribution,
			m.OS.Kernel,
			m.OS.Release,
			m.OS.Series,
			m.OS.Type,
		).Scan(&machineType)
	return machineType, errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareUpsertMachine() error {
	return st.prepare(stmtUpsertMachine, `
INSERT into machines (
	id,
	machine_type
) VALUES (
	$1, $2
) ON CONFLICT(id) DO UPDATE SET
	machine_type = $2
`)
}

// upsertMachine inserts an entry into the machine table
// or updates it if it already exists.
// The machineType entry must already exist.
func (st *pgStore) upsertMachine(ctx context.Context, machineID string, machineType int) error {
	_, err := st.stmt(stmtUpsertMachine).ExecContext(ctx,
		machineID,
		machineType,
	)
	return errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareAddProduct() error {
	return st.prepare(stmtAddProduct, `
	INSERT INTO products (
		id,
		name
	) VALUES ($1, $2)
`)
}

func (st *pgStore) addProduct(ctx context.Context, p store.Product) error {
	_, err := st.stmt(stmtAddProduct).ExecContext(ctx,
		p.ID,
		p.Name,
	)
	return errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareAddProductResource() error {
	return st.prepare(stmtAddProductResource, `
	INSERT INTO product_resources (
		product_id,
		resource_type,
		data
	) VALUES ($1, $2, $3)
`)
}

func (st *pgStore) addProductResource(ctx context.Context, productID string, r store.ProductResource) error {
	_, err := st.stmt(stmtAddProductResource).ExecContext(ctx,
		productID,
		r.ResourceType,
		r.Data,
	)
	return errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareGetProduct() error {
	return st.prepare(stmtGetProduct, `
	SELECT name FROM products 
	WHERE id = $1
`)
}

func (st *pgStore) getProduct(ctx context.Context, id string) (store.Product, error) {
	p := store.Product{
		ID: id,
	}
	err := st.stmt(stmtGetProduct).QueryRowContext(ctx, id).Scan(
		&p.Name,
	)
	return p, errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareGetProductResources() error {
	return st.prepare(stmtGetProductResources, `
	SELECT resource_type, data FROM product_resources
	WHERE product_id = $1
`)
}

func (st *pgStore) getProductResources(ctx context.Context, id string) ([]store.ProductResource, error) {
	rows, err := st.stmt(stmtGetProductResources).QueryContext(ctx, id)
	if err != nil {
		return nil, errgo.Mask(err, isDBError)
	}
	defer rows.Close()
	var resources []store.ProductResource
	for rows.Next() {
		var r store.ProductResource
		err := rows.Scan(
			&r.ResourceType,
			&r.Data,
		)
		if err != nil {
			return nil, errgo.Mask(err, isDBError)
		}
		resources = append(resources, r)
	}
	if err := rows.Err(); err != nil {
		return nil, errgo.Mask(err, isDBError)
	}
	return resources, nil
}

func (st *pgStore) prepareEnsureServiceAccount() error {
	return st.prepare(stmtEnsureServiceAccount, `
	INSERT INTO service_accounts (name, password_salt, password_hash)
	VALUES ($1, $2, $3)
	ON CONFLICT (name)
	DO UPDATE SET
		password_salt=$2,
		password_hash=$3
`)
}

func (st *pgStore) ensureServiceAccount(ctx context.Context, account store.ServiceAccount) error {
	_, err := st.stmt(stmtEnsureServiceAccount).ExecContext(
		ctx,
		account.Name,
		account.PasswordSalt,
		account.PasswordHash,
	)
	return errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareGetServiceAccount() error {
	return st.prepare(stmtGetServiceAccount, `
	SELECT name, password_salt, password_hash
	FROM service_accounts
	WHERE name = $1
`)
}

func (st *pgStore) getServiceAccount(ctx context.Context, name string) (store.ServiceAccount, error) {
	row := st.stmt(stmtGetServiceAccount).QueryRowContext(ctx, name)
	var account store.ServiceAccount
	err := row.Scan(
		&account.Name,
		&account.PasswordSalt,
		&account.PasswordHash,
	)
	return account, errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareRemoveContractProducts() error {
	return st.prepare(stmtRemoveContractProducts, `
	DELETE FROM contract_products WHERE contract_id = $1;
`)
}

func (st *pgStore) removeContractProducts(ctx context.Context, contractID token.ContractID) error {
	_, err := st.stmt(stmtRemoveContractProducts).ExecContext(ctx, contractID[:])
	return errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareUpdateProduct() error {
	return st.prepare(stmtUpdateProduct, `
	UPDATE products
	SET name = $1
	WHERE id = $2
`)
}

func (st *pgStore) updateProduct(ctx context.Context, p store.Product) error {
	_, err := st.stmt(stmtUpdateProduct).ExecContext(ctx,
		p.Name,
		p.ID,
	)
	return errgo.Mask(err, isDBError)
}

func (st *pgStore) prepareDeleteProductResources() error {
	return st.prepare(stmtDeleteProductResources, `
	DELETE FROM product_resources
	WHERE product_id = $1
`)
}

func (st *pgStore) deleteProductResources(ctx context.Context, id string) error {
	_, err := st.stmt(stmtDeleteProductResources).ExecContext(ctx, id)
	return errgo.Mask(err, isDBError)
}
