package postgresstore

import (
	"context"
	"database/sql"
	"fmt"
	"strings"

	"github.com/lib/pq"
	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
)

type pgStore struct {
	// db holds the underlying database.
	db *sql.DB
	// tx holds the current transaction if any.
	// This is only non-nil within withTx.
	tx    *sql.Tx
	stmts []*sql.Stmt
}

// New returns a new Store implementation that uses the given Postgres
// SQL database for storage.
func New(db *sql.DB) (store.Store, error) {
	st := &pgStore{
		db: db,
	}
	st.stmts = make([]*sql.Stmt, numStmts)
	for _, prepare := range prepareMethods {
		if err := prepare(st); err != nil {
			return nil, errgo.Mask(err)
		}
	}
	var missing []string
	// Check that all statements have been prepared.
	for i, stmt := range st.stmts {
		if stmt == nil {
			missing = append(missing, stmtID(i).String())
		}
	}
	if len(missing) > 0 {
		panic(errgo.Newf("unprepared statements: %s", strings.Join(missing, ", ")))
	}
	return st, nil
}

// withTx starts a transaction and executes the given function
// with an instance of the store that uses statements taken
// from that.
func (st *pgStore) withTx(ctx context.Context, f func(*pgStore) error) error {
	if st.tx != nil {
		panic("nested transaction")
	}
	tx, err := st.db.BeginTx(ctx, nil)
	if err != nil {
		return errgo.Mask(err, isDBError)
	}
	err = f(&pgStore{
		stmts: st.stmts,
		tx:    tx,
	})
	if err == nil {
		if err := tx.Commit(); err != nil {
			return errgo.Mask(err, isDBError)
		}
		return nil
	}
	rbErr := tx.Rollback()
	if rbErr == nil {
		// Successful rollback; preserve the original error.
		return errgo.Mask(err, errgo.Any)
	}
	return errgo.NoteMask(err, fmt.Sprintf("failed to roll back (error: %v) after", rbErr))
}

func (st *pgStore) prepare(id stmtID, stmtStr string) error {
	if st.stmts[id] != nil {
		panic(fmt.Sprintf("statement %v prepared twice", id))
	}
	stmt, err := st.db.Prepare(stmtStr)
	if err != nil {
		return errgo.Notef(err, "invalid prepared statement %v (%s)", id, stmtStr)
	}
	st.stmts[id] = stmt
	return nil
}

// bulkInsert runs a bulk insert of the given columns within the given table.
// The addRows argument is expected to add all the rows by calling
// its argument function.
func (st *pgStore) bulkInsert(ctx context.Context, addRows func(addRow func(...interface{}) error) error, table string, columns ...string) error {
	// Note: we can't use a previously prepared statement because CopyIn
	// statements can only be prepared inside a transaction.
	add := func(st *pgStore) error {
		stmt, err := st.tx.Prepare(pq.CopyIn(table, columns...))
		if err != nil {
			return errgo.Mask(err)
		}
		defer stmt.Close()
		if err := addRows(func(args ...interface{}) error {
			_, err := stmt.ExecContext(ctx, args...)
			return errgo.Mask(err)
		}); err != nil {
			return errgo.Mask(err, errgo.Any)
		}
		if _, err := stmt.ExecContext(ctx); err != nil {
			return errgo.Mask(err, isDBError)
		}
		return nil
	}
	// If we're not in a transaction, we need to be.
	if st.tx == nil {
		err := st.withTx(ctx, func(st *pgStore) error {
			return add(st)
		})
		return errgo.Mask(err, errgo.Any)
	}
	return errgo.Mask(add(st), errgo.Any)
}

func (st *pgStore) stmt(id stmtID) *sql.Stmt {
	stmt := st.stmts[id]
	if st.tx != nil {
		stmt = st.tx.Stmt(stmt)
	}
	return stmt
}

func (st *pgStore) Account(ctx context.Context, id token.AccountID) (store.Account, error) {
	a, err := st.getAccount(ctx, id)
	if errgo.Cause(err) == sql.ErrNoRows {
		return a, errgo.WithCausef(err, store.ErrAccountNotFound, "account %q not found in store", id)
	}
	return a, errgo.Mask(err)
}

func (st *pgStore) AccountForSalesforceID(ctx context.Context, id string) (store.Account, error) {
	a, err := st.getAccountWithSalesforceID(ctx, id)
	if errgo.Cause(err) == sql.ErrNoRows {
		return a, errgo.WithCausef(err, store.ErrAccountNotFound, "account for %q not found in store", id)
	}
	return a, errgo.Mask(err)
}

func (st *pgStore) AccountAccess(ctx context.Context, accountID token.AccountID, userID string) (store.AccountAccess, error) {
	access, err := st.getAccountAccess(ctx, accountID, userID)
	if errgo.Cause(err) == sql.ErrNoRows {
		return store.NoAccountAccess, nil
	}
	return access, errgo.Mask(err)
}

func (st *pgStore) AccountContracts(ctx context.Context, accountID token.AccountID) ([]store.FullContract, error) {
	contracts, err := st.getAccountContracts(ctx, accountID)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	for i := range contracts {
		var err error
		contracts[i].Contract.ProductIDs, err = st.getContractProducts(ctx, contracts[i].Contract.ID)
		if err != nil {
			return nil, errgo.Mask(err)
		}
	}

	return contracts, nil
}

func (st *pgStore) AddAccount(ctx context.Context, a store.Account, userID string) error {
	err := st.withTx(ctx, func(st *pgStore) error {
		if err := st.addAccount(ctx, a); err != nil {
			return errgo.Mask(err, isDBError)
		}
		if err := st.addAccountAccess(ctx, a.ID, userID, store.AdminAccountAccess); err != nil {
			return errgo.Mask(err)
		}
		return nil
	})
	if err == nil {
		return nil
	}
	if pqErr, ok := pqError(err); ok && pqErr.Code.Name() == "unique_violation" {
		switch pqErr.Constraint {
		case "accounts_pkey":
			return errgo.WithCausef(nil, store.ErrAccountExists, "")
		case "accounts_salesforce_account_id_key":
			return errgo.WithCausef(nil, store.ErrDuplicateSalesforceAccount, "")
		}
	}
	return errgo.Mask(err)
}

func (st *pgStore) AddContract(ctx context.Context, c store.Contract) (token.ContractID, error) {
	id, err := token.NewContractID()
	if err != nil {
		return id, errgo.Mask(err)
	}
	c.ID = id
	err = st.withTx(ctx, func(st *pgStore) error {
		if err := st.addContract(ctx, c); err != nil {
			if isDuplicateError(err) {
				return errgo.WithCausef(nil, store.ErrContractExists, "")
			}
			return errgo.Mask(err)
		}
		err := st.addContractProducts(ctx, func(addProduct func(contractID token.ContractID, pid string) error) error {
			for _, p := range c.ProductIDs {
				if err := addProduct(id, p); err != nil {
					return errgo.Mask(err)
				}
			}
			return nil
		})
		return errgo.Mask(err)
	})
	if err != nil {
		return token.ContractID{}, errgo.Mask(err, errgo.Is(store.ErrContractExists))
	}
	return c.ID, nil
}

func (st *pgStore) AddContractMachine(ctx context.Context, contractID token.ContractID, machineID string) error {
	err := st.addContractMachine(ctx, contractID, machineID)
	return errgo.Mask(err)
}

func (st *pgStore) AddContractToken(ctx context.Context, info store.ContractTokenInfo) error {
	err := st.addContractToken(ctx, info)
	if err != nil {
		if isDuplicateError(err) {
			panic("duplicated contract token")
		}
		return errgo.Mask(err)
	}
	return nil
}

func (st *pgStore) ConfigureContract(ctx context.Context, contractID token.ContractID, resources []store.ProductResource) error {
	err := st.withTx(ctx, func(st *pgStore) error {
		if err := st.removeContractConfig(ctx, contractID); err != nil {
			return errgo.Mask(err)
		}
		err := st.addContractConfigResources(ctx, func(addResource func(cid token.ContractID, r store.ProductResource) error) error {
			for _, r := range resources {
				if err := addResource(contractID, r); err != nil {
					return errgo.Mask(err, isDBError)
				}
			}
			return nil
		})
		if err != nil {
			switch {
			case isDuplicateError(err):
				return errgo.WithCausef(err, store.ErrProductConflict, "")
			case isForeignKeyViolation(err):
				return errgo.WithCausef(err, store.ErrContractNotFound, "")
			}
			return errgo.Mask(err)
		}
		return nil
	})
	return errgo.Mask(err, errgo.Any)
}

func (st *pgStore) AddProduct(ctx context.Context, p store.Product) error {
	err := st.withTx(ctx, func(st *pgStore) error {
		if err := st.addProduct(ctx, p); err != nil {
			if isDuplicateError(err) {
				return errgo.WithCausef(err, store.ErrDuplicateProduct, "")
			}
			return errgo.Mask(err)
		}
		for _, r := range p.Resources {
			if err := st.addProductResource(ctx, p.ID, r); err != nil {
				if isDuplicateError(err) {
					return errgo.WithCausef(err, store.ErrProductConflict, "")
				}
				return errgo.Mask(err)
			}
		}
		return nil
	})
	if err != nil {
		return errgo.Mask(err, errgo.Any)
	}
	return nil
}

func (st *pgStore) UpdateContractProducts(ctx context.Context, contractID token.ContractID, productIDs []string) error {
	err := st.withTx(ctx, func(st *pgStore) error {
		err := st.removeContractProducts(ctx, contractID)
		if err != nil {
			return errgo.Mask(err)
		}
		for _, productID := range productIDs {
			err := st.addContractProduct(ctx, contractID, productID)
			if err != nil {
				return errgo.Mask(err)
			}
		}
		return nil
	})
	if err != nil {
		return errgo.Mask(err)
	}
	return nil
}

func (st *pgStore) UpdateProduct(ctx context.Context, p store.Product) error {
	err := st.withTx(ctx, func(st *pgStore) error {
		if _, err := st.getProduct(ctx, p.ID); err != nil {
			if errgo.Cause(err) == sql.ErrNoRows {
				return errgo.WithCausef(nil, store.ErrProductNotFound, "")
			}
			return errgo.Mask(err)
		}
		if err := st.updateProduct(ctx, p); err != nil {
			return errgo.Mask(err)
		}
		if err := st.deleteProductResources(ctx, p.ID); err != nil {
			return errgo.Mask(err)
		}
		for _, r := range p.Resources {
			if err := st.addProductResource(ctx, p.ID, r); err != nil {
				if isDuplicateError(err) {
					return errgo.WithCausef(err, store.ErrProductConflict, "")
				}
				return errgo.Mask(err)
			}
		}
		return nil
	})
	if err != nil {
		return errgo.Mask(err, errgo.Any)
	}
	return nil
}

func (st *pgStore) Contract(ctx context.Context, id token.ContractID) (store.FullContract, error) {
	c, err := st.getContract(ctx, id)
	if err != nil {
		if errgo.Cause(err) == sql.ErrNoRows {
			return store.FullContract{}, errgo.WithCausef(nil, store.ErrContractNotFound, "")
		}
		return store.FullContract{}, errgo.Mask(err)
	}
	return c, nil
}

func (st *pgStore) ContractToken(ctx context.Context, token token.ContractToken) (store.ContractTokenInfo, error) {
	info, err := st.getContractToken(ctx, token)
	if err != nil {
		if errgo.Cause(err) == sql.ErrNoRows {
			return store.ContractTokenInfo{}, errgo.WithCausef(nil, store.ErrTokenNotFound, "")
		}
		return store.ContractTokenInfo{}, errgo.Mask(err)
	}
	return info, nil
}

// ContractTokens implements store.Store.ContractTokens.
func (st *pgStore) ContractTokens(ctx context.Context, id token.ContractID) ([]store.ContractTokenInfo, error) {
	tokens, err := st.getContractTokens(ctx, id)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return tokens, nil
}

// EnsureMachine implements store.Store.EnsureMachine.
func (st *pgStore) EnsureMachine(ctx context.Context, m store.MachineInfo) error {
	machineType, err := st.ensureMachineType(ctx, m)
	if err != nil {
		return errgo.Notef(err, "cannot create machine type")
	}
	if err := st.upsertMachine(ctx, m.MachineID, machineType); err != nil {
		return errgo.Notef(err, "cannot ensure machine")
	}
	return nil
}

// EnsureServiceAccount implements store.Store.EnsureServiceAccount.
func (st *pgStore) EnsureServiceAccount(ctx context.Context, account store.ServiceAccount) error {
	err := st.ensureServiceAccount(ctx, account)
	return errgo.Mask(err)
}

func (st *pgStore) EnsureUser(ctx context.Context, user store.User) error {
	if err := st.ensureUser(ctx, user); err != nil {
		return errgo.Mask(err)
	}
	return nil
}

func (st *pgStore) MachineHasContract(ctx context.Context, contractID token.ContractID, machineID string) (bool, error) {
	ok, err := st.getMachineHasContract(ctx, contractID, machineID)
	return ok, errgo.Mask(err)
}

// MachineInfo implements store.Store.MachineInfo.
func (st *pgStore) MachineInfo(ctx context.Context, id string) (store.MachineInfo, error) {
	m, err := st.getMachine(ctx, id)
	if err == nil {
		return m, nil
	}
	if errgo.Cause(err) == sql.ErrNoRows {
		return store.MachineInfo{}, errgo.WithCausef(nil, store.ErrMachineNotFound, "machine %q not found", id)
	}
	return store.MachineInfo{}, errgo.Mask(err)
}

func (st *pgStore) Product(ctx context.Context, id string) (store.Product, error) {
	var p store.Product
	p, err := st.getProduct(ctx, id)
	if errgo.Cause(err) == sql.ErrNoRows {
		return store.Product{}, errgo.WithCausef(err, store.ErrProductNotFound, "")
	}
	p.Resources, err = st.getProductResources(ctx, id)
	if err != nil {
		return store.Product{}, errgo.Mask(err)
	}
	return p, nil
}

func (st *pgStore) AllProducts(ctx context.Context) ([]store.Product, error) {
	products, err := st.getAllProducts(ctx)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return products, nil
}

func (st *pgStore) RemoveContractMachine(ctx context.Context, cid token.ContractID, machineID string) error {
	err := st.removeContractMachine(ctx, cid, machineID)
	return errgo.Mask(err)
}

func (st *pgStore) SetAccountAccess(ctx context.Context, accountID token.AccountID, userID string, a store.AccountAccess) error {
	err := st.setAccountAccess(ctx, accountID, userID, a)
	return errgo.Mask(err)
}

func (st *pgStore) ServiceAccount(ctx context.Context, name string) (store.ServiceAccount, error) {
	a, err := st.getServiceAccount(ctx, name)
	if errgo.Cause(err) == sql.ErrNoRows {
		return store.ServiceAccount{}, errgo.WithCausef(nil, store.ErrServiceAccountNotFound, "")
	}
	if err != nil {
		return store.ServiceAccount{}, errgo.Mask(err)
	}
	return a, nil
}

func (st *pgStore) User(ctx context.Context, id string) (store.User, error) {
	u, err := st.getUser(ctx, id)
	if err != nil {
		if errgo.Cause(err) == sql.ErrNoRows {
			return store.User{}, errgo.WithCausef(nil, store.ErrUserNotFound, "")
		}
	}
	return u, nil
}

func (st *pgStore) UserAccounts(ctx context.Context, userID string) ([]store.AccountWithAccess, error) {
	accounts, err := st.getUserAccounts(ctx, userID)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return accounts, nil
}

func (st *pgStore) ensureMachineType(ctx context.Context, m store.MachineInfo) (machineType int, err error) {
	for i := 0; i < 2; i++ {
		id, err := st.getMachineType(ctx, m)
		if err == nil {
			return id, nil
		}
		if errgo.Cause(err) != sql.ErrNoRows {
			return 0, errgo.Mask(err)
		}
		id, err = st.insertMachineType(ctx, m)
		if err != nil {
			if errgo.Cause(err) == sql.ErrNoRows {
				// No rows were returned because there
				// was a conflict (ignored), so go round the
				// loop again to pick up the row.
				continue
			}
			return 0, errgo.Mask(err)
		}
		return id, nil
	}
	// This should never happen (we should always find a machine
	// type if we've got a duplicate error trying to insert one),
	// but be defensive in case someone's deleting machine type
	// entries for some reason.
	return 0, errgo.Newf("too many iterations trying to create machine type")
}

// isDBError reports whether the error represents
// a database error that the code might want to
// inspect.
func isDBError(err error) bool {
	_, ok := err.(*pq.Error)
	return ok || err == sql.ErrNoRows
}

func pqError(err error) (*pq.Error, bool) {
	pqerr, ok := errgo.Cause(err).(*pq.Error)
	return pqerr, ok
}

func pqErrCode(err error) string {
	if pqerr, ok := pqError(err); ok {
		return pqerr.Code.Name()
	}
	return ""
}

func isDuplicateError(err error) bool {
	return pqErrCode(err) == "unique_violation"
}

func isForeignKeyViolation(err error) bool {
	return pqErrCode(err) == "foreign_key_violation"
}
