// Copyright 2019 Canonical Ltd.

package saauth_test

import (
	"context"
	"io/ioutil"
	"net/http"
	"net/http/httptest"
	"testing"

	qt "github.com/frankban/quicktest"
	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/internal/memstore"
	"github.com/CanonicalLtd/ua-contracts/internal/saauth"
	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/params"
)

func TestRoundTripPassword(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	s := memstore.New()

	err := saauth.SetPassword(ctx, s, "test-account", "test-password")
	c.Assert(err, qt.Equals, nil)
	id, err := saauth.Validate(ctx, s, "test-account", "test-password")
	c.Assert(err, qt.Equals, nil)
	c.Assert(id.Account(), qt.Equals, "test-account")
	allowed, err := id.Allow(ctx, []string{"test-account", "other-test-account"})
	c.Assert(err, qt.Equals, nil)
	c.Assert(allowed, qt.Equals, true)
	allowed, err = id.Allow(ctx, []string{"other-test-account"})
	c.Assert(err, qt.Equals, nil)
	c.Assert(allowed, qt.Equals, false)
}

func TestNoAccount(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	s := memstore.New()

	id, err := saauth.Validate(ctx, s, "test-account", "test-password")
	c.Assert(err, qt.Equals, nil)
	c.Assert(id.Account(), qt.Equals, "")
	allowed, err := id.Allow(ctx, []string{"test-account", ""})
	c.Assert(err, qt.Equals, nil)
	c.Assert(allowed, qt.Equals, false)
}

func TestBadPassword(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	s := memstore.New()

	err := saauth.SetPassword(ctx, s, "test-account", "test-password")
	c.Assert(err, qt.Equals, nil)
	id, err := saauth.Validate(ctx, s, "test-account", "bad-password")
	c.Assert(err, qt.Equals, nil)
	c.Assert(id.Account(), qt.Equals, "")
	allowed, err := id.Allow(ctx, []string{"test-account", ""})
	c.Assert(err, qt.Equals, nil)
	c.Assert(allowed, qt.Equals, false)
}

func TestAuthenticateACL(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	s := memstore.New()

	err := saauth.SetPassword(ctx, s, "test-user", "test-password")
	c.Assert(err, qt.Equals, nil)

	req, err := http.NewRequest("GET", "/", nil)
	c.Assert(err, qt.Equals, nil)
	req.SetBasicAuth("test-user", "test-password")

	id, err := saauth.AuthenticateACL(s)(ctx, nil, req)
	c.Assert(err, qt.Equals, nil)
	allow, err := id.Allow(ctx, []string{"test-user"})
	c.Assert(err, qt.Equals, nil)
	c.Assert(allow, qt.Equals, true)
}

func TestAuthenticateACLNoAuth(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	s := memstore.New()

	w := httptest.NewRecorder()
	req, err := http.NewRequest("GET", "/", nil)
	c.Assert(err, qt.Equals, nil)

	id, err := saauth.AuthenticateACL(s)(ctx, w, req)
	c.Assert(errgo.Cause(err), qt.Equals, params.ErrUnauthorized)
	c.Assert(err, qt.ErrorMatches, "unauthorized")
	c.Assert(id, qt.IsNil)

	resp := w.Result()
	c.Assert(resp.StatusCode, qt.Equals, http.StatusUnauthorized)
	b, err := ioutil.ReadAll(resp.Body)
	c.Assert(err, qt.Equals, nil)
	c.Assert(string(b), qt.Equals, `{"code":"unauthorized","message":"unauthorized"}`)
}

func TestAuthenticateACLValidateError(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	s := memstore.New()

	w := httptest.NewRecorder()
	req, err := http.NewRequest("GET", "/", nil)
	c.Assert(err, qt.Equals, nil)
	req.SetBasicAuth("test-user", "test-password")

	id, err := saauth.AuthenticateACL(testStore{s})(ctx, w, req)
	c.Assert(err, qt.ErrorMatches, `test-error`)
	c.Assert(id, qt.IsNil)

	resp := w.Result()
	c.Assert(resp.StatusCode, qt.Equals, http.StatusInternalServerError)
	b, err := ioutil.ReadAll(resp.Body)
	c.Assert(err, qt.Equals, nil)
	c.Assert(string(b), qt.Equals, `{"message":"test-error"}`)
}

type testStore struct {
	store.Store
}

func (s testStore) ServiceAccount(ctx context.Context, name string) (store.ServiceAccount, error) {
	return store.ServiceAccount{}, errgo.New("test-error")
}
