// Copyright 2019 Canonical Ltd.

package storeutil_test

import (
	"bytes"
	"context"
	"testing"
	"time"

	"github.com/CanonicalLtd/salesforce"
	sfparams "github.com/CanonicalLtd/salesforce/params"
	qt "github.com/frankban/quicktest"
	"github.com/google/go-cmp/cmp"
	"github.com/google/go-cmp/cmp/cmpopts"
	"github.com/juju/zaputil/zapctx"
	"go.uber.org/zap/zaptest"
	"gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/internal/memstore"
	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/storeutil"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
)

func TestGetAccounts(t *testing.T) {
	c := qt.New(t)

	ctx := testContext(c)
	st := memstore.New()
	sfclient := storeutil.SalesforceDisabledClient{}
	err := storeutil.EnsureDefaultProducts(ctx, st)
	c.Assert(err, qt.Equals, nil)

	u := store.User{
		ID:          "accountID",
		Email:       "mine@email.com",
		DisplayName: "mine",
	}
	// check the user has no account stored.
	as, err := st.UserAccounts(ctx, u.ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(as, qt.HasLen, 0)

	as, err = storeutil.GetAccounts(ctx, sfclient, st, u)
	c.Assert(err, qt.Equals, nil)
	c.Assert(as, qt.HasLen, 1)

	// check the new account has been stored.
	as2, err := st.UserAccounts(ctx, u.ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(as, qt.HasLen, 1)
	c.Assert(as2, qt.DeepEquals, as)

	a := store.Account{
		ID:      token.MustNewAccountID(),
		Name:    "test account 2",
		Created: time.Now(),
	}
	err = st.AddAccount(ctx, a, u.ID)
	c.Assert(err, qt.Equals, nil)

	// check that GetAccounts returns the stored accounts
	as3, err := storeutil.GetAccounts(ctx, sfclient, st, u)
	c.Assert(err, qt.Equals, nil)
	c.Assert(as3, qt.CmpEquals(cmpopts.SortSlices(sortAccounts)), []store.AccountWithAccess{
		as2[0],
		{
			Account: a,
			Access:  store.AdminAccountAccess,
		},
	})
}

func TestGetAccountsSalesforce(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	st := memstore.New()
	err := storeutil.EnsureDefaultProducts(ctx, st)
	c.Assert(err, qt.Equals, nil)

	sfclient := testSalesforceClient{
		User: &sfparams.User{
			Account: sfparams.Account{
				ID:   "sfacc1",
				Name: "sf account name",
			},
		},
	}
	u := store.User{
		ID:          "accountID",
		Email:       "mine@email.com",
		DisplayName: "mine",
	}
	// check the user has no account stored.
	as, err := st.UserAccounts(ctx, u.ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(as, qt.HasLen, 0)

	as, err = storeutil.GetAccounts(ctx, sfclient, st, u)
	c.Assert(err, qt.Equals, nil)
	c.Assert(as, qt.HasLen, 1)

	// check the new account has been stored with the salesforce
	// account ID.
	a, err := st.AccountForSalesforceID(ctx, "sfacc1")
	c.Assert(err, qt.Equals, nil)
	c.Assert(a.SalesforceAccountID, qt.Equals, "sfacc1")

	a = store.Account{
		ID:      token.MustNewAccountID(),
		Name:    "test account 2",
		Created: time.Now(),
	}
	err = st.AddAccount(ctx, a, u.ID)
	c.Assert(err, qt.Equals, nil)

	// check that GetAccounts returns the stored accounts
	as3, err := storeutil.GetAccounts(ctx, sfclient, st, u)
	c.Assert(err, qt.Equals, nil)
	c.Assert(as3, qt.CmpEquals(cmpopts.SortSlices(sortAccounts)), []store.AccountWithAccess{
		as[0],
		{
			Account: a,
			Access:  store.AdminAccountAccess,
		},
	})
}

func sortAccounts(x, y store.AccountWithAccess) bool {
	if x.Access == y.Access {
		return bytes.Compare(x.Account.ID[:], y.Account.ID[:]) < 0
	}
	return x.Access < y.Access
}

func TestGetAccountsExistingSalesforce(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	st := memstore.New()
	err := storeutil.EnsureDefaultProducts(ctx, st)
	c.Assert(err, qt.Equals, nil)

	a := store.Account{
		ID:                  token.MustNewAccountID(),
		Name:                "test salesforce account 1",
		Created:             time.Now(),
		SalesforceAccountID: "sfacc1",
	}
	err = st.AddAccount(ctx, a, "user0")
	c.Assert(err, qt.Equals, nil)

	sfclient := testSalesforceClient{
		User: &sfparams.User{
			Account: sfparams.Account{
				ID:   "sfacc1",
				Name: "sf account name",
			},
		},
	}
	u := store.User{
		ID:          "user1",
		Email:       "mine@email.com",
		DisplayName: "mine",
	}
	// check the user has no account stored.
	as, err := st.UserAccounts(ctx, u.ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(as, qt.HasLen, 0)

	as, err = storeutil.GetAccounts(ctx, sfclient, st, u)
	c.Assert(err, qt.Equals, nil)
	c.Assert(as, qt.HasLen, 1)

	// check the existing salesforce account has been used
	c.Assert(as[0].Account.ID.String(), qt.Equals, a.ID.String())
}

func TestGetAccountsSalesforceError(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	st := memstore.New()
	err := storeutil.EnsureDefaultProducts(ctx, st)
	c.Assert(err, qt.Equals, nil)

	sfclient := testSalesforceClient{
		Error: errgo.New("some error"),
	}
	u := store.User{
		ID:          "accountID",
		Email:       "mine@email.com",
		DisplayName: "mine",
	}

	_, err = storeutil.GetAccounts(ctx, sfclient, st, u)
	c.Assert(err, qt.ErrorMatches, `some error`)
}

func TestEnsureContract(t *testing.T) {
	c := qt.New(t)

	ctx := testContext(c)
	st := memstore.New()
	sfclient := storeutil.SalesforceDisabledClient{}
	err := storeutil.EnsureDefaultProducts(ctx, st)
	c.Assert(err, qt.Equals, nil)

	acc, err := storeutil.GetAccounts(ctx, sfclient, st, store.User{
		ID:          "accountID",
		Email:       "mine@email.com",
		DisplayName: "mine",
	})
	c.Assert(err, qt.Equals, nil)
	contracts, err := storeutil.GetContractsForAccount(ctx, sfclient, st, acc[0].ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(contracts, qt.HasLen, 1)
	c.Assert(
		contracts[0],
		qt.CmpEquals(equateApproxTime(time.Second),
			cmpopts.IgnoreFields(store.Contract{}, "ID")),
		store.FullContract{
			Account: store.Account{
				ID:      acc[0].ID,
				Name:    "mine@email.com",
				Created: time.Now(),
			},
			Contract: store.Contract{
				Name:          "mine@email.com",
				AccountID:     acc[0].ID,
				ProductIDs:    []string{storeutil.DefaultProductID},
				Created:       time.Now(),
				EffectiveFrom: newTime(time.Now()),
			},
		},
	)
}

func TestEnsureContractWithSFAsset(t *testing.T) {
	c := qt.New(t)

	ctx := testContext(c)
	st := memstore.New()
	err := storeutil.EnsureDefaultProducts(ctx, st)
	c.Assert(err, qt.Equals, nil)

	sfclient := testSalesforceClient{
		Assets: []salesforce.Asset{{
			ID: "Not Active",
			Product: salesforce.Product{
				Name: "Canonical FIPS Service",
			},
			UsageEndDate: salesforce.Time{
				T: time.Now().Add(time.Minute),
			},
			Status: "Inactive",
		}, {
			Product: salesforce.Product{
				Family:       "Support",
				ServiceLevel: "essential",
			},
			UsageEndDate: salesforce.Time{
				T: time.Now().Add(time.Hour),
			},
			Status: "Active",
		}, {
			Product: salesforce.Product{
				Name: "Canonical FIPS Service",
			},
			UsageEndDate: salesforce.Time{
				T: time.Now().Add(time.Hour),
			},
			Status: "Active",
		}, {
			ID: "Old one",
			Product: salesforce.Product{
				Name: "Canonical FIPS Service",
			},
			UsageEndDate: salesforce.Time{
				T: time.Now().Add(-time.Hour),
			},
			Status: "Active",
		}},
		User: &sfparams.User{
			Account: sfparams.Account{
				ID:   "sfacc2",
				Name: "sf account name",
			},
		},
	}

	acc, err := storeutil.GetAccounts(ctx, sfclient, st, store.User{
		ID:          "accountID",
		Email:       "mine@email.com",
		DisplayName: "mine",
	})
	c.Assert(err, qt.Equals, nil)

	contracts, err := storeutil.GetContractsForAccount(ctx, sfclient, st, acc[0].ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(contracts, qt.HasLen, 1)
	c.Assert(
		contracts[0],
		qt.CmpEquals(
			equateApproxTime(time.Second),
			cmpopts.IgnoreFields(store.Contract{}, "ID"),
		),
		store.FullContract{
			Account: store.Account{
				ID:                  acc[0].ID,
				Name:                "sf account name",
				Created:             time.Now(),
				SalesforceAccountID: "sfacc2",
			},
			Contract: store.Contract{
				Name:          "sf account name",
				AccountID:     acc[0].ID,
				ProductIDs:    []string{"support-essential", "fips"},
				Created:       time.Now(),
				EffectiveFrom: newTime(time.Now()),
				EffectiveTo:   newTime(time.Now().Add(time.Hour)),
			},
		},
	)

	sfclient.Assets = nil
	sfclient.Error = errgo.New("something wrong")

	// Once a contract is created, it doesn't check Salesforce again and return the same contract.
	contracts2, err := storeutil.GetContractsForAccount(ctx, sfclient, st, acc[0].ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(contracts2, qt.HasLen, 1)
	c.Assert(
		contracts2[0],
		qt.CmpEquals(equateApproxTime(time.Second)),
		contracts[0],
	)
}

func TestEnsureContractWithSFNoAssets(t *testing.T) {
	c := qt.New(t)

	ctx := testContext(c)
	st := memstore.New()
	err := storeutil.EnsureDefaultProducts(ctx, st)
	c.Assert(err, qt.Equals, nil)

	sfclient := testSalesforceClient{
		User: &sfparams.User{
			Account: sfparams.Account{
				ID:   "sfacc2",
				Name: "sf account name",
			},
		},
	}

	acc, err := storeutil.GetAccounts(ctx, sfclient, st, store.User{
		ID:          "accountID",
		Email:       "mine@email.com",
		DisplayName: "mine",
	})
	c.Assert(err, qt.Equals, nil)

	contracts, err := storeutil.GetContractsForAccount(ctx, sfclient, st, acc[0].ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(contracts, qt.HasLen, 1)
	c.Assert(
		contracts[0],
		qt.CmpEquals(
			equateApproxTime(time.Second),
			cmpopts.IgnoreFields(store.Contract{}, "ID"),
		),
		store.FullContract{
			Account: store.Account{
				ID:                  acc[0].ID,
				Name:                "sf account name",
				Created:             time.Now(),
				SalesforceAccountID: "sfacc2",
			},
			Contract: store.Contract{
				Name:          "sf account name",
				AccountID:     acc[0].ID,
				ProductIDs:    []string{storeutil.DefaultProductID},
				Created:       time.Now(),
				EffectiveFrom: newTime(time.Now()),
			},
		},
	)

	sfclient.Assets = nil
	sfclient.Error = errgo.New("something wrong")

	// Once a contract is created, it doesn't check Salesforce again and return the same contract.
	contracts2, err := storeutil.GetContractsForAccount(ctx, sfclient, st, acc[0].ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(contracts2, qt.HasLen, 1)
	c.Assert(
		contracts2[0],
		qt.CmpEquals(equateApproxTime(time.Second)),
		contracts[0],
	)
}

func TestEnsureContractWithSFError(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	st := memstore.New()
	err := storeutil.EnsureDefaultProducts(ctx, st)
	c.Assert(err, qt.Equals, nil)

	sfclient := testSalesforceClient{
		User: &sfparams.User{
			Account: sfparams.Account{
				ID:   "sfacc2",
				Name: "sf account name",
			},
		},
	}

	acc, err := storeutil.GetAccounts(ctx, sfclient, st, store.User{
		ID:          "accountID",
		Email:       "mine@email.com",
		DisplayName: "mine",
	})
	c.Assert(err, qt.Equals, nil)

	// With an error from Salesforce
	sfclient.Error = errgo.New("something wrong")
	_, err = storeutil.GetContractsForAccount(ctx, sfclient, st, acc[0].ID)
	c.Assert(err, qt.ErrorMatches, "cannot ensure contract: something wrong")
}

func TestUpdateContractsForSalesforceAccount(t *testing.T) {
	c := qt.New(t)

	tests := []struct {
		name                    string
		user                    *store.User
		account                 *store.Account
		sfID                    string
		sfError                 error
		assets                  []salesforce.Asset
		expectedContractsLength int
		expectedProductIDs      []string
		expectedErrorCause      error
	}{{
		name: "no-contract",
		user: &store.User{
			ID:          "accountID",
			Email:       "mine@email.com",
			DisplayName: "mine",
		},
		account: &store.Account{
			ID:                  token.MustNewAccountID(),
			Name:                "my account",
			Created:             time.Now(),
			SalesforceAccountID: "sfid",
		},
		sfID:                    "sfid",
		assets:                  []salesforce.Asset{},
		expectedContractsLength: 1,
		expectedProductIDs:      []string{"free"},
	}, {
		name: "new-contract",
		user: &store.User{
			ID:          "accountID",
			Email:       "mine@email.com",
			DisplayName: "mine",
		},
		account: &store.Account{
			ID:                  token.MustNewAccountID(),
			Name:                "my account",
			Created:             time.Now(),
			SalesforceAccountID: "sfid",
		},
		sfID: "sfid",
		assets: []salesforce.Asset{{
			UsageEndDate: salesforce.Time{time.Now().Add(24 * time.Hour)},
			Status:       "Active",
			Product: salesforce.Product{
				ESM: salesforce.Bool{true},
			},
		}},
		expectedContractsLength: 1,
		expectedProductIDs:      []string{"esm"},
	}, {
		name:                    "no local user",
		expectedContractsLength: 0,
	}, {
		name:               "no-salesforce-account",
		sfError:            sfparams.ErrNotFound,
		expectedErrorCause: storeutil.ErrSalesforceAccountNotFound,
	}}

	for i, test := range tests {
		c.Logf("test %d: %q", i, test.name)
		ctx := context.Background()
		st := memstore.New()
		err := storeutil.EnsureDefaultProducts(ctx, st)
		c.Assert(err, qt.Equals, nil)
		sfClient := testSalesforceClient{
			Assets: test.assets,
			Error:  test.sfError,
		}

		if test.user != nil {
			err = st.EnsureUser(ctx, *test.user)
			c.Assert(err, qt.Equals, nil)
		}
		if test.account != nil {
			err = st.AddAccount(ctx, *test.account, test.user.ID)
			c.Assert(err, qt.Equals, nil)
		}
		sfClient.Assets = test.assets
		contracts, err := storeutil.UpdateContractsForSalesforceAccount(ctx, sfClient, st, test.sfID)

		if test.expectedErrorCause != nil {
			c.Assert(errgo.Cause(err), qt.Equals, test.expectedErrorCause)
			continue
		}
		c.Assert(err, qt.Equals, nil)
		c.Assert(contracts, qt.HasLen, test.expectedContractsLength)
		if test.expectedContractsLength == 1 {
			c.Assert(contracts[0].Contract.ProductIDs, qt.DeepEquals, test.expectedProductIDs)
		}
	}
}

type testSalesforceClient struct {
	User   *sfparams.User
	Assets []salesforce.Asset
	Error  error
}

func (t testSalesforceClient) GetUser(ctx context.Context, email string) (*sfparams.User, error) {
	return t.User, t.Error
}

func (t testSalesforceClient) GetAssetsForAccount(ctx context.Context, salesforceID string) ([]salesforce.Asset, error) {
	return t.Assets, t.Error
}

func equateApproxTime(margin time.Duration) cmp.Option {
	return cmp.Comparer(func(t0, t1 time.Time) bool {
		if t0.IsZero() && t1.IsZero() {
			return true
		}
		if t0.IsZero() || t1.IsZero() {
			return false
		}
		diff := t0.Sub(t1)
		if diff < 0 {
			diff = -diff
		}
		return diff <= margin
	})
}

func newTime(t time.Time) *time.Time {
	return &t
}

func testContext(c *qt.C) context.Context {
	return zapctx.WithLogger(context.Background(), zaptest.NewLogger(c))
}
