package storeutil

import (
	"context"
	"fmt"

	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/params"
)

// EnsureDefaultProducts adds some default products to the store,
// corresponding to those products that can be enabled in Salesforce.
func EnsureDefaultProducts(ctx context.Context, st store.Store) (err error) {
	addProduct := func(p store.Product) {
		if err != nil {
			return
		}
		err = st.AddProduct(ctx, p)
		if errgo.Cause(err) == store.ErrDuplicateProduct {
			err = nil
		}
		if err != nil {
			err = errgo.Notef(err, "cannot add product %q", p.ID)
		}
	}
	addProduct(store.Product{
		ID:   "esm",
		Name: "esm",
		Resources: []store.ProductResource{{
			ResourceType: params.ResourceESM,
			// TODO: Is aptKey a useful thing here?
			Data: `{
				"aptURL": "https://esm.ubuntu.com",
				"aptKey": "67C7A026",
				"series": ["precise", "trusty"],
				"suites": ["precise", "trusty-security", "trusty-updates"],
				"enableByDefault": true
			}`,
		}},
	})
	addProduct(store.Product{
		ID:   "livepatch",
		Name: "livepatch",
		Resources: []store.ProductResource{{
			ResourceType: params.ResourceLivepatch,
			Data: `{
				"architectures": ["x86_64"],
				"kernelFlavors": ["generic", "lowlatency"],
				"minKernelVersion": "4.4",
				"series": ["trusty", "xenial", "bionic"],
				"remoteServer": "https://livepatch.canonical.com",
				"tier": "updates",
				"enableByDefault": true
			}`,
		}},
	})
	addProduct(store.Product{
		ID:   "fips",
		Name: "fips",
		Resources: []store.ProductResource{{
			ResourceType: params.ResourceFIPS,
			Data: `{
				"additionalPackages": ["fips-initramfs", "libssl1.0.0", "libssl1.0.0-hmac", "linux-fips", "openssh-client", "openssh-client-hmac", "openssh-server", "openssh-server-hmac", "openssl",  "strongswan", "strongswan-hmac"],
				"aptURL": "https://esm.ubuntu.com/fips",
				"aptKey": "A166877412DAC26E73CEBF3FF6C280178D13028C",
				"series": ["xenial"],
				"suites": ["xenial"]
			}`,
		}, {
			ResourceType: params.ResourceFIPSUpdates,
			Data: `{
				"additionalPackages": ["fips-initramfs", "libssl1.0.0", "libssl1.0.0-hmac", "linux-fips", "openssh-client", "openssh-client-hmac", "openssh-server", "openssh-server-hmac", "openssl",  "strongswan", "strongswan-hmac"],
				"aptURL": "https://esm.ubuntu.com/fips-updates",
				"aptKey": "A166877412DAC26E73CEBF3FF6C280178D13028C",
				"series": ["xenial"],
				"suites": ["xenial"]
			}`,
		}},
	})
	addProduct(store.Product{
		ID:   "cc",
		Name: "cc",
		Resources: []store.ProductResource{{
			ResourceType: params.ResourceCC,
			Data: `{
				"additionalPackages": ["ubuntu-commoncriteria"],
				"aptURL": "https://esm.ubuntu.com/cc",
				"aptKey": "A166877412DAC26E73CEBF3FF6C280178D13028C",
				"architectures": ["x86_64", "ppc64le", "s390x"],
				"series": ["xenial"],
				"suites": ["xenial"]
			}`,
		}},
	})
	for _, level := range []string{"essential", "standard", "advanced"} {
		addProduct(store.Product{
			ID:   "support-" + level,
			Name: "support-" + level,
			Resources: []store.ProductResource{{
				ResourceType: params.ResourceSupport,
				Data: fmt.Sprintf(`{
					"supportLevel": %q
				}`,
					level),
			}},
		})
	}
	addProduct(store.Product{
		ID:   DefaultProductID,
		Name: DefaultProductID,
		Resources: []store.ProductResource{{
			ResourceType: params.ResourceESM,
			Data: `{
				"aptURL": "https://esm.ubuntu.com",
				"aptKey": "67C7A026",
				"series": ["precise", "trusty"],
				"suites": ["precise", "trusty-security", "trusty-updates"],
				"enableByDefault": true
			}`,
		}, {
			ResourceType: params.ResourceLivepatch,
			Data: `{
				"remoteServer": "https://livepatch.canonical.com",
				"architectures": ["x86_64"],
				"kernelFlavors": ["generic", "lowlatency"],
				"minKernelVersion": "4.4",
				"series": ["trusty", "xenial", "bionic"],
				"tier": "stable",
				"enableByDefault": true
			}`,
		}, {
			ResourceType: params.ResourceSupport,
			Data: `{
				"supportLevel": "community",
				"enableByDefault": true
			}`,
		}},
	})
	addProduct(store.Product{
		ID:   "aws",
		Name: "aws",
		Resources: []store.ProductResource{{
			ResourceType: params.ResourceESM,
			Data: `{
				"aptURL": "https://esm.ubuntu.com",
				"aptKey": "67C7A026",
				"series": ["precise", "trusty"],
				"suites": ["precise", "trusty-security", "trusty-updates"],
				"enableByDefault": true
			}`,
		}, {
			ResourceType: params.ResourceLivepatch,
			Data: `{
				"remoteServer": "https://livepatch.canonical.com",
				"architectures": ["x86_64"],
				"kernelFlavors": ["aws", "generic", "lowlatency"],
				"minKernelVersion": "4.4",
				"series": ["trusty", "xenial", "bionic"],
				"tier": "stable",
				"enableByDefault": true
			}`,
		}, {
			ResourceType: params.ResourceSupport,
			Data: `{
				"supportLevel": "essential",
				"enableByDefault": true
			}`,
		}},
	})
	return err
}
