package token

import (
	"context"
	"crypto/rand"
	"crypto/sha256"
	"database/sql"
	"encoding/base64"
	"strings"

	"github.com/btcsuite/btcutil/base58"
	errgo "gopkg.in/errgo.v1"
	"gopkg.in/macaroon-bakery.v2/bakery"
	macaroon "gopkg.in/macaroon.v2"
)

const (
	accountIDType     = 'a'
	contractIDType    = 'c'
	contractTokenType = 'C'
	machineTokenType  = 'm'
)

// A ContractToken is a token that is used to authenticate a particular
// contract.
type ContractToken [16]byte

// NewContractToken creates a new random contract token.
func NewContractToken() (ContractToken, error) {
	var ct ContractToken
	_, err := rand.Read(ct[:])
	return ct, errgo.Mask(err)
}

var _ sql.Scanner = (*ContractToken)(nil)

// Scan implements sql.Scanner.
func (t *ContractToken) Scan(src interface{}) error {
	return scanArray(t[:], src, "contract token")
}

// MustNewContractToken is like NewContractToken except that it will
// panic when NewContractToken returns an error.
func MustNewContractToken() ContractToken {
	ct, err := NewContractToken()
	if err != nil {
		panic(err)
	}
	return ct
}

// ParseContractToken parses the given string as a contract token.
func ParseContractToken(s string) (ContractToken, error) {
	if s == "" || s[0] != contractTokenType {
		return ContractToken{}, errgo.New("invalid token")
	}
	var ct ContractToken
	body, version, err := base58.CheckDecode(s[1:])
	if err != nil {
		return ct, errgo.Notef(err, "invalid token")
	}
	if version != 0 || len(body) != len(ct) {
		return ct, errgo.Notef(err, "invalid token")
	}
	copy(ct[:], body)
	return ct, nil
}

// String returns a string representation of the contract token.
func (t ContractToken) String() string {
	var b strings.Builder
	b.WriteByte(contractTokenType)
	b.WriteString(base58.CheckEncode(t[:], 0))
	return b.String()
}

// A MachineTokenBody contains the body of a MachineToken.
type MachineTokenBody struct {
	MachineID  string
	ContractID ContractID
}

// MarshalBinary encodes the MachineTokenBody as a byte slice.
// The data is encoded as:
//      1 byte - version (0)
//     32 byte - contract ID
//      x byte - machine ID
func (b MachineTokenBody) MarshalBinary() ([]byte, error) {
	body := make([]byte, 0, 1+len(b.ContractID)+len(b.MachineID))
	body = append(body, 0)
	body = append(body, b.ContractID[:]...)
	body = append(body, b.MachineID...)
	return body, nil
}

// UnmarshalBinary decodes version 0 body bytes into the MachineTokenBody.
func (b *MachineTokenBody) UnmarshalBinary(buf []byte) error {
	// Check minimum length and version.
	if len(buf) < 1+1+len(b.ContractID) || buf[0] != 0 {
		return errgo.New("invalid machine token")
	}
	copy(b.ContractID[:], buf[1:1+len(b.ContractID)])
	b.MachineID = string(buf[1+len(b.ContractID):])
	return nil
}

// IsZero reports whether this MachineTokenBody is the zero value.
func (b MachineTokenBody) IsZero() bool {
	return b.MachineID == "" && b.ContractID.IsZero()
}

type MachineToken []byte

// ParseMachineToken parses the given string as a machine token.
func ParseMachineToken(s string) (MachineToken, error) {
	if s == "" || s[0] != machineTokenType {
		return nil, errgo.New("invalid token")
	}
	b := make([]byte, base64.RawURLEncoding.DecodedLen(len(s[1:])))
	if _, err := base64.RawURLEncoding.Decode(b, []byte(s[1:])); err != nil {
		return nil, errgo.Notef(err, "invalid token")
	}
	return MachineToken(b), nil
}

// String returns a string representation of the machine token.
func (t MachineToken) String() string {
	b := make([]byte, 1+base64.RawURLEncoding.EncodedLen(len(t)))
	b[0] = machineTokenType
	base64.RawURLEncoding.Encode(b[1:], t)
	return string(b)
}

// MachineTokenCodec is a MachineTokenEncoder/Decoder that wraps the
// machine token in a macaroon.
type MachineTokenCodec struct {
	oven *bakery.Oven
}

// NewMachineTokenCodec creates a new MachineTokenCodec using the given
// shared secret.
func NewMachineTokenCodec(secret []byte) MachineTokenCodec {
	rootKeys := rootKeyStore{secret}
	return MachineTokenCodec{
		oven: bakery.NewOven(bakery.OvenParams{
			RootKeyStoreForOps: func([]bakery.Op) bakery.RootKeyStore {
				return rootKeys
			},
		}),
	}
}

// Encode implements MachineTokenEncoder.Encode.
func (b MachineTokenCodec) Encode(body MachineTokenBody) (MachineToken, error) {
	tok, err := body.MarshalBinary()
	if err != nil {
		return MachineToken{}, errgo.Mask(err)
	}
	m, err := b.oven.NewMacaroon(context.Background(), bakery.Version3, nil,
		bakery.Op{
			Entity: "mt-" + string(tok),
			Action: "*",
		},
	)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	data, err := macaroon.Slice{m.M()}.MarshalBinary()
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return data, nil
}

// Decode implements MachineTokenEncoder.Decode.
func (b MachineTokenCodec) Decode(t MachineToken) (MachineTokenBody, error) {
	var ms macaroon.Slice
	if err := ms.UnmarshalBinary(t); err != nil {
		return MachineTokenBody{}, errgo.Mask(err)
	}
	ops, conds, err := b.oven.VerifyMacaroon(context.Background(), ms)
	if err != nil {
		return MachineTokenBody{}, errgo.Mask(err)
	}
	if len(conds) > 0 {
		return MachineTokenBody{}, errgo.Newf("unrecognized caveats")
	}
	var r MachineTokenBody
	for _, op := range ops {
		if !strings.HasPrefix(op.Entity, "mt-") {
			continue
		}
		if err := r.UnmarshalBinary([]byte(op.Entity[3:])); err != nil {
			return MachineTokenBody{}, errgo.Notef(err, "bad machine token operation")
		}
	}
	if len(r.MachineID) == 0 {
		return MachineTokenBody{}, errgo.Newf("machine token macaroon not authorized")
	}
	return r, nil
}

// A ContractID is a unique identifier for a contract.
type ContractID [32]byte

var _ sql.Scanner = (*ContractID)(nil)

// NewContractID creates a new random contract ID.
func NewContractID() (ContractID, error) {
	var cid ContractID
	_, err := rand.Read(cid[1:])
	return cid, errgo.Mask(err)
}

// MustNewContractID is like NewContractID except that it will panic when
// NewContractID returns an error.
func MustNewContractID() ContractID {
	cid, err := NewContractID()
	if err != nil {
		panic(err)
	}
	return cid
}

var contractIDEncodedLen = 1 + base64.RawURLEncoding.EncodedLen(len(ContractID{}))

// ParseContractID parses the given string as a contract ID.
func ParseContractID(s string) (ContractID, error) {
	if len(s) != contractIDEncodedLen || s[0] != contractIDType {
		return ContractID{}, errgo.New("invalid id")
	}
	var cid ContractID
	if _, err := base64.RawURLEncoding.Decode(cid[:], []byte(s[1:])); err != nil || cid[0] != 0 {
		return ContractID{}, errgo.Notef(err, "invalid id")
	}
	return cid, nil
}

// IsZero reports whether the contract ID is the zero value.
func (t ContractID) IsZero() bool {
	for i := range t {
		if t[i] != 0 {
			return false
		}
	}
	return true
}

// Scan implements sql.Scanner.
func (t *ContractID) Scan(src interface{}) error {
	return scanArray(t[:], src, "contract id")
}

// String returns a string representation of the contract ID.
func (t ContractID) String() string {
	b := make([]byte, contractIDEncodedLen)
	b[0] = contractIDType
	base64.RawURLEncoding.Encode(b[1:], t[:])
	return string(b)
}

var contractIDStdEncodedLen = base64.StdEncoding.EncodedLen(len(ContractID{}))

// MarshalText implements the TextMarshaler interface.
func (t ContractID) MarshalText() ([]byte, error) {
	return []byte(t.String()), nil
}

// MarshalText implements the TextUnmarshaler interface.
func (t *ContractID) UnmarshalText(data []byte) error {
	cid, err := ParseContractID(string(data))
	if err != nil {
		return err
	}
	copy(t[:], cid[:])
	return nil
}

// An AccountID is a unique identifier for an account.
type AccountID [32]byte

const (
	userNamespace = iota + 1
	awsNamespace
	numNamespaces
)

var accountNamespaces = [numNamespaces][]byte{
	userNamespace: {0x37, 0xfd, 0xe7, 0xa0, 0x9c, 0xf2, 0x7d, 0x5b, 0xe7, 0x5f, 0x91, 0x03, 0xbc, 0x0c, 0x7a, 0x70, 0xea, 0x99, 0x62, 0xf8, 0x70, 0x06, 0xcb, 0x11, 0x00, 0x30, 0x0e, 0xb4, 0x68, 0x59, 0x1f, 0xad},
	awsNamespace:  {0x13, 0x81, 0x23, 0xd9, 0x7e, 0x2b, 0x3b, 0x18, 0x55, 0xfa, 0x23, 0x1a, 0xf6, 0x2f, 0x37, 0xf6, 0x16, 0x75, 0x96, 0xe6, 0x45, 0x7b, 0x8a, 0x32, 0x6b, 0x96, 0x6a, 0x68, 0xab, 0x85, 0x2f, 0x42},
}

// nsAccountID creates a new AccountID within the given namespace.
func nsAccountID(ns int, data []byte) AccountID {
	h := sha256.New()
	h.Write(accountNamespaces[ns])
	h.Write(data)
	var aid AccountID
	h.Sum(aid[:0])
	// Mark this type of account ID.
	aid[0] = byte(ns)
	return aid
}

// AccountIDForUserID creates a new new account ID for the given user ID.
// The account ID will always be the same for the same user ID.
func AccountIDForUserID(uid string) AccountID {
	return nsAccountID(userNamespace, []byte(uid))
}

// AccountIDForAWSAccountID creates a new new account ID for the given
// AWS account ID. The account ID will always be the same for the same
// AWS account ID.
func AccountIDForAWSAccountID(accID string) AccountID {
	return nsAccountID(awsNamespace, []byte(accID))
}

// NewAccountID creates a new random account ID.
func NewAccountID() (AccountID, error) {
	var aid AccountID
	_, err := rand.Read(aid[1:])
	return aid, errgo.Mask(err)
}

// MustNewAccountID is like NewAccountID except that it will panic when
// NewAccountID returns an error.
func MustNewAccountID() AccountID {
	aid, err := NewAccountID()
	if err != nil {
		panic(err)
	}
	return aid
}

var accountIDEncodedLen = 1 + base64.RawURLEncoding.EncodedLen(len(AccountID{}))

// ParseAccountID parses the given string as a contract ID.
func ParseAccountID(s string) (AccountID, error) {
	if len(s) != accountIDEncodedLen || s[0] != accountIDType {
		return AccountID{}, errgo.New("invalid id")
	}
	var aid AccountID
	if _, err := base64.RawURLEncoding.Decode(aid[:], []byte(s[1:])); err != nil || aid[0] > 1 {
		return AccountID{}, errgo.Notef(err, "invalid id")
	}
	return aid, nil
}

// String returns a string representation of the contract ID.
func (t AccountID) String() string {
	b := make([]byte, accountIDEncodedLen)
	b[0] = accountIDType
	base64.RawURLEncoding.Encode(b[1:], t[:])
	return string(b)
}

// Scan implements sql.Scanner.
func (t *AccountID) Scan(src interface{}) error {
	return scanArray(t[:], src, "account id")
}

var accountIDStdEncodedLen = base64.StdEncoding.EncodedLen(len(AccountID{}))

// MarshalText implements the TextMarshaler interface.
func (t AccountID) MarshalText() ([]byte, error) {
	return []byte(t.String()), nil
}

// MarshalText implements the TextUnmarshaler interface.
func (t *AccountID) UnmarshalText(data []byte) error {
	aid, err := ParseAccountID(string(data))
	if err != nil {
		return err
	}
	copy(t[:], aid[:])
	return nil
}

func scanArray(dst []byte, src interface{}, what string) error {
	data, ok := src.([]byte)
	if !ok {
		return errgo.Newf("invalid type %T for %s", src, what)
	}
	if len(data) != len(dst) {
		return errgo.Newf("invalid length for %s (got %d want %d)", what, len(data), len(dst))
	}
	copy(dst, data)
	return nil
}

type rootKeyStore struct {
	key []byte
}

// Get implements bakery.Store.Get.
func (s rootKeyStore) Get(_ context.Context, id []byte) ([]byte, error) {
	if string(id) != "0" {
		return nil, bakery.ErrNotFound
	}
	return s.key, nil
}

// RootKey implements bakery.Store.RootKey by always returning the same root
// key.
func (s rootKeyStore) RootKey(context.Context) (rootKey, id []byte, err error) {
	return s.key, []byte("0"), nil
}
