package v1

import (
	"context"
	"encoding/base64"
	"time"

	aclstore "github.com/juju/aclstore/v2"
	"github.com/juju/zaputil/zapctx"
	"go.uber.org/zap"
	errgo "gopkg.in/errgo.v1"
	httprequest "gopkg.in/httprequest.v1"

	"github.com/CanonicalLtd/ua-contracts/internal/awsauth"
	"github.com/CanonicalLtd/ua-contracts/internal/kafkaauditor"
	"github.com/CanonicalLtd/ua-contracts/internal/saauth"
	"github.com/CanonicalLtd/ua-contracts/internal/ssoauth"
	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/storeutil"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
	"github.com/CanonicalLtd/ua-contracts/params"
)

type handler struct {
	p                 Params
	machineTokenCodec token.MachineTokenCodec
}

const (
	// contractTokenExpiryDuration holds how long before
	// an issued contract token expires.
	// TODO make this configurable.
	contractTokenExpiryDuration = 365 * 24 * time.Hour

	// machineTokenExpiryDuration holds how long before
	// an issued machine token expires.
	// TODO make this configurable.
	machineTokenExpiryDuration = 100 * 365 * 24 * time.Hour
)

// Params holds the parameters for a v1 API server.
type Params struct {
	ACLManager *aclstore.Manager
	AWSDecoder *awsauth.Decoder
	Logger     *zap.Logger
	Store      store.Store
	SSOAuth    *ssoauth.Authenticator
	// TokenSecret holds the secret key that's used to generate
	// client tokens.
	TokenSecret      []byte
	SalesforceClient storeutil.SalesforceClient
}

var reqServer = &httprequest.Server{
	ErrorMapper: errorMapper,
}

// Handlers returns a set of handlers for the v1 HTTP endpoints.
func Handlers(ctx context.Context, p Params) ([]httprequest.Handler, error) {
	if err := ensureACLs(ctx, p.ACLManager); err != nil {
		return nil, errgo.Mask(err)
	}
	return reqServer.Handlers(func(hp httprequest.Params, arg interface{}) (*handler, context.Context, error) {
		if p.Logger != nil {
			hp.Context = zapctx.WithLogger(hp.Context, p.Logger)
		}
		h := &handler{
			p:                 p,
			machineTokenCodec: token.NewMachineTokenCodec(p.TokenSecret),
		}
		ctx, err := h.authorize(hp, arg)
		if err != nil {
			return nil, nil, errgo.Mask(err, isAPIError)
		}

		ctx = zapctx.WithFields(
			ctx,
			zap.String("method", hp.Request.Method),
			zap.String("path", hp.Request.URL.Path),
		)
		zapctx.Info(ctx, "handling request")
		return h, ctx, nil
	}), nil
}

func isAPIError(err error) bool {
	_, ok := err.(params.ErrorCode)
	return ok
}

// GetContractToken creates a contract token and returns it. If a
// contract token exists for the contract then return that.
// TODO(mhilton) we should probably have a way to force the creation of a
// new token.
func (h *handler) GetContractToken(p httprequest.Params, req *params.GetContractTokenRequest) (*params.GetContractTokenResponse, error) {
	ctx := p.Context

	cid, err := token.ParseContractID(req.Contract)
	if err != nil {
		return nil, errgo.WithCausef(err, params.ErrContractNotFound, "")
	}
	contract, err := h.getContract(ctx, cid)
	if err != nil {
		return nil, errgo.Mask(err, errgo.Is(params.ErrContractNotFound), errgo.Is(params.ErrForbidden))
	}
	return h.getContractToken(ctx, contract)
}

func (h *handler) getContractToken(ctx context.Context, c store.FullContract) (*params.GetContractTokenResponse, error) {
	// Find any existing tokens
	tokens, err := h.p.Store.ContractTokens(ctx, c.Contract.ID)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	var contractToken token.ContractToken
	if len(tokens) > 0 {
		contractToken = tokens[0].ContractToken
	} else {
		contractToken = token.MustNewContractToken()
		if err := h.p.Store.AddContractToken(ctx, store.ContractTokenInfo{
			ContractToken: contractToken,
			ContractID:    c.Contract.ID,
			Expires:       time.Now().Add(contractTokenExpiryDuration),
		}); err != nil {
			return nil, errgo.Mask(err)
		}
	}

	info, err := h.contractInfo(ctx, c)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return &params.GetContractTokenResponse{
		ContractToken: contractToken.String(),
		ContractTokenInfo: params.ContractTokenInfo{
			ContractInfo: info,
		},
	}, nil
}

// GetContractTokenForAWSInstance creates, or retrieves, an appropriate
// account and contract for the specified aws instance, returning a
// contract token that can then be used to attach the machine.
func (h *handler) GetContractTokenForAWSInstance(p httprequest.Params, req *params.GetContractTokenForAWSInstanceRequest) (*params.GetContractTokenResponse, error) {
	ctx := p.Context

	doc := awsDocFromContext(ctx)
	c, err := storeutil.GetAWSContract(ctx, h.p.Store, doc)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return h.getContractToken(ctx, c)
}

// GetContract gets information on a contract.
func (h *handler) GetContract(p httprequest.Params, req *params.GetContractRequest) (*params.AccountContractInfo, error) {
	ctx := p.Context

	cid, err := token.ParseContractID(req.Contract)
	if err != nil {
		return nil, errgo.WithCausef(err, params.ErrContractNotFound, "")
	}
	contract, err := h.getContract(ctx, cid)
	if err != nil {
		return nil, errgo.Mask(err, errgo.Is(params.ErrContractNotFound), errgo.Is(params.ErrForbidden))
	}
	info, err := h.contractInfo(ctx, contract)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return &params.AccountContractInfo{
		AccountInfo:  accountInfo(contract.Account),
		ContractInfo: info,
	}, nil
}

// AddContractMachine associates a machine with a contract and
// returns a machine token that allows the machine to request
// resources available through that contract.
func (h *handler) AddContractMachine(p httprequest.Params, req *params.AddContractMachineRequest) (*params.AddContractMachineResponse, error) {
	ctx := p.Context
	contractID := contractIDFromContext(ctx)
	if contractID.IsZero() {
		panic("no contract ID found in context")
	}
	c, err := h.p.Store.Contract(ctx, contractID)
	if err != nil {
		// Note that the contract *should* be there because
		// if a contract is removed, its contract token
		// should be removed too. So this is always
		// an internal server error.
		return nil, errgo.Mask(err)
	}

	// TODO check that the machine is valid for the given contract.

	body := req.Body
	if err := validateAddContractMachineBody(body); err != nil {
		return nil, errgo.Mask(err, isAPIError)
	}

	mt, err := h.machineTokenCodec.Encode(token.MachineTokenBody{
		MachineID:  body.MachineId,
		ContractID: contractID,
	})
	if err != nil {
		return nil, errgo.Mask(err)
	}
	if err := h.p.Store.EnsureMachine(ctx, store.MachineInfo{
		MachineID:    body.MachineId,
		Architecture: body.Architecture,
		OS:           body.Os,
	}); err != nil {
		return nil, errgo.Mask(err)
	}
	if err := h.p.Store.AddContractMachine(ctx, contractID, body.MachineId); err != nil {
		return nil, errgo.Mask(err)
	}
	info, err := h.contractInfo(ctx, c)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return &params.AddContractMachineResponse{
		MachineToken: mt.String(),
		MachineTokenInfo: params.MachineTokenInfo{
			MachineId:    body.MachineId,
			ContractInfo: info,
			AccountInfo:  accountInfo(c.Account),
		},
	}, nil
}

// AddProduct adds a product.
func (h *handler) AddProduct(p httprequest.Params, req *params.AddProductRequest) error {
	req.Body.Id = req.ProductID

	// TODO verify that the product parameters produce a valid
	// set of entitlements.

	p1 := storeProduct(req.Body)
	if req.Update {
		if err := h.p.Store.UpdateProduct(p.Context, p1); err != nil {
			if errgo.Cause(err) == store.ErrProductNotFound {
				return errgo.WithCausef(err, params.ErrNotFound, "")
			}
			return errgo.Mask(err)
		}
	} else {
		if err := h.p.Store.AddProduct(p.Context, p1); err != nil {
			if errgo.Cause(err) == store.ErrDuplicateProduct {
				return errgo.WithCausef(err, params.ErrAlreadyExists, "")
			}
			return errgo.Mask(err)
		}
	}
	return nil
}

// GetProduct returns information on one product.
func (h *handler) GetProduct(p httprequest.Params, req *params.GetProductRequest) (*params.Product, error) {
	prod, err := h.p.Store.Product(p.Context, req.ProductID)
	if err != nil {
		if errgo.Cause(err) == store.ErrProductNotFound {
			return nil, errgo.WithCausef(err, params.ErrNotFound, "")
		}
		return nil, errgo.Mask(err)
	}
	prod1 := paramsProduct(prod)
	return &prod1, nil
}

// GetProducts gets information on all available products.
func (h *handler) GetProducts(p httprequest.Params, req *params.GetProductsRequest) (*params.GetProductsResponse, error) {
	products, err := h.p.Store.AllProducts(p.Context)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	products1 := make([]params.Product, len(products))
	for i, product := range products {
		products1[i] = paramsProduct(product)
	}
	return &params.GetProductsResponse{
		Products: products1,
	}, nil
}

// GetResourceMachineAccess returns information on a specific resource
// for a machine, including a resource token that acts as an authorization
// token when accessing the resource-specific service.
func (h *handler) GetResourceMachineAccess(p httprequest.Params, req *params.GetResourceMachineAccessRequest) (*params.GetResourceMachineAccessResponse, error) {
	ctx := p.Context
	// TODO: we could use a resource specific token instead of using the machine token.
	tok := machineTokenFromContext(ctx)
	if tok.body.IsZero() {
		// This should never happen because the authorization code
		// should have filled this in for us.
		return nil, errgo.Newf("no machine token found")
	}
	contract, err := h.getContract(ctx, tok.body.ContractID)
	if err != nil {
		return nil, errgo.Mask(err, errgo.Is(params.ErrContractNotFound))
	}

	// TODO check that the contract hasn't yet expired?

	entitlements, err := h.contractEntitlements(ctx, contract)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	ent, ok := entitlements[req.Resource]
	if !ok {
		return nil, errgo.WithCausef(nil, params.ErrResourceForbidden, "contract %q does not allow access to %q", tok.body.ContractID, req.Resource)
	}
	auditor, ok := h.p.Store.(kafkaauditor.Auditor)
	if ok {
		auditor.Audit(ctx, time.Now().UTC().Truncate(time.Millisecond), &kafkaauditor.AuditResourceToken{
			ContractID: tok.body.ContractID,
			MachineID:  tok.body.MachineID,
			Resource:   req.Resource,
		})
	}
	return &params.GetResourceMachineAccessResponse{
		Entitlement:   ent,
		ResourceToken: tok.token.String(),
	}, nil
}

// GetResourceTokenInfo returns information on a specific resource for a machine.
// This end point is called by a resource server.
func (h *handler) GetResourceTokenInfo(p httprequest.Params, req *params.GetResourceTokenInfoRequest) (*params.GetResourceTokenInfoResponse, error) {
	ctx := p.Context
	mt, err := token.ParseMachineToken(req.Token)
	if err != nil {
		return nil, errgo.WithCausef(err, params.ErrUnauthorized, "")
	}
	body, err := h.machineTokenCodec.Decode(mt)
	if err != nil {
		return nil, errgo.WithCausef(err, params.ErrUnauthorized, "")
	}
	// Check that the machine is still attached to the  contract.
	ok, err := h.p.Store.MachineHasContract(ctx, body.ContractID, body.MachineID)
	if err != nil {
		return nil, errgo.WithCausef(err, params.ErrUnauthorized, "")
	}
	if !ok {
		return nil, forbiddenErrorf(nil, "machine %q no longer attached to contract %q", body.MachineID, body.ContractID)
	}
	contract, err := h.p.Store.Contract(ctx, body.ContractID)
	if err != nil {
		if errgo.Cause(err) == store.ErrContractNotFound {
			return nil, errgo.WithCausef(err, params.ErrContractNotFound, "")
		}
		return nil, errgo.Mask(err)
	}
	entitlements, err := h.contractEntitlements(ctx, contract)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	ent, ok := entitlements[req.Resource]
	if !ok {
		return nil, errgo.WithCausef(nil, params.ErrResourceForbidden, "")
	}
	return &params.GetResourceTokenInfoResponse{
		Affordances: getAffordances(ent),
		MachineId:   body.MachineID,
		ContractId:  body.ContractID.String(),
	}, nil
}

// GetCanonicalSSOMacaroon returns a newly minted macaroon with a
// third-party caveat address to canonical SSO. Once discharged this
// caveat can be used to authenticate requests.
func (h *handler) GetCanonicalSSOMacaroon(p httprequest.Params, req *params.GetCanonicalSSOMacaroonRequest) (*params.GetCanonicalSSOMacaroonResponse, error) {
	m, err := h.p.SSOAuth.Macaroon(p.Context)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	buf, err := m.M().MarshalBinary()
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return &params.GetCanonicalSSOMacaroonResponse{
		Macaroon: base64.RawURLEncoding.EncodeToString(buf),
	}, nil
}

// GetAccounts gets the available accounts for the logged in user.
func (h *handler) GetAccounts(p httprequest.Params, req *params.GetAccountsRequest) (*params.GetAccountsResponse, error) {
	ctx := p.Context
	accounts, err := storeutil.GetAccounts(ctx, h.p.SalesforceClient, h.p.Store, userFromContext(ctx))
	if err != nil {
		return nil, errgo.Mask(err)
	}
	respAccounts := make([]params.AccountInfo, len(accounts))
	for i, acc := range accounts {
		respAccounts[i] = accountInfo(acc.Account)
	}
	return &params.GetAccountsResponse{
		Accounts: respAccounts,
	}, nil
}

// GetContractConfig returns per-contract resource configuration data.
func (h *handler) GetContractConfig(p httprequest.Params, req *params.GetContractConfigRequest) (*params.GetContractConfigResponse, error) {
	ctx := p.Context
	cid, err := token.ParseContractID(req.Contract)
	if err != nil {
		return nil, errgo.WithCausef(err, params.ErrContractNotFound, "")
	}
	contract, err := h.p.Store.Contract(ctx, cid)
	if err != nil {
		if errgo.Cause(err) == store.ErrContractNotFound {
			return nil, errgo.WithCausef(nil, params.ErrContractNotFound, "")
		}
		return nil, errgo.Mask(err)
	}
	return &params.GetContractConfigResponse{
		Resources: paramsProductResources(contract.Config),
	}, nil
}

// ConfigureContract sets per-contract resource configuration data.
func (h *handler) ConfigureContract(p httprequest.Params, req *params.ConfigureContractRequest) error {
	ctx := p.Context
	cid, err := token.ParseContractID(req.Contract)
	if err != nil {
		return errgo.WithCausef(err, params.ErrContractNotFound, "")
	}
	contract, err := h.p.Store.Contract(ctx, cid)
	if err != nil {
		if errgo.Cause(err) == store.ErrContractNotFound {
			return errgo.WithCausef(nil, params.ErrContractNotFound, "")
		}
		return errgo.Mask(err)
	}
	rs, err := h.productResources(ctx, contract.Contract.ProductIDs)
	if err != nil {
		return errgo.Mask(err)
	}
	// Technically this isn't atomic, but that shouldn't matter too much, as products will
	// change infrequently and we're resilient when products change.
	for _, r := range req.Body.Resources {
		if !rs[r.ResourceType].CanOverride {
			return errgo.WithCausef(nil, params.ErrProductConflict, "cannot override resource %q", r.ResourceType)
		}
	}
	if err := h.p.Store.ConfigureContract(p.Context, cid, storeProductResources(req.Body.Resources)); err != nil {
		return errgo.Mask(err, errgo.Is(store.ErrProductConflict), errgo.Is(store.ErrContractNotFound))
	}
	return nil
}

// GetAccountContracts gets the available contracts for the requested account.
func (h *handler) GetAccountContracts(p httprequest.Params, req *params.GetAccountContractsRequest) (*params.GetContractsResponse, error) {
	ctx := p.Context

	aid, err := token.ParseAccountID(req.Account)
	if err != nil {
		return nil, errgo.WithCausef(err, params.ErrNotFound, "account %q not found", req.Account)
	}

	contracts, err := storeutil.GetContractsForAccount(ctx, h.p.SalesforceClient, h.p.Store, aid)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	respContracts := make([]params.AccountContractInfo, len(contracts))
	for i, ct := range contracts {
		info, err := h.contractInfo(ctx, ct)
		if err != nil {
			return nil, errgo.Mask(err)
		}
		respContracts[i] = params.AccountContractInfo{
			AccountInfo:  accountInfo(ct.Account),
			ContractInfo: info,
		}
	}
	return &params.GetContractsResponse{
		Contracts: respContracts,
	}, nil
}

// GetContractMachine retrieves the information about a machine under a
// contract.
func (h *handler) GetContractMachine(p httprequest.Params, req *params.GetContractMachineRequest) (*params.AddContractMachineResponse, error) {
	ctx := p.Context

	cid, err := token.ParseContractID(req.Contract)
	if err != nil {
		return nil, errgo.WithCausef(err, params.ErrContractNotFound, "")
	}
	c, err := h.getContract(ctx, cid)
	if err != nil {
		return nil, errgo.Mask(err, errgo.Is(params.ErrContractNotFound))
	}
	info, err := h.contractInfo(ctx, c)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	// TODO(mhilton) generate a new machine token if we were
	// authorized by another method?
	var mt string
	tok := machineTokenFromContext(ctx)
	if !tok.body.IsZero() {
		mt = tok.token.String()
	}
	return &params.AddContractMachineResponse{
		MachineToken: mt,
		MachineTokenInfo: params.MachineTokenInfo{
			MachineId:    req.Machine,
			ContractInfo: info,
			AccountInfo:  accountInfo(c.Account),
		},
	}, nil
}

// RevokeContractMachine revokes a machine from a contract. Subsequent
// requests to access resources associated with the contract will fail
// until the machine is re-attached to the contract.
func (h *handler) RevokeContractMachine(p httprequest.Params, req *params.RevokeContractMachineRequest) (*params.MachineTokenInfo, error) {
	ctx := p.Context

	cid, err := token.ParseContractID(req.Contract)
	if err != nil {
		return nil, errgo.WithCausef(err, params.ErrContractNotFound, "")
	}
	c, err := h.getContract(ctx, cid)
	if err != nil {
		return nil, errgo.Mask(err, errgo.Is(params.ErrContractNotFound))
	}
	if err := h.p.Store.RemoveContractMachine(ctx, cid, req.Machine); err != nil {
		return nil, errgo.Mask(err)
	}
	info, err := h.contractInfo(ctx, c)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return &params.MachineTokenInfo{
		AccountInfo:  accountInfo(c.Account),
		ContractInfo: info,
		MachineId:    req.Machine,
	}, nil
}

func validateAddContractMachineBody(body params.AddContractMachineBody) error {
	if body.MachineId == "" {
		// TODO better check whether it looks like a valid machine ID?
		return badRequestf(nil, "no machine ID")
	}
	if body.Architecture == "" {
		return badRequestf(nil, "no machine architecture supplied")
	}
	if body.Os.Distribution == "" {
		return badRequestf(nil, "no OS distribution supplied")
	}
	if body.Os.Kernel == "" {
		return badRequestf(nil, "no OS kernel supplied")
	}
	if body.Os.Release == "" {
		return badRequestf(nil, "no OS release supplied")
	}
	if body.Os.Series == "" {
		return badRequestf(nil, "no OS series supplied")
	}
	if body.Os.Type == "" {
		return badRequestf(nil, "no OS type supplied")
	}
	return nil
}

// getContract reads the contract with the given ID from the store.
func (h *handler) getContract(ctx context.Context, id token.ContractID) (store.FullContract, error) {
	contract, err := h.p.Store.Contract(ctx, id)
	if err != nil {
		if errgo.Cause(err) == store.ErrContractNotFound {
			return store.FullContract{}, errgo.WithCausef(nil, params.ErrContractNotFound, "")
		}
		return store.FullContract{}, errgo.Mask(err)
	}
	return contract, nil
}

// SetServiceAccountPassword sets the password for a service account.
func (h *handler) SetServiceAccountPassword(p httprequest.Params, req *params.SetServiceAccountPasswordRequest) error {
	ctx := p.Context
	if req.Account == "" {
		return badRequestf(nil, "no account")
	}
	if req.Body.Password == "" {
		return badRequestf(nil, "no password")
	}
	return errgo.Mask(saauth.SetPassword(ctx, h.p.Store, req.Account, req.Body.Password))
}

// UpdateAccountContracts updates account's contracts by synching its state with Salesforce.
func (h *handler) UpdateAccountContracts(p httprequest.Params, req *params.UpdateAccountContractsRequest) (*params.GetContractsResponse, error) {
	ctx := p.Context
	contracts, err := storeutil.UpdateContractsForSalesforceAccount(ctx, h.p.SalesforceClient, h.p.Store, req.Sfid)
	if err != nil {
		if errgo.Cause(err) == storeutil.ErrSalesforceAccountNotFound {
			return nil, errgo.WithCausef(err, params.ErrNotFound, "cannot find a salesforce account with id %q", req.Sfid)
		}
		return nil, errgo.Mask(err)
	}
	respContracts := make([]params.AccountContractInfo, len(contracts))
	for i, ct := range contracts {
		info, err := h.contractInfo(ctx, ct)
		if err != nil {
			return nil, errgo.Mask(err)
		}
		respContracts[i] = params.AccountContractInfo{
			AccountInfo:  accountInfo(ct.Account),
			ContractInfo: info,
		}
	}
	return &params.GetContractsResponse{
		Contracts: respContracts,
	}, nil
}
