package v1_test

import (
	"context"
	"encoding/base64"
	"net/http/httptest"
	"testing"
	"time"

	"github.com/CanonicalLtd/salesforce"
	sfparams "github.com/CanonicalLtd/salesforce/params"
	qt "github.com/frankban/quicktest"
	"github.com/google/go-cmp/cmp"
	"github.com/google/go-cmp/cmp/cmpopts"
	aclstore "github.com/juju/aclstore/v2"
	"github.com/juju/simplekv/memsimplekv"
	"github.com/juju/zaputil/zapctx"
	"github.com/julienschmidt/httprouter"
	"go.uber.org/zap/zaptest"
	"golang.org/x/crypto/argon2"
	errgo "gopkg.in/errgo.v1"
	httprequest "gopkg.in/httprequest.v1"
	"gopkg.in/macaroon-bakery.v2/bakery"
	macaroon "gopkg.in/macaroon.v2"

	"github.com/CanonicalLtd/ua-contracts/contractclient"
	"github.com/CanonicalLtd/ua-contracts/internal/memstore"
	"github.com/CanonicalLtd/ua-contracts/internal/saauth"
	"github.com/CanonicalLtd/ua-contracts/internal/ssoauth"
	"github.com/CanonicalLtd/ua-contracts/internal/ssoauth/ssoauthtest"
	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/storeutil"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
	v1 "github.com/CanonicalLtd/ua-contracts/internal/v1"
	"github.com/CanonicalLtd/ua-contracts/params"
)

const testAccountID = "aAYQ_ZZc4w8vp4fnYo6HavRb6N4N_g_HrChY-b7r_DKo"

var (
	ssoDischarger = new(ssoauthtest.Discharger)
	entriesEqual  = qt.CmpEquals(cmpopts.IgnoreTypes(time.Time{}), cmpopts.IgnoreFields(params.AuditLogEntry{}, "Data"))
)

var testAccount = ssoauth.Account{
	OpenID:      "AAAAAA",
	Username:    "test",
	DisplayName: "test-user",
	Email:       "test@example.com",
	IsVerified:  true,
}

func TestGetContractToken(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	client := contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: getMacaroonAuthorizer(ctx, c, u, &testAccount),
	})

	resp, err := client.GetAccounts(ctx, &params.GetAccountsRequest{})
	c.Assert(err, qt.Equals, nil)
	c.Assert(resp.Accounts, qt.HasLen, 1)

	resp2, err := client.GetAccountContracts(ctx, &params.GetAccountContractsRequest{
		Account: resp.Accounts[0].Id,
	})
	c.Assert(err, qt.Equals, nil, qt.Commentf("error: %v", err))
	c.Assert(resp2.Contracts, qt.HasLen, 1)

	resp3, err := client.GetContractToken(ctx, &params.GetContractTokenRequest{
		Contract: resp2.Contracts[0].ContractInfo.Id,
	})
	c.Assert(err, qt.Equals, nil)

	// Check that "Adding" a second token returns the same token as before.
	resp4, err := client.GetContractToken(ctx, &params.GetContractTokenRequest{
		Contract: resp2.Contracts[0].ContractInfo.Id,
	})
	c.Assert(err, qt.Equals, nil)
	c.Assert(resp3, qt.DeepEquals, resp4)

	// Check that the token has been stored correctly.
	// TestAddContractMachine checks that it can actually be
	// used appropriately.

	ct, err := token.ParseContractToken(resp3.ContractToken)
	c.Assert(err, qt.Equals, nil)
	info, err := srvStore.ContractToken(ctx, ct)
	c.Assert(err, qt.Equals, nil)
	c.Assert(info.ContractID.String(), qt.Equals, resp2.Contracts[0].ContractInfo.Id)
	if info.Expires.Before(time.Now().Add(time.Second)) {
		c.Fatalf("expiry time too early")
	}
}

func TestAddContractMachine(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)

	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	cid, ct := getContractToken(ctx, c, u)
	// Use a client that authorizes with the newly acquired contract token.
	client := contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(ct),
	})
	resp, err := client.AddContractMachine(ctx, &params.AddContractMachineRequest{
		Body: params.AddContractMachineBody{
			Architecture: "amd64",
			MachineId:    "somemachine",
			Os: params.OS{
				Distribution: "ubuntu",
				Kernel:       "0.01",
				Release:      "aRelease",
				Series:       "someSeries",
				Type:         "osType",
			},
		},
	})
	c.Assert(err, qt.Equals, nil)
	cmpEq := qt.CmpEquals(
		equateApproxTime(time.Second),
		cmpopts.IgnoreFields(params.AddContractMachineResponse{}, "MachineToken"),
	)
	c.Assert(resp, cmpEq, &params.AddContractMachineResponse{
		MachineToken: "checked below",
		MachineTokenInfo: params.MachineTokenInfo{
			MachineId: "somemachine",
			AccountInfo: params.AccountInfo{
				Id:        testAccountID,
				CreatedAt: time.Now(),
				Name:      "test@example.com",
			},
			ContractInfo: params.ContractInfo{
				Id:            cid,
				Name:          "test@example.com",
				CreatedAt:     time.Now(),
				EffectiveFrom: newTime(time.Now()),
				ResourceEntitlements: []params.Entitlement{
					params.EntitlementAptRepository{
						Type:     "cc",
						Entitled: true,
						Affordances: params.AffordancesAptRepository{
							Architectures: []string{"x86_64", "ppc64le", "s390x"},
							Series:        []string{"xenial"},
						},
						Directives: params.DirectivesAptRepository{
							AdditionalPackages: []string{"ubuntu-commoncriteria"},
							AptKey:             "A166877412DAC26E73CEBF3FF6C280178D13028C",
							AptURL:             "https://cc.ubuntu.com",
							Suites:             []string{"xenial-cc"},
						},
					},
					params.EntitlementAptRepository{
						Type:     "esm",
						Entitled: true,
						Affordances: params.AffordancesAptRepository{
							Series: []string{"precise", "trusty"},
						},
						Directives: params.DirectivesAptRepository{
							AptKey: "67C7A026",
							AptURL: "https://esm.ubuntu.com",
							Suites: []string{"precise", "trusty-security", "trusty-updates"},
						},
						Obligations: params.Obligations{EnableByDefault: true},
					},
					params.EntitlementAptRepository{
						Type:     "fips",
						Entitled: true,
						Affordances: params.AffordancesAptRepository{
							Series: []string{"xenial"},
						},
						Directives: params.DirectivesAptRepository{
							AdditionalPackages: []string{
								"fips-initramfs",
								"libssl1.0.0",
								"libssl1.0.0-hmac",
								"linux-fips",
								"openssh-client",
								"openssh-client-hmac",
								"openssh-server",
								"openssh-server-hmac",
								"openssl",
								"strongswan",
								"strongswan-hmac",
							},
							AptKey: "A166877412DAC26E73CEBF3FF6C280178D13028C",
							AptURL: "https://private-ppa.launchpad.net/ubuntu-advantage/fips",
							Suites: []string{"xenial"},
						},
					},
					params.EntitlementAptRepository{
						Type:     "fips-updates",
						Entitled: false,
					},
					params.EntitlementLivepatch{
						Type:     "livepatch",
						Entitled: true,
						Affordances: params.AffordancesLivepatch{
							Architectures:    []string{"x86_64"},
							KernelFlavors:    []string{"generic", "lowlatency"},
							MinKernelVersion: "4.4",
							Series:           []string{"xenial", "bionic"},
						},
						Directives: params.DirectivesLivepatch{},
					},
					params.EntitlementSupport{
						Type:        "support",
						Entitled:    true,
						Affordances: params.AffordancesSupport{SupportLevel: "community"},
					},
				},
			},
		},
	})
	machineToken := resp.MachineToken
	// Check that the machine token actually encodes valid contents.
	m, err := token.ParseMachineToken(machineToken)
	c.Assert(err, qt.Equals, nil)
	codec := token.NewMachineTokenCodec([]byte("highlysecret"))
	body, err := codec.Decode(m)
	c.Assert(err, qt.Equals, nil)
	c.Assert(body.MachineID, qt.Equals, "somemachine")
	c.Assert(body.ContractID.String(), qt.Equals, cid)
}

func TestGetContract(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)

	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	cid, _ := getContractToken(ctx, c, u)
	client := newBasicAuthClient(c, u)
	resp, err := client.GetContract(ctx, &params.GetContractRequest{
		Contract: cid,
	})
	c.Assert(err, qt.Equals, nil)

	cmpEq := qt.CmpEquals(equateApproxTime(time.Second))
	c.Assert(resp, cmpEq, &params.AccountContractInfo{
		AccountInfo: params.AccountInfo{
			CreatedAt: time.Now().UTC(),
			Id:        testAccountID,
			Name:      "test@example.com",
		},
		ContractInfo: params.ContractInfo{
			Id:            cid,
			Name:          "test@example.com",
			CreatedAt:     time.Now(),
			EffectiveFrom: newTime(time.Now()),
			ResourceEntitlements: []params.Entitlement{
				params.EntitlementAptRepository{
					Type:     "cc",
					Entitled: true,
					Affordances: params.AffordancesAptRepository{
						Architectures: []string{"x86_64", "ppc64le", "s390x"},
						Series:        []string{"xenial"},
					},
					Directives: params.DirectivesAptRepository{
						AdditionalPackages: []string{"ubuntu-commoncriteria"},
						AptKey:             "A166877412DAC26E73CEBF3FF6C280178D13028C",
						AptURL:             "https://cc.ubuntu.com",
						Suites:             []string{"xenial-cc"},
					},
				},
				params.EntitlementAptRepository{
					Type:     "esm",
					Entitled: true,
					Affordances: params.AffordancesAptRepository{
						Series: []string{"precise", "trusty"},
					},
					Directives: params.DirectivesAptRepository{
						AptKey: "67C7A026",
						AptURL: "https://esm.ubuntu.com",
						Suites: []string{"precise", "trusty-security", "trusty-updates"},
					},
					Obligations: params.Obligations{EnableByDefault: true},
				},
				params.EntitlementAptRepository{
					Type:     "fips",
					Entitled: true,
					Affordances: params.AffordancesAptRepository{
						Series: []string{"xenial"},
					},
					Directives: params.DirectivesAptRepository{
						AdditionalPackages: []string{
							"fips-initramfs",
							"libssl1.0.0",
							"libssl1.0.0-hmac",
							"linux-fips",
							"openssh-client",
							"openssh-client-hmac",
							"openssh-server",
							"openssh-server-hmac",
							"openssl",
							"strongswan",
							"strongswan-hmac",
						},
						AptKey: "A166877412DAC26E73CEBF3FF6C280178D13028C",
						AptURL: "https://private-ppa.launchpad.net/ubuntu-advantage/fips",
						Suites: []string{"xenial"},
					},
				},
				params.EntitlementAptRepository{
					Type:     "fips-updates",
					Entitled: false,
				},
				params.EntitlementLivepatch{
					Type:     "livepatch",
					Entitled: true,
					Affordances: params.AffordancesLivepatch{
						Architectures:    []string{"x86_64"},
						KernelFlavors:    []string{"generic", "lowlatency"},
						MinKernelVersion: "4.4",
						Series:           []string{"xenial", "bionic"},
					},
					Directives: params.DirectivesLivepatch{},
				},
				params.EntitlementSupport{
					Type:        "support",
					Entitled:    true,
					Affordances: params.AffordancesSupport{SupportLevel: "community"},
				},
			},
		},
	})
}

func TestConfigureContract(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	// Add a product that has a resource that can be configured.
	err := srvStore.AddProduct(ctx, store.Product{
		ID: "on-prem-livepatch",
		Resources: []store.ProductResource{{
			ResourceType: "esm",
			Data: `
				{
					"aptURL": "https://esm.ubuntu.com",
					"aptKey": "67C7A026",
					"series": ["precise", "trusty"],
					"enableByDefault": true
				}
			`,
		}, {
			ResourceType: "livepatch",
			CanOverride:  true,
			Data: `{
					"remoteServer":     "https://livepatch.ubuntu.com",
					"kernelFlavors":    ["lowlatency"],
					"minKernelVersion": "4.4",
					"enableByDefault":  true
			}`,
		}},
	})
	c.Assert(err, qt.Equals, nil)
	accountID := getAccountID(ctx, c, newTestAccountClient(ctx, c, u))
	// Add a contract that uses this product.
	contractID, err := srvStore.AddContract(ctx, store.Contract{
		AccountID:  accountID,
		ProductIDs: []string{"on-prem-livepatch"},
		Created:    time.Now(),
	})
	c.Assert(err, qt.Equals, nil)

	client := newBasicAuthClient(c, u)
	err = client.ConfigureContract(ctx, &params.ConfigureContractRequest{
		Contract: contractID.String(),
		Body: params.ConfigureContractBody{
			Resources: []params.ProductResource{{
				ResourceType: "livepatch",
				Data: map[string]interface{}{
					"remoteServer": "https://livepatch.onprem",
				},
			}},
		},
	})
	c.Assert(err, qt.Equals, nil)

	// Check that we can get the contract configuration.
	configResp, err := client.GetContractConfig(ctx, &params.GetContractConfigRequest{
		Contract: contractID.String(),
	})
	c.Assert(err, qt.Equals, nil)
	c.Assert(configResp, qt.DeepEquals, &params.GetContractConfigResponse{
		Resources: []params.ProductResource{{
			ResourceType: "livepatch",
			Data: map[string]interface{}{
				"remoteServer": "https://livepatch.onprem",
			},
		}},
	})

	ctResp, err := newTestAccountClient(ctx, c, u).GetContractToken(ctx, &params.GetContractTokenRequest{
		Contract: contractID.String(),
	})
	c.Assert(err, qt.Equals, nil)

	// Use a client that authorizes with the newly acquired contract token.
	client = contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(ctResp.ContractToken),
	})

	resp, err := client.AddContractMachine(ctx, &params.AddContractMachineRequest{
		Body: params.AddContractMachineBody{
			Architecture: "amd64",
			MachineId:    "somemachine",
			Os: params.OS{
				Distribution: "ubuntu",
				Kernel:       "0.01",
				Release:      "aRelease",
				Series:       "someSeries",
				Type:         "osType",
			},
		},
	})
	c.Assert(err, qt.Equals, nil)
	cmpEq := qt.CmpEquals(
		equateApproxTime(time.Second),
		cmpopts.IgnoreFields(params.AddContractMachineResponse{}, "MachineToken"),
	)
	c.Assert(resp, cmpEq, &params.AddContractMachineResponse{
		MachineToken: "checked below",
		MachineTokenInfo: params.MachineTokenInfo{
			MachineId: "somemachine",
			AccountInfo: params.AccountInfo{
				CreatedAt: time.Now(),
				Id:        accountID.String(),
				Name:      "test@example.com",
			},
			ContractInfo: params.ContractInfo{
				Id:        contractID.String(),
				CreatedAt: time.Now(),
				ResourceEntitlements: []params.Entitlement{
					params.EntitlementAptRepository{
						Type:     "cc",
						Entitled: false,
					},
					params.EntitlementAptRepository{
						Type:     "esm",
						Entitled: true,
						Affordances: params.AffordancesAptRepository{
							Series: []string{"precise", "trusty"},
						},
						Directives: params.DirectivesAptRepository{
							AptKey: "67C7A026",
							AptURL: "https://esm.ubuntu.com",
						},
						Obligations: params.Obligations{EnableByDefault: true},
					},
					params.EntitlementAptRepository{
						Type:     "fips",
						Entitled: false,
					},
					params.EntitlementAptRepository{
						Type:     "fips-updates",
						Entitled: false,
					},
					params.EntitlementLivepatch{
						Type:     "livepatch",
						Entitled: true,
						Affordances: params.AffordancesLivepatch{
							KernelFlavors:    []string{"lowlatency"},
							MinKernelVersion: "4.4",
						},
						Directives: params.DirectivesLivepatch{
							RemoteServer: "https://livepatch.onprem",
						},
						Obligations: params.Obligations{EnableByDefault: true},
					},
					params.EntitlementSupport{
						Type:     "support",
						Entitled: false,
					},
				},
			},
		},
	})
}

func TestGetResourceMachineAccess(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	_, token := getContractToken(ctx, c, u)
	// Use a client that authorizes with the newly acquired contract token
	// to acquire a machine token.
	client := contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(token),
	})
	resp1, err := client.AddContractMachine(ctx, &params.AddContractMachineRequest{
		Body: params.AddContractMachineBody{
			Architecture: "amd64",
			MachineId:    "somemachine",
			Os: params.OS{
				Distribution: "ubuntu",
				Kernel:       "0.01",
				Release:      "aRelease",
				Series:       "someSeries",
				Type:         "osType",
			},
		},
	})
	c.Assert(err, qt.Equals, nil)

	// Now we've got the machine token, we can actually
	// test GetResourceMachineAccess.

	// Use a client that authorizes with the newly acquired machine
	// token to acquire a resource token.
	client = contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(resp1.MachineToken),
	})

	resp2, err := client.GetResourceMachineAccess(ctx, &params.GetResourceMachineAccessRequest{
		Resource: "esm",
		Machine:  "somemachine",
	})
	c.Assert(err, qt.Equals, nil)
	c.Assert(resp2, qt.DeepEquals, &params.GetResourceMachineAccessResponse{
		ResourceToken: resp1.MachineToken,
		Entitlement: params.EntitlementAptRepository{
			Type:     "esm",
			Entitled: true,
			Affordances: params.AffordancesAptRepository{
				Series: []string{"precise", "trusty"},
			},
			Obligations: params.Obligations{
				EnableByDefault: true,
			},
			Directives: params.DirectivesAptRepository{
				AptKey: "67C7A026",
				AptURL: "https://esm.ubuntu.com",
				Suites: []string{"precise", "trusty-security", "trusty-updates"},
			},
		},
	})

	// If we try to acquire a resource token for a resource
	// we're not entitled to, we get an appropriate not-found error.
	_, err = client.GetResourceMachineAccess(ctx, &params.GetResourceMachineAccessRequest{
		Resource: "other",
		Machine:  "somemachine",
	})
	c.Assert(errgo.Cause(err), qt.Equals, params.ErrResourceForbidden)

	// If we use a client without the token, we get an unauthorized error.
	client = contractclient.New(contractclient.NewParams{
		BaseURL: u,
	})
	_, err = client.GetResourceMachineAccess(ctx, &params.GetResourceMachineAccessRequest{
		Resource: "esm",
		Machine:  "somemachine",
	})
	c.Assert(errgo.Cause(err), qt.Equals, params.ErrUnauthorized)
}

func TestGetResourceInfo(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	cid, token := getContractToken(ctx, c, u)
	// Use a client that authorizes with the newly acquired contract token
	// to acquire a machine token.
	client := contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(token),
	})
	resp1, err := client.AddContractMachine(ctx, &params.AddContractMachineRequest{
		Body: params.AddContractMachineBody{
			Architecture: "amd64",
			MachineId:    "somemachine",
			Os: params.OS{
				Distribution: "ubuntu",
				Kernel:       "0.01",
				Release:      "aRelease",
				Series:       "someSeries",
				Type:         "osType",
			},
		},
	})
	c.Assert(err, qt.Equals, nil)

	client = contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(resp1.MachineToken),
	})

	resp, err := client.GetResourceMachineAccess(ctx, &params.GetResourceMachineAccessRequest{
		Resource: "esm",
		Machine:  "somemachine",
	})
	c.Assert(err, qt.Equals, nil)

	client = newBasicAuthClient(c, u)
	resp2, err := client.GetResourceTokenInfo(ctx, &params.GetResourceTokenInfoRequest{
		Token:    resp.ResourceToken,
		Resource: "esm",
	})
	c.Assert(err, qt.Equals, nil)
	c.Assert(resp2.MachineId, qt.Equals, "somemachine")
	c.Assert(resp2.ContractId, qt.Equals, cid)
	c.Assert(resp2.Affordances, qt.DeepEquals, map[string]interface{}{
		"architectures": nil,
		"series":        []interface{}{"precise", "trusty"},
	})
}

func TestGetResourceInfoRevokedMachine(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	cid, token := getContractToken(ctx, c, u)
	// Use a client that authorizes with the newly acquired contract token
	// to acquire a machine token.
	client := contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(token),
	})
	resp1, err := client.AddContractMachine(ctx, &params.AddContractMachineRequest{
		Body: params.AddContractMachineBody{
			Architecture: "amd64",
			MachineId:    "somemachine",
			Os: params.OS{
				Distribution: "ubuntu",
				Kernel:       "0.01",
				Release:      "aRelease",
				Series:       "someSeries",
				Type:         "osType",
			},
		},
	})
	c.Assert(err, qt.Equals, nil)

	client = contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(resp1.MachineToken),
	})

	resp, err := client.GetResourceMachineAccess(ctx, &params.GetResourceMachineAccessRequest{
		Resource: "esm",
		Machine:  "somemachine",
	})
	c.Assert(err, qt.Equals, nil)
	_, err = client.RevokeContractMachine(ctx, &params.RevokeContractMachineRequest{
		Contract: cid,
		Machine:  "somemachine",
	})
	c.Assert(err, qt.Equals, nil)
	client = newBasicAuthClient(c, u)
	_, err = client.GetResourceTokenInfo(ctx, &params.GetResourceTokenInfoRequest{
		Token:    resp.ResourceToken,
		Resource: "esm",
	})
	c.Assert(err, qt.ErrorMatches, `Get http://.*/v1/resources/esm\?token=`+resp.ResourceToken+`: machine "somemachine" no longer attached to contract ".*"`)
	c.Assert(errgo.Cause(err), qt.Equals, params.ErrForbidden)
}

func TestGetResourceInfoUnknownToken(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	client := newBasicAuthClient(c, u)
	_, err := client.GetResourceTokenInfo(ctx, &params.GetResourceTokenInfoRequest{
		Token:    "unknown",
		Resource: "esm",
	})
	c.Assert(errgo.Cause(err), qt.Equals, params.ErrUnauthorized)
	c.Assert(err, qt.ErrorMatches, `Get http://.*/v1/resources/esm\?token=unknown: invalid token`)
}

func TestGetCanonicalSSOMacaroon(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	u := newServer(ctx, c, nil, nil)

	client := contractclient.New(contractclient.NewParams{
		BaseURL: u,
	})

	resp, err := client.GetCanonicalSSOMacaroon(ctx, &params.GetCanonicalSSOMacaroonRequest{})
	c.Assert(err, qt.Equals, nil)
	buf, err := base64.RawURLEncoding.DecodeString(resp.Macaroon)
	c.Assert(err, qt.Equals, nil)
	var m macaroon.Macaroon
	err = m.UnmarshalBinary(buf)
	c.Assert(err, qt.Equals, nil)

	// Just check that the discharger thinks it can discharge the macaroon.
	_, err = ssoauthtest.Discharge(ssoDischarger, &m, nil, time.Time{}, time.Time{})
	c.Assert(err, qt.Equals, nil)
}

func TestGetAccounts(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	u := newServer(ctx, c, nil, nil)

	client := contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: getMacaroonAuthorizer(ctx, c, u, &testAccount),
	})

	resp, err := client.GetAccounts(ctx, &params.GetAccountsRequest{})
	c.Assert(err, qt.Equals, nil)
	now := time.Now().UTC().Truncate(time.Second)
	c.Assert(resp, qt.CmpEquals(equateApproxTime(time.Second)), &params.GetAccountsResponse{
		Accounts: []params.AccountInfo{{
			CreatedAt: now,
			Id:        testAccountID,
			Name:      "test@example.com",
		}},
	})
}

func TestGetAccountContracts(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	u := newServer(ctx, c, nil, nil)

	client := contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: getMacaroonAuthorizer(ctx, c, u, &testAccount),
	})

	resp, err := client.GetAccounts(ctx, &params.GetAccountsRequest{})
	c.Assert(err, qt.Equals, nil)

	resp2, err := client.GetAccountContracts(ctx, &params.GetAccountContractsRequest{
		Account: resp.Accounts[0].Id,
	})
	c.Assert(err, qt.Equals, nil)
	c.Assert(resp2.Contracts[0].ContractInfo.Id, qt.Not(qt.Equals), "")
	resp2.Contracts[0].ContractInfo.Id = ""
	c.Assert(resp2, qt.CmpEquals(equateApproxTime(time.Second)), &params.GetContractsResponse{
		Contracts: []params.AccountContractInfo{{
			AccountInfo: params.AccountInfo{
				CreatedAt: time.Now().UTC(),
				Id:        testAccountID,
				Name:      "test@example.com",
			},
			ContractInfo: params.ContractInfo{
				Name:          "test@example.com",
				CreatedAt:     time.Now(),
				EffectiveFrom: newTime(time.Now()),
				ResourceEntitlements: []params.Entitlement{
					params.EntitlementAptRepository{
						Type:     "cc",
						Entitled: true,
						Affordances: params.AffordancesAptRepository{
							Architectures: []string{"x86_64", "ppc64le", "s390x"},
							Series:        []string{"xenial"},
						},
						Directives: params.DirectivesAptRepository{
							AdditionalPackages: []string{"ubuntu-commoncriteria"},
							AptKey:             "A166877412DAC26E73CEBF3FF6C280178D13028C",
							AptURL:             "https://cc.ubuntu.com",
							Suites:             []string{"xenial-cc"},
						},
					},
					params.EntitlementAptRepository{
						Type:     "esm",
						Entitled: true,
						Affordances: params.AffordancesAptRepository{
							Series: []string{"precise", "trusty"},
						},
						Directives: params.DirectivesAptRepository{
							AptKey: "67C7A026",
							AptURL: "https://esm.ubuntu.com",
							Suites: []string{"precise", "trusty-security", "trusty-updates"},
						},
						Obligations: params.Obligations{EnableByDefault: true},
					},
					params.EntitlementAptRepository{
						Type:     "fips",
						Entitled: true,
						Affordances: params.AffordancesAptRepository{
							Series: []string{"xenial"},
						},
						Directives: params.DirectivesAptRepository{
							AdditionalPackages: []string{
								"fips-initramfs",
								"libssl1.0.0",
								"libssl1.0.0-hmac",
								"linux-fips",
								"openssh-client",
								"openssh-client-hmac",
								"openssh-server",
								"openssh-server-hmac",
								"openssl",
								"strongswan",
								"strongswan-hmac",
							},
							AptKey: "A166877412DAC26E73CEBF3FF6C280178D13028C",
							AptURL: "https://private-ppa.launchpad.net/ubuntu-advantage/fips",
							Suites: []string{"xenial"},
						},
					},
					params.EntitlementAptRepository{
						Type:     "fips-updates",
						Entitled: false,
					},
					params.EntitlementLivepatch{
						Type:     "livepatch",
						Entitled: true,
						Affordances: params.AffordancesLivepatch{
							Architectures:    []string{"x86_64"},
							KernelFlavors:    []string{"generic", "lowlatency"},
							MinKernelVersion: "4.4",
							Series:           []string{"xenial", "bionic"},
						},
						Directives:  params.DirectivesLivepatch{},
						Obligations: params.Obligations{},
					},
					params.EntitlementSupport{
						Type:        "support",
						Entitled:    true,
						Affordances: params.AffordancesSupport{SupportLevel: "community"},
					},
				},
			},
		}},
	})
}

func TestGetContractMachine(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	cid, token := getContractToken(ctx, c, u)
	// Use a client that authorizes with the newly acquired contract token.
	client := contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(token),
	})
	resp, err := client.AddContractMachine(ctx, &params.AddContractMachineRequest{
		Body: params.AddContractMachineBody{
			Architecture: "amd64",
			MachineId:    "somemachine",
			Os: params.OS{
				Distribution: "ubuntu",
				Kernel:       "0.01",
				Release:      "aRelease",
				Series:       "someSeries",
				Type:         "osType",
			},
		},
	})
	c.Assert(err, qt.Equals, nil)

	client = contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(resp.MachineToken),
	})
	resp2, err := client.GetContractMachine(ctx, &params.GetContractMachineRequest{
		Contract: cid,
		Machine:  "somemachine",
	})
	c.Assert(err, qt.Equals, nil)
	resp2.MachineTokenInfo.Expires = resp.MachineTokenInfo.Expires
	c.Assert(resp2, qt.CmpEquals(equateApproxTime(time.Second)), resp)
}

func TestGetContractMachineServiceAccountAuth(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	cid, token := getContractToken(ctx, c, u)
	// Use a client that authorizes with the newly acquired contract token.
	client := contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(token),
	})
	resp, err := client.AddContractMachine(ctx, &params.AddContractMachineRequest{
		Body: params.AddContractMachineBody{
			Architecture: "amd64",
			MachineId:    "somemachine",
			Os: params.OS{
				Distribution: "ubuntu",
				Kernel:       "0.01",
				Release:      "aRelease",
				Series:       "someSeries",
				Type:         "osType",
			},
		},
	})
	c.Assert(err, qt.Equals, nil)

	client = newBasicAuthClient(c, u)
	resp2, err := client.GetContractMachine(ctx, &params.GetContractMachineRequest{
		Contract: cid,
		Machine:  "somemachine",
	})
	c.Assert(err, qt.Equals, nil)
	c.Assert(resp2.MachineToken, qt.Equals, "")
	resp2.MachineToken = resp.MachineToken
	resp2.MachineTokenInfo.Expires = resp.MachineTokenInfo.Expires
	c.Assert(resp2, qt.CmpEquals(equateApproxTime(time.Second)), resp)
}

func TestGetContractMachineRevokedMachine(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	cid, token := getContractToken(ctx, c, u)
	// Use a client that authorizes with the newly acquired contract token.
	client := contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(token),
	})
	resp, err := client.AddContractMachine(ctx, &params.AddContractMachineRequest{
		Body: params.AddContractMachineBody{
			Architecture: "amd64",
			MachineId:    "somemachine",
			Os: params.OS{
				Distribution: "ubuntu",
				Kernel:       "0.01",
				Release:      "aRelease",
				Series:       "someSeries",
				Type:         "osType",
			},
		},
	})
	c.Assert(err, qt.Equals, nil)

	client = contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(resp.MachineToken),
	})
	_, err = client.RevokeContractMachine(ctx, &params.RevokeContractMachineRequest{
		Contract: cid,
		Machine:  "somemachine",
	})
	c.Assert(err, qt.Equals, nil)
	_, err = client.GetContractMachine(ctx, &params.GetContractMachineRequest{
		Contract: cid,
		Machine:  "somemachine",
	})
	c.Assert(err, qt.ErrorMatches, `Get .*/v1/contracts/c.*/context/machines/somemachine: machine "somemachine" no longer attached to contract "c.*"`)
}

func TestRevokeContractMachine(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	cid, token := getContractToken(ctx, c, u)
	// Use a client that authorizes with the newly acquired contract token.
	client := contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(token),
	})
	resp, err := client.AddContractMachine(ctx, &params.AddContractMachineRequest{
		Body: params.AddContractMachineBody{
			Architecture: "amd64",
			MachineId:    "somemachine",
			Os: params.OS{
				Distribution: "ubuntu",
				Kernel:       "0.01",
				Release:      "aRelease",
				Series:       "someSeries",
				Type:         "osType",
			},
		},
	})
	c.Assert(err, qt.Equals, nil)

	client = contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(resp.MachineToken),
	})
	resp2, err := client.RevokeContractMachine(ctx, &params.RevokeContractMachineRequest{
		Contract: cid,
		Machine:  "somemachine",
	})
	c.Assert(err, qt.Equals, nil)
	c.Assert(resp2, qt.CmpEquals(equateApproxTime(time.Second)), &params.MachineTokenInfo{
		MachineId: "somemachine",
		AccountInfo: params.AccountInfo{
			Id:        testAccountID,
			CreatedAt: time.Now(),
			Name:      "test@example.com",
		},
		ContractInfo: params.ContractInfo{
			Id:            cid,
			Name:          "test@example.com",
			CreatedAt:     time.Now(),
			EffectiveFrom: newTime(time.Now()),
			ResourceEntitlements: []params.Entitlement{
				params.EntitlementAptRepository{
					Type:     "cc",
					Entitled: true,
					Affordances: params.AffordancesAptRepository{
						Architectures: []string{"x86_64", "ppc64le", "s390x"},
						Series:        []string{"xenial"},
					},
					Directives: params.DirectivesAptRepository{
						AdditionalPackages: []string{"ubuntu-commoncriteria"},
						AptKey:             "A166877412DAC26E73CEBF3FF6C280178D13028C",
						AptURL:             "https://cc.ubuntu.com",
						Suites:             []string{"xenial-cc"},
					},
				},
				params.EntitlementAptRepository{
					Type:     "esm",
					Entitled: true,
					Affordances: params.AffordancesAptRepository{
						Series: []string{"precise", "trusty"},
					},
					Directives: params.DirectivesAptRepository{
						AptKey: "67C7A026",
						AptURL: "https://esm.ubuntu.com",
						Suites: []string{"precise", "trusty-security", "trusty-updates"},
					},
					Obligations: params.Obligations{EnableByDefault: true},
				},
				params.EntitlementAptRepository{
					Type:     "fips",
					Entitled: true,
					Affordances: params.AffordancesAptRepository{
						Series: []string{"xenial"},
					},
					Directives: params.DirectivesAptRepository{
						AdditionalPackages: []string{
							"fips-initramfs",
							"libssl1.0.0",
							"libssl1.0.0-hmac",
							"linux-fips",
							"openssh-client",
							"openssh-client-hmac",
							"openssh-server",
							"openssh-server-hmac",
							"openssl",
							"strongswan",
							"strongswan-hmac",
						},
						AptKey: "A166877412DAC26E73CEBF3FF6C280178D13028C",
						AptURL: "https://private-ppa.launchpad.net/ubuntu-advantage/fips",
						Suites: []string{"xenial"},
					},
				},
				params.EntitlementAptRepository{
					Type:     "fips-updates",
					Entitled: false,
				},
				params.EntitlementLivepatch{
					Affordances: params.AffordancesLivepatch{
						Architectures:    []string{"x86_64"},
						KernelFlavors:    []string{"generic", "lowlatency"},
						MinKernelVersion: "4.4",
						Series:           []string{"xenial", "bionic"},
					},
					Directives:  params.DirectivesLivepatch{},
					Entitled:    true,
					Obligations: params.Obligations{},
					Type:        "livepatch",
				},
				params.EntitlementSupport{
					Entitled:    true,
					Type:        "support",
					Affordances: params.AffordancesSupport{SupportLevel: "community"},
				},
			},
		},
	})
	_, err = client.GetResourceMachineAccess(ctx, &params.GetResourceMachineAccessRequest{
		Machine:  "somemachine",
		Resource: "esm",
	})
	c.Assert(err, qt.ErrorMatches, `Get http://.*/v1/resources/esm/context/machines/somemachine: machine "somemachine" no longer attached to contract ".*"`)
	c.Assert(errgo.Cause(err), qt.Equals, params.ErrForbidden)
}

func TestRevokeContractMachineSSOAuth(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	cid, token := getContractToken(ctx, c, u)
	// Use a client that authorizes with the newly acquired contract token.
	client := contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(token),
	})
	// Adding the contract machine also creates the contract itself.
	resp, err := client.AddContractMachine(ctx, &params.AddContractMachineRequest{
		Body: params.AddContractMachineBody{
			Architecture: "amd64",
			MachineId:    "somemachine",
			Os: params.OS{
				Distribution: "ubuntu",
				Kernel:       "0.01",
				Release:      "aRelease",
				Series:       "someSeries",
				Type:         "osType",
			},
		},
	})
	c.Assert(err, qt.Equals, nil)

	client = contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: getMacaroonAuthorizer(ctx, c, u, &testAccount),
	})
	resp2, err := client.RevokeContractMachine(ctx, &params.RevokeContractMachineRequest{
		Contract: cid,
		Machine:  "somemachine",
	})
	c.Assert(err, qt.Equals, nil)
	c.Assert(resp2, qt.CmpEquals(equateApproxTime(time.Second)), &params.MachineTokenInfo{
		MachineId: "somemachine",
		AccountInfo: params.AccountInfo{
			Id:        testAccountID,
			CreatedAt: time.Now(),
			Name:      "test@example.com",
		},
		ContractInfo: params.ContractInfo{
			Id:            cid,
			Name:          "test@example.com",
			CreatedAt:     time.Now(),
			EffectiveFrom: newTime(time.Now()),
			ResourceEntitlements: []params.Entitlement{
				params.EntitlementAptRepository{
					Type:     "cc",
					Entitled: true,
					Affordances: params.AffordancesAptRepository{
						Architectures: []string{"x86_64", "ppc64le", "s390x"},
						Series:        []string{"xenial"},
					},
					Directives: params.DirectivesAptRepository{
						AdditionalPackages: []string{"ubuntu-commoncriteria"},
						AptKey:             "A166877412DAC26E73CEBF3FF6C280178D13028C",
						AptURL:             "https://cc.ubuntu.com",
						Suites:             []string{"xenial-cc"},
					},
				},
				params.EntitlementAptRepository{
					Type:     "esm",
					Entitled: true,
					Affordances: params.AffordancesAptRepository{
						Series: []string{"precise", "trusty"},
					},
					Directives: params.DirectivesAptRepository{
						AptKey: "67C7A026",
						AptURL: "https://esm.ubuntu.com",
						Suites: []string{"precise", "trusty-security", "trusty-updates"},
					},
					Obligations: params.Obligations{EnableByDefault: true},
				},
				params.EntitlementAptRepository{
					Type:     "fips",
					Entitled: true,
					Affordances: params.AffordancesAptRepository{
						Series: []string{"xenial"},
					},
					Directives: params.DirectivesAptRepository{
						AdditionalPackages: []string{
							"fips-initramfs",
							"libssl1.0.0",
							"libssl1.0.0-hmac",
							"linux-fips",
							"openssh-client",
							"openssh-client-hmac",
							"openssh-server",
							"openssh-server-hmac",
							"openssl",
							"strongswan",
							"strongswan-hmac",
						},
						AptKey: "A166877412DAC26E73CEBF3FF6C280178D13028C",
						AptURL: "https://private-ppa.launchpad.net/ubuntu-advantage/fips",
						Suites: []string{"xenial"},
					},
				},
				params.EntitlementAptRepository{
					Type:     "fips-updates",
					Entitled: false,
				},
				params.EntitlementLivepatch{
					Type:     "livepatch",
					Entitled: true,
					Affordances: params.AffordancesLivepatch{
						Architectures:    []string{"x86_64"},
						KernelFlavors:    []string{"generic", "lowlatency"},
						MinKernelVersion: "4.4",
						Series:           []string{"xenial", "bionic"},
					},
					Directives:  params.DirectivesLivepatch{},
					Obligations: params.Obligations{},
				},
				params.EntitlementSupport{
					Type:        "support",
					Entitled:    true,
					Affordances: params.AffordancesSupport{SupportLevel: "community"},
				},
			},
		},
	})

	client = contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: contractclient.BearerAuthorizer(resp.MachineToken),
	})
	_, err = client.GetResourceMachineAccess(ctx, &params.GetResourceMachineAccessRequest{
		Machine:  "somemachine",
		Resource: "esm",
	})
	c.Assert(err, qt.ErrorMatches, `Get http://.*/v1/resources/esm/context/machines/somemachine: machine "somemachine" no longer attached to contract ".*"`)
	c.Assert(errgo.Cause(err), qt.Equals, params.ErrForbidden)
}

// TestBugWithNoContractOnAccount tests the case we had before where we were creating account and contract together
// but when Salesforce was failing it leaves us with an account without contract and no possibilities to add a
// contract later
func TestBugWithNoContractOnAccount(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)

	mStore := memstore.New()
	st := &mutableStore{mStore}
	sfClient := &testSalesforceClient{}
	u := newServer(ctx, c, st, sfClient)

	client := contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: getMacaroonAuthorizer(ctx, c, u, &testAccount),
	})

	sfClient.User = &sfparams.User{
		Account: sfparams.Account{
			ID:   "sfacc2",
			Name: "sf account name",
		},
	}
	resp, err := client.GetAccounts(ctx, &params.GetAccountsRequest{})
	c.Assert(err, qt.Equals, nil)
	c.Assert(resp.Accounts, qt.HasLen, 1)

	sfClient.E = errgo.New("something wrong")
	_, err = client.GetAccountContracts(ctx, &params.GetAccountContractsRequest{
		Account: resp.Accounts[0].Id,
	})
	c.Assert(err, qt.ErrorMatches, `Get http://.*/v1/accounts/.*/contracts: cannot ensure contract: something wrong`)
	// Working Salesforce.
	sfClient.E = nil
	resp2, err := client.GetAccountContracts(ctx, &params.GetAccountContractsRequest{
		Account: resp.Accounts[0].Id,
	})
	c.Assert(err, qt.Equals, nil)
	c.Assert(resp2.Contracts, qt.HasLen, 1)
}

func TestSetServiceAccountPassword(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	client := newBasicAuthClient(c, u)

	err := client.SetServiceAccountPassword(ctx, &params.SetServiceAccountPasswordRequest{
		Account: "test-service-account",
		Body: params.SetServiceAccountPasswordBody{
			Password: "new-password",
		},
	})
	c.Assert(err, qt.Equals, nil)

	sa, err := srvStore.ServiceAccount(ctx, "test-service-account")
	c.Assert(err, qt.Equals, nil)
	passwordHash := argon2.IDKey([]byte("new-password"), sa.PasswordSalt, 1, 64*1024, 4, 32)
	c.Assert(passwordHash, qt.DeepEquals, sa.PasswordHash)
}

func TestGetProduct(t *testing.T) {
	c := qt.New(t)
	defer c.Done()
	ctx := testContext(c)

	u := newServer(ctx, c, nil, nil)

	client := newBasicAuthClient(c, u)

	prod, err := client.GetProduct(ctx, &params.GetProductRequest{
		ProductID: defaultProduct.ID,
	})
	c.Assert(err, qt.Equals, nil)
	c.Assert(prod, qt.DeepEquals, &params.Product{
		Id: storeutil.DefaultProductID,
		Resources: []params.ProductResource{{
			ResourceType: "esm",
			Data: map[string]interface{}{
				"aptURL":          "https://esm.ubuntu.com",
				"aptKey":          "67C7A026",
				"series":          []interface{}{"precise", "trusty"},
				"suites":          []interface{}{"precise", "trusty-security", "trusty-updates"},
				"enableByDefault": true,
			},
		}, {
			ResourceType: "fips",
			Data: map[string]interface{}{
				"additionalPackages": []interface{}{
					"fips-initramfs",
					"libssl1.0.0",
					"libssl1.0.0-hmac",
					"linux-fips",
					"openssh-client",
					"openssh-client-hmac",
					"openssh-server",
					"openssh-server-hmac",
					"openssl",
					"strongswan",
					"strongswan-hmac",
				},
				"aptKey": string("A166877412DAC26E73CEBF3FF6C280178D13028C"),
				"aptURL": string("https://private-ppa.launchpad.net/ubuntu-advantage/fips"),
				"series": []interface{}{string("xenial")},
				"suites": []interface{}{"xenial"},
			},
		}, {
			ResourceType: "cc",
			Data: map[string]interface{}{
				"additionalPackages": []interface{}{"ubuntu-commoncriteria"},
				"aptURL":             "https://cc.ubuntu.com",
				"aptKey":             "A166877412DAC26E73CEBF3FF6C280178D13028C",
				"architectures":      []interface{}{"x86_64", "ppc64le", "s390x"},
				"series":             []interface{}{"xenial"},
				"suites":             []interface{}{"xenial-cc"},
			},
		}, {
			ResourceType: "livepatch",
			Data: map[string]interface{}{
				"architectures":    []interface{}{"x86_64"},
				"kernelFlavors":    []interface{}{"generic", "lowlatency"},
				"minKernelVersion": "4.4",
				"series":           []interface{}{"xenial", "bionic"},
			},
		}, {
			ResourceType: "support",
			Data: map[string]interface{}{
				"supportLevel": "community",
			},
		}},
	})
}

func TestGetProductNotFound(t *testing.T) {
	c := qt.New(t)
	defer c.Done()
	ctx := testContext(c)

	u := newServer(ctx, c, nil, nil)

	client := newBasicAuthClient(c, u)

	_, err := client.GetProduct(ctx, &params.GetProductRequest{
		ProductID: "unknownProduct",
	})
	c.Assert(err, qt.ErrorMatches, `Get .*: product not found`)
	c.Assert(errgo.Cause(err), qt.Equals, params.ErrNotFound)
}

func TestAddProduct(t *testing.T) {
	c := qt.New(t)
	defer c.Done()
	ctx := testContext(c)

	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	client := newBasicAuthClient(c, u)

	prod := &params.Product{
		Name: "new product name",
		Resources: []params.ProductResource{{
			ResourceType: "esm",
			Data: map[string]interface{}{
				"aptURL": "https://example.com",
			},
		}},
	}
	err := client.AddProduct(ctx, &params.AddProductRequest{
		ProductID: "newProduct",
		Body:      *prod,
	})
	c.Assert(err, qt.Equals, nil)

	prod1, err := client.GetProduct(ctx, &params.GetProductRequest{
		ProductID: "newProduct",
	})
	c.Assert(err, qt.Equals, nil)
	prod.Id = "newProduct"
	c.Assert(prod1, qt.DeepEquals, prod)

	prod2 := params.Product{
		Name: "new product name",
		Resources: []params.ProductResource{{
			ResourceType: "esm",
			Data: map[string]interface{}{
				"aptURL": "https://other.com",
			},
		}},
	}
	// Try to update the product details.
	err = client.AddProduct(ctx, &params.AddProductRequest{
		ProductID: "newProduct",
		Update:    true,
		Body:      prod2,
	})
	c.Assert(err, qt.Equals, nil)

	prod3, err := client.GetProduct(ctx, &params.GetProductRequest{
		ProductID: "newProduct",
	})
	c.Assert(err, qt.Equals, nil)
	prod2.Id = "newProduct"
	c.Assert(prod3, qt.DeepEquals, &prod2)
}

func TestGetProducts(t *testing.T) {
	c := qt.New(t)
	defer c.Done()
	ctx := testContext(c)

	srvStore := memstore.New()
	u := newServer(ctx, c, srvStore, nil)

	client := newBasicAuthClient(c, u)

	prod := &params.Product{
		Id:   "newProduct",
		Name: "new product name",
		Resources: []params.ProductResource{{
			ResourceType: "esm",
			Data: map[string]interface{}{
				"aptURL": "https://example.com",
			},
		}},
	}
	err := client.AddProduct(ctx, &params.AddProductRequest{
		ProductID: "newProduct",
		Body:      *prod,
	})
	c.Assert(err, qt.Equals, nil)

	resp, err := client.GetProducts(ctx, &params.GetProductsRequest{})
	c.Assert(err, qt.Equals, nil)
	c.Assert(resp, qt.DeepEquals, &params.GetProductsResponse{
		Products: []params.Product{
			v1.ParamsProduct(defaultProduct),
			*prod,
		},
	})
}

func TestUpdateUserContractsWhenSFAccountIsNotFound(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	sfClient := &testSalesforceClient{}
	u := newServer(ctx, c, nil, sfClient)
	sfClient.E = sfparams.ErrNotFound
	client := newBasicAuthClient(c, u)

	_, err := client.UpdateAccountContracts(ctx, &params.UpdateAccountContractsRequest{
		Sfid: "salesforceAccountId",
	})
	c.Assert(errgo.Cause(err), qt.Equals, params.ErrNotFound)
}

func TestUpdateUserContracts(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	sfClient := &testSalesforceClient{
		User: &sfparams.User{
			Account: sfparams.Account{
				ID:   "sfaccid1",
				Name: "sf account name",
			},
		},
	}
	u := newServer(ctx, c, nil, sfClient)
	client1 := contractclient.New(contractclient.NewParams{
		BaseURL:    u,
		Authorizer: getMacaroonAuthorizer(ctx, c, u, &testAccount),
	})
	accId := getAccountID(ctx, c, client1)
	client := newBasicAuthClient(c, u)
	resp, err := client.UpdateAccountContracts(ctx, &params.UpdateAccountContractsRequest{
		Sfid: "sfaccid1",
	})
	c.Assert(err, qt.Equals, nil)
	resp.Contracts[0].ContractInfo.Id = ""
	c.Assert(resp, qt.CmpEquals(equateApproxTime(time.Second)), &params.GetContractsResponse{
		Contracts: []params.AccountContractInfo{{
			AccountInfo: params.AccountInfo{
				CreatedAt: time.Now().UTC(),
				Id:        accId.String(),
				Name:      "sf account name",
			},
			ContractInfo: params.ContractInfo{
				Name:          "sf account name",
				CreatedAt:     time.Now(),
				EffectiveFrom: newTime(time.Now()),
				ResourceEntitlements: []params.Entitlement{
					params.EntitlementAptRepository{
						Type:     "cc",
						Entitled: true,
						Affordances: params.AffordancesAptRepository{
							Architectures: []string{"x86_64", "ppc64le", "s390x"},
							Series:        []string{"xenial"},
						},
						Directives: params.DirectivesAptRepository{
							AdditionalPackages: []string{"ubuntu-commoncriteria"},
							AptKey:             "A166877412DAC26E73CEBF3FF6C280178D13028C",
							AptURL:             "https://cc.ubuntu.com",
							Suites:             []string{"xenial-cc"},
						},
					},
					params.EntitlementAptRepository{
						Type:     "esm",
						Entitled: true,
						Affordances: params.AffordancesAptRepository{
							Series: []string{"precise", "trusty"},
						},
						Directives: params.DirectivesAptRepository{
							AptKey: "67C7A026",
							AptURL: "https://esm.ubuntu.com",
							Suites: []string{"precise", "trusty-security", "trusty-updates"},
						},
						Obligations: params.Obligations{EnableByDefault: true},
					},
					params.EntitlementAptRepository{
						Type:     "fips",
						Entitled: true,
						Affordances: params.AffordancesAptRepository{
							Series: []string{"xenial"},
						},
						Directives: params.DirectivesAptRepository{
							AdditionalPackages: []string{
								"fips-initramfs",
								"libssl1.0.0",
								"libssl1.0.0-hmac",
								"linux-fips",
								"openssh-client",
								"openssh-client-hmac",
								"openssh-server",
								"openssh-server-hmac",
								"openssl",
								"strongswan",
								"strongswan-hmac",
							},
							AptKey: "A166877412DAC26E73CEBF3FF6C280178D13028C",
							AptURL: "https://private-ppa.launchpad.net/ubuntu-advantage/fips",
							Suites: []string{"xenial"},
						},
					},
					params.EntitlementAptRepository{
						Type:     "fips-updates",
						Entitled: false,
					},
					params.EntitlementLivepatch{
						Type:     "livepatch",
						Entitled: true,
						Affordances: params.AffordancesLivepatch{
							Architectures:    []string{"x86_64"},
							KernelFlavors:    []string{"generic", "lowlatency"},
							MinKernelVersion: "4.4",
							Series:           []string{"xenial", "bionic"},
						},
						Directives:  params.DirectivesLivepatch{},
						Obligations: params.Obligations{},
					},
					params.EntitlementSupport{
						Type:        "support",
						Entitled:    true,
						Affordances: params.AffordancesSupport{SupportLevel: "community"},
					},
				},
			},
		}},
	})
}

func TestUpdateUserContractsUnauthorized(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	ctx := testContext(c)
	u := newServer(ctx, c, nil, nil)
	client := contractclient.New(contractclient.NewParams{
		BaseURL: u,
	})

	_, err := client.UpdateAccountContracts(ctx, &params.UpdateAccountContractsRequest{
		Sfid: "salesforceAccountId",
	})
	c.Assert(errgo.Cause(err), qt.Equals, params.ErrUnauthorized)
}

var defaultProduct = store.Product{
	ID: storeutil.DefaultProductID,
	Resources: []store.ProductResource{{
		ResourceType: params.ResourceESM,
		Data: `{
			"aptURL": "https://esm.ubuntu.com",
			"aptKey": "67C7A026",
			"series": ["precise", "trusty"],
			"suites": ["precise", "trusty-security", "trusty-updates"],
			"enableByDefault": true
		}`,
	}, {
		ResourceType: params.ResourceFIPS,
		Data: `{
			"additionalPackages": ["fips-initramfs", "libssl1.0.0", "libssl1.0.0-hmac", "linux-fips", "openssh-client", "openssh-client-hmac", "openssh-server", "openssh-server-hmac", "openssl",  "strongswan", "strongswan-hmac"],
			"aptURL": "https://private-ppa.launchpad.net/ubuntu-advantage/fips",
			"aptKey": "A166877412DAC26E73CEBF3FF6C280178D13028C",
			"series": ["xenial"],
			"suites": ["xenial"]
		}`,
	}, {
		ResourceType: params.ResourceCC,
		Data: `{
			"additionalPackages": ["ubuntu-commoncriteria"],
			"aptURL": "https://cc.ubuntu.com",
			"aptKey": "A166877412DAC26E73CEBF3FF6C280178D13028C",
			"architectures": ["x86_64", "ppc64le", "s390x"],
			"series": ["xenial"],
			"suites": ["xenial-cc"]
		}`,
	}, {
		ResourceType: params.ResourceLivepatch,
		Data: `{
			"architectures": ["x86_64"],
			"kernelFlavors": ["generic", "lowlatency"],
			"minKernelVersion": "4.4",
			"series": ["xenial", "bionic"]
		}`,
	}, {
		ResourceType: params.ResourceSupport,
		Data: `{
			"supportLevel": "community"
		}`,
	}},
}

// newServer starts an API server and returns its URL. The server will be
// stopped when c.Done is called. If srvStore is nil, memstore.New() will
// be used. If sfclient is nil then a testSalesforceClient will be used.
func newServer(ctx context.Context, c *qt.C, srvStore store.Store, sfclient storeutil.SalesforceClient) string {
	if srvStore == nil {
		srvStore = memstore.New()
	}
	if sfclient == nil {
		sfclient = storeutil.SalesforceDisabledClient{}
	}

	// Add the default product.
	err := srvStore.AddProduct(ctx, defaultProduct)
	if errgo.Cause(err) == store.ErrDuplicateProduct {
		err = nil
	}
	c.Assert(err, qt.Equals, nil)

	aclStore := aclstore.NewACLStore(memsimplekv.NewStore())
	m, err := aclstore.NewManager(ctx, aclstore.Params{
		Store:             aclStore,
		InitialAdminUsers: []string{"admin"},
	})
	c.Assert(err, qt.Equals, nil)

	router := httprouter.New()
	hs, err := v1.Handlers(ctx, v1.Params{
		ACLManager: m,
		Logger:     zapctx.Logger(ctx),
		Store:      srvStore,
		SSOAuth: ssoauth.New(ssoauth.Params{
			Oven:      bakery.NewOven(bakery.OvenParams{}),
			Location:  ssoDischarger.Location(),
			PublicKey: ssoDischarger.PublicKey(),
		}),
		TokenSecret:      []byte("highlysecret"),
		SalesforceClient: sfclient,
	})
	c.Assert(err, qt.Equals, nil)

	// Add the admin user.
	err = saauth.SetPassword(ctx, srvStore, "admin", "testpassword")
	c.Assert(err, qt.Equals, nil)
	// Grant the admin user access to all ACLs.
	for _, name := range v1.ACLs {
		err := aclStore.Add(ctx, name, []string{"admin"})
		c.Assert(err, qt.Equals, nil)
	}

	httprequest.AddHandlers(router, hs)
	srv := httptest.NewServer(router)
	c.Defer(srv.Close)
	return srv.URL
}

// newTestAccountClient returns a client instance that authorizes
// as the test user.
func newTestAccountClient(ctx context.Context, c *qt.C, url string) *contractclient.Client {
	return contractclient.New(contractclient.NewParams{
		BaseURL:    url,
		Authorizer: getMacaroonAuthorizer(ctx, c, url, &testAccount),
	})
}

func newBasicAuthClient(c *qt.C, url string) *contractclient.Client {
	return contractclient.New(contractclient.NewParams{
		BaseURL:    url,
		Authorizer: contractclient.BasicAuthorizer("admin", "testpassword"),
	})
}

// getAccountID returns the account ID of the user
// that the client authorizes as.
func getAccountID(ctx context.Context, c *qt.C, client *contractclient.Client) token.AccountID {
	resp, err := client.GetAccounts(ctx, &params.GetAccountsRequest{})
	c.Assert(err, qt.Equals, nil)
	c.Assert(resp.Accounts, qt.HasLen, 1)
	id, err := token.ParseAccountID(resp.Accounts[0].Id)
	c.Assert(err, qt.Equals, nil)
	return id
}

func getContractToken(ctx context.Context, c *qt.C, url string) (id, token string) {
	client := newTestAccountClient(ctx, c, url)

	accountID := getAccountID(ctx, c, client)

	resp2, err := client.GetAccountContracts(ctx, &params.GetAccountContractsRequest{
		Account: accountID.String(),
	})
	c.Assert(err, qt.Equals, nil)
	c.Assert(resp2.Contracts, qt.HasLen, 1)

	resp3, err := client.GetContractToken(ctx, &params.GetContractTokenRequest{
		Contract: resp2.Contracts[0].ContractInfo.Id,
	})
	c.Assert(err, qt.Equals, nil)

	return resp2.Contracts[0].ContractInfo.Id, resp3.ContractToken
}

func getMacaroonAuthorizer(ctx context.Context, c *qt.C, url string, acc *ssoauth.Account) contractclient.Authorizer {
	m, err := contractclient.GetCanonicalSSOMacaroon(ctx, url)
	c.Assert(err, qt.Equals, nil)
	ms, err := ssoauthtest.Discharge(ssoDischarger, m, acc, time.Time{}, time.Time{})
	c.Assert(err, qt.Equals, nil)
	auth, err := contractclient.MacaroonAuthorizer(ms)
	c.Assert(err, qt.Equals, nil)
	return auth
}

func equateApproxTime(margin time.Duration) cmp.Option {
	return cmp.Comparer(func(t0, t1 time.Time) bool {
		if t0.IsZero() && t1.IsZero() {
			return true
		}
		if t0.IsZero() || t1.IsZero() {
			return false
		}
		diff := t0.Sub(t1)
		if diff < 0 {
			diff = -diff
		}
		return diff <= margin
	})
}

type mutableStore struct {
	store.Store
}

type testSalesforceClient struct {
	User   *sfparams.User
	Assets []salesforce.Asset
	E      error
}

func (t testSalesforceClient) GetUser(ctx context.Context, email string) (*sfparams.User, error) {
	return t.User, t.E
}

func (t testSalesforceClient) GetAssetsForAccount(ctx context.Context, id string) ([]salesforce.Asset, error) {
	return t.Assets, t.E
}

func testContext(c *qt.C) context.Context {
	return zapctx.WithLogger(context.Background(), zaptest.NewLogger(c))
}

func newTime(t time.Time) *time.Time {
	return &t
}
