package v1

import (
	"context"
	"encoding/json"
	"fmt"
	"net/http"
	"strings"
	"time"

	aclstore "github.com/juju/aclstore/v2"
	errgo "gopkg.in/errgo.v1"
	httprequest "gopkg.in/httprequest.v1"
	macaroon "gopkg.in/macaroon.v2"

	"github.com/CanonicalLtd/ua-contracts/internal/awsauth"
	"github.com/CanonicalLtd/ua-contracts/internal/saauth"
	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
	"github.com/CanonicalLtd/ua-contracts/params"
)

var acls = []string{
	"audit",
	"contract",
	"product",
	"resource",
	"service-account",
}

// ensureACls creates all required ACLs if they don't exist.
func ensureACLs(ctx context.Context, m *aclstore.Manager) error {
	for _, name := range acls {
		if err := m.CreateACL(ctx, name); err != nil {
			return errgo.Mask(err)
		}
	}
	return nil
}

func (h *handler) authorize(p httprequest.Params, arg interface{}) (context.Context, error) {
	switch arg := arg.(type) {
	case *params.GetContractTokenRequest:
		return h.authSSOMacaroon(p)
	case *params.AddContractMachineRequest:
		return h.authContractToken(p)
	case *params.GetResourceMachineAccessRequest:
		return h.authMachineToken(p, "", arg.Machine)
	case *params.GetResourceTokenInfoRequest:
		return h.authBasic(p.Context, p.Request, "resource")
	case *params.GetCanonicalSSOMacaroonRequest, *DebugInfoRequest:
		// No authentication is required for this endpoint.
		return p.Context, nil
	case *params.GetAccountsRequest:
		return h.authSSOMacaroon(p)
	case *params.GetAccountContractsRequest:
		return h.authSSOMacaroon(p)
	case *params.GetContractMachineRequest:
		ctx, err := h.authMachineToken(p, arg.Contract, arg.Machine)
		if errgo.Cause(err) != params.ErrUnauthorized {
			return ctx, errgo.Mask(err, isAPIError)
		}
		return h.authBasic(p.Context, p.Request, "contract")
	case *params.GetContractRequest:
		return h.authBasic(p.Context, p.Request, "contract")
	case *params.RevokeContractMachineRequest:
		ctx, err := h.authMachineToken(p, arg.Contract, arg.Machine)
		if errgo.Cause(err) != params.ErrUnauthorized {
			return ctx, errgo.Mask(err, isAPIError)
		}
		return h.authSSOMacaroon(p)
	case *params.AuditLogQueryRequest:
		return h.authBasic(p.Context, p.Request, "audit")
	case *params.SetServiceAccountPasswordRequest:
		return h.authBasic(p.Context, p.Request, "service-account")
	case *params.AddProductRequest:
		return h.authBasic(p.Context, p.Request, "product")
	case *params.GetProductRequest:
		return h.authBasic(p.Context, p.Request, "product")
	case *params.GetProductsRequest:
		return h.authBasic(p.Context, p.Request, "product")
	case *params.ConfigureContractRequest:
		return h.authBasic(p.Context, p.Request, "contract")
	case *params.GetContractConfigRequest:
		return h.authBasic(p.Context, p.Request, "contract")
	case *params.UpdateAccountContractsRequest:
		return h.authBasic(p.Context, p.Request, "contract")
	case *params.GetContractTokenForAWSInstanceRequest:
		return h.authAWSSignature(p.Context, arg)
	default:
		return p.Context, errgo.Newf("cannot authorize unknown API entry point")
	}
}

// authBasic authorizes a service-account by checking the request for
// basic auth credentials. If aclName is non-zero then the ACL with that
// name will be checked to ensure the service-account has access.
func (h *handler) authBasic(ctx context.Context, req *http.Request, aclName string) (context.Context, error) {
	var id saauth.Identity
	username, password, ok := req.BasicAuth()
	if !ok {
		return ctx, errgo.WithCausef(nil, params.ErrUnauthorized, "")
	}
	id, err := saauth.Validate(ctx, h.p.Store, username, password)
	if err != nil {
		return ctx, errgo.Mask(err)
	}
	if aclName != "" {
		return ctx, errgo.Mask(h.checkACL(ctx, id, aclName), errgo.Is(params.ErrForbidden))
	}
	return ctx, nil
}

// checkACL checks that the the given id has access to the ACL with the
// given name.
func (h *handler) checkACL(ctx context.Context, id saauth.Identity, aclName string) error {
	acl, err := h.p.ACLManager.ACL(ctx, aclName)
	if err != nil {
		return errgo.Mask(err)
	}
	ok, err := id.Allow(ctx, acl)
	if err != nil {
		return errgo.Mask(err)
	}
	if !ok {
		return errgo.WithCausef(nil, params.ErrForbidden, "")
	}
	return nil
}

func (h *handler) authContractToken(p httprequest.Params) (context.Context, error) {
	ctx := p.Context
	raw, err := extractTokenFromHeader(p.Request)
	if err != nil {
		return ctx, errgo.Mask(err, isAPIError)
	}
	ct, err := token.ParseContractToken(raw)
	if err != nil {
		return ctx, errgo.Mask(err, isAPIError)
	}
	info, err := h.p.Store.ContractToken(ctx, ct)
	if err != nil {
		if errgo.Cause(err) == store.ErrTokenNotFound {
			return nil, unauthorizedErrorf(nil, "")
		}
		return ctx, errgo.Mask(err)
	}
	if time.Now().After(info.Expires) {
		return nil, unauthorizedErrorf(nil, "contract token has expired")
	}
	return contextWithContractID(ctx, info.ContractID), nil
}

func (h *handler) authMachineToken(p httprequest.Params, contractID, machineID string) (context.Context, error) {
	ctx := p.Context
	raw, err := extractTokenFromHeader(p.Request)
	if err != nil {
		return ctx, errgo.Mask(err, isAPIError)
	}
	mt, err := token.ParseMachineToken(raw)
	if err != nil {
		return ctx, errgo.Mask(err, isAPIError)
	}
	body, err := h.machineTokenCodec.Decode(mt)
	if err != nil {
		return ctx, errgo.Mask(err, isAPIError)
	}

	// Check the token is still valid.
	ok, err := h.p.Store.MachineHasContract(ctx, body.ContractID, body.MachineID)
	if err != nil {
		return ctx, errgo.Mask(err)
	}
	if !ok {
		return ctx, forbiddenErrorf(nil, "machine %q no longer attached to contract %q", body.MachineID, body.ContractID)
	}

	if contractID != "" {
		cid, err := token.ParseContractID(contractID)
		if err != nil {
			return ctx, errgo.WithCausef(err, params.ErrContractNotFound, "")
		}
		if body.ContractID != cid {
			return ctx, forbiddenErrorf(nil, fmt.Sprintf("token invalid for contract %q", contractID))
		}
	}

	if body.MachineID != machineID {
		return ctx, forbiddenErrorf(nil, fmt.Sprintf("token invalid for machine %q", machineID))
	}

	return contextWithMachineToken(ctx, mt, body), nil
}

func (h *handler) authSSOMacaroon(p httprequest.Params) (context.Context, error) {
	auth := p.Request.Header.Get("Authorization")
	if auth == "" || !strings.HasPrefix(auth, "Macaroon ") {
		// TODO add WWW-Authenticate response header.
		return p.Context, unauthorizedErrorf(nil, "no authorization macaroon provided")
	}
	token := strings.TrimPrefix(auth, "Macaroon ")

	ms, err := unmarshalMacaroon(token)
	if err != nil {
		// TODO add WWW-Authenticate response header.
		return p.Context, unauthorizedErrorf(err, "invalid macaroon")
	}

	account, err := h.p.SSOAuth.Authenticate(p.Context, ms)
	if err != nil {
		// TODO add WWW-Authenticate response header.
		return p.Context, unauthorizedErrorf(err, "invalid macaroon")
	}

	if account.OpenID == "" {
		// TODO add WWW-Authenticate response header.
		return p.Context, unauthorizedErrorf(nil, "invalid account")
	}

	u := store.User{
		ID:          account.OpenID,
		Username:    account.Username,
		DisplayName: account.DisplayName,
		Email:       account.Email,
		Verified:    account.IsVerified,
	}
	if err := h.p.Store.EnsureUser(p.Context, u); err != nil {
		return p.Context, errgo.Mask(err)
	}

	return contextWithUser(p.Context, u), nil
}

func (h *handler) authAWSSignature(ctx context.Context, req *params.GetContractTokenForAWSInstanceRequest) (context.Context, error) {
	if h.p.AWSDecoder == nil {
		return ctx, errgo.WithCausef(nil, params.ErrForbidden, "AWS instance registration not supported")
	}
	doc, err := h.p.AWSDecoder.Decode(req.Body.Pkcs7)
	if err != nil {
		return ctx, errgo.WithCausef(err, params.ErrForbidden, "forbidden")
	}
	return contextWithAWSDoc(ctx, *doc), nil
}

func extractTokenFromHeader(req *http.Request) (string, error) {
	h := req.Header.Get("Authorization")
	if h == "" {
		// TODO add WWW-Authenticate response header.
		return "", unauthorizedErrorf(nil, "no authorization token provided")
	}
	token := strings.TrimPrefix(h, "Bearer ")
	if len(token) == len(h) {
		return "", unauthorizedErrorf(nil, "bearer token expected")
	}
	return token, nil
}

func unmarshalMacaroon(token string) (macaroon.Slice, error) {
	buf, err := macaroon.Base64Decode([]byte(token))
	if err != nil {
		return nil, errgo.Mask(err)
	}
	if len(buf) == 0 {
		return nil, errgo.New("macaroon not found")
	}
	var ms macaroon.Slice
	switch buf[0] {
	case '[':
		if err := json.Unmarshal(buf, &ms); err != nil {
			return nil, errgo.Mask(err)
		}
	default:
		if err := ms.UnmarshalBinary(buf); err != nil {
			return nil, errgo.Mask(err)
		}
	}
	return ms, nil
}

type contextMachineToken struct {
	token token.MachineToken
	body  token.MachineTokenBody
}

type machineTokenKey struct{}

func machineTokenFromContext(ctx context.Context) contextMachineToken {
	tok, _ := ctx.Value(machineTokenKey{}).(contextMachineToken)
	return tok
}

// contextWithMachineToken returns the given context associated with the
// given machine token.
func contextWithMachineToken(ctx context.Context, t token.MachineToken, body token.MachineTokenBody) context.Context {
	return context.WithValue(ctx, machineTokenKey{}, contextMachineToken{
		token: t,
		body:  body,
	})
}

type contextContractIDKey struct{}

func contextWithContractID(ctx context.Context, id token.ContractID) context.Context {
	return context.WithValue(ctx, contextContractIDKey{}, id)
}

func contractIDFromContext(ctx context.Context) token.ContractID {
	id, _ := ctx.Value(contextContractIDKey{}).(token.ContractID)
	return id
}

type contextUserKey struct{}

func contextWithUser(ctx context.Context, user store.User) context.Context {
	return context.WithValue(ctx, contextUserKey{}, user)
}

func userFromContext(ctx context.Context) store.User {
	user, _ := ctx.Value(contextUserKey{}).(store.User)
	return user
}

type contextAWSDocKey struct{}

func contextWithAWSDoc(ctx context.Context, doc awsauth.InstanceInformationDocument) context.Context {
	return context.WithValue(ctx, contextAWSDocKey{}, doc)
}

func awsDocFromContext(ctx context.Context) awsauth.InstanceInformationDocument {
	doc, _ := ctx.Value(contextAWSDocKey{}).(awsauth.InstanceInformationDocument)
	return doc
}
