package params

import (
	"encoding/json"
	"fmt"

	errgo "gopkg.in/errgo.v1"
)

const (
	ResourceLivepatch   = "livepatch"
	ResourceESM         = "esm"
	ResourceFIPS        = "fips"
	ResourceFIPSUpdates = "fips-updates"
	ResourceCC          = "cc"
	ResourceSupport     = "support"
)

// EntitlementUnknown is the entitlement that is returned when the server
// returns an entitlement that isn't recognised.
type EntitlementUnknown struct {
	Type        string
	Entitlement json.RawMessage
}

func (c *ContractInfo) UnmarshalJSON(data []byte) error {
	type noUnmarshal ContractInfo
	var info struct {
		noUnmarshal
		ResourceEntitlements []json.RawMessage `json:"resourceEntitlements"`
	}
	if err := json.Unmarshal(data, &info); err != nil {
		return errgo.Mask(err)
	}
	var entitlements []Entitlement
	if info.ResourceEntitlements != nil {
		entitlements = make([]Entitlement, len(info.ResourceEntitlements))

		for i, m := range info.ResourceEntitlements {
			var e Entitlement
			if err := unmarshalEntitlement(m, &e); err != nil {
				return errgo.Mask(err)
			}
			entitlements[i] = e
		}
	}
	*c = ContractInfo(info.noUnmarshal)
	c.ResourceEntitlements = entitlements

	return nil
}

func (r *GetResourceMachineAccessResponse) UnmarshalJSON(data []byte) error {
	type noUnmarshal GetResourceMachineAccessResponse
	var ru struct {
		noUnmarshal
		Entitlement json.RawMessage `json:"entitlement"`
	}
	if err := json.Unmarshal(data, &ru); err != nil {
		return errgo.Mask(err)
	}
	if ru.Entitlement == nil {
		*r = GetResourceMachineAccessResponse(ru.noUnmarshal)
		return nil
	}
	if err := unmarshalEntitlement(ru.Entitlement, &ru.noUnmarshal.Entitlement); err != nil {
		return errgo.Mask(err)
	}
	*r = GetResourceMachineAccessResponse(ru.noUnmarshal)
	return nil
}

func unmarshalEntitlement(data []byte, e *Entitlement) error {
	var discrim struct {
		Type string `json:"type"`
	}
	if err := json.Unmarshal(data, &discrim); err != nil {
		return errgo.Mask(err)
	}
	switch discrim.Type {
	case ResourceLivepatch:
		var r EntitlementLivepatch
		if err := json.Unmarshal(data, &r); err != nil {
			return errgo.Mask(err)
		}
		*e = r
	case ResourceCC, ResourceESM, ResourceFIPS, ResourceFIPSUpdates:
		var r EntitlementAptRepository
		if err := json.Unmarshal(data, &r); err != nil {
			return errgo.Mask(err)
		}
		*e = r
	case ResourceSupport:
		var r EntitlementSupport
		if err := json.Unmarshal(data, &r); err != nil {
			return errgo.Mask(err)
		}
		*e = r
	default:
		*e = EntitlementUnknown{
			Type:        discrim.Type,
			Entitlement: append([]byte(nil), data...),
		}
	}
	return nil
}

func (e EntitlementAptRepository) IsEntitled() bool {
	return e.Entitled
}

func (e EntitlementAptRepository) ResourceName() string {
	return e.Type
}

func (e EntitlementAptRepository) GetAffordances() interface{} {
	return e.Affordances
}

func (e EntitlementLivepatch) IsEntitled() bool {
	return e.Entitled
}

func (e EntitlementLivepatch) ResourceName() string {
	return e.Type
}

func (e EntitlementLivepatch) GetAffordances() interface{} {
	return e.Affordances
}

func (e EntitlementSupport) IsEntitled() bool {
	return e.Entitled
}

func (e EntitlementSupport) ResourceName() string {
	return e.Type
}

func (e EntitlementSupport) GetAffordances() interface{} {
	return e.Affordances
}

// MarshalYAML implements yaml.Marshaler.
func (r ProductResource) MarshalYAML() (interface{}, error) {
	return yamlProductResource(r), nil
}

// UnmarshalYAML implements yaml.Unmarshaler.
func (r *ProductResource) UnmarshalYAML(unmarshal func(interface{}) error) error {
	var ypr yamlProductResource
	if err := unmarshal(&ypr); err != nil {
		return errgo.Mask(err, errgo.Any)
	}
	data, err := rewriteMaps("data", ypr.Data)
	if err != nil {
		return errgo.Mask(err)
	}
	ypr.Data = data
	*r = ProductResource(ypr)
	return nil
}

// A yamlProductResource is identical to a ProductResource, except that
// it has tags for marshaling and unmarshaling as YAML.
type yamlProductResource struct {
	CanOverride  bool `yaml:"canOverride"`
	Data         interface{}
	ResourceType string `yaml:"resourceType"`
}

// rewriteMaps rewrites maps of type map[interface{}]interface{} as
// produced by the yaml decoder into map[string]interface{} which can be
// encoded by JSON.
func rewriteMaps(path string, x interface{}) (interface{}, error) {
	switch x := x.(type) {
	case map[interface{}]interface{}:
		nm := make(map[string]interface{})
		for k, v := range x {
			k1, ok := k.(string)
			if !ok {
				return nil, errgo.Newf("map key at %s.%v (type %T) is not string", path, x, k)
			}
			v1, err := rewriteMaps(path+"."+k1, v)
			if err != nil {
				return nil, err
			}
			nm[k1] = v1
		}
		return nm, nil
	case []interface{}:
		nx := make([]interface{}, len(x))
		for i := range x {
			v, err := rewriteMaps(fmt.Sprintf("%s[%d]", path, i), x[i])
			if err != nil {
				return nil, err
			}
			nx[i] = v
		}
		return nx, nil
	default:
		return x, nil
	}
}
