// Copyright 2019 Canonical Ltd.

// Package main contains an implementation of a command used to backload
// KPI metrics, It pull entities from postgresql and creates appropriate
// audit messages.
package main

import (
	"context"
	"database/sql"
	"io"
	"log"
	"os"
	"time"

	"github.com/CanonicalLtd/ua-contracts/internal/kafkaauditor"
	"github.com/CanonicalLtd/ua-contracts/internal/postgresstore"
	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
	"github.com/Shopify/sarama"
	errgo "gopkg.in/errgo.v1"
)

func init() {
	sarama.Logger = log.New(os.Stderr, "", log.LstdFlags)
}

// Auditor defines an interface used by the command to send audit messages.
type Auditor interface {
	Audit(ctx context.Context, _ time.Time, entry kafkaauditor.AuditEntry)
}

func main() {
	ctx := context.Background()

	cfg, err := kafkaauditor.KafkaConfig()
	if err != nil {
		log.Fatalf("failed to configure a kafka auditor: %v ", err)
	}
	auditor, err := cfg.Auditor()
	if err != nil {
		log.Fatalf("failed to configure new auditor: %v", err)
	}
	defer func() {
		closer, ok := auditor.(io.Closer)
		if ok {
			err := closer.Close()
			if err != nil {
				log.Fatalf("failed to close auditor: %v", err)
			}
		}
	}()

	// open the postgres connection
	db, err := sql.Open("postgres", "")
	if err != nil {
		log.Fatalf("failed to open the postgres connection: %v", err)
	}
	defer func() {
		err := db.Close()
		if err != nil {
			log.Fatalf("failed to close the postgres connection: %v", err)
		}
	}()

	if err := sendAuditMessages(ctx, auditor, db); err != nil {
		log.Fatalf("failed to backload kpi data: %v", err)
	}
}

func sendAuditMessages(ctx context.Context, auditor Auditor, db *sql.DB) error {
	// we use the postgresstore whereever possible
	pgstore, err := postgresstore.New(db)
	if err != nil {
		return errgo.Notef(err, "failed to create a postgres store")
	}

	// first we collect all products
	if err := auditProducts(ctx, db, pgstore, auditor); err != nil {
		log.Printf("failed to send product data: %v", err)
	}

	// accounts
	if err := auditAccounts(ctx, db, pgstore, auditor); err != nil {
		log.Printf("failed to send account data: %v", err)
	}

	// users
	if err := auditUsers(ctx, db, pgstore, auditor); err != nil {
		log.Printf("failed to send user data: %v", err)
	}

	// contracts
	if err := auditContracts(ctx, db, pgstore, auditor); err != nil {
		log.Printf("failed to send contract data: %v", err)
	}

	// machines
	if err := auditMachines(ctx, db, pgstore, auditor); err != nil {
		log.Printf("failed to send machine data: %v", err)
	}

	// service accounts
	if err := auditServiceAccounts(ctx, db, pgstore, auditor); err != nil {
		log.Printf("failed to send machine data: %v", err)
	}

	return nil
}

func auditProducts(ctx context.Context, db *sql.DB, pgstore store.Store, auditor Auditor) error {
	products, err := pgstore.AllProducts(ctx)
	if err != nil {
		return errgo.Notef(err, "failed to retrieve products")
	}
	// and send AuditAddProduct messages
	for _, p := range products {
		resources := make([]kafkaauditor.ProductResource, len(p.Resources))
		for i, r := range p.Resources {
			resources[i] = kafkaauditor.ProductResource{
				Type: r.ResourceType,
				// Data is not relevant to KPIs atm.
			}
		}
		// audit time does not really matter here as we have no
		// KPIs based on product creation time
		auditor.Audit(ctx, time.Now(), &kafkaauditor.AuditAddProduct{
			ProductID:   p.ID,
			ProductName: p.Name,
			Resources:   resources,
		})
	}
	return nil
}

func auditAccounts(ctx context.Context, db *sql.DB, pgstore store.Store, auditor Auditor) error {
	// then we find all accounts
	rows, err := db.Query(`SELECT id, name, created FROM accounts;`)
	if err != nil {
		return errgo.Notef(err, "failed to retrieve accounts")
	}
	defer rows.Close()

	// and send AuditAddAccount messages
	for rows.Next() {
		var a store.Account
		err := rows.Scan(
			&a.ID,
			&a.Name,
			&a.Created,
		)
		if err != nil {
			return errgo.Notef(err, "failed scan an account row")
		}

		auditor.Audit(ctx, a.Created, &kafkaauditor.AuditAddAccount{
			AccountID:      a.ID,
			AccountName:    a.Name,
			AccountCreated: a.Created,
		})
	}
	if err := rows.Err(); err != nil {
		return errgo.Notef(err, "account rows error")
	}
	return nil
}

func auditUsers(ctx context.Context, db *sql.DB, pgstore store.Store, auditor Auditor) error {
	rows, err := db.Query(`SELECT id, name, display_name, email, verified FROM users;`)
	if err != nil {
		return errgo.Notef(err, "failed to retrieve accounts")
	}
	defer rows.Close()

	for rows.Next() {
		var u store.User
		err := rows.Scan(
			&u.ID,
			&u.Username,
			&u.DisplayName,
			&u.Email,
			&u.Verified,
		)
		if err != nil {
			return errgo.Notef(err, "failed to scan a user row")
		}

		// audit time does not really matter here as we have no
		// KPIs based on user creation time
		auditor.Audit(ctx, time.Now(), &kafkaauditor.AuditEnsureUser{
			ID:          u.ID,
			Username:    u.Username,
			DisplayName: u.DisplayName,
			Email:       u.Email,
			Verified:    u.Verified,
		})
	}
	if err := rows.Err(); err != nil {
		return errgo.Notef(err, "user rows error")
	}
	return nil
}

func auditContracts(ctx context.Context, db *sql.DB, pgstore store.Store, auditor Auditor) error {
	// since postgresstore does not
	// have a method to fetch all contracts, we fetch all contract
	// ids and the use the Contract method on the postgresstore
	// to fetch individual contracts
	rows, err := db.Query(`SELECT id FROM contracts;`)
	if err != nil {
		return errgo.Notef(err, "failed to retrieve contracts")
	}
	defer rows.Close()

	for rows.Next() {
		var contractID token.ContractID
		err := rows.Scan(&contractID)
		if err != nil {
			return errgo.Notef(err, "failed to retreve contract %s", contractID)
		}
		c, err := pgstore.Contract(ctx, contractID)
		if err != nil {
			return errgo.Notef(err, "failed to retrieve contract %s", contractID)
		}
		auditor.Audit(ctx, c.Contract.Created, &kafkaauditor.AuditAddContract{
			ContractID:    c.Contract.ID,
			Name:          c.Contract.Name,
			AccountID:     c.Account.ID,
			ProductIDs:    c.Contract.ProductIDs,
			Created:       c.Contract.Created,
			CreatedBy:     c.Contract.CreatedBy,
			EffectiveFrom: c.Contract.EffectiveFrom,
			EffectiveTo:   c.Contract.EffectiveTo,
		})

		// we also fetch any contract tokens
		tokens, err := pgstore.ContractTokens(ctx, contractID)
		if err != nil {
			log.Printf("failed to retrieve contract tokens %s: %v", contractID, err)
			continue
		}
		for _, t := range tokens {
			// unfortunately we do not store the token creation time..
			auditor.Audit(ctx, time.Now(), &kafkaauditor.AuditAddContractToken{
				ContractToken: t.ContractToken,
				ContractID:    t.ContractID,
				Expires:       t.Expires,
			})
		}
	}
	if err := rows.Err(); err != nil {
		return errgo.Notef(err, "contract rows error")
	}
	return nil
}

func auditMachines(ctx context.Context, db *sql.DB, pgstore store.Store, auditor Auditor) error {
	rows, err := db.Query(`SELECT id FROM machines;`)
	if err != nil {
		return errgo.Notef(err, "failed to retrieve machines")
	}
	defer rows.Close()

	for rows.Next() {
		var machineID string
		err := rows.Scan(&machineID)
		if err != nil {
			return errgo.Notef(err, "failed to scan a machine row")
		}
		m, err := pgstore.MachineInfo(ctx, machineID)
		if err != nil {
			return errgo.Notef(err, "failed to retrieve machine info %s", machineID)
		}
		// unfortunately we do not store the machine creation time
		auditor.Audit(ctx, time.Now(), &kafkaauditor.AuditEnsureMachine{
			MachineID:    m.MachineID,
			Architecture: m.Architecture,
			OS:           m.OS,
		})

		// fetch the contract this machine belongs to
		var contractID token.ContractID
		err = db.QueryRow(`SELECT contract_id FROM contract_machines WHERE machine_id = $1;`, machineID).Scan(&contractID)
		if err != nil {
			log.Printf("cannot find contract for machine %s: %v", machineID, err)
			continue
		}
		// again, we do not store the machine creation time
		auditor.Audit(ctx, time.Now(), &kafkaauditor.AuditAddContractMachine{
			ContractID: contractID,
			MachineID:  machineID,
		})

	}
	if err := rows.Err(); err != nil {
		return errgo.Notef(err, "machine rows error")
	}
	return nil
}

func auditServiceAccounts(ctx context.Context, db *sql.DB, pgstore store.Store, auditor Auditor) error {
	rows, err := db.Query(`SELECT name FROM service_accounts;`)
	if err != nil {
		return errgo.Notef(err, "failed to retrieve service_accounts")
	}
	defer rows.Close()

	for rows.Next() {
		var name string
		err := rows.Scan(&name)
		if err != nil {
			return errgo.Notef(err, "failed to scan service account row")
		}

		// unfortunately we do not store the service account creation time
		auditor.Audit(ctx, time.Now(), &kafkaauditor.AuditEnsureServiceAccount{
			Name: name,
		})
	}
	if err := rows.Err(); err != nil {
		return errgo.Notef(err, "service account rows error")
	}
	return nil
}
