// Copyright 2019 Canonical Ltd.

package main_test

import (
	"context"
	"io/ioutil"
	"path/filepath"
	"testing"
	"time"

	qt "github.com/frankban/quicktest"
	"github.com/juju/postgrestest"
	errgo "gopkg.in/errgo.v1"

	backload "github.com/CanonicalLtd/ua-contracts/cmd/backload-kpis"
	"github.com/CanonicalLtd/ua-contracts/internal/kafkaauditor"
	"github.com/CanonicalLtd/ua-contracts/internal/postgresstore"
	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
	"github.com/CanonicalLtd/ua-contracts/params"
)

func TestAuditProducts(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	defer c.Done()

	f := newFixture(c)

	err := f.store.AddProduct(ctx, store.Product{
		ID:   "test-product-1",
		Name: "test-product-1",
		Resources: []store.ProductResource{{
			ResourceType: "test-resource-type1",
			Data:         `{"test":"data"}`,
		}},
	})
	c.Assert(err, qt.IsNil)
	err = f.store.AddProduct(ctx, store.Product{
		ID:   "test-product-2",
		Name: "test-product-2",
		Resources: []store.ProductResource{{
			ResourceType: "test-resource-type2",
			Data:         `{"test":"data"}`,
		}},
	})
	c.Assert(err, qt.IsNil)

	err = backload.AuditProducts(ctx, f.pg.DB, f.store, f.auditor)
	c.Assert(err, qt.IsNil)

	c.Assert(f.auditor.entries, qt.DeepEquals, []kafkaauditor.AuditEntry{
		&kafkaauditor.AuditAddProduct{
			ProductID:   "test-product-1",
			ProductName: "test-product-1",
			Resources:   []kafkaauditor.ProductResource{{Type: "test-resource-type1"}},
		},
		&kafkaauditor.AuditAddProduct{
			ProductID:   "test-product-2",
			ProductName: "test-product-2",
			Resources:   []kafkaauditor.ProductResource{{Type: "test-resource-type2"}},
		},
	},
	)
}

func TestAuditUsers(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	defer c.Done()

	f := newFixture(c)

	err := f.store.EnsureUser(ctx, store.User{
		ID:          "test-id1",
		Username:    "test-user1",
		DisplayName: "test-user-1",
		Email:       "user@test.com",
		Verified:    true,
	})
	c.Assert(err, qt.IsNil)
	err = f.store.EnsureUser(ctx, store.User{
		ID:          "test-id2",
		Username:    "test-user2",
		DisplayName: "test-user-2",
		Email:       "user@test.com",
		Verified:    false,
	})
	c.Assert(err, qt.IsNil)

	err = backload.AuditUsers(ctx, f.pg.DB, f.store, f.auditor)
	c.Assert(err, qt.IsNil)

	c.Assert(f.auditor.entries, qt.DeepEquals, []kafkaauditor.AuditEntry{
		&kafkaauditor.AuditEnsureUser{
			ID:          "test-id1",
			Username:    "test-user1",
			DisplayName: "test-user-1",
			Email:       "user@test.com",
			Verified:    true,
		},
		&kafkaauditor.AuditEnsureUser{
			ID: "test-id2", Username: "test-user2",
			DisplayName: "test-user-2",
			Email:       "user@test.com",
			Verified:    false,
		},
	})
}

func TestAuditAccounts(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	defer c.Done()

	f := newFixture(c)

	tok1 := token.MustNewAccountID()
	t1 := time.Now().Truncate(time.Millisecond)
	tok2 := token.MustNewAccountID()
	t2 := time.Now().Truncate(time.Millisecond)
	err := f.store.AddAccount(ctx, store.Account{
		ID:                  tok1,
		Name:                "test-account-1",
		Created:             t1,
		SalesforceAccountID: "test-sf-id-1",
		Origin:              "test",
	}, "test-user-1")
	c.Assert(err, qt.IsNil)
	err = f.store.AddAccount(ctx, store.Account{
		ID:                  tok2,
		Name:                "test-account-2",
		Created:             t2,
		SalesforceAccountID: "test-sf-id-2",
		Origin:              "test",
	}, "test-user-2")
	c.Assert(err, qt.IsNil)

	err = backload.AuditAccounts(ctx, f.pg.DB, f.store, f.auditor)
	c.Assert(err, qt.IsNil)

	c.Assert(f.auditor.entries, qt.DeepEquals, []kafkaauditor.AuditEntry{
		&kafkaauditor.AuditAddAccount{
			AccountID:      tok1,
			AccountName:    "test-account-1",
			AccountCreated: t1,
		},
		&kafkaauditor.AuditAddAccount{
			AccountID:      tok2,
			AccountName:    "test-account-2",
			AccountCreated: t2,
		},
	},
	)
}

func TestAuditContracts(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	defer c.Done()

	f := newFixture(c)

	accountID1 := token.MustNewAccountID()
	t1 := time.Now().Truncate(time.Millisecond)
	accountID2 := token.MustNewAccountID()
	t2 := time.Now().Truncate(time.Millisecond)

	err := f.store.AddProduct(ctx, store.Product{
		ID:   "test-product-1",
		Name: "test-product-1",
		Resources: []store.ProductResource{{
			ResourceType: "test-resource-type1",
			Data:         `{"test":"data"}`,
		}},
	})
	c.Assert(err, qt.IsNil)

	err = f.store.AddAccount(ctx, store.Account{
		ID:                  accountID1,
		Name:                "test-account-1",
		Created:             t1,
		SalesforceAccountID: "test-sf-id-1",
		Origin:              "test",
	}, "test-user-1")
	c.Assert(err, qt.IsNil)

	err = f.store.AddAccount(ctx, store.Account{
		ID:                  accountID2,
		Name:                "test-account-2",
		Created:             t2,
		SalesforceAccountID: "test-sf-id-2",
		Origin:              "test",
	}, "test-user-2")
	c.Assert(err, qt.IsNil)

	contractID1, err := f.store.AddContract(ctx, store.Contract{
		Name:      "test-contract-1",
		AccountID: accountID1,
		ProductIDs: []string{
			"test-product-1",
		},
		Created: t1,
	})
	c.Assert(err, qt.IsNil)

	contractID2, err := f.store.AddContract(ctx, store.Contract{
		Name:      "test-contract-2",
		AccountID: accountID2,
		ProductIDs: []string{
			"test-product-1",
		},
		Created: t2,
	})
	c.Assert(err, qt.IsNil)

	err = backload.AuditContracts(ctx, f.pg.DB, f.store, f.auditor)
	c.Assert(err, qt.IsNil)

	c.Assert(f.auditor.entries, qt.DeepEquals, []kafkaauditor.AuditEntry{
		&kafkaauditor.AuditAddContract{
			ContractID: contractID1,
			Name:       "test-contract-1",
			AccountID:  accountID1,
			ProductIDs: []string{"test-product-1"},
			Created:    t1,
		},
		&kafkaauditor.AuditAddContract{
			ContractID: contractID2,
			Name:       "test-contract-2",
			AccountID:  accountID2,
			ProductIDs: []string{"test-product-1"},
			Created:    t2,
		},
	})
}

func TestAuditMachines(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	defer c.Done()

	f := newFixture(c)

	accountID1 := token.MustNewAccountID()
	t1 := time.Now().Truncate(time.Millisecond)
	accountID2 := token.MustNewAccountID()
	t2 := time.Now().Truncate(time.Millisecond)

	err := f.store.AddProduct(ctx, store.Product{
		ID:   "test-product-1",
		Name: "test-product-1",
		Resources: []store.ProductResource{{
			ResourceType: "test-resource-type1",
			Data:         `{"test":"data"}`,
		}},
	})
	c.Assert(err, qt.IsNil)

	err = f.store.AddAccount(ctx, store.Account{
		ID:                  accountID1,
		Name:                "test-account-1",
		Created:             t1,
		SalesforceAccountID: "test-sf-id-1",
		Origin:              "test",
	}, "test-user-1")
	c.Assert(err, qt.IsNil)

	err = f.store.AddAccount(ctx, store.Account{
		ID:                  accountID2,
		Name:                "test-account-2",
		Created:             t2,
		SalesforceAccountID: "test-sf-id-2",
		Origin:              "test",
	}, "test-user-2")
	c.Assert(err, qt.IsNil)

	contractID1, err := f.store.AddContract(ctx, store.Contract{
		Name:      "test-contract-1",
		AccountID: accountID1,
		ProductIDs: []string{
			"test-product-1",
		},
		Created: t1,
	})
	c.Assert(err, qt.IsNil)

	contractID2, err := f.store.AddContract(ctx, store.Contract{
		Name:      "test-contract-2",
		AccountID: accountID2,
		ProductIDs: []string{
			"test-product-1",
		},
		Created: t2,
	})
	c.Assert(err, qt.IsNil)

	err = f.store.EnsureMachine(ctx, store.MachineInfo{
		MachineID:    "test-machine-1",
		Architecture: "amd64",
		OS:           params.OS{},
	})
	c.Assert(err, qt.IsNil)
	err = f.store.EnsureMachine(ctx, store.MachineInfo{
		MachineID:    "test-machine-2",
		Architecture: "arm64",
		OS:           params.OS{},
	})
	c.Assert(err, qt.IsNil)

	err = f.store.AddContractMachine(ctx, contractID1, "test-machine-1")
	c.Assert(err, qt.IsNil)
	err = f.store.AddContractMachine(ctx, contractID2, "test-machine-2")
	c.Assert(err, qt.IsNil)

	err = backload.AuditMachines(ctx, f.pg.DB, f.store, f.auditor)
	c.Assert(err, qt.IsNil)

	c.Assert(f.auditor.entries, qt.DeepEquals, []kafkaauditor.AuditEntry{
		&kafkaauditor.AuditEnsureMachine{MachineID: "test-machine-1", Architecture: "amd64"},
		&kafkaauditor.AuditAddContractMachine{
			ContractID: contractID1,
			MachineID:  "test-machine-1",
		},
		&kafkaauditor.AuditEnsureMachine{MachineID: "test-machine-2", Architecture: "arm64"},
		&kafkaauditor.AuditAddContractMachine{
			ContractID: contractID2,
			MachineID:  "test-machine-2",
		},
	})
}

func TestServiceAccounts(t *testing.T) {
	c := qt.New(t)

	ctx := context.Background()
	defer c.Done()

	f := newFixture(c)

	err := f.store.EnsureServiceAccount(ctx, store.ServiceAccount{
		Name: "test-service-account-1",
	})
	c.Assert(err, qt.IsNil)
	err = f.store.EnsureServiceAccount(ctx, store.ServiceAccount{
		Name: "test-service-account-2",
	})
	c.Assert(err, qt.IsNil)

	err = backload.AuditServiceAccounts(ctx, f.pg.DB, f.store, f.auditor)
	c.Assert(err, qt.IsNil)

	c.Assert(f.auditor.entries, qt.DeepEquals, []kafkaauditor.AuditEntry{
		&kafkaauditor.AuditEnsureServiceAccount{Name: "test-service-account-1"},
		&kafkaauditor.AuditEnsureServiceAccount{Name: "test-service-account-2"},
	})
}

// test implementations below this line

type testAuditor struct {
	entries []kafkaauditor.AuditEntry
}

func (t *testAuditor) Audit(ctx context.Context, _ time.Time, entry kafkaauditor.AuditEntry) {
	t.entries = append(t.entries, entry)
}

type fixture struct {
	store   store.Store
	pg      *postgrestest.DB
	auditor *testAuditor
}

func newFixture(c *qt.C) *fixture {
	pg, err := postgrestest.New()
	if errgo.Cause(err) == postgrestest.ErrDisabled {
		c.Skip(err.Error())
	}
	c.Assert(err, qt.Equals, nil)
	c.Defer(func() {
		if err := pg.Close(); err != nil {
			c.Error(err)
		}
	})
	// Load any SQL initialization files.
	files, err := filepath.Glob("../../internal/postgresstore/sql/*.sql")
	c.Assert(err, qt.Equals, nil)
	for _, fn := range files {
		b, err := ioutil.ReadFile(fn)
		c.Assert(err, qt.Equals, nil)
		_, err = pg.DB.Exec(string(b))
		c.Assert(err, qt.Equals, nil)
	}
	pgstore, err := postgresstore.New(pg.DB)
	c.Assert(err, qt.Equals, nil)

	return &fixture{
		pg:      pg,
		store:   pgstore,
		auditor: &testAuditor{},
	}
}
