// Copyright 2019 Canonical Ltd.

package main

import (
	"context"

	"github.com/juju/cmd"
	"github.com/juju/gnuflag"
	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/params"
)

// addAccountCommand implements the add-account command.
type addAccountCommand struct {
	cmd.CommandBase

	name   string
	output cmd.Output
}

const addAccountDoc = `
Create a new account with the specified name.

    contract add-account <name>
`

// Info implements cmd.Command.Info.
func (c *addAccountCommand) Info() *cmd.Info {
	return &cmd.Info{
		Name:           "add-account",
		Purpose:        "Create an account.",
		Args:           "<name>",
		Doc:            addAccountDoc[1:],
		ShowSuperFlags: commonFlags,
	}
}

// Init implements cmd.Command.Init.
func (c *addAccountCommand) Init(args []string) error {
	if len(args) < 1 {
		return errgo.New("account name not specified")
	}
	c.name = args[0]
	if len(args) > 1 {
		return errgo.Newf("too many args")
	}
	return nil
}

// SetFlags implements cmd.Command.SetFlags.
func (c *addAccountCommand) SetFlags(f *gnuflag.FlagSet) {
	formatters := map[string]cmd.Formatter{
		"json": cmd.FormatJson,
		"yaml": formatYAML,
	}
	c.output.AddFlags(f, "yaml", formatters)
}

// Run implements cmd.Command.Run.
func (c *addAccountCommand) Run(ctxt *cmd.Context) error {
	ctx := context.Background()

	info, err := client().AddAccount(ctx, &params.AddAccountRequest{
		Body: params.AddAccountRequestBody{
			Name: c.name,
		},
	})
	if err != nil {
		switch errorCode(err) {
		case params.ErrForbidden:
			return errgo.New("permission denied")
		default:
			return errgo.Mask(err)
		}
	}

	return errgo.Mask(c.output.Write(ctxt, info))
}
