// Copyright 2019 Canonical Ltd.

package main

import (
	"context"

	"github.com/juju/cmd"
	"github.com/juju/gnuflag"
	errgo "gopkg.in/errgo.v1"
	yaml "gopkg.in/yaml.v2"

	"github.com/CanonicalLtd/ua-contracts/params"
)

// addContractCommand implements the add-contract command.
type addContractCommand struct {
	cmd.CommandBase

	accountID string
	file      cmd.FileVar
	output    cmd.Output
}

const addContractDoc = `
Create the specified contract. The contract will be set to the value in
the specified file. The filename "-" will use the standard input.

    contract add-contract <account-id> new-contract.yaml

The input file should be a YAML document like:

    name: <contract name>
    createdBy: <creator>
    effectiveFrom: <time>
    effectiveTo: <time>
    products:
     - <product id>
     - <product id>
`

// Info implements cmd.Command.Info.
func (c *addContractCommand) Info() *cmd.Info {
	return &cmd.Info{
		Name:           "add-contract",
		Purpose:        "Create a contract.",
		Args:           "<account-id> <filename>",
		Doc:            addContractDoc[1:],
		ShowSuperFlags: commonFlags,
	}
}

// Init implements cmd.Command.Init.
func (c *addContractCommand) Init(args []string) error {
	if len(args) < 1 {
		return errgo.New("account ID not specified")
	}
	c.accountID = args[0]
	if len(args) < 2 {
		return errgo.New("filename not specified")
	}
	c.file.Path = args[1]
	if len(args) > 2 {
		return errgo.Newf("too many args")
	}
	return nil
}

// SetFlags implements cmd.Command.SetFlags.
func (c *addContractCommand) SetFlags(f *gnuflag.FlagSet) {
	formatters := map[string]cmd.Formatter{
		"json": cmd.FormatJson,
		"yaml": formatYAML,
	}
	c.output.AddFlags(f, "yaml", formatters)
}

// Run implements cmd.Command.Run.
func (c *addContractCommand) Run(ctxt *cmd.Context) error {
	ctx := context.Background()

	buf, err := c.file.Read(ctxt)
	if err != nil {
		return errgo.Mask(err)
	}

	var contract params.AddAccountContractRequestBody
	if err := yaml.UnmarshalStrict(buf, &contract); err != nil {
		return errgo.Notef(err, "invalid contract definition")
	}

	if contract.Name == "" {
		return errgo.New("invalid contract definition: name must be specified")
	}
	if contract.CreatedBy == "" {
		return errgo.New("invalid contract definition: createdby must be specified")
	}
	if len(contract.Products) == 0 {
		return errgo.New("invalid contract definition: no products specified")
	}
	info, err := client().AddAccountContract(ctx, &params.AddAccountContractRequest{
		Account: c.accountID,
		Body:    contract,
	})
	if err != nil {
		switch errorCode(err) {
		case params.ErrNotFound:
			return errgo.Newf("account %q not found", c.accountID)
		case params.ErrForbidden:
			return errgo.New("permission denied")
		}
		return errgo.Mask(err)
	}
	return errgo.Mask(c.output.Write(ctxt, info))
}
