// Copyright 2019 Canonical Ltd.

package main

import (
	"context"

	"github.com/juju/cmd"
	"github.com/juju/gnuflag"
	errgo "gopkg.in/errgo.v1"
	yaml "gopkg.in/yaml.v2"

	"github.com/CanonicalLtd/ua-contracts/params"
)

// addProductCommand implements the add-product command.
type addProductCommand struct {
	productSet
}

const addProductDoc = `
Create the specified product. The product will be set to the value in
the specified file. The filename "-" will use the standard input.

    contract add-product new-product.yaml

The input file should be a YAML document like:

    id: <product id>
    name: <product name>
    resources:
     - resourceType: <resource type>
       canOverride: (true|false)
       data:
         key1: value1
         key2: value2
`

// Info implements cmd.Command.Info.
func (c *addProductCommand) Info() *cmd.Info {
	return &cmd.Info{
		Name:           "add-product",
		Purpose:        "Create a product.",
		Args:           "<filename>",
		Doc:            addProductDoc[1:],
		ShowSuperFlags: commonFlags,
	}
}

// Run implements cmd.Command.Run.
func (c *addProductCommand) Run(ctxt *cmd.Context) error {
	return c.run(ctxt, false)
}

// productSet provides shared infrastructure for the add-product and
// update-product commands.
type productSet struct {
	cmd.CommandBase

	file cmd.FileVar
}

func (c *productSet) SetFlags(f *gnuflag.FlagSet) {
	c.file.StdinMarkers = stdinMarkers
}

// Init implements cmd.Command.Init.
func (c *productSet) Init(args []string) error {
	if len(args) < 1 {
		return errgo.New("filename not specified")
	}
	c.file.Path = args[0]
	if len(args) > 1 {
		return errgo.Newf("too many args")
	}
	return nil
}

// run performs the AddProduct API call.
func (c *productSet) run(ctxt *cmd.Context, update bool) error {
	ctx := context.Background()

	buf, err := c.file.Read(ctxt)
	if err != nil {
		return errgo.Mask(err)
	}

	var p params.Product
	if err := yaml.UnmarshalStrict(buf, &p); err != nil {
		return errgo.Notef(err, "invalid product definition")
	}

	if p.Id == "" {
		return errgo.New("invalid product definition: empty ID")
	}

	if update {
		err = client().UpdateProduct(ctx, &params.UpdateProductRequest{
			ProductID: p.Id,
			Body:      p,
		})
	} else {
		err = client().AddProduct(ctx, &params.AddProductRequest{
			Body: p,
		})
	}
	if err != nil {
		switch errorCode(err) {
		case params.ErrNotFound:
			return errgo.Newf("product %q not found", p.Id)
		case params.ErrAlreadyExists:
			return errgo.Newf("product %q already exists", p.Id)
		case params.ErrForbidden:
			return errgo.New("permission denied")
		}
	}
	return errgo.Mask(err)
}
