// Copyright 2019 Canonical Ltd.

package main

import (
	"encoding/json"
	"io"
	"os"
	"strings"
	"text/tabwriter"

	"github.com/juju/cmd"
	"github.com/juju/gnuflag"
	errgo "gopkg.in/errgo.v1"
	yaml "gopkg.in/yaml.v2"

	"github.com/CanonicalLtd/ua-contracts/contractclient"
	"github.com/CanonicalLtd/ua-contracts/params"
)

const (
	cmdName             = "contract"
	defaultContractsURL = "https://contracts.canonical.com"
)

var (
	url  string
	auth string

	// stdinMarkers contains file names that are taken to be stdin.
	stdinMarkers = []string{"-"}

	// commonFlags shows super-flags that should always be shown in
	// help output.
	commonFlags = []string{"a", "auth", "u", "url"}
)

type globalFlagAdder struct{}

func (globalFlagAdder) AddFlags(f *gnuflag.FlagSet) {
	f.StringVar(&url, "u", "", "URL of the contracts service.")
	f.StringVar(&url, "url", "", "")
	f.StringVar(&auth, "a", "", "Contract service credentials, in the format <username>:<password>.")
	f.StringVar(&auth, "auth", "", "")
}

// contractsURL returns the URL of the contracts service.
func contractsURL() string {
	url := url
	if url == "" {
		url = os.Getenv("CONTRACTS_URL")
	}
	if url == "" {
		url = defaultContractsURL
	}
	return url
}

// authorizer creates an authorizer for accessing the contracts service.
func authorizer() contractclient.Authorizer {
	var username, password string
	parts := strings.SplitN(auth, ":", 2)
	username = parts[0]
	if len(parts) > 1 {
		password = parts[1]
	}
	return contractclient.BasicAuthorizer(username, password)
}

// client creates a contractclient.Client for use by commands.
func client() *contractclient.Client {
	return contractclient.New(contractclient.NewParams{
		BaseURL:    contractsURL(),
		Authorizer: authorizer(),
	})
}

// command creates the top-level command for contracts-admin.
func command() cmd.Command {
	sc := cmd.NewSuperCommand(cmd.SuperCommandParams{
		Name:        cmdName,
		Purpose:     "Administration tool for the contracts service.",
		GlobalFlags: globalFlagAdder{},
	})

	sc.Register(new(addAccountCommand))
	sc.Register(new(addContractCommand))
	sc.Register(new(addProductCommand))
	sc.Register(new(configureContractCommand))
	sc.Register(new(grantCommand))
	sc.Register(new(listProductsCommand))
	sc.Register(new(revokeCommand))
	sc.Register(new(setPasswordCommand))
	sc.Register(new(showACLCommand))
	sc.Register(new(showContractCommand))
	sc.Register(new(showProductCommand))
	sc.Register(new(statusCommand))
	sc.Register(new(syncSalesforceCommand))
	sc.Register(new(updateProductCommand))
	return sc
}

// errorcode returns the params.ErrorCode specified in the given error,
// if any.
func errorCode(err error) params.ErrorCode {
	if ec, ok := errgo.Cause(err).(params.ErrorCode); ok {
		return ec
	}
	return ""
}

// formatTabular creates a new table formatter. The given function is
// called to write the table contents, any error returned from the
// function will not have its cause masked.
func formatTabular(f func(w io.Writer, v interface{}) error) cmd.Formatter {
	return func(w io.Writer, v interface{}) error {
		tw := tabwriter.NewWriter(w, 0, 0, 2, ' ', 0)
		if err := f(tw, v); err != nil {
			return errgo.Mask(err, errgo.Any)
		}
		return errgo.Mask(tw.Flush())
	}
}

// formatYAML is similar to cmd.FormatYaml, but
// does not rely on YAML-specific tags in structs
// in the value - instead it always uses the same field
// names as in the JSON.
//
// It also doesn't truncate a final run of newlines,
// which seems like an error in the original.
func formatYAML(w io.Writer, value interface{}) error {
	if value == nil {
		return nil
	}
	jdata, err := json.Marshal(value)
	if err != nil {
		return errgo.Mask(err)
	}
	// Unmarshal the JSON-marshaled data into
	// a generic interface value.
	var x interface{}
	if err := json.Unmarshal(jdata, &x); err != nil {
		return errgo.Mask(err)
	}
	data, err := yaml.Marshal(x)
	if err != nil {
		return errgo.Mask(err)
	}
	if _, err := w.Write(data); err != nil {
		return errgo.Mask(err)
	}
	return nil
}
