// Copyright 2019 Canonical Ltd.

package main

import (
	"context"

	"github.com/juju/cmd"
	errgo "gopkg.in/errgo.v1"
	yaml "gopkg.in/yaml.v2"

	"github.com/CanonicalLtd/ua-contracts/params"
)

// configureContractCommand implements the configure-contract command.
type configureContractCommand struct {
	cmd.CommandBase

	file cmd.FileVar
}

const configureContractDoc = `
This command configures resources for a specific contract.
It is similar to the update-product command except that the
top level id key refers to a contract id instead of a product id.

The resources specified in the file will override the resources
associated with the contract. Fields specified
by the products but not by the contract configuration will remain
as they are.

The resources being configured must have been specified
with canOverride true.

	contract configure-contract contact-config.yaml

The input file should be a YAML document like:

    id: <contract id>
    resources:
     - resourceType: <resource type>
       data:
         key1: value1
         key2: value2
`

// Info implements cmd.Command.Info.
func (c *configureContractCommand) Info() *cmd.Info {
	return &cmd.Info{
		Name:           "configure-contract",
		Purpose:        "Add specific configuration to a contract.",
		Args:           "<filename>",
		Doc:            configureContractDoc[1:],
		ShowSuperFlags: commonFlags,
	}
}

// Init implements cmd.Command.Init.
func (c *configureContractCommand) Init(args []string) error {
	if len(args) < 1 {
		return errgo.New("filename not specified")
	}
	c.file.Path = args[0]
	if len(args) > 1 {
		return errgo.Newf("too many args")
	}
	return nil
}

// Run implements cmd.Command.Run.
func (c *configureContractCommand) Run(ctxt *cmd.Context) error {
	ctx := context.Background()

	buf, err := c.file.Read(ctxt)
	if err != nil {
		return errgo.Mask(err)
	}

	var p struct {
		ID        string                   `yaml:"id"`
		Resources []params.ProductResource `yaml:"resources"`
	}
	if err := yaml.UnmarshalStrict(buf, &p); err != nil {
		return errgo.Notef(err, "invalid product definition")
	}
	if p.ID == "" {
		return errgo.New("invalid contract configuration: empty contract ID")
	}
	err = client().ConfigureContract(ctx, &params.ConfigureContractRequest{
		Contract: p.ID,
		Body: params.ConfigureContractBody{
			Resources: p.Resources,
		},
	})
	if err != nil {
		switch errorCode(err) {
		case params.ErrForbidden:
			return errgo.New("permission denied")
		}
	}
	return errgo.Mask(err)
}
