// Copyright 2019 Canonical Ltd.

package main

import (
	"context"
	"fmt"
	"io"

	"github.com/juju/cmd"
	"github.com/juju/gnuflag"
	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/params"
)

// listProductsCommand implements the list-product command.
type listProductsCommand struct {
	cmd.CommandBase

	output cmd.Output
}

const listProductsDoc = `
List the products known to the contract service.

    contract list-products
`

// Info implements cmd.Command.Info.
func (c *listProductsCommand) Info() *cmd.Info {
	return &cmd.Info{
		Name:           "list-products",
		Purpose:        "List all products.",
		Doc:            listProductsDoc[1:],
		ShowSuperFlags: commonFlags,
	}
}

// SetFlags implements cmd.Command.SetFlags.
func (c *listProductsCommand) SetFlags(f *gnuflag.FlagSet) {
	formatters := map[string]cmd.Formatter{
		"json":    cmd.FormatJson,
		"yaml":    formatYAML,
		"tabular": formatTabular(productsTable),
	}
	c.output.AddFlags(f, "tabular", formatters)
}

// Init implements cmd.Command.Init.
func (c *listProductsCommand) Init(args []string) error {
	if len(args) > 0 {
		return errgo.Newf("too many args")
	}
	return nil
}

// Run implements cmd.Command.Run.
func (c *listProductsCommand) Run(ctxt *cmd.Context) error {
	ctx := context.Background()

	resp, err := client().GetProducts(ctx, &params.GetProductsRequest{})
	if err != nil {
		switch errorCode(err) {
		case params.ErrForbidden:
			return errgo.New("permission denied")
		default:
			return errgo.Mask(err)
		}
	}
	return errgo.Mask(c.output.Write(ctxt, resp.Products))
}

func productsTable(w io.Writer, v interface{}) error {
	if _, err := fmt.Fprint(w, "ID\tNAME"); err != nil {
		return errgo.Mask(err)
	}
	for _, p := range v.([]params.Product) {
		if _, err := fmt.Fprintf(w, "\n%s\t%s", p.Id, p.Name); err != nil {
			return errgo.Mask(err)
		}
	}
	return nil
}
