// Copyright 2019 Canonical Ltd.

package main

import (
	"bytes"
	"context"

	"github.com/juju/cmd"
	"github.com/juju/gnuflag"
	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/params"
)

type setPasswordCommand struct {
	cmd.CommandBase

	username string
	password string
	file     cmd.FileVar
}

const setPasswordDoc = `
Sets the password for a service account. If the service account does
not exist, setting its password will create it.

    contract set-password admin password123
    contract set-password -f password.txt admin
`

func (c *setPasswordCommand) Info() *cmd.Info {
	return &cmd.Info{
		Name:           "set-password",
		Args:           "<username> <password>",
		Purpose:        "Set a service-account password.",
		Doc:            setPasswordDoc[1:],
		ShowSuperFlags: commonFlags,
	}
}

func (c *setPasswordCommand) SetFlags(f *gnuflag.FlagSet) {
	c.file.StdinMarkers = stdinMarkers
	f.Var(&c.file, "f", "File to read password from.")
}

func (c *setPasswordCommand) Init(args []string) error {
	if len(args) < 1 {
		return errgo.New("username not specified")
	}
	c.username = args[0]
	if len(args) < 2 {
		if c.file.Path == "" {
			return errgo.New("password not specified")
		}
		return nil
	}
	c.password = args[1]
	if len(args) > 2 {
		return errgo.New("too many args")
	}
	return nil
}

func (c *setPasswordCommand) Run(ctxt *cmd.Context) error {
	if c.password == "" {
		buf, err := c.file.Read(ctxt)
		if err != nil {
			return errgo.Notef(err, "cannot read password")
		}
		c.password = string(bytes.TrimSuffix(buf, []byte{'\n'}))
	}
	err := client().SetServiceAccountPassword(context.Background(), &params.SetServiceAccountPasswordRequest{
		Account: c.username,
		Body: params.SetServiceAccountPasswordBody{
			Password: c.password,
		},
	})
	switch errorCode(err) {
	case params.ErrForbidden:
		return errgo.New("permission denied")
	}
	return errgo.Mask(err)
}
