// Copyright 2019 Canonical Ltd.

package main

import (
	"context"
	"io"
	"net/http"
	"strings"

	aclstore "github.com/juju/aclstore/v2"
	"github.com/juju/aclstore/v2/aclclient"
	"github.com/juju/cmd"
	"github.com/juju/gnuflag"
	errgo "gopkg.in/errgo.v1"
	httprequest "gopkg.in/httprequest.v1"

	"github.com/CanonicalLtd/ua-contracts/contractclient"
)

// showACLCommand implements the show-acl command.
type showACLCommand struct {
	cmd.CommandBase

	name   string
	output cmd.Output
}

const showACLDoc = `
Show the current value of an ACL.

    contract show-acl service-account
`

// Info implements cmd.Command.Info.
func (c *showACLCommand) Info() *cmd.Info {
	return &cmd.Info{
		Name:           "show-acl",
		Args:           "<acl>",
		Purpose:        "Show the current value of an ACL.",
		Doc:            showACLDoc[1:],
		ShowSuperFlags: commonFlags,
	}
}

// SetFlags implements cmd.Command.SetFlags.
func (c *showACLCommand) SetFlags(f *gnuflag.FlagSet) {
	formatters := map[string]cmd.Formatter{
		"json":    cmd.FormatJson,
		"yaml":    formatYAML,
		"tabular": formatTabular(stringsTable),
	}
	c.output.AddFlags(f, "tabular", formatters)
}

// Init implements cmd.Command.Init.
func (c *showACLCommand) Init(args []string) error {
	if len(args) < 1 {
		return errgo.New("ACL not specified")
	}
	c.name = args[0]
	if len(args) > 1 {
		return errgo.Newf("too many args")
	}
	return nil
}

// Run implements cmd.Command.Run.
func (c *showACLCommand) Run(ctxt *cmd.Context) error {
	ctx := context.Background()

	acl, err := aclClient().Get(ctx, c.name)
	if err != nil {
		return aclError(err, c.name)
	}
	return errgo.Mask(c.output.Write(ctxt, acl))
}

func stringsTable(w io.Writer, v interface{}) error {
	_, err := w.Write([]byte(strings.Join(v.([]string), "\n")))
	return errgo.Mask(err)
}

// aclClient returns a client for doing ACL requests with the contracts
// service.
func aclClient() *aclclient.Client {
	return aclclient.New(aclclient.NewParams{
		BaseURL: contractsURL() + "/acl",
		Doer:    authorizingDoer{authorizer()},
	})
}

type authorizingDoer struct {
	authorizeRequest contractclient.Authorizer
}

func (d authorizingDoer) Do(req *http.Request) (*http.Response, error) {
	d.authorizeRequest(req)
	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return resp, errgo.Mask(err)
	}
	return resp, nil
}

func aclError(err error, aclName string) error {
	if err == nil {
		return nil
	}
	if rerr, ok := errgo.Cause(err).(*httprequest.RemoteError); ok {
		switch rerr.Code {
		case aclstore.CodeACLNotFound:
			err = errgo.Newf("ACL %q not found", aclName)
		case httprequest.CodeForbidden:
			err = errgo.New("permission denied")
		default:
			err = errgo.Mask(err)
		}
	} else {
		err = errgo.Mask(err)
	}
	err.(*errgo.Err).SetLocation(1)
	return err
}
