// Copyright 2019 Canonical Ltd.

package main

import (
	"context"
	"fmt"
	"io"
	"strings"
	"text/tabwriter"
	"time"

	"github.com/juju/cmd"
	"github.com/juju/gnuflag"
	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/params"
)

type statusCommand struct {
	cmd.CommandBase

	id     string
	output cmd.Output
}

const statusDoc = `
Show a status summary for a user, account, contract, or machine.

    contract status joe.bloggs@example.com
    contract status aAE3B0rhpBSbT5EqNsf4r_OHO32yk_HCVPZ6srq4tcUA
    contract status cAGF2FXdGx--kDLprt9M1npJ8TDFfIgHU9ywgVF15iW4
    contract status b90683a4fa754887b36bb72852149aa9
`

// Info implements cmd.Command.Info.
func (c *statusCommand) Info() *cmd.Info {
	return &cmd.Info{
		Name:           "status",
		Purpose:        "Show status summary.",
		Args:           "<email>|<account-id>|<contract-id>|<machine-id>",
		Doc:            showContractDoc[1:],
		ShowSuperFlags: commonFlags,
	}
}

// SetFlags implements cmd.Command.SetFlags.
func (c *statusCommand) SetFlags(f *gnuflag.FlagSet) {
	formatters := map[string]cmd.Formatter{
		"json":    cmd.FormatJson,
		"tabular": formatStatus,
		"yaml":    formatYAML,
	}
	c.output.AddFlags(f, "tabular", formatters)
}

// Init implements cmd.Command.Init.
func (c *statusCommand) Init(args []string) error {
	if len(args) < 1 {
		return errgo.New("ID not specified")
	}
	c.id = args[0]
	if len(args) > 1 {
		return errgo.Newf("too many args")
	}
	return nil
}

// Run implements cmd.Command.Run.
func (c *statusCommand) Run(ctxt *cmd.Context) error {
	ctx := context.Background()

	var status interface{}
	var err error
	switch {
	case strings.Contains(c.id, "@"):
		status, err = c.userStatus(ctx, c.id)
	default:
		return errgo.Newf("unsupported ID %q", c.id)
	}
	if err != nil {
		switch errorCode(errgo.Cause(err)) {
		case params.ErrNotFound:
			return errgo.New("not found")
		case params.ErrForbidden:
			return errgo.New("permission denied")
		}
		return errgo.Mask(err)
	}
	return errgo.Mask(c.output.Write(ctxt, status))
}

func (c *statusCommand) userStatus(ctx context.Context, email string) (userStatus, error) {
	v := userStatus{
		Email: email,
		// TODO(mhilton) get the user name.
	}

	accs, err := client().GetAccounts(ctx, &params.GetAccountsRequest{
		Email: email,
	})
	if err != nil {
		return userStatus{}, errgo.Mask(err, errgo.Any)
	}
	for _, a := range accs.Accounts {
		acc, err := c.userAccount(ctx, a.Id)
		acc.Name = a.Name
		if err != nil {
			return userStatus{}, errgo.Mask(err, errgo.Any)
		}
		v.Accounts = append(v.Accounts, acc)
	}
	return v, nil
}

func (c *statusCommand) userAccount(ctx context.Context, accountID string) (userAccount, error) {
	v := userAccount{
		ID: accountID,
	}
	contracts, err := client().GetAccountContracts(ctx, &params.GetAccountContractsRequest{
		Account: accountID,
	})
	if err != nil {
		return userAccount{}, errgo.Mask(err, errgo.Any)
	}

	if len(contracts.Contracts) > 0 {
		v.Name = contracts.Contracts[0].AccountInfo.Name
		// TODO(mhilton) get the origin.
	}

	for _, contract := range contracts.Contracts {
		v.Contracts = append(v.Contracts, accountContract{
			ID:     contract.ContractInfo.Id,
			Name:   contract.ContractInfo.Name,
			Origin: contract.ContractInfo.Origin,
		})
		// TODO(mhilton) get contract machines.
	}
	return v, nil
}

func formatStatus(w io.Writer, v interface{}) error {
	switch v1 := v.(type) {
	case userStatus:
		formatUserStatus(w, v1)
	default:
		return errgo.Newf("unsupported status type %T", v)
	}
	return nil
}

type userStatus struct {
	Email    string        `json:"email" yaml:"email"`
	Name     string        `json:"name" yaml:"name"`
	Accounts []userAccount `json:"accounts" yaml:"accounts"`
}

type userAccount struct {
	ID        string            `json:"id" yaml:"id"`
	Name      string            `json:"name" yaml:"name"`
	Origin    string            `json:"origin" yaml:"origin"`
	Contracts []accountContract `json:"contracts" yaml:"contracts"`
}

type accountContract struct {
	ID       string            `json:"id" yaml:"id"`
	Name     string            `json:"name" yaml:"name"`
	Origin   string            `json:"origin" yaml:"origin"`
	Machines []contractMachine `json:"machines" yaml:"machines"`
}

type contractMachine struct {
	ID     string         `json:"id" yaml:"id"`
	Errors []machineError `json:"errors" yaml:"errors"`
}

type machineError struct {
	Time     time.Time `json:"time" yaml:"time"`
	Resource string    `json:"resource" yaml:"resource"`
	Error    string    `json:"error" yaml:"error"`
}

func formatUserStatus(w io.Writer, v userStatus) error {
	tw := tabwriter.NewWriter(w, 0, 0, 2, ' ', 0)
	if _, err := fmt.Fprintf(tw, "Name\tEmail\t\n%s\t%s\t\n", v.Email, v.Name); err != nil {
		return errgo.Mask(err)
	}
	if err := tw.Flush(); err != nil {
		return errgo.Mask(err)
	}
	if _, err := w.Write([]byte{'\n'}); err != nil {
		return errgo.Mask(err)
	}

	var contracts []accountContract
	tw.Init(w, 0, 0, 2, ' ', 0)
	if _, err := fmt.Fprintln(tw, "Account ID\tAccount Name\tOrigin\tContracts\t"); err != nil {
		return errgo.Mask(err)
	}
	for _, a := range v.Accounts {
		contracts = append(contracts, a.Contracts...)
		if _, err := fmt.Fprintf(tw, "%s\t%s\t%s\t%d\t\n", a.ID, a.Name, a.Origin, len(a.Contracts)); err != nil {
			return errgo.Mask(err)
		}
	}
	if err := tw.Flush(); err != nil {
		return errgo.Mask(err)
	}
	if _, err := w.Write([]byte{'\n'}); err != nil {
		return errgo.Mask(err)
	}

	if err := formatContracts(w, contracts); err != nil {
		return errgo.Mask(err)
	}
	if _, err := w.Write([]byte{'\n'}); err != nil {
		return errgo.Mask(err)
	}

	// TODO(mhilton) work out what errors to display, once we've
	// worked out how to fetch them.
	if err := formatErrors(w, []machineError{}); err != nil {
		return errgo.Mask(err)
	}
	return nil
}

func formatContracts(w io.Writer, contracts []accountContract) error {
	tw := tabwriter.NewWriter(w, 0, 0, 2, ' ', 0)
	if _, err := fmt.Fprintln(tw, "Contract ID\tContract Name\tOrigin\tMachines\t"); err != nil {
		return errgo.Mask(err)
	}
	for _, c := range contracts {
		if _, err := fmt.Fprintf(tw, "%s\t%s\t%s\t%d\t\n", c.ID, c.Name, c.Origin, len(c.Machines)); err != nil {
			return errgo.Mask(err)
		}
	}
	if err := tw.Flush(); err != nil {
		return errgo.Mask(err)
	}
	return nil
}

func formatErrors(w io.Writer, errors []machineError) error {
	tw := tabwriter.NewWriter(w, 0, 0, 2, ' ', 0)
	if _, err := fmt.Fprintln(tw, "Time\tResource\tError\t"); err != nil {
		return errgo.Mask(err)
	}
	for _, e := range errors {
		if _, err := fmt.Fprintf(tw, "%s\t%s\t%s\t\n", e.Time.Format("2006-01-02 15:04:05Z07:00"), e.Resource, e.Error); err != nil {
			return errgo.Mask(err)
		}
	}
	if err := tw.Flush(); err != nil {
		return errgo.Mask(err)
	}
	return nil
}
