// Copyright 2019 Canonical Ltd.

package main

import (
	"context"
	"strings"

	"github.com/juju/cmd"
	"github.com/juju/gnuflag"
	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/params"
)

// syncSalesforceCommand implements the sync-contract command.
type syncSalesforceCommand struct {
	cmd.CommandBase

	sfId   string
	mail   string
	output cmd.Output
}

const syncSalesforceDoc = `
Synchronize the contract with the specified email or salesforce id.

    contract sync <email or salesforceid>
`

// Info implements cmd.Command.Info.
func (c *syncSalesforceCommand) Info() *cmd.Info {
	return &cmd.Info{
		Name:           "sync-salesforce",
		Purpose:        "Synchronize a salesforce account.",
		Doc:            syncSalesforceDoc[1:],
		ShowSuperFlags: commonFlags,
	}
}

// SetFlags implements cmd.Command.SetFlags.
func (c *syncSalesforceCommand) SetFlags(f *gnuflag.FlagSet) {
	formatters := map[string]cmd.Formatter{
		"json": cmd.FormatJson,
		"yaml": formatYAML,
	}
	c.output.AddFlags(f, "yaml", formatters)
}

// Init implements cmd.Command.Init.
func (c *syncSalesforceCommand) Init(args []string) error {
	if len(args) != 1 {
		return errgo.New("specify either an email or a salesforce account id")
	}
	if strings.Contains(args[0], "@") {
		c.mail = args[0]
	} else {
		c.sfId = args[0]
	}
	return nil
}

// Run implements cmd.Command.Run.
func (c *syncSalesforceCommand) Run(ctxt *cmd.Context) error {
	ctx := context.Background()

	resp, err := client().UpdateAccountContracts(ctx, &params.UpdateAccountContractsRequest{
		Body: params.UpdateAccountContractsRequestBody{
			Sfid: c.sfId,
			Mail: c.mail,
		},
	})
	if err != nil {
		return errgo.Mask(err)
	}
	return errgo.Mask(c.output.Write(ctxt, resp))
}
