package main

import (
	"encoding/base64"
	"html/template"
	"io"
	"log"
	"net/http"
	"os"
	"runtime/debug"

	errgo "gopkg.in/errgo.v1"
	macaroon "gopkg.in/macaroon.v2"

	"github.com/CanonicalLtd/ua-contracts/contractclient"
	"github.com/CanonicalLtd/ua-contracts/params"
)

var (
	listenAddr  = os.Getenv("LISTEN_ADDR")
	contractURL = os.Getenv("CONTRACT_URL")
	serviceURL  = os.Getenv("SERVICE_URL")
)

func main() {
	log.Println("starting the server")
	log.Printf("contracts service: %s\n", contractURL)
	http.Handle("/login", http.HandlerFunc(recoverPanic(login)))
	http.Handle("/get-resource-token", http.HandlerFunc(recoverPanic(serveResourceToken)))
	http.Handle("/", http.HandlerFunc(recoverPanic(root)))
	err := http.ListenAndServe(listenAddr, nil)
	log.Printf("exiting server: %v\n", err)
}

// recoverPanic decorates a http.HandlerFunc such that if the handler
// panics an appropriate response is returned.
func recoverPanic(f func(w http.ResponseWriter, req *http.Request)) func(w http.ResponseWriter, req *http.Request) {
	return func(w http.ResponseWriter, req *http.Request) {
		defer func() {
			if err := recover(); err != nil {
				log.Println("Internal Server error", err)
				debug.PrintStack()
				w.WriteHeader(http.StatusInternalServerError)
				mustRender(w, internalServerErrorTemplate, err)
			}
		}()
		f(w, req)
	}
}

var internalServerErrorTemplate = template.Must(template.New("internal-server-error").Parse(`
<!doctype html>
<html>
<head><title>Contracts</title></head>
<body><em>Error: {{.}}</em></body>
</html>
`[1:]))

// root handles the default endpoint of the service. If the browser is
// not logged in it displays a link to start the login process, otherwise
// it attempts to get a contract token for the user and display that.
func root(w http.ResponseWriter, req *http.Request) {
	ctx := req.Context()

	ms, err := readMacaroon(req, "contracts-auth")
	if err != nil {
		w.Header().Set("Content-Type", "text/html")
		mustRender(w, unauthorizedTemplate, nil)
		return
	}
	auth, err := contractclient.MacaroonAuthorizer(ms)
	if err != nil {
		panic(err)
	}
	client := contractclient.New(contractclient.NewParams{
		BaseURL:    contractURL,
		Authorizer: auth,
	})
	resp, err := client.GetAccounts(ctx, &params.GetAccountsRequest{})
	if err != nil {
		panic(err)
	}
	resp2, err := client.GetAccountContracts(ctx, &params.GetAccountContractsRequest{
		Account: resp.Accounts[0].Id,
	})
	if err != nil {
		panic(err)
	}
	resp3, err := client.GetContractToken(ctx, &params.GetContractTokenRequest{
		Contract: resp2.Contracts[0].ContractInfo.Id,
	})
	if err != nil {
		panic(err)
	}
	w.Header().Set("Content-Type", "text/html")
	mustRender(w, authorizedTemplate, struct {
		ContractName    string
		ContractToken   string
		ResourceChoices []string
	}{
		ContractName:    resp3.ContractTokenInfo.ContractInfo.Name,
		ContractToken:   resp3.ContractToken,
		ResourceChoices: resourceChoices,
	})
}

var resourceChoices = []string{"esm", "fips", "cc-eal", "livepatch"}

var unauthorizedTemplate = template.Must(template.New("unauthorized").Parse(`
<!doctype html>
<html>
<head><title>Contracts</title></head>
<body><p><a href="/login?next=/">Login...</a></p></body>
</html>
`[1:]))

var authorizedTemplate = template.Must(template.New("authorized").Parse(`
<!doctype html>
<html>
<head><title>Contracts</title></head>
<body>
<table>
  <tr><th>Contract</th><th>Token</th></tr>
  <tr><td>{{.ContractName}}<td>{{.ContractToken}}</td></tr>
</table>

{{range $resource := .ResourceChoices}}
<form method="get" action="/get-resource-token">
<input type="hidden" name="token" value="{{$.ContractToken}}"/>
<input type="hidden" name="resource" value="{{$resource}}"/>
<table>
  <tr><th colspan="2">{{$resource}}</th></tr>
	<tr>
		<td><input type="text" name="machine-id" value="" placeholder="machine id"/></td>
		<td><input type="submit" value="Get Token"/></td>
  </tr>
</table>
</form>
{{end}}

</body>
</html>
`[1:]))

func readMacaroon(req *http.Request, name string) (macaroon.Slice, error) {
	c, err := req.Cookie(name)
	if err != nil {
		return nil, errgo.Mask(err, errgo.Is(http.ErrNoCookie))
	}
	buf, err := base64.RawURLEncoding.DecodeString(c.Value)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	var ms macaroon.Slice
	if err := ms.UnmarshalBinary(buf); err != nil {
		return nil, errgo.Mask(err)
	}
	return ms, nil
}

func mustRender(w io.Writer, t *template.Template, data interface{}) {
	if err := t.Execute(w, data); err != nil {
		panic(err)
	}
}
