// Copyright 2019 Canonical Ltd.

package config_test

import (
	"encoding/base64"
	"os"
	"testing"

	qt "github.com/frankban/quicktest"
	"go.uber.org/zap/zapcore"

	"github.com/CanonicalLtd/ua-contracts/config"
)

func TestListenAddr(t *testing.T) {
	c := qt.New(t)

	os.Setenv("PORT", "1234")
	c.Assert(config.ListenAddr(), qt.Equals, ":1234")
}

func TestLogLevel(t *testing.T) {
	c := qt.New(t)

	os.Setenv("LOGLEVEL", "DEBUG")
	c.Assert(config.LogLevel(), qt.Equals, zapcore.DebugLevel)

	os.Setenv("LOGLEVEL", "ERROR")
	c.Assert(config.LogLevel(), qt.Equals, zapcore.ErrorLevel)

	os.Unsetenv("LOGLEVEL")
	c.Assert(config.LogLevel(), qt.Equals, zapcore.InfoLevel)
}

func TestSSOLocation(t *testing.T) {
	c := qt.New(t)

	c.Assert(config.SSOLocation(), qt.Equals, "")
	os.Setenv("SSO_LOCATION", "login.example.com")
	c.Assert(config.SSOLocation(), qt.Equals, "login.example.com")
}

func TestTokenSecret(t *testing.T) {
	c := qt.New(t)

	token := []byte("7BR#%fjkHt-S8TCK")
	b64 := base64.StdEncoding.EncodeToString(token)
	c.Setenv("TOKEN_SECRET", b64)
	res, err := config.TokenSecret()
	c.Assert(err, qt.Equals, nil)
	c.Assert(res, qt.DeepEquals, token)

	c.Setenv("TOKEN_SECRET", "notb64!!")
	res, err = config.TokenSecret()
	c.Assert(err, qt.ErrorMatches, "illegal base64 data at input byte 6")

	c.Setenv("TOKEN_SECRET", "")
	res, err = config.TokenSecret()
	c.Assert(err, qt.Equals, nil)
	c.Assert(res, qt.IsNil)
}

var ssoPublicKeyTests = []struct {
	name        string
	env         string
	expectError string
}{{
	name: "default",
}, {
	name: "invalid-block",
	env: `MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAtO9bUDClVY5VKD8cLRbo
bw/ruGTTepdvhDc3oewHrnog/i7ymQpB7fCbn3iryNs704p9Yg1UaQM/iFJ75ojB
4gsGHF0yZWj4vf0Wt8vjpY7wdy/7MHerizED132ncoSuc676ulydVzyOPUGKQbPe
/EkYJhfoxyJHrBPvIHj445EK//uScrrKmwXteu0sOytDSLU+EPJgTQQ38sKak9FM
+MAW4R4wPvGZibUsm4JWfGGwf68FPXwZ0a/BIJXL8hBvuGaQlFy9jrafXlJKqY+I
XAQrHqnBTYxhyhhXCQrwBJOxcVTZe/zowGC3W/YO2yIWpqgkgkg4KkmyzayvhpQ2
4wIDAQAB
-----END PUBLIC KEY-----
`,
	expectError: `failed to read PEM block from SSO_PUBLICKEY`,
}, {
	name: "invalid-key",
	env: `-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAtO9bUDClVY5VKD8cLRbo
bw/ruGTTepdvhDc3oewHrnog/i7ymQpB7fCbn3iryNs704p9Yg1UaQM/iFJ75ojB
4gsGHF0yZWj4vf0Wt8vjpY7wdy/7MHerizED132ncoSuc676ulydVzyOPUGKQbPe
/EkYJhfoxyJHrBPvIHj445EK//uScrrKmwXteu0sOytDSLU+EPJgTQQ38sKak9FM
+MAW4R4wPvGZibUsm4JWfGGwf68FPXwZ0a/BIJXL8hBvuGaQlFy9jrafXlJKqY+I
XAQrHqnBTYxhyhhXCQrwBJOxcVTZe/zowGC3W/YO2yIWpqgkgkg4KkmyzayvhpQ2
4wIDAQABAA==
-----END PUBLIC KEY-----
`,
	expectError: `x509: trailing data after ASN.1 of public-key`,
}, {
	name: "invalid-key-type",
	env: `-----BEGIN PUBLIC KEY-----
MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAEpS0TN4JLay9g9sjSqIdf4AhlIpsi
nMfnogcF/val8eELGiavmOj9lfvoUHS76vlfAmXniWg5/nslChsq6hYlVQ==
-----END PUBLIC KEY-----
`,
	expectError: `unsupported key type \*ecdsa.PublicKey`,
}}

func TestSSOPublicKey(t *testing.T) {
	c := qt.New(t)

	for _, test := range ssoPublicKeyTests {
		c.Run(test.name, func(c *qt.C) {
			os.Setenv("SSO_PUBLICKEY", test.env)
			pkey, err := config.SSOPublicKey()
			if test.expectError != "" {
				c.Assert(err, qt.ErrorMatches, test.expectError)
				return
			}
			c.Assert(err, qt.Equals, nil)
			c.Assert(pkey, qt.IsNil)
		})
	}
}

func TestUseMemstore(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	c.Setenv("USE_MEMSTORE", "")
	c.Assert(config.UseMemStore(), qt.Equals, false)
	c.Setenv("USE_MEMSTORE", "x")
	c.Assert(config.UseMemStore(), qt.Equals, true)
}

func TestKafkaBrokerURLs(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	os.Unsetenv("KAFKA_BROKERS")
	c.Assert(config.KafkaBrokerURLs(), qt.IsNil)
	os.Setenv("KAFKA_BROKERS", "localhost:1234")
	c.Assert(config.KafkaBrokerURLs(), qt.DeepEquals, []string{"localhost:1234"})
}

func TestKafkaTLS(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	cfg, err := config.KafkaTLS()
	c.Assert(err, qt.IsNil)
	c.Assert(cfg, qt.IsNil)

	certPem := []byte(`-----BEGIN CERTIFICATE-----
MIIBhTCCASugAwIBAgIQIRi6zePL6mKjOipn+dNuaTAKBggqhkjOPQQDAjASMRAw
DgYDVQQKEwdBY21lIENvMB4XDTE3MTAyMDE5NDMwNloXDTE4MTAyMDE5NDMwNlow
EjEQMA4GA1UEChMHQWNtZSBDbzBZMBMGByqGSM49AgEGCCqGSM49AwEHA0IABD0d
7VNhbWvZLWPuj/RtHFjvtJBEwOkhbN/BnnE8rnZR8+sbwnc/KhCk3FhnpHZnQz7B
5aETbbIgmuvewdjvSBSjYzBhMA4GA1UdDwEB/wQEAwICpDATBgNVHSUEDDAKBggr
BgEFBQcDATAPBgNVHRMBAf8EBTADAQH/MCkGA1UdEQQiMCCCDmxvY2FsaG9zdDo1
NDUzgg4xMjcuMC4wLjE6NTQ1MzAKBggqhkjOPQQDAgNIADBFAiEA2zpJEPQyz6/l
Wf86aX6PepsntZv2GYlA5UpabfT2EZICICpJ5h/iI+i341gBmLiAFQOyTDT+/wQc
6MF9+Yw1Yy0t
-----END CERTIFICATE-----`)
	keyPem := []byte(`-----BEGIN EC PRIVATE KEY-----
MHcCAQEEIIrYSSNQFaA2Hwf1duRSxKtLYX5CB04fSeQ6tF1aY/PuoAoGCCqGSM49
AwEHoUQDQgAEPR3tU2Fta9ktY+6P9G0cWO+0kETA6SFs38GecTyudlHz6xvCdz8q
EKTcWGekdmdDPsHloRNtsiCa697B2O9IFA==
-----END EC PRIVATE KEY-----`)
	caCertPem := []byte(`-----BEGIN CERTIFICATE-----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-----END CERTIFICATE-----`)

	os.Setenv("KAFKA_CLIENT_CERT", string(certPem))
	os.Setenv("KAFKA_CLIENT_KEY", string(keyPem))
	os.Setenv("KAFKA_CA_CERT", string(caCertPem))
	cfg, err = config.KafkaTLS()
	c.Assert(err, qt.IsNil)
	c.Assert(cfg.Certificate, qt.Not(qt.IsNil))
	c.Assert(cfg.CACertificate, qt.Not(qt.IsNil))
}

func TestBootstrapAdminPassword(t *testing.T) {
	c := qt.New(t)
	defer c.Done()

	c.Setenv("BOOTSTRAP_ADMIN_PASSWORD", "")
	c.Assert(config.BootstrapAdminPassword(), qt.Equals, "")
	c.Setenv("BOOTSTRAP_ADMIN_PASSWORD", "x")
	c.Assert(config.BootstrapAdminPassword(), qt.Equals, "x")
}

var awsCertificatesTests = []struct {
	name           string
	envvar         string
	expectSubjects []string
}{{
	name: "empty",
}, {
	name: "one-cert",
	envvar: `
-----BEGIN CERTIFICATE-----
MIIC7TCCAq0CCQCWukjZ5V4aZzAJBgcqhkjOOAQDMFwxCzAJBgNVBAYTAlVTMRkw
FwYDVQQIExBXYXNoaW5ndG9uIFN0YXRlMRAwDgYDVQQHEwdTZWF0dGxlMSAwHgYD
VQQKExdBbWF6b24gV2ViIFNlcnZpY2VzIExMQzAeFw0xMjAxMDUxMjU2MTJaFw0z
ODAxMDUxMjU2MTJaMFwxCzAJBgNVBAYTAlVTMRkwFwYDVQQIExBXYXNoaW5ndG9u
IFN0YXRlMRAwDgYDVQQHEwdTZWF0dGxlMSAwHgYDVQQKExdBbWF6b24gV2ViIFNl
cnZpY2VzIExMQzCCAbcwggEsBgcqhkjOOAQBMIIBHwKBgQCjkvcS2bb1VQ4yt/5e
ih5OO6kK/n1Lzllr7D8ZwtQP8fOEpp5E2ng+D6Ud1Z1gYipr58Kj3nssSNpI6bX3
VyIQzK7wLclnd/YozqNNmgIyZecN7EglK9ITHJLP+x8FtUpt3QbyYXJdmVMegN6P
hviYt5JH/nYl4hh3Pa1HJdskgQIVALVJ3ER11+Ko4tP6nwvHwh6+ERYRAoGBAI1j
k+tkqMVHuAFcvAGKocTgsjJem6/5qomzJuKDmbJNu9Qxw3rAotXau8Qe+MBcJl/U
hhy1KHVpCGl9fueQ2s6IL0CaO/buycU1CiYQk40KNHCcHfNiZbdlx1E9rpUp7bnF
lRa2v1ntMX3caRVDdbtPEWmdxSCYsYFDk4mZrOLBA4GEAAKBgEbmeve5f8LIE/Gf
MNmP9CM5eovQOGx5ho8WqD+aTebs+k2tn92BBPqeZqpWRa5P/+jrdKml1qx4llHW
MXrs3IgIb6+hUIB+S8dz8/mmO0bpr76RoZVCXYab2CZedFut7qc3WUH9+EUAH5mw
vSeDCOUMYQR7R9LINYwouHIziqQYMAkGByqGSM44BAMDLwAwLAIUWXBlk40xTwSw
7HX32MxXYruse9ACFBNGmdX2ZBrVNGrN9N2f6ROk0k9K
-----END CERTIFICATE-----
`,
	expectSubjects: []string{"O=Amazon Web Services LLC,L=Seattle,ST=Washington State,C=US"},
}, {
	name: "many-certs",
	envvar: `
-----BEGIN CERTIFICATE-----
MIIC7TCCAq0CCQCWukjZ5V4aZzAJBgcqhkjOOAQDMFwxCzAJBgNVBAYTAlVTMRkw
FwYDVQQIExBXYXNoaW5ndG9uIFN0YXRlMRAwDgYDVQQHEwdTZWF0dGxlMSAwHgYD
VQQKExdBbWF6b24gV2ViIFNlcnZpY2VzIExMQzAeFw0xMjAxMDUxMjU2MTJaFw0z
ODAxMDUxMjU2MTJaMFwxCzAJBgNVBAYTAlVTMRkwFwYDVQQIExBXYXNoaW5ndG9u
IFN0YXRlMRAwDgYDVQQHEwdTZWF0dGxlMSAwHgYDVQQKExdBbWF6b24gV2ViIFNl
cnZpY2VzIExMQzCCAbcwggEsBgcqhkjOOAQBMIIBHwKBgQCjkvcS2bb1VQ4yt/5e
ih5OO6kK/n1Lzllr7D8ZwtQP8fOEpp5E2ng+D6Ud1Z1gYipr58Kj3nssSNpI6bX3
VyIQzK7wLclnd/YozqNNmgIyZecN7EglK9ITHJLP+x8FtUpt3QbyYXJdmVMegN6P
hviYt5JH/nYl4hh3Pa1HJdskgQIVALVJ3ER11+Ko4tP6nwvHwh6+ERYRAoGBAI1j
k+tkqMVHuAFcvAGKocTgsjJem6/5qomzJuKDmbJNu9Qxw3rAotXau8Qe+MBcJl/U
hhy1KHVpCGl9fueQ2s6IL0CaO/buycU1CiYQk40KNHCcHfNiZbdlx1E9rpUp7bnF
lRa2v1ntMX3caRVDdbtPEWmdxSCYsYFDk4mZrOLBA4GEAAKBgEbmeve5f8LIE/Gf
MNmP9CM5eovQOGx5ho8WqD+aTebs+k2tn92BBPqeZqpWRa5P/+jrdKml1qx4llHW
MXrs3IgIb6+hUIB+S8dz8/mmO0bpr76RoZVCXYab2CZedFut7qc3WUH9+EUAH5mw
vSeDCOUMYQR7R9LINYwouHIziqQYMAkGByqGSM44BAMDLwAwLAIUWXBlk40xTwSw
7HX32MxXYruse9ACFBNGmdX2ZBrVNGrN9N2f6ROk0k9K
-----END CERTIFICATE-----
-----BEGIN CERTIFICATE-----
MIICuzCCAiQCCQDrSGnlRgvSazANBgkqhkiG9w0BAQUFADCBoTELMAkGA1UEBhMC
VVMxCzAJBgNVBAgTAldBMRAwDgYDVQQHEwdTZWF0dGxlMRMwEQYDVQQKEwpBbWF6
b24uY29tMRYwFAYDVQQLEw1FQzIgQXV0aG9yaXR5MRowGAYDVQQDExFFQzIgQU1J
IEF1dGhvcml0eTEqMCgGCSqGSIb3DQEJARYbZWMyLWluc3RhbmNlLWlpZEBhbWF6
b24uY29tMB4XDTExMDgxMjE3MTgwNVoXDTIxMDgwOTE3MTgwNVowgaExCzAJBgNV
BAYTAlVTMQswCQYDVQQIEwJXQTEQMA4GA1UEBxMHU2VhdHRsZTETMBEGA1UEChMK
QW1hem9uLmNvbTEWMBQGA1UECxMNRUMyIEF1dGhvcml0eTEaMBgGA1UEAxMRRUMy
IEFNSSBBdXRob3JpdHkxKjAoBgkqhkiG9w0BCQEWG2VjMi1pbnN0YW5jZS1paWRA
YW1hem9uLmNvbTCBnzANBgkqhkiG9w0BAQEFAAOBjQAwgYkCgYEAqaIcGFFTx/SO
1W5G91jHvyQdGP25n1Y91aXCuOOWAUTvSvNGpXrI4AXNrQF+CmIOC4beBASnHCx0
82jYudWBBl9Wiza0psYc9flrczSzVLMmN8w/c78F/95NfiQdnUQPpvgqcMeJo82c
gHkLR7XoFWgMrZJqrcUK0gnsQcb6kakCAwEAATANBgkqhkiG9w0BAQUFAAOBgQDF
VH0+UGZr1LCQ78PbBH0GreiDqMFfa+W8xASDYUZrMvY3kcIelkoIazvi4VtPO7Qc
yAiLr6nkk69Tr/MITnmmsZJZPetshqBndRyL+DaTRnF0/xvBQXj5tEh+AmRjvGtp
6iS1rQoNanN8oEcT2j4b48rmCmnDhRoBcFHwCYs/3w==
-----END CERTIFICATE-----
`,
	expectSubjects: []string{
		"O=Amazon Web Services LLC,L=Seattle,ST=Washington State,C=US",
		"CN=EC2 AMI Authority,OU=EC2 Authority,O=Amazon.com,L=Seattle,ST=WA,C=US",
	},
}, {
	name: "wrong-block",
	envvar: `
-----BEGIN PUBLIC KEY-----
MIIC
-----END PUBLIC KEY-----
-----BEGIN CERTIFICATE-----
MIIC7TCCAq0CCQCWukjZ5V4aZzAJBgcqhkjOOAQDMFwxCzAJBgNVBAYTAlVTMRkw
FwYDVQQIExBXYXNoaW5ndG9uIFN0YXRlMRAwDgYDVQQHEwdTZWF0dGxlMSAwHgYD
VQQKExdBbWF6b24gV2ViIFNlcnZpY2VzIExMQzAeFw0xMjAxMDUxMjU2MTJaFw0z
ODAxMDUxMjU2MTJaMFwxCzAJBgNVBAYTAlVTMRkwFwYDVQQIExBXYXNoaW5ndG9u
IFN0YXRlMRAwDgYDVQQHEwdTZWF0dGxlMSAwHgYDVQQKExdBbWF6b24gV2ViIFNl
cnZpY2VzIExMQzCCAbcwggEsBgcqhkjOOAQBMIIBHwKBgQCjkvcS2bb1VQ4yt/5e
ih5OO6kK/n1Lzllr7D8ZwtQP8fOEpp5E2ng+D6Ud1Z1gYipr58Kj3nssSNpI6bX3
VyIQzK7wLclnd/YozqNNmgIyZecN7EglK9ITHJLP+x8FtUpt3QbyYXJdmVMegN6P
hviYt5JH/nYl4hh3Pa1HJdskgQIVALVJ3ER11+Ko4tP6nwvHwh6+ERYRAoGBAI1j
k+tkqMVHuAFcvAGKocTgsjJem6/5qomzJuKDmbJNu9Qxw3rAotXau8Qe+MBcJl/U
hhy1KHVpCGl9fueQ2s6IL0CaO/buycU1CiYQk40KNHCcHfNiZbdlx1E9rpUp7bnF
lRa2v1ntMX3caRVDdbtPEWmdxSCYsYFDk4mZrOLBA4GEAAKBgEbmeve5f8LIE/Gf
MNmP9CM5eovQOGx5ho8WqD+aTebs+k2tn92BBPqeZqpWRa5P/+jrdKml1qx4llHW
MXrs3IgIb6+hUIB+S8dz8/mmO0bpr76RoZVCXYab2CZedFut7qc3WUH9+EUAH5mw
vSeDCOUMYQR7R9LINYwouHIziqQYMAkGByqGSM44BAMDLwAwLAIUWXBlk40xTwSw
7HX32MxXYruse9ACFBNGmdX2ZBrVNGrN9N2f6ROk0k9K
-----END CERTIFICATE-----
`,
	expectSubjects: []string{"O=Amazon Web Services LLC,L=Seattle,ST=Washington State,C=US"},
}, {
	name: "invalid-cert",
	envvar: `
-----BEGIN CERTIFICATE-----
MIIC7TCCAq0CCQCWukjZ5V4aZzAJBgcqhkjOOAQDMFwxCzAJBgNVBAYTAlVTMRkw
FwYDVQQIExBXYXNoaW5ndG9uIFN0YXRlMRAwDgYDVQQHEwdTZWF0dGxlMSAwHgYD
VQQKExdBbWF6b24gV2ViIFNlcnZpY2VzIExMQzAeFw0xMjAxMDUxMjU2MTJaFw0z
ODAxMDUxMjU2MTJaMFwxCzAJBgNVBAYTAlVTMRkwFwYDVQQIExBXYXNoaW5ndG9u
IFN0YXRlMRAwDgYDVQQHEwdTZWF0dGxlMSAwHgYDVQQKExdBbWF6b24gV2ViIFNl
cnZpY2VzIExMQzCCAbcwggEsBgcqhkjOOAQBMIIBHwKBgQCjkvcS2bb1VQ4yt/5e
ih5OO6kK/n1Lzllr7D8ZwtQP8fOEpp5E2ng+D6Ud1Z1gYipr58Kj3nssSNpI6bX3
VyIQzK7wLclnd/YozqNNmgIyZecN7EglK9ITHJLP+x8FtUpt3QbyYXJdmVMegN6P
hviYt5JH/nYl4hh3Pa1HJdskgQIVALVJ3ER11+Ko4tP6nwvHwh6+ERYRAoGBAI1j
k+tkqMVHuAFcvAGKocTgsjJem6/5qomzJuKDmbJNu9Qxw3rAotXau8Qe+MBcJl/U
hhy1KHVpCGl9fueQ2s6IL0CaO/buycU1CiYQk40KNHCcHfNiZbdlx1E9rpUp7bnF
lRa2v1ntMX3caRVDdbtPEWmdxSCYsYFDk4mZrOLBA4GEAAKBgEbmeve5f8LIE/Gf
MNmP9CM5eovQOGx5ho8WqD+aTebs+k2tn92BBPqeZqpWRa5P/+jrdKml1qx4llHW
MXrs3IgIb6+hUIB+S8dz8/mmO0bpr76RoZVCXYab2CZedFut7qc3WUH9+EUAH5mw
vSeDCOUMYQR7R9LINYwouHIziqQYMAkGByqGSM44BAMDLwAwLAIUWXBlk40xTwSw
7HX32MxXYruse9ACFBNGmdX2ZBrVNGrN9N2f6ROk
-----END CERTIFICATE-----
-----BEGIN CERTIFICATE-----
MIICuzCCAiQCCQDrSGnlRgvSazANBgkqhkiG9w0BAQUFADCBoTELMAkGA1UEBhMC
VVMxCzAJBgNVBAgTAldBMRAwDgYDVQQHEwdTZWF0dGxlMRMwEQYDVQQKEwpBbWF6
b24uY29tMRYwFAYDVQQLEw1FQzIgQXV0aG9yaXR5MRowGAYDVQQDExFFQzIgQU1J
IEF1dGhvcml0eTEqMCgGCSqGSIb3DQEJARYbZWMyLWluc3RhbmNlLWlpZEBhbWF6
b24uY29tMB4XDTExMDgxMjE3MTgwNVoXDTIxMDgwOTE3MTgwNVowgaExCzAJBgNV
BAYTAlVTMQswCQYDVQQIEwJXQTEQMA4GA1UEBxMHU2VhdHRsZTETMBEGA1UEChMK
QW1hem9uLmNvbTEWMBQGA1UECxMNRUMyIEF1dGhvcml0eTEaMBgGA1UEAxMRRUMy
IEFNSSBBdXRob3JpdHkxKjAoBgkqhkiG9w0BCQEWG2VjMi1pbnN0YW5jZS1paWRA
YW1hem9uLmNvbTCBnzANBgkqhkiG9w0BAQEFAAOBjQAwgYkCgYEAqaIcGFFTx/SO
1W5G91jHvyQdGP25n1Y91aXCuOOWAUTvSvNGpXrI4AXNrQF+CmIOC4beBASnHCx0
82jYudWBBl9Wiza0psYc9flrczSzVLMmN8w/c78F/95NfiQdnUQPpvgqcMeJo82c
gHkLR7XoFWgMrZJqrcUK0gnsQcb6kakCAwEAATANBgkqhkiG9w0BAQUFAAOBgQDF
VH0+UGZr1LCQ78PbBH0GreiDqMFfa+W8xASDYUZrMvY3kcIelkoIazvi4VtPO7Qc
yAiLr6nkk69Tr/MITnmmsZJZPetshqBndRyL+DaTRnF0/xvBQXj5tEh+AmRjvGtp
6iS1rQoNanN8oEcT2j4b48rmCmnDhRoBcFHwCYs/3w==
-----END CERTIFICATE-----
`,
	expectSubjects: []string{
		"CN=EC2 AMI Authority,OU=EC2 Authority,O=Amazon.com,L=Seattle,ST=WA,C=US",
	},
}}

func TestAWSCertificates(t *testing.T) {
	c := qt.New(t)

	for _, test := range awsCertificatesTests {
		c.Run(test.name, func(c *qt.C) {
			c.Setenv("AWS_CERTIFICATES", test.envvar)
			certs := config.AWSCertificates()
			c.Assert(len(certs), qt.Equals, len(test.expectSubjects))
			for i, cert := range certs {
				c.Assert(cert.Subject.String(), qt.Equals, test.expectSubjects[i])
			}
		})
	}
}
