// Copyright 2019 Canonical Ltd.

package main

import (
	"crypto/dsa"
	"crypto/x509"
	"encoding/base64"
	"encoding/pem"
	"flag"
	"fmt"
	"io/ioutil"
	"os"

	"github.com/CanonicalLtd/ua-contracts/internal/pkcs7"
)

var (
	cert = flag.String("cert", "cert.pem", "Certificate `file` used to sign data.")
	key  = flag.String("key", "key.pem", "Private key `file` used to sign data.")
)

func main() {
	flag.Parse()
	if err := run(); err != nil {
		fmt.Fprintf(os.Stderr, "sign: %s\n", err)
		os.Exit(1)
	}
}

func run() error {
	key, err := loadKey()
	if err != nil {
		return err
	}
	cert, err := loadCert()
	if err != nil {
		return err
	}
	data, err := ioutil.ReadAll(os.Stdin)
	if err != nil {
		return err
	}

	sd := pkcs7.NewSignedData(data)
	if err := sd.Sign(x509.DSAWithSHA1, cert, key); err != nil {
		return err
	}

	fmt.Println(base64.StdEncoding.EncodeToString(sd.Encode()))
	return nil
}

func loadKey() (*dsa.PrivateKey, error) {
	buf, err := ioutil.ReadFile(*key)
	if err != nil {
		return nil, err
	}
	block, _ := pem.Decode(buf)
	if block.Type != "DSA PRIVATE KEY" {
		return nil, fmt.Errorf("unsupported key type %q", block.Type)
	}
	return pkcs7.ParseDSAPrivateKey(block.Bytes)
}

func loadCert() (*x509.Certificate, error) {
	buf, err := ioutil.ReadFile(*cert)
	if err != nil {
		return nil, err
	}
	block, _ := pem.Decode(buf)
	if block.Type != "CERTIFICATE" {
		return nil, fmt.Errorf("unsupported certificate type %q", block.Type)
	}
	return x509.ParseCertificate(block.Bytes)
}
