// Copyright 2019 Canonical Ltd.

// Package azureauth provides authentication for Azure PKCS#7 signed
// documents.
package azureauth

import (
	"crypto/x509"
	"encoding/base64"
	"encoding/json"

	"github.com/fullsailor/pkcs7"
	errgo "gopkg.in/errgo.v1"
)

type Decoder struct {
	Intermediates *x509.CertPool
}

// Decode verifies and decodes a Azure Attested Instance Metadata
// document from a PKCS#7 signature, see
// https://docs.microsoft.com/en-us/azure/virtual-machines/windows/instance-metadata-service#attested-data.
func (d Decoder) Decode(msg string) (*AttestedData, error) {
	buf, err := base64.StdEncoding.DecodeString(msg)
	if err != nil {
		return nil, errgo.Notef(err, "cannot decode PKCS#7 document")
	}
	data, err := pkcs7.Parse(buf)
	if err != nil {
		return nil, errgo.Notef(err, "cannot parse PKCS#7 document")
	}
	if cert := data.GetOnlySigner(); cert != nil {
		// Verify signing certificate against intermediate and root CAs.
		_, err = cert.Verify(x509.VerifyOptions{
			DNSName:       "metadata.azure.com",
			Intermediates: d.Intermediates,
		})
		if err != nil {
			return nil, errgo.Notef(err, "cannot verify signing certificate")
		}
	} else {
		return nil, errgo.Newf("expected 1 certificate, found %d", len(data.Certificates))
	}
	if err := data.Verify(); err != nil {
		return nil, errgo.Notef(err, "cannot verify PKCS#7 document")
	}
	var doc AttestedData
	if err := json.Unmarshal(data.Content, &doc); err != nil {
		return nil, errgo.Notef(err, "invalid instance metadata document")
	}
	return &doc, nil
}

// AttestedData provides information about virtual machine instances in Azure.
type AttestedData struct {
	Nonce string `json:"nonce"` // "1234567890",
	Plan  struct {
		Name      string `json:"name"`      //  "",
		Product   string `json:"product"`   //  "",
		Publisher string `json:"publisher"` //  ""
	} `json:"plan"` //  {,
	Timestamp struct {
		CreatedOn string `json:"createdOn"` // "04/22/19 17:34:15 -0000",
		ExpiresOn string `json:"expiresOn"` // "04/22/19 23:34:15 -0000"
	} `json:"timeStamp"`
	SubscriptionID string `json:"subscriptionId"` //  "xxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxx",
	VMID           string `json:"vmId"`           //  "453945c8-3923-4366-b2d3-ea4c80e9b70e",
}
