// Copyright 2019 Canonical Ltd.

package kafkaauditor_test

import (
	"bytes"
	"context"
	"encoding/json"
	"sync"
	"testing"
	"time"

	"github.com/Shopify/sarama"
	qt "github.com/frankban/quicktest"
	"github.com/google/go-cmp/cmp/cmpopts"
	"github.com/juju/zaputil/zapctx"
	"go.uber.org/zap/zaptest"

	"github.com/CanonicalLtd/ua-contracts/internal/kafkaauditor"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
)

var (
	msgEqual = qt.CmpEquals(cmpopts.IgnoreUnexported(sarama.ProducerMessage{}))
)

var tests = []struct {
	expectedTopic string
	payload       kafkaauditor.AuditEntry
}{{
	expectedTopic: "ua-resource-token",
	payload: &kafkaauditor.AuditResourceToken{
		ContractID: token.MustNewContractID(),
		MachineID:  "test-machine",
		Resource:   "test-resource",
		Product:    "test-product",
	},
}, {
	expectedTopic: "ua-contract",
	payload: &kafkaauditor.AuditAddContract{
		ContractID: token.MustNewContractID(),
		Name:       "test-contract",
		AccountID:  token.MustNewAccountID(),
		ProductIDs: []string{"p1", "p2"},
		Created:    time.Now(),
		CreatedBy:  "test-user",
	},
}, {
	expectedTopic: "ua-contract-machine",
	payload: &kafkaauditor.AuditAddContractMachine{
		ContractID: token.MustNewContractID(),
		MachineID:  "test-machine",
	},
}, {
	expectedTopic: "ua-contract-token",
	payload: &kafkaauditor.AuditAddContractToken{
		ContractID: token.MustNewContractID(),
		Expires:    time.Now(),
	},
}, {
	expectedTopic: "ua-account",
	payload: &kafkaauditor.AuditAddAccount{
		AccountID:      token.MustNewAccountID(),
		AccountName:    "test-account",
		AccountCreated: time.Now(),
		UserID:         "test-user",
	},
}, {
	expectedTopic: "ua-machine",
	payload: &kafkaauditor.AuditEnsureMachine{
		MachineID: "test-id",
	},
}, {
	expectedTopic: "ua-user",
	payload: &kafkaauditor.AuditEnsureUser{
		ID: "test-id",
	},
}, {
	expectedTopic: "ua-contract-machine-remove",
	payload: &kafkaauditor.AuditRemoveContractMachine{
		ContractID: token.MustNewContractID(),
		MachineID:  "test-id",
	},
}, {
	expectedTopic: "ua-account-access",
	payload: &kafkaauditor.AuditSetAccountAccess{
		AccountID: token.MustNewAccountID(),
		UserID:    "test-id",
	},
}, {
	expectedTopic: "ua-product",
	payload: &kafkaauditor.AuditAddProduct{
		ProductID:   "123",
		ProductName: "test-product",
	},
}, {
	expectedTopic: "ua-contract-configure",
	payload: &kafkaauditor.AuditConfigureContract{
		ContractID: token.MustNewContractID(),
	},
}, {
	expectedTopic: "ua-product-update",
	payload: &kafkaauditor.AuditUpdateProduct{
		AuditAddProduct: kafkaauditor.AuditAddProduct{
			ProductID:   "123",
			ProductName: "test-product",
		},
	},
}, {
	expectedTopic: "ua-contract-products",
	payload: &kafkaauditor.AuditUpdateContractProducts{
		ContractID: token.MustNewContractID(),
		Products:   []string{"p1", "p2", "p3"},
	},
}, {
	expectedTopic: "ua-service-account",
	payload: &kafkaauditor.AuditEnsureServiceAccount{
		Name: "test-service-account",
	},
}, {
	expectedTopic: "ua-resource-token-info",
	payload: &kafkaauditor.AuditResourceTokenInfo{
		ContractID: token.MustNewContractID(),
		MachineID:  "test-machine-id",
		Resource:   "test-resource",
	},
}}

func TestClient(t *testing.T) {
	c := qt.New(t)

	testProducer := newTestAsyncProducer()
	client := kafkaauditor.NewClient(testProducer)
	defer client.Close()

	skipped := make(chan bool, 10)
	*kafkaauditor.MessageSkipped = func() {
		select {
		case skipped <- true:
		default:
		}
	}

	t0 := time.Now().UTC().Truncate(time.Second)

	for i, test := range tests {
		c.Logf("running test %d", i)

		c.Run(test.expectedTopic, func(c *qt.C) {
			ctx := zapctx.WithLogger(context.Background(), zaptest.NewLogger(c))

			client.Audit(ctx, t0, test.payload)

			if test.expectedTopic != "" {
				data, err := json.Marshal(test.payload)
				c.Assert(err, qt.Equals, nil)
				data = bytes.TrimPrefix(data, []byte("{"))
				payloadData := append(t0.AppendFormat([]byte(`{ "timestamp": "`), time.RFC3339Nano), []byte(`", `)...)
				payloadData = append(payloadData, data...)

				info := test.payload.KafkaInfo()

				select {
				case msg := <-testProducer.producerMessages:
					assertEqualMessage(c, msg, &sarama.ProducerMessage{
						Topic:     test.expectedTopic,
						Key:       sarama.ByteEncoder(info.Key),
						Value:     sarama.ByteEncoder(payloadData),
						Offset:    1,
						Partition: 1,
					})
					msgValue, err := msg.Value.Encode()
					c.Assert(err, qt.Equals, nil)
					assertIDFormats(c, msgValue)
				case <-time.After(time.Second):
					c.Fatalf("timed out")
				}
			} else {
				select {
				case <-skipped:
				case <-time.After(time.Second):
					c.Fatalf("timed out")
				}
			}
		})
	}
}

// assertIDFormats asserts that account and contract IDs in kafka messages are
// JSON-serialized according to their string representation, and not as an
// array of byte-sized integers.
func assertIDFormats(c *qt.C, data []byte) {
	m := struct {
		AccountID  string `json:"accountID"`
		ContractID string `json:"contractID"`
	}{}
	c.Log(string(data))
	err := json.Unmarshal(data, &m)
	c.Assert(err, qt.Equals, nil)
	if m.AccountID != "" {
		_, err = token.ParseAccountID(m.AccountID)
		c.Assert(err, qt.Equals, nil)
	}
	if m.ContractID != "" {
		_, err = token.ParseContractID(m.ContractID)
		c.Assert(err, qt.Equals, nil)
	}
}

func assertEqualMessage(c *qt.C, m1, m2 *sarama.ProducerMessage) {
	c.Assert(m1.Topic, qt.Equals, m2.Topic)
	c.Assert(m1.Key, qt.DeepEquals, m2.Key)
	c.Assert(m1.Value, qt.DeepEquals, m2.Value)
	c.Assert(m1.Partition, qt.DeepEquals, m2.Partition)
	c.Assert(m1.Offset, qt.DeepEquals, m2.Offset)
}

func newTestAsyncProducer() *testAsyncProducer {
	p := &testAsyncProducer{
		messages:         make(chan *sarama.ProducerMessage),
		producerMessages: make(chan *sarama.ProducerMessage, 10),
		successes:        make(chan *sarama.ProducerMessage),
		errors:           make(chan *sarama.ProducerError),
		close:            make(chan bool, 1),
	}
	p.w.Add(1)
	go p.loop()
	go p.closeChannels()

	return p
}

type testAsyncProducer struct {
	chanT     sync.Mutex
	messages  chan *sarama.ProducerMessage
	successes chan *sarama.ProducerMessage
	errors    chan *sarama.ProducerError

	w                sync.WaitGroup
	m                sync.Mutex
	producerErrors   []error
	producerMessages chan *sarama.ProducerMessage
	close            chan bool
}

func (a *testAsyncProducer) AsyncClose() {
	a.close <- true
}

func (a *testAsyncProducer) Close() error {
	panic("not impemented")
}

func (a *testAsyncProducer) Input() chan<- *sarama.ProducerMessage {
	return a.messages
}

func (a *testAsyncProducer) Successes() <-chan *sarama.ProducerMessage {
	return a.successes
}

func (a *testAsyncProducer) Errors() <-chan *sarama.ProducerError {
	return a.errors
}

func (a *testAsyncProducer) loop() {
	for {
		msg := <-a.messages
		if msg == nil {
			a.w.Done()
			return
		}
		err := a.nextError()
		if err != nil {
			a.w.Add(1)
			go func() {
				a.errors <- &sarama.ProducerError{
					Err: err,
				}
				a.w.Done()
			}()
		} else {
			a.w.Add(1)
			go func(msg sarama.ProducerMessage) {
				msg.Partition = 1
				msg.Offset = 1
				a.successes <- &msg
				a.producerMessages <- &msg
				a.w.Done()
			}(*msg)
		}
	}
}

func (a *testAsyncProducer) closeChannels() {
	<-a.close
	close(a.messages)
	a.w.Wait()
	close(a.successes)
	close(a.errors)
}

func (a *testAsyncProducer) nextError() error {
	a.m.Lock()
	defer a.m.Unlock()
	if len(a.producerErrors) > 0 {
		var e error
		e, a.producerErrors = a.producerErrors[0], a.producerErrors[1:]
		return e
	}
	return nil
}

func (a *testAsyncProducer) setErrors(errors ...error) {
	a.m.Lock()
	defer a.m.Unlock()
	a.producerErrors = errors
}
