// Copyright 2019 Canonical Ltd.

package kafkaauditor

import (
	"context"
	"time"

	"github.com/Shopify/sarama"
	"github.com/juju/zaputil"
	"github.com/juju/zaputil/zapctx"
	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/config"
	"github.com/CanonicalLtd/ua-contracts/internal/monitoring"
	"github.com/CanonicalLtd/ua-contracts/internal/store"
)

const defaultKafkaVersion = "2.1.0"

var (
	defaultBackoff = 5 * time.Minute
)

// AuditStore represents the store implementation implemented by kafkaauditor.
type AuditStore interface {
	Auditor
	store.Store
}

// Auditor allows a custom audit message to be sent using the kafkaauditor
// Store implementation. Usually, messages are generated automatically
// when the Store methods are called.
type Auditor interface {
	// Audit send an audit entry to Kafka.
	Audit(context.Context, time.Time, AuditEntry)
}

// AuditEntry represents an interface implemented by each audit entry type.
type AuditEntry interface {
	// KafkaInfo returns information on how the message should
	// sent to Kafka.
	KafkaInfo() KafkaInfo
}

// KafkaInfo holds information on how a message should be
// held within Kafka.
type KafkaInfo struct {
	// Topic holds the topic that the message will be sent to.
	Topic string
	// Key holds the key of the message sent to Kafka. The partition
	// a message gets sent to depends on the message Key and the
	// chose partitioner (client side configuration).
	Key string
}

// KafkaConfig returns the configuration for the Kafka auditor.
func KafkaConfig() (*Config, error) {
	kafkaBrokerURLs := config.KafkaBrokerURLs()
	if len(kafkaBrokerURLs) > 0 {
		kafkaConfig := &Config{
			Brokers: kafkaBrokerURLs,
		}
		tlsConfig, err := config.KafkaTLS()
		if err != nil {
			return nil, errgo.Mask(err)
		}
		kafkaConfig.TLSConfig = tlsConfig
		return kafkaConfig, nil
	}
	return nil, nil
}

// Config holds the configuration data needed by
// the client to connect to kafka.
type Config struct {
	Brokers         []string          `yaml:"brokers"`
	VersionOverride string            `yaml:"version-override"`
	TLSConfig       *config.TLSConfig `yaml:"tls"`

	Version sarama.KafkaVersion
}

// Validate validates the kafka configuration
func (c *Config) Validate() error {
	if len(c.Brokers) == 0 {
		return errgo.New("kafka brokers must be specified")
	}
	var (
		version sarama.KafkaVersion
		err     error
	)
	if c.VersionOverride == "" {
		version, err = sarama.ParseKafkaVersion(defaultKafkaVersion)
	} else {
		version, err = sarama.ParseKafkaVersion(c.VersionOverride)
	}
	if err != nil {
		return errgo.Notef(err, "unable to parse kafka version")
	}
	c.Version = version

	return nil
}

// Auditor returns an implementation of store.Auditor
// that sends selected audit messages to kafka.
func (c *Config) Auditor() (Auditor, error) {
	if err := c.Validate(); err != nil {
		return nil, errgo.Mask(err)
	}

	config := sarama.NewConfig()
	config.ClientID = "ua-contracts-service"
	config.Producer.Return.Successes = true
	config.Producer.Partitioner = sarama.NewHashPartitioner
	config.Version = c.Version

	if c.TLSConfig != nil {
		cfg, err := c.TLSConfig.Config()
		if err != nil {
			return nil, errgo.Mask(err)
		}
		config.Net.TLS.Config = cfg
		config.Net.TLS.Enable = true
	}

	if err := config.Validate(); err != nil {
		return nil, errgo.Mask(err)
	}

	client, err := sarama.NewClient(c.Brokers, config)
	if err != nil {
		return nil, errgo.Mask(err)
	}

	producer, err := sarama.NewAsyncProducerFromClient(client)
	if err != nil {
		return nil, errgo.Mask(err)
	}

	return newClient(producer), nil
}

// Store returns a store implementation that wraps the given store so
// that method calls send appropriate kafka events to the Kafka server
// specified by the config.
func (c *Config) Store(ctx context.Context, store store.Store) (AuditStore, error) {
	return c.store(ctx, store, c.Auditor, defaultBackoff)
}

func (c *Config) store(ctx context.Context, store store.Store, auditorFactory func() (Auditor, error), backoffDuration time.Duration) (AuditStore, error) {
	s := &auditStore{
		store: store,
	}
	go func() {
		for {
			// try to start the kafka auditor
			auditor, err := auditorFactory()
			if err == nil {
				s.setAuditor(auditor)
				zapctx.Info(ctx, "started the kafka auditor")
				monitoring.KafkaConnection.Set(1)
				return
			}
			// if we fail to start the auditor, we log
			// and try again after a backoff period
			zapctx.Warn(ctx, "failed to start kafka auditor", zaputil.Error(err))
			monitoring.KafkaConnection.Set(0)
			select {
			case <-time.After(backoffDuration):
			case <-ctx.Done():
				return
			}
		}
	}()
	return s, nil
}
