// Copyright 2019 Canonical Ltd.

package kafkaauditor_test

import (
	"context"
	"testing"
	"time"

	"github.com/Shopify/sarama"
	qt "github.com/frankban/quicktest"
	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/internal/kafkaauditor"
	"github.com/CanonicalLtd/ua-contracts/internal/memstore"
	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
)

func TestConfig(t *testing.T) {
	c := qt.New(t)

	cfg := kafkaauditor.Config{
		Brokers:         []string{"127.0.0.1:9092"},
		VersionOverride: "0.10.1.0",
	}
	c.Assert(cfg.Validate(), qt.Equals, nil)
	c.Assert(cfg.Version, qt.Equals, sarama.V0_10_1_0)

	cfg.VersionOverride = ""
	c.Assert(cfg.Validate(), qt.Equals, nil)
	ver, err := sarama.ParseKafkaVersion("2.1.0")
	c.Assert(err, qt.Equals, nil)
	c.Assert(cfg.Version, qt.Equals, ver)
}

func TestConfigAsyncKafkaConnect(t *testing.T) {
	c := qt.New(t)

	cfg := kafkaauditor.Config{
		Brokers:         []string{"127.0.0.1:9092"},
		VersionOverride: "0.10.1.0",
	}
	c.Assert(cfg.Validate(), qt.Equals, nil)

	mstore := memstore.New()
	auditor := &testAuditor{}
	factory := newTestKafkaFactory(auditor)
	ctx := context.Background()
	now := time.Now().UTC()

	st, err := cfg.NewStore(ctx, mstore, factory.new, time.Millisecond)
	c.Assert(err, qt.IsNil)

	account := store.Account{
		ID:      token.MustNewAccountID(),
		Name:    "test-account",
		Created: now,
	}

	err = st.AddAccount(ctx, account, "test-user")
	c.Assert(err, qt.Equals, nil)
	retrievedAccount, err := mstore.Account(ctx, account.ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(retrievedAccount, qt.DeepEquals, account)
	c.Assert(auditor.entries, qt.HasLen, 0)

	factory.unblock <- true
	select {
	case <-factory.created:
	case <-time.After(time.Second):
		c.Fatal("time out")
	}

	account = store.Account{
		ID:      token.MustNewAccountID(),
		Name:    "new-test-account",
		Created: now,
	}
	err = st.AddAccount(ctx, account, "test-user")
	c.Assert(err, qt.Equals, nil)
	retrievedAccount, err = mstore.Account(ctx, account.ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(retrievedAccount, qt.DeepEquals, account)
	c.Assert(auditor.entries, qt.DeepEquals, []kafkaauditor.AuditEntry{
		&kafkaauditor.AuditAddAccount{
			AccountID:      account.ID,
			AccountName:    account.Name,
			AccountCreated: now,
			UserID:         "test-user",
		},
	})
}

func newTestKafkaFactory(a kafkaauditor.Auditor) *testKafkaFactory {
	return &testKafkaFactory{
		auditor: a,
		unblock: make(chan bool, 1),
		created: make(chan bool, 1),
	}
}

type testKafkaFactory struct {
	auditor kafkaauditor.Auditor
	unblock chan bool
	created chan bool
}

func (t *testKafkaFactory) new() (kafkaauditor.Auditor, error) {
	select {
	case <-t.unblock:
		select {
		case t.created <- true:
		default:
		}
		return t.auditor, nil
	default:
		return nil, errgo.New("not ready")
	}

}
