// Copyright 2019 Canonical Ltd.

package monitoring

import (
	"context"
	"encoding/base64"
	"net/http"

	"github.com/juju/zaputil/zapctx"
	"github.com/julienschmidt/httprouter"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	"github.com/rogpeppe/fastuuid"
	"go.uber.org/zap"
	"gopkg.in/httprequest.v1"
)

type traceIdContextKey string

const traceIdKey = traceIdContextKey("traceId")

var (
	RequestDuration = prometheus.NewSummaryVec(prometheus.SummaryOpts{
		Namespace: "http",
		Subsystem: "request",
		Name:      "duration_seconds",
		Help:      "The duration of API HTTP requests.",
	}, []string{"method", "handler", "code"})

	SalesforceErrorCount = prometheus.NewCounter(prometheus.CounterOpts{
		Namespace: "salesforce",
		Name:      "error_count",
		Help:      "A counter for errors when interacting with Salesforce.",
	})

	DatabaseErrorCount = prometheus.NewCounter(prometheus.CounterOpts{
		Namespace: "database",
		Name:      "error_count",
		Help:      "A counter for errors when interacting with postgres.",
	})

	KafkaConnection = prometheus.NewGauge(prometheus.GaugeOpts{
		Namespace: "kafka",
		Name:      "connection",
		Help:      "A gauge indicating connection to kafka",
	})

	KafkaErrorCount = prometheus.NewCounter(prometheus.CounterOpts{
		Namespace: "kafka",
		Name:      "error_count",
		Help:      "A counter for errors when sending to kafka.",
	})

	uuidGenerator = fastuuid.MustNewGenerator()
)

func init() {
	prometheus.MustRegister(
		DatabaseErrorCount,
		RequestDuration,
		SalesforceErrorCount,
		KafkaConnection,
		KafkaErrorCount,
	)
}

// AddHandlers adds all the handlers in the given slice to the given *httprouter.Router,
// wrapping the handlers in Prometheus metrics middleware to track request details,
// durations, and status.
func AddHandlers(r *httprouter.Router, hs []httprequest.Handler) {
	for _, h := range hs {
		AddHTTPHandler(r, h.Method, h.Path, httprequest.ToHTTP(h.Handle))
	}
}

// AddHTTPHandler adds the handler with the given method and path to the
// given *httprouter.Router, wrapping the handler in a Prometheus metrics
// middleware to track request details, durations, and status. A unique trace ID
// is also added to the request context and as a response header.
func AddHTTPHandler(r *httprouter.Router, method, path string, hnd http.Handler) {
	h := promhttp.InstrumentHandlerDuration(
		RequestDuration.MustCurryWith(
			prometheus.Labels{"handler": path},
		),
		hnd,
	)
	r.HandlerFunc(method, path, func(rw http.ResponseWriter, r *http.Request) {
		id := NewTraceID()
		rw.Header().Set("x-trace-id", id)
		ctx := context.WithValue(r.Context(), traceIdKey, id)
		ctx = zapctx.WithFields(
			ctx,
			zap.String("method", r.Method),
			zap.String("path", r.URL.Path),
			zap.String("trace-id", id),
		)
		h(rw, r.WithContext(ctx))
	})
}

// AddMetricsHandler adds a prometheus metrics handler to the given router on the given
// path (usually /metrics).
func AddMetricsHandler(r *httprouter.Router, path string) {
	r.Handler("GET", path, promhttp.Handler())
}

// NewTraceID generates a new base64 encoded trace ID string.
func NewTraceID() string {
	b := uuidGenerator.Next()
	return base64.RawURLEncoding.EncodeToString(b[:])
}

// TraceIDFromContext returns the trace ID for a given context, or an empty string if not set.
func TraceIDFromContext(ctx context.Context) string {
	id, _ := ctx.Value(traceIdKey).(string)
	return id
}
