// Copyright 2019 Caonical Ltd.

// Package pkcs7 contains the minimal parts of the PKCS#7 standard
// required for the contract service.
package pkcs7

import (
	"crypto"
	"crypto/dsa"
	"crypto/rand"
	"crypto/rsa"
	"crypto/x509"
	"crypto/x509/pkix"
	"encoding/asn1"
	"fmt"
	"math/big"

	errgo "gopkg.in/errgo.v1"
)

// A SignedData holds a data packet signed by an authority.
type SignedData struct {
	data  []byte
	certs []*x509.Certificate
	sigs  []signature
}

// NewSignedData creates a new SignedData value with the given contents.
func NewSignedData(b []byte) *SignedData {
	return &SignedData{
		data: b,
	}
}

// Sign signs the data with the given algorithm and key, identified by
// the given certificate.
func (s *SignedData) Sign(alg x509.SignatureAlgorithm, cert *x509.Certificate, key interface{}) error {
	var hash crypto.Hash
	var sign func(key interface{}, hash crypto.Hash, hashed []byte) ([]byte, error)
	switch alg {
	case x509.DSAWithSHA1:
		hash = crypto.SHA1
		sign = signDSA
	case x509.SHA256WithRSA:
		hash = crypto.SHA256
		sign = signRSA
	default:
		return errgo.WithCausef(nil, x509.ErrUnsupportedAlgorithm, "")
	}

	if !hash.Available() {
		return errgo.WithCausef(nil, x509.ErrUnsupportedAlgorithm, "")
	}
	h := hash.New()
	h.Write(s.data)

	sig, err := sign(key, hash, h.Sum(nil))
	if err != nil {
		return errgo.Mask(err)
	}

	if err := cert.CheckSignature(alg, s.data, sig); err != nil {
		return errgo.Notef(err, "private key does not match certificate")
	}

	s.sigs = append(s.sigs, signature{
		Certificate:        cert,
		SignatureAlgorithm: alg,
		EncryptedDigest:    sig,
		data:               s.data,
	})
	return nil
}

func signDSA(key interface{}, _ crypto.Hash, hashed []byte) ([]byte, error) {
	dsakey, ok := key.(*dsa.PrivateKey)
	if !ok {
		return nil, errgo.Newf("incorrect key type for DSA encryption (%T)", key)
	}
	r, s, err := dsa.Sign(rand.Reader, dsakey, hashed)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	sig, err := asn1.Marshal(dsaSignature{R: r, S: s})
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return sig, nil
}

type dsaSignature struct {
	R, S *big.Int
}

func signRSA(key interface{}, hash crypto.Hash, hashed []byte) ([]byte, error) {
	rsakey, ok := key.(*rsa.PrivateKey)
	if !ok {
		return nil, errgo.Newf("incorrect key type for RSA encryption (%T)", key)
	}
	sig, err := rsa.SignPKCS1v15(rand.Reader, rsakey, hash, hashed)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return sig, nil
}

// Encode creates a DER encoded PKCS#7 signed data packet.
func (s *SignedData) Encode() []byte {
	dataBytes, err := asn1.Marshal(s.data)
	if err != nil {
		panic(errgo.Mask(err))
	}
	data := contentInfo{
		ContentType: oidData,
		Content: asn1.RawValue{
			Class:      asn1.ClassContextSpecific,
			Tag:        0,
			IsCompound: true,
			Bytes:      dataBytes,
		},
	}

	digests := make([]pkix.AlgorithmIdentifier, 0, len(s.sigs))
	signers := make([]signerInfo, len(s.sigs))

	for i, sig := range s.sigs {
		signers[i] = signerInfo{
			Version: 1,
			IssuerAndSerialNumber: issuerAndSerialNumber{
				Name: asn1.RawValue{
					FullBytes: sig.Certificate.RawIssuer,
				},
				SerialNumber: sig.Certificate.SerialNumber,
			},
			DigestAlgorithm:           digestAlgorithm(sig.SignatureAlgorithm),
			DigestEncryptionAlgorithm: digestEncryptionAlgorithm(sig.SignatureAlgorithm),
			EncryptedDigest:           sig.EncryptedDigest,
		}
		var found bool
		for _, alg := range digests {
			if alg.Algorithm.Equal(signers[i].DigestAlgorithm.Algorithm) {
				found = true
				break
			}
		}
		if !found {
			digests = append(digests, signers[i].DigestAlgorithm)
		}
	}

	sd := signedData{
		Version:          1,
		DigestAlgorithms: digests,
		Content:          data,
		Certificates:     s.certs,
		SignerInfos:      signers,
	}

	sdBytes, err := asn1.Marshal(sd)
	if err != nil {
		panic(errgo.Mask(err))
	}

	content := contentInfo{
		ContentType: oidSignedData,
		Content: asn1.RawValue{
			Class:      asn1.ClassContextSpecific,
			Tag:        0,
			IsCompound: true,
			Bytes:      sdBytes,
		},
	}

	b, err := asn1.Marshal(content)
	if err != nil {
		panic(errgo.Mask(err))
	}
	return b
}

func digestAlgorithm(alg x509.SignatureAlgorithm) pkix.AlgorithmIdentifier {
	switch alg {
	case x509.DSAWithSHA1:
		return pkix.AlgorithmIdentifier{Algorithm: oidSHA1}
	case x509.SHA256WithRSA:
		return pkix.AlgorithmIdentifier{Algorithm: oidSHA256}
	default:
		panic(fmt.Errorf("unsupported algorithm %s", alg))
	}
}

func digestEncryptionAlgorithm(alg x509.SignatureAlgorithm) pkix.AlgorithmIdentifier {
	switch alg {
	case x509.DSAWithSHA1:
		return pkix.AlgorithmIdentifier{Algorithm: oidSHA1WithDSA}
	case x509.SHA256WithRSA:
		return pkix.AlgorithmIdentifier{Algorithm: oidSHA256WithRSA}
	default:
		panic(fmt.Errorf("unsupported algorithm %s", alg))
	}
}

// A signature is a signature of some signed data.
type signature struct {
	data               []byte
	Certificate        *x509.Certificate
	SignatureAlgorithm x509.SignatureAlgorithm
	EncryptedDigest    []byte
}
