// Copyright 2019 Canonical Ltd.

package pkcs7_test

import (
	"crypto/dsa"
	"crypto/x509"
	"encoding/pem"
	"testing"

	qt "github.com/frankban/quicktest"
	fspkcs7 "github.com/fullsailor/pkcs7"

	"github.com/CanonicalLtd/ua-contracts/internal/pkcs7"
)

var testDSAKey = mustParseDSAKey(`
-----BEGIN DSA PRIVATE KEY-----
MIIBvAIBAAKBgQDV58Din14EN5OyKZ6QHnobCpAS3KRLsLS87rbiolVrbxWLl98H
9uhPxx78KMEoDsxwEv87R64Plp1vkXz4QNfobMbxopmnTqbyGZZKm3EW2jyzHuxL
bI0yWUR/IRaaUgSpTsYNg78DxhqT5WzezMZLqRrEIr+AvHt/K+ad5wI3gwIVAO19
giKrbg0pfdi3qKJQ/IQJ02hFAoGBAJisnrvK+lrVuzot7DCy1YrpW9Y5d5TzriEs
tx3H4PsbyeT68hOqoGer1BdN1RQ3dxlfZrBm/XGEKfFRdV8y6Idnsq9SPLZox29z
gj0eEqXZ6aTIJsUSf6BgWNKwD06u3xrxkKEMRI0SK34i2HxtqhQzfpLwoRoWzvOV
GT9WbEbaAoGAWAq8KwJA+B4dR3Mz0suoG5M2uSbaGds7J7OeagfvDi1BOtv3a0E4
0FOmxXCYpZWEJXrL2CMLlW0q9N5SeSLGvjtDsuO6uqOeFlPsXPFlpVgOY5fVICJN
+C7FYzQQSEmLzVPVG6EJalV7xo80fONd7CTOpBY+V/P4sUlLU2R1+xUCFQDZIhbG
2vu4y3zK+FIUDSSVZwBplA==
-----END DSA PRIVATE KEY-----
`)

func mustParseDSAKey(p string) *dsa.PrivateKey {
	block, _ := pem.Decode([]byte(p))
	key, err := pkcs7.ParseDSAPrivateKey(block.Bytes)
	if err != nil {
		panic(err)
	}
	return key
}

var testDSACertificate = mustParseCert(`
-----BEGIN CERTIFICATE-----
MIICsTCCAm+gAwIBAgIJAJaZspYyDBfaMAsGCWCGSAFlAwQDAjAPMQ0wCwYDVQQD
DARUZXN0MB4XDTE5MDQxODA4MzgwM1oXDTM5MDQxMzA4MzgwM1owDzENMAsGA1UE
AwwEVGVzdDCCAbcwggEsBgcqhkjOOAQBMIIBHwKBgQDV58Din14EN5OyKZ6QHnob
CpAS3KRLsLS87rbiolVrbxWLl98H9uhPxx78KMEoDsxwEv87R64Plp1vkXz4QNfo
bMbxopmnTqbyGZZKm3EW2jyzHuxLbI0yWUR/IRaaUgSpTsYNg78DxhqT5WzezMZL
qRrEIr+AvHt/K+ad5wI3gwIVAO19giKrbg0pfdi3qKJQ/IQJ02hFAoGBAJisnrvK
+lrVuzot7DCy1YrpW9Y5d5TzriEstx3H4PsbyeT68hOqoGer1BdN1RQ3dxlfZrBm
/XGEKfFRdV8y6Idnsq9SPLZox29zgj0eEqXZ6aTIJsUSf6BgWNKwD06u3xrxkKEM
RI0SK34i2HxtqhQzfpLwoRoWzvOVGT9WbEbaA4GEAAKBgFgKvCsCQPgeHUdzM9LL
qBuTNrkm2hnbOyeznmoH7w4tQTrb92tBONBTpsVwmKWVhCV6y9gjC5VtKvTeUnki
xr47Q7LjurqjnhZT7FzxZaVYDmOX1SAiTfguxWM0EEhJi81T1RuhCWpVe8aPNHzj
XewkzqQWPlfz+LFJS1NkdfsVo1MwUTAdBgNVHQ4EFgQUthZKFYU5gOkvaf1mI4sZ
fdxicZYwHwYDVR0jBBgwFoAUthZKFYU5gOkvaf1mI4sZfdxicZYwDwYDVR0TAQH/
BAUwAwEB/zALBglghkgBZQMEAwIDLwAwLAIUGJz0hVq34AAvTlUWvnk+N227mtkC
FD2NA5R+eVU2P9Nwws+GvYmex2KP
-----END CERTIFICATE-----
`)

func mustParseCert(p string) *x509.Certificate {
	block, _ := pem.Decode([]byte(p))
	crt, err := x509.ParseCertificate(block.Bytes)
	if err != nil {
		panic(err)
	}
	return crt
}

func TestDSAWithSHA1(t *testing.T) {
	c := qt.New(t)

	sd := pkcs7.NewSignedData([]byte("TEST"))
	err := sd.Sign(x509.DSAWithSHA1, testDSACertificate, testDSAKey)
	c.Assert(err, qt.Equals, nil)

	p7, err := fspkcs7.Parse(sd.Encode())
	c.Assert(err, qt.Equals, nil)

	p7.Certificates = append(p7.Certificates, testDSACertificate)
	err = p7.Verify()
	c.Assert(err, qt.Equals, nil)
}
