// Copyright 2019 Canonical Ltd.

// Package saauth provides methods for authenticating service-accounts.
package saauth

import (
	"context"
	"crypto/rand"
	"crypto/subtle"
	"net/http"

	aclstore "github.com/juju/aclstore/v2"
	"golang.org/x/crypto/argon2"
	errgo "gopkg.in/errgo.v1"
	httprequest "gopkg.in/httprequest.v1"

	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/params"
)

// An Identity holds the identity of a service account.
type Identity struct {
	account string
}

// Account returns the name of this service account.
func (i Identity) Account() string {
	return i.account
}

// Allow determines whether this identity can access the given ACL. Allow
// implements aclstore.Identity.
func (i Identity) Allow(_ context.Context, acl []string) (bool, error) {
	if i.account == "" {
		return false, nil
	}
	for _, v := range acl {
		if v == i.account {
			return true, nil
		}
	}
	return false, nil
}

// Validate checks that the given account has the given password. If bad
// credentials are provided the returned Identity will not allow access
// to any ACL. An error is only returned if there is an unexpected
// problem communicating with the store.
func Validate(ctx context.Context, s store.Store, account, password string) (Identity, error) {
	var i Identity
	sa, err := s.ServiceAccount(ctx, account)
	if errgo.Cause(err) == store.ErrServiceAccountNotFound {
		return i, nil
	}
	if err != nil {
		return i, errgo.Mask(err)
	}
	hash := argon2.IDKey([]byte(password), sa.PasswordSalt, 1, 64*1024, 4, 32)
	if subtle.ConstantTimeCompare(hash, sa.PasswordHash) == 1 {
		// 1 means they are the same.
		i.account = account
	}
	return i, nil
}

// SetPassword sets the password for a service account.
func SetPassword(ctx context.Context, s store.Store, account, password string) error {
	salt := make([]byte, 16)
	if _, err := rand.Read(salt); err != nil {
		return errgo.Mask(err)
	}
	hash := argon2.IDKey([]byte(password), salt, 1, 64*1024, 4, 32)
	return errgo.Mask(s.EnsureServiceAccount(ctx, store.ServiceAccount{
		Name:         account,
		PasswordSalt: salt,
		PasswordHash: hash,
	}))
}

// AuthenticateACL creates an authenticator suitable for use with an
// ACLManager.
func AuthenticateACL(s store.Store) func(ctx context.Context, w http.ResponseWriter, req *http.Request) (aclstore.Identity, error) {
	return func(ctx context.Context, w http.ResponseWriter, req *http.Request) (aclstore.Identity, error) {
		username, password, ok := req.BasicAuth()
		if !ok {
			err := &params.Error{
				Code:    params.ErrUnauthorized,
				Message: string(params.ErrUnauthorized),
			}
			httprequest.WriteJSON(w, http.StatusUnauthorized, err)
			return nil, err
		}
		id, err := Validate(ctx, s, username, password)
		if err != nil {
			httprequest.WriteJSON(w, http.StatusInternalServerError, params.Error{
				Message: err.Error(),
			})
			return nil, err
		}
		return id, nil
	}
}
