// Copyright 2019 Canonical Ltd.

package store

import (
	"context"
	"time"

	"github.com/CanonicalLtd/ua-contracts/internal/token"
	"github.com/CanonicalLtd/ua-contracts/params"
)

// ResourceTypes holds all the known resource types in alphabetical order.
var ResourceTypes = []string{
	"cc-eal",
	"esm",
	"fips",
	"fips-updates",
	"livepatch",
	"support",
	"uca",
}

type Origin string

const (
	AwsOrigin        Origin = "AWS"
	AzureOrigin      Origin = "Azure"
	SalesforceOrigin Origin = "Salesforce"
	FreeOrigin       Origin = "Free"
)

// A Contract represents a product associated with an account.
type Contract struct {
	// ID holds the ID of the contract.
	ID token.ContractID

	// The name associated with the contract.
	Name string

	// AccountID contains the account which is associated with this
	// contract.
	AccountID token.AccountID

	// ProductIDs contains the products to which this contract
	// pertains. The contract resources implied by these products
	// may be overridden by contract-specific data added with
	// AddContractInfo.
	ProductIDs []string

	// Created contains the time the contract was created.
	Created time.Time

	// CreatedBy contains an identifier for the user who created the
	// contract.
	CreatedBy string

	// EffectiveFrom and EffectiveTo contain the range of times
	// that the contract is valid. For the free contract, they will
	// be nil (implying unlimited effectiveness over all time)
	EffectiveFrom *time.Time
	EffectiveTo   *time.Time

	// Origin represents where the contract was originated from.
	Origin Origin
}

func timeEqual(t1, t2 *time.Time) bool {
	if t1 != nil && t2 != nil {
		return t1.Round(time.Microsecond).Equal(t2.Round(time.Microsecond))
	}
	return (t1 == nil) == (t2 == nil)
}

// EffectivelyEqual reports whether c1 is effectively equivalent to c,
// taking into account only the account ID and effective dates.
func (c Contract) EffectivelyEqual(c1 Contract) bool {
	return c.AccountID == c1.AccountID &&
		timeEqual(c.EffectiveFrom, c1.EffectiveFrom) &&
		timeEqual(c.EffectiveTo, c1.EffectiveTo)
}

// A FullContract contains all the information about a contract including
// the associated account and product.
type FullContract struct {
	Contract Contract
	// Account holds the account information for the contract.
	Account Account
	// Config holds product configuration overridden for the contract.
	Config []ProductResource
}

type MachineInfo struct {
	MachineID    string
	Architecture string
	OS           params.OS
}

type ContractTokenInfo struct {
	ContractToken token.ContractToken
	ContractID    token.ContractID
	Expires       time.Time
}

// A User is an individual user that can log-in to the system.
type User struct {
	// ID contains a unique ID for the user.
	ID          string
	Username    string
	DisplayName string
	Email       string
	Verified    bool
}

// An Account is an entity that can have associated contracts.
type Account struct {
	ID      token.AccountID
	Name    string
	Created time.Time

	// SalesforceAccountID contains the account ID of the represented
	// salesforce account.
	SalesforceAccountID string

	// Origin holds where the account originated.
	Origin Origin
	// OriginID holds the account ID on the origin of the account
	// this represents.
	OriginID string
}

// AccountAccess describes the possible levels of access a user might
// have on an account.
type AccountAccess int

const (
	NoAccountAccess        AccountAccess = 0
	AdminAccountAccess     AccountAccess = 1
	DelegatedAccountAccess AccountAccess = 2
)

// AccountWithAccess contains account information with the access level
// of the requested account user.
type AccountWithAccess struct {
	Account
	Access AccountAccess
}

// A Product contains the details of a product.
type Product struct {
	// ID holds the ID of the product.
	ID string

	// Name holds the name of the product.
	Name string

	// Resources contains an entry for each resource
	// associated with the product.
	// A product should have only a maximum of one
	// resource entry for any given resource type.
	Resources []ProductResource
}

// ProductResource represents a resource that a product enables.
type ProductResource struct {
	// ResourceType holds the name of the resource type.
	ResourceType string
	// CanOverride specifies whether the product can
	// be overridden. If this is true and Data is empty,
	// this resource must be overridden for a contract
	// using the product to be used.
	CanOverride bool
	// Data holds entitlement data for the resource for this
	// product, in JSON format.
	Data string
}

// A ServiceAccount is an account used by a service to authenticate with
// the contracts service. These accounts authenticate using HTTP Basic
// authentication.
type ServiceAccount struct {
	Name         string
	PasswordSalt []byte
	PasswordHash []byte
}

// Store represents the storage backend required by the contracts
// service.
type Store interface {
	// Account retrieves the account with the given ID. If the
	// account is not found then the returned error will have a cause
	// of ErrAccountNotFound.
	Account(ctx context.Context, id token.AccountID) (Account, error)

	// AccountForSalesforceID retrieves the account with the given
	// Salesforce ID. If there is no such account then the returned
	// error will have a cause of ErrAccountNotFound.
	AccountForSalesforceID(ctx context.Context, salesforceID string) (Account, error)

	// AccountAccess gets the level of access the given user has to
	// the given account. An error is only returned if the lookup
	// fails. A return value of NoAccountAccess is used to indicate
	// that the user cannot access the account.
	AccountAccess(ctx context.Context, accountID token.AccountID, userID string) (AccountAccess, error)

	// AccountContracts returns the full details of all the contracts
	// belonging to the given account.
	AccountContracts(ctx context.Context, accountID token.AccountID) ([]FullContract, error)

	// AddContract adds the given contract to the store and returns
	// its newly created identifier. If the contract has the same
	// account ID, product ID and effective from and to dates, then
	// the contract is considered a duplicate and will fail with an
	// ErrContractExists error.
	//
	// AddContract will return an error if the contract's account
	// does not exist.
	//
	// If the products associated with the contract have clashing resources,
	// this returns an error with a ErrProductConflict cause.
	AddContract(ctx context.Context, c Contract) (token.ContractID, error)

	// AddContractMachine associates a machine with a contract.
	AddContractMachine(ctx context.Context, cid token.ContractID, machineID string) error

	// AddContractToken adds the given contract token to the store
	// associated with the given info. The store may remove the
	// token entry after info.Expires.
	// The token must be unique.
	AddContractToken(ctx context.Context, info ContractTokenInfo) error

	// Product returns the product with the given ID.
	// If the product was not found, it returns an error with a ErrProductNotFound
	// cause.
	Product(ctx context.Context, productID string) (Product, error)

	// AllProducts returns all products from the store, sorted by ID.
	AllProducts(ctx context.Context) ([]Product, error)

	// AddProduct adds a product to the store. It returns an error
	// with an ErrDuplicateProduct cause if there is already a
	// product with the same ID. The product's resource data applies
	// to all contracts with this product ID, but if CanOverride is
	// true, a product's resource may be overridden by
	// contract-specific info added by ConfigureContract.
	AddProduct(ctx context.Context, p Product) error

	// UpdateProduct updates an existing product. The product with
	// the id p.ID is updated to be the same as p.
	// UpdateProduct returns an error with an ErrProductNotFound cause
	// if the product does not exist.
	UpdateProduct(ctx context.Context, p Product) error

	// ConfigureContract configures the given contract to
	// override information for the given resources. The
	// caller is responsible for checking that the products
	// associated with the contract contain the given
	// resources and that they're overridable.
	ConfigureContract(ctx context.Context, contractID token.ContractID, resources []ProductResource) error

	// Contract returns a contract given its id. If the contract was
	// not found, it returns an error with an ErrContractNotFound
	// cause.
	Contract(ctx context.Context, id token.ContractID) (FullContract, error)

	// ContractToken returns information for the given contract
	// token. If an entry doesn't exist, it returns an error with
	// a ErrTokenNotFound cause.
	ContractToken(ctx context.Context, ct token.ContractToken) (ContractTokenInfo, error)

	// ContractTokens returns any contract tokens associated with the
	// given contract.
	ContractTokens(ctx context.Context, id token.ContractID) ([]ContractTokenInfo, error)

	// AddAccount creates the given account. The given userID is
	// given admin access to the account. If an account already
	// exists with the same ID then an error with a cause of
	// ErrAccountExists is returned. If the account specifies a
	// SalesforceAccountID and there is already an account for that
	// Salesforce account then an error with a cause of
	// ErrDuplicateSalesforceAccount is returned. If the account
	// specifies an Origin and OriginID and an account with that
	// Origin ID exists then an error with a cause of
	// ErrDuplicateOriginAccount is returned.
	AddAccount(ctx context.Context, a Account, userID string) error

	// EnsureMachine ensures that an entry for the machine with
	// the given info exists, creating it if needed.
	EnsureMachine(ctx context.Context, info MachineInfo) error

	// EnsureServiceAccount creates, or updates, the given service
	// account information.
	EnsureServiceAccount(ctx context.Context, account ServiceAccount) error

	// EnsureUser ensures that an entry for the user with the given
	// info exists, creating it if needed, and updating
	// it if the user id already exists.
	EnsureUser(ctx context.Context, user User) error

	// FindUserByMail find a user by email address.
	FindUserByMail(ctx context.Context, email string) (User, error)

	// MachineHasContract checks that the given machine is associated
	// with the given contract.
	MachineHasContract(ctx context.Context, cid token.ContractID, machineID string) (bool, error)

	// MachineInfo returns information on the machine with the given ID.
	// If the machine is not found, and error with an ErrMachineNotFound
	// cause is returned.
	MachineInfo(ctx context.Context, machineID string) (MachineInfo, error)

	// RemoveContractMachine disassociates a machine with a contract.
	RemoveContractMachine(ctx context.Context, cid token.ContractID, machineID string) error

	// ServiceAccount gets the service account with the given name.
	// If there is no such service account then an error with a cause
	// of ErrServiceAccountNotFound will be returned.
	ServiceAccount(ctx context.Context, name string) (ServiceAccount, error)

	// SetAccountAccess gives the given user the given access level
	// on the given account.
	SetAccountAccess(ctx context.Context, accountID token.AccountID, userID string, a AccountAccess) error

	// UpdateContractProducts updates the products linked to a contract.
	UpdateContractProducts(ctx context.Context, contractID token.ContractID, productIDs []string) error

	// User returns the details of the given user. If the user cannot
	// be found an error with a cause of ErrUserNotFound will be
	// returned.
	User(ctx context.Context, id string) (User, error)

	// UserAccounts returns all the accounts a user has access to.
	UserAccounts(ctx context.Context, id string) ([]AccountWithAccess, error)
}
