package storetest

import (
	"bytes"
	"context"
	"fmt"
	"sort"
	"time"

	qt "github.com/frankban/quicktest"
	"github.com/frankban/quicktest/qtsuite"
	"github.com/google/go-cmp/cmp"
	"github.com/juju/zaputil/zapctx"
	"go.uber.org/zap/zaptest"
	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
	"github.com/CanonicalLtd/ua-contracts/params"
)

// StoreSuite contains a set of tests for Store implementations.
type StoreSuite struct {
	newStore func(c *qt.C) store.Store

	store store.Store
	ctx   context.Context
}

func TestStore(c *qt.C, newStore func(c *qt.C) store.Store) {
	qtsuite.Run(c, &StoreSuite{
		newStore: newStore,
	})
}

// NewStoreSuite returns an instance of the suite that will use the
// given store. The returned value should only
// be used for a single test.
//
// This can be used to run a specific test from the suite.
func NewStoreSuite(ctx context.Context, st store.Store) *StoreSuite {
	return &StoreSuite{
		store: st,
		ctx:   ctx,
	}
}

func (s *StoreSuite) Init(c *qt.C) {
	c.Parallel()
	s.store = s.newStore(c)
	s.ctx = zapctx.WithLogger(context.Background(), zaptest.NewLogger(c))
}

func (s *StoreSuite) TestAddContractNoAccount(c *qt.C) {
	ctx := s.ctx
	s.addTestProduct(ctx, c, s.store, "prodID")
	_, err := s.store.AddContract(ctx, store.Contract{
		Name:       "contract name",
		ProductIDs: []string{"prodID"},
	})
	c.Assert(err, qt.Not(qt.Equals), nil)
}

func (s *StoreSuite) TestAddProduct(c *qt.C) {
	ctx := s.ctx
	p := store.Product{
		ID:   "someProduct",
		Name: "some product name",
		Resources: []store.ProductResource{{
			ResourceType: "fips",
			Data:         `{"AptURL": "https://0.1.2.3/fips"}`,
		}, {
			ResourceType: "esm",
			Data:         `{"AptURL": "https://0.1.2.3/esm"}`,
		}},
	}
	err := s.store.AddProduct(ctx, p)
	c.Assert(err, qt.Equals, nil)
	p1, err := s.store.Product(ctx, "someProduct")
	c.Assert(p1, qt.DeepEquals, p)
}

func (s *StoreSuite) TestAddProductWithInvalidData(c *qt.C) {
	err := s.store.AddProduct(s.ctx, store.Product{
		ID:   "someProduct",
		Name: "some product name",
		Resources: []store.ProductResource{{
			ResourceType: "fips",
			Data:         `{`,
		}},
	})
	c.Assert(err, qt.Not(qt.IsNil))
}

func (s *StoreSuite) TestAddProductWithDuplicateResource(c *qt.C) {
	err := s.store.AddProduct(s.ctx, store.Product{
		ID:   "someProduct",
		Name: "some product name",
		Resources: []store.ProductResource{{
			ResourceType: "fips",
			Data:         `{}`,
		}, {
			ResourceType: "fips",
			Data:         `{}`,
		}},
	})
	c.Assert(errgo.Cause(err), qt.Equals, store.ErrProductConflict)
}

func (s *StoreSuite) TestAllProducts(c *qt.C) {
	ctx := s.ctx
	p1 := store.Product{
		ID:   "prod1",
		Name: "some product name",
		Resources: []store.ProductResource{{
			ResourceType: "fips",
			Data:         `{"AptURL": "https://0.1.2.3/fips"}`,
		}, {
			ResourceType: "esm",
			Data:         `{"AptURL": "https://0.1.2.3/esm"}`,
		}},
	}
	err := s.store.AddProduct(ctx, p1)
	p2 := store.Product{
		ID:   "prod2",
		Name: "other product name",
		Resources: []store.ProductResource{{
			ResourceType: "fips",
			Data:         `{"AptURL": "https://0.1.2.4/fips"}`,
		}, {
			ResourceType: "esm",
			Data:         `{"AptURL": "https://0.1.2.4/esm"}`,
		}},
	}
	err = s.store.AddProduct(ctx, p2)
	c.Assert(err, qt.Equals, nil)
	ps, err := s.store.AllProducts(ctx)
	c.Assert(ps, qt.DeepEquals, []store.Product{p1, p2})
}

func (s *StoreSuite) TestUpdateProduct(c *qt.C) {
	ctx := s.ctx

	p := store.Product{
		ID:   "someProduct",
		Name: "some product name",
		Resources: []store.ProductResource{{
			ResourceType: "fips",
			Data:         `{"AptURL": "https://0.1.2.3/fips"}`,
		}, {
			ResourceType: "esm",
			Data:         `{"AptURL": "https://0.1.2.3/esm"}`,
		}},
	}
	err := s.store.UpdateProduct(ctx, p)
	c.Assert(errgo.Cause(err), qt.Equals, store.ErrProductNotFound)

	err = s.store.AddProduct(ctx, p)
	c.Assert(err, qt.Equals, nil)

	p1 := store.Product{
		ID:   "someProduct",
		Name: "other product name",
		Resources: []store.ProductResource{{
			ResourceType: "fips-updates",
			Data:         `{"AptURL": "https://0.1.2.3/fips-update"}`,
		}, {
			ResourceType: "esm",
			Data:         `{"AptURL": "https://0.1.2.3/esm"}`,
		}},
	}
	err = s.store.UpdateProduct(ctx, p1)
	c.Assert(err, qt.Equals, nil)

	p2, err := s.store.Product(ctx, "someProduct")
	c.Assert(p2, qt.DeepEquals, p1)
}

func (s *StoreSuite) TestConfigureContractNotFound(c *qt.C) {
	ctx := s.ctx
	contractID, err := token.NewContractID()
	c.Assert(err, qt.Equals, nil)
	err = s.store.ConfigureContract(ctx, contractID, []store.ProductResource{{
		ResourceType: "fips-updates",
		Data:         `{"AptURL": "https://0.1.2.3/fips-update"}`,
	}})
	c.Assert(errgo.Cause(err), qt.Equals, store.ErrContractNotFound, qt.Commentf("%#v", err))
}

func (s *StoreSuite) TestConfigureContract(c *qt.C) {
	ctx := s.ctx

	accID := token.MustNewAccountID()
	err := s.store.AddAccount(ctx, store.Account{
		ID:   accID,
		Name: "test account",
	}, "userid")
	c.Assert(err, qt.Equals, nil)
	s.addTestProduct(ctx, c, s.store, "productID")
	contractID, err := s.store.AddContract(ctx, store.Contract{
		Name:       "contract-1",
		AccountID:  accID,
		ProductIDs: []string{"productID"},
	})
	c.Assert(err, qt.Equals, nil)

	err = s.store.ConfigureContract(ctx, contractID, []store.ProductResource{{
		ResourceType: "fips",
		Data:         `{"AptURL": "https://0.1.2.3/onsite-fips"}`,
	}, {
		ResourceType: "esm",
		Data:         `{"AptURL": "https://0.1.2.3/onsite-esm"}`,
	}})
	c.Assert(err, qt.Equals, nil)

	contract, err := s.store.Contract(ctx, contractID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(contract, qt.DeepEquals, store.FullContract{
		Account: store.Account{
			ID:   accID,
			Name: "test account",
		},
		Contract: store.Contract{
			ID:         contractID,
			Name:       "contract-1",
			AccountID:  accID,
			ProductIDs: []string{"productID"},
		},
		Config: []store.ProductResource{{
			ResourceType: "fips",
			Data:         `{"AptURL": "https://0.1.2.3/onsite-fips"}`,
		}, {
			ResourceType: "esm",
			Data:         `{"AptURL": "https://0.1.2.3/onsite-esm"}`,
		}},
	})
}

func (s *StoreSuite) TestConfigureContractWithDuplicateResource(c *qt.C) {
	ctx := s.ctx

	accID := token.MustNewAccountID()
	err := s.store.AddAccount(ctx, store.Account{
		ID:   accID,
		Name: "test account",
	}, "userid")
	c.Assert(err, qt.Equals, nil)
	s.addTestProduct(ctx, c, s.store, "productID")
	contractID, err := s.store.AddContract(ctx, store.Contract{
		Name:       "contract-1",
		AccountID:  accID,
		ProductIDs: []string{"productID"},
	})
	c.Assert(err, qt.Equals, nil)

	err = s.store.ConfigureContract(ctx, contractID, []store.ProductResource{{
		ResourceType: "fips",
		Data:         `{"AptURL": "https://0.1.2.3/onsite-fips"}`,
	}, {
		ResourceType: "fips",
		Data:         `{"AptURL": "https://0.1.2.3/onsite-esm"}`,
	}})
	c.Assert(errgo.Cause(err), qt.Equals, store.ErrProductConflict)
}

func (s *StoreSuite) TestAddContractDuplicateNoEffectiveDates(c *qt.C) {
	ctx := s.ctx
	accID := token.MustNewAccountID()
	err := s.store.AddAccount(ctx, store.Account{
		ID:   accID,
		Name: "test account",
	}, "userid")
	c.Assert(err, qt.Equals, nil)
	s.addTestProduct(ctx, c, s.store, "productID")
	_, err = s.store.AddContract(ctx, store.Contract{
		Name:       "contract-1",
		AccountID:  accID,
		ProductIDs: []string{"productID"},
	})
	c.Assert(err, qt.Equals, nil)
	// Try to add another contract with the same account,
	// product and effective dates.
	_, err = s.store.AddContract(ctx, store.Contract{
		Name:       "contract-2",
		AccountID:  accID,
		ProductIDs: []string{"productID"},
	})
	c.Assert(errgo.Cause(err), qt.Equals, store.ErrContractExists)
}

func (s *StoreSuite) TestAddContractBadProduct(c *qt.C) {
	ctx := s.ctx
	accID := token.MustNewAccountID()
	err := s.store.AddAccount(ctx, store.Account{
		ID:   accID,
		Name: "test account",
	}, "userid")
	c.Assert(err, qt.Equals, nil)
	s.addTestProduct(ctx, c, s.store, "productID")
	_, err = s.store.AddContract(ctx, store.Contract{
		Name:       "contract-1",
		AccountID:  accID,
		ProductIDs: []string{"productID", "badProductID"},
	})
	c.Assert(errgo.Cause(err), qt.Not(qt.Equals), nil)
}

func (s *StoreSuite) TestAddContractMultiple(c *qt.C) {
	ctx := s.ctx
	ids := make(map[token.ContractID]int)
	accIDs := make(map[token.ContractID]token.AccountID)
	for i := 0; i < 5; i++ {
		accID := token.MustNewAccountID()
		err := s.store.AddAccount(ctx, store.Account{
			ID:   accID,
			Name: "test account",
		}, "userid")
		c.Assert(err, qt.Equals, nil)
		prodID := fmt.Sprintf("product%d", i)
		s.addTestProduct(ctx, c, s.store, prodID)
		id, err := s.store.AddContract(ctx, store.Contract{
			Name:       fmt.Sprintf("contract %d", i),
			AccountID:  accID,
			ProductIDs: []string{prodID},
		})
		c.Assert(err, qt.Equals, nil)
		_, ok := ids[id]
		c.Assert(ok, qt.Equals, false)
		ids[id] = i
		accIDs[id] = accID
	}
	// Check we can retrieve them all.
	for id, index := range ids {
		ct, err := s.store.Contract(ctx, id)
		c.Assert(err, qt.Equals, nil, qt.Commentf("contract id %v", id))

		c.Assert(ct, deepEquals, store.FullContract{
			Account: store.Account{
				ID:   accIDs[id],
				Name: "test account",
			},
			Contract: store.Contract{
				ID:         id,
				Name:       fmt.Sprintf("contract %d", index),
				AccountID:  accIDs[id],
				ProductIDs: []string{fmt.Sprintf("product%d", index)},
			},
		})
	}
}

func (s *StoreSuite) TestContractNotFound(c *qt.C) {
	ctx := s.ctx
	_, err := s.store.Contract(ctx, token.MustNewContractID())
	c.Assert(err, qt.ErrorMatches, `contract not found in store`)
	c.Assert(errgo.Cause(err), qt.Equals, store.ErrContractNotFound)
}

func (s *StoreSuite) TestEnsureMachine(c *qt.C) {
	ctx := s.ctx

	err := s.store.EnsureMachine(ctx, store.MachineInfo{
		MachineID: "foo",
		OS: params.OS{
			Distribution: "ubuntu",
			Kernel:       "1.0.0",
			Release:      "14.04",
			Series:       "trusty",
			Type:         "Linux",
		},
	})
	c.Assert(err, qt.Equals, nil)

	mi, err := s.store.MachineInfo(ctx, "foo")
	c.Assert(err, qt.Equals, nil)
	c.Assert(mi, deepEquals, store.MachineInfo{
		MachineID: "foo",
		OS: params.OS{
			Distribution: "ubuntu",
			Kernel:       "1.0.0",
			Release:      "14.04",
			Series:       "trusty",
			Type:         "Linux",
		},
	})

	// Check that we can update some machine details.
	err = s.store.EnsureMachine(ctx, store.MachineInfo{
		MachineID: "foo",
		OS: params.OS{
			Distribution: "ubuntu",
			Kernel:       "2.0.0",
			Release:      "14.04",
			Series:       "xenial",
			Type:         "Linux",
		},
	})
	c.Assert(err, qt.Equals, nil)
	mi, err = s.store.MachineInfo(ctx, "foo")
	c.Assert(err, qt.Equals, nil)
	c.Assert(mi, deepEquals, store.MachineInfo{
		MachineID: "foo",
		OS: params.OS{
			Distribution: "ubuntu",
			Kernel:       "2.0.0",
			Release:      "14.04",
			Series:       "xenial",
			Type:         "Linux",
		},
	})
}

func (s *StoreSuite) TestMachineInfo(c *qt.C) {
	ctx := s.ctx

	err := s.store.EnsureMachine(ctx, store.MachineInfo{
		MachineID: "foo",
	})
	c.Assert(err, qt.Equals, nil)

	info, err := s.store.MachineInfo(ctx, "foo")
	c.Assert(err, qt.Equals, nil)
	c.Assert(info, deepEquals, store.MachineInfo{
		MachineID: "foo",
	})
}

func (s *StoreSuite) TestAddContractMachine(c *qt.C) {
	ctx := s.ctx
	accountID := token.MustNewAccountID()

	err := s.store.AddAccount(ctx, store.Account{
		ID:      accountID,
		Name:    "test account",
		Created: time.Now(),
	}, "AAAAAA")
	c.Assert(err, qt.Equals, nil)
	s.addTestProduct(ctx, c, s.store, "p1")
	ctID, err := s.store.AddContract(ctx, store.Contract{
		AccountID:  accountID,
		Name:       "blah",
		ProductIDs: []string{"p1"},
	})
	c.Assert(err, qt.Equals, nil)

	s.addTestProduct(ctx, c, s.store, "p2")
	ctID2, err := s.store.AddContract(ctx, store.Contract{
		AccountID:   accountID,
		Name:        "blah2",
		ProductIDs:  []string{"p2"},
		EffectiveTo: newTime(time.Now().Add(time.Hour)),
	})
	c.Assert(err, qt.Equals, nil)

	err = s.store.EnsureMachine(ctx, store.MachineInfo{
		MachineID: "foo",
	})
	c.Assert(err, qt.Equals, nil)

	err = s.store.AddContractMachine(ctx, ctID, "foo")
	c.Assert(err, qt.Equals, nil)

	b, err := s.store.MachineHasContract(ctx, ctID, "foo")
	c.Assert(err, qt.Equals, nil)
	c.Assert(b, qt.Equals, true)

	b, err = s.store.MachineHasContract(ctx, ctID2, "foo")
	c.Assert(err, qt.Equals, nil)
	c.Assert(b, qt.Equals, false)

	err = s.store.AddContractMachine(ctx, ctID, "foo")
	c.Assert(err, qt.Equals, nil)

	err = s.store.AddContractMachine(ctx, ctID2, "foo")
	c.Assert(err, qt.Equals, nil)

	b, err = s.store.MachineHasContract(ctx, ctID, "foo")
	c.Assert(err, qt.Equals, nil)
	c.Assert(b, qt.Equals, true)

	b, err = s.store.MachineHasContract(ctx, ctID2, "foo")
	c.Assert(err, qt.Equals, nil)
	c.Assert(b, qt.Equals, true)
}

func (s *StoreSuite) TestRemoveContractMachine(c *qt.C) {
	ctx := s.ctx
	accountID := token.MustNewAccountID()
	err := s.store.AddAccount(ctx, store.Account{
		ID:      accountID,
		Name:    "test account",
		Created: time.Now(),
	}, "AAAAAA")
	c.Assert(err, qt.Equals, nil)
	s.addTestProduct(ctx, c, s.store, "p1")
	ctID, err := s.store.AddContract(ctx, store.Contract{
		AccountID:  accountID,
		Name:       "blah",
		ProductIDs: []string{"p1"},
	})
	c.Assert(err, qt.Equals, nil)

	s.addTestProduct(ctx, c, s.store, "p2")
	ctID2, err := s.store.AddContract(ctx, store.Contract{
		AccountID:   accountID,
		Name:        "blah2",
		ProductIDs:  []string{"p2"},
		EffectiveTo: newTime(time.Now().Add(time.Hour)),
	})
	c.Assert(err, qt.Equals, nil)

	err = s.store.EnsureMachine(ctx, store.MachineInfo{
		MachineID: "foo",
	})
	c.Assert(err, qt.Equals, nil)

	err = s.store.AddContractMachine(ctx, ctID, "foo")
	c.Assert(err, qt.Equals, nil)

	b, err := s.store.MachineHasContract(ctx, ctID, "foo")
	c.Assert(err, qt.Equals, nil)
	c.Assert(b, qt.Equals, true)

	b, err = s.store.MachineHasContract(ctx, ctID2, "foo")
	c.Assert(err, qt.Equals, nil)
	c.Assert(b, qt.Equals, false)

	err = s.store.AddContractMachine(ctx, ctID, "foo")
	c.Assert(err, qt.Equals, nil)

	err = s.store.AddContractMachine(ctx, ctID2, "foo")
	c.Assert(err, qt.Equals, nil)

	b, err = s.store.MachineHasContract(ctx, ctID, "foo")
	c.Assert(err, qt.Equals, nil)
	c.Assert(b, qt.Equals, true)

	b, err = s.store.MachineHasContract(ctx, ctID2, "foo")
	c.Assert(err, qt.Equals, nil)
	c.Assert(b, qt.Equals, true)

	err = s.store.RemoveContractMachine(ctx, ctID, "foo")
	c.Assert(err, qt.Equals, nil)

	b, err = s.store.MachineHasContract(ctx, ctID, "foo")
	c.Assert(err, qt.Equals, nil)
	c.Assert(b, qt.Equals, false)

	b, err = s.store.MachineHasContract(ctx, ctID2, "foo")
	c.Assert(err, qt.Equals, nil)
	c.Assert(b, qt.Equals, true)
}

func (s *StoreSuite) TestMachineInfoNotFound(c *qt.C) {
	ctx := s.ctx
	_, err := s.store.MachineInfo(ctx, "foo")
	c.Assert(errgo.Cause(err), qt.Equals, store.ErrMachineNotFound)
}

func (s *StoreSuite) TestContractTokens(c *qt.C) {
	ctx := s.ctx
	t0 := time.Now()
	accountID := token.MustNewAccountID()
	err := s.store.AddAccount(ctx, store.Account{
		ID:      accountID,
		Name:    "test account",
		Created: time.Now(),
	}, "AAAAAA")
	c.Assert(err, qt.Equals, nil)
	s.addTestProduct(ctx, c, s.store, "p1")
	cid0, err := s.store.AddContract(ctx, store.Contract{
		AccountID:  accountID,
		Name:       "blah",
		ProductIDs: []string{"p1"},
	})
	c.Assert(err, qt.Equals, nil)
	ct0 := token.MustNewContractToken()
	err = s.store.AddContractToken(ctx, store.ContractTokenInfo{
		ContractToken: ct0,
		ContractID:    cid0,
		Expires:       t0,
	})
	c.Assert(err, qt.Equals, nil)

	t1 := t0.Add(5 * time.Minute)
	ct1 := token.MustNewContractToken()
	s.addTestProduct(ctx, c, s.store, "p2")
	cid1, err := s.store.AddContract(ctx, store.Contract{
		AccountID:   accountID,
		Name:        "blah",
		ProductIDs:  []string{"p2"},
		EffectiveTo: newTime(time.Now().Add(time.Hour)),
	})
	c.Assert(err, qt.Equals, nil)
	err = s.store.AddContractToken(ctx, store.ContractTokenInfo{
		ContractToken: ct1,
		ContractID:    cid1,
		Expires:       t1,
	})
	c.Assert(err, qt.Equals, nil)

	// Adding an existing token panics.

	c.Assert(func() {
		s.store.AddContractToken(ctx, store.ContractTokenInfo{
			ContractToken: ct1,
		})
	}, qt.PanicMatches, `duplicated contract token.*`)

	info0, err := s.store.ContractToken(ctx, ct0)
	c.Assert(err, qt.Equals, nil)
	c.Assert(info0, deepEquals, store.ContractTokenInfo{
		ContractToken: ct0,
		ContractID:    cid0,
		Expires:       t0,
	})

	tokens0, err := s.store.ContractTokens(ctx, cid0)
	c.Assert(err, qt.Equals, nil)
	c.Assert(tokens0, deepEquals, []store.ContractTokenInfo{{
		ContractToken: ct0,
		ContractID:    cid0,
		Expires:       t0,
	}})

	info1, err := s.store.ContractToken(ctx, ct1)
	c.Assert(err, qt.Equals, nil)
	c.Assert(info1, deepEquals, store.ContractTokenInfo{
		ContractToken: ct1,
		ContractID:    cid1,
		Expires:       t1,
	})

	_, err = s.store.ContractToken(ctx, token.MustNewContractToken())
	c.Assert(err, qt.ErrorMatches, `no token found`)
	c.Assert(errgo.Cause(err), qt.Equals, store.ErrTokenNotFound)
}

func (s *StoreSuite) TestEnsureUser(c *qt.C) {
	ctx := s.ctx

	u1 := store.User{
		ID:          "AAAAAA",
		Username:    "test-user",
		DisplayName: "Test User",
		Email:       "test@example.com",
		Verified:    true,
	}
	err := s.store.EnsureUser(ctx, u1)
	c.Assert(err, qt.Equals, nil)

	// Check we can update an existing user.
	u1.DisplayName = "Test User 1"
	err = s.store.EnsureUser(ctx, u1)
	c.Assert(err, qt.Equals, nil)

	// Check we can insert another user with the same
	// email address (email addresses do not necessarily
	// have to be unique)
	u2 := store.User{
		ID:          "AAAAAB",
		Username:    "test-user",
		DisplayName: "Test User",
		Email:       "test@example.com",
		Verified:    true,
	}
	err = s.store.EnsureUser(ctx, u2)
	c.Assert(err, qt.Equals, nil)
}

func (s *StoreSuite) TestUser(c *qt.C) {
	ctx := s.ctx

	u, err := s.store.User(ctx, "AAAAAA")
	c.Assert(errgo.Cause(err), qt.Equals, store.ErrUserNotFound)
	c.Assert(u, deepEquals, store.User{})

	u1 := store.User{
		ID:          "AAAAAA",
		Username:    "test-user",
		DisplayName: "Test User",
		Email:       "test@example.com",
		Verified:    true,
	}
	err = s.store.EnsureUser(ctx, u1)
	c.Assert(err, qt.Equals, nil)

	u, err = s.store.User(ctx, "AAAAAA")
	c.Assert(err, qt.Equals, nil)
	c.Assert(u, deepEquals, u1)
}

func (s *StoreSuite) TestAccount(c *qt.C) {
	ctx := s.ctx

	id := token.MustNewAccountID()
	_, err := s.store.Account(ctx, id)
	c.Assert(errgo.Cause(err), qt.Equals, store.ErrAccountNotFound)

	a := store.Account{
		ID:      id,
		Name:    "test account",
		Created: time.Now(),
	}
	err = s.store.AddAccount(ctx, a, "AAAAAA")
	c.Assert(err, qt.Equals, nil)

	a2, err := s.store.Account(ctx, id)
	c.Assert(err, qt.Equals, nil)
	c.Assert(a2, deepEquals, a)
}

func (s *StoreSuite) TestAccountForSalesforceID(c *qt.C) {
	ctx := s.ctx

	sfid := "sfacc1"
	_, err := s.store.AccountForSalesforceID(ctx, sfid)
	c.Assert(errgo.Cause(err), qt.Equals, store.ErrAccountNotFound)

	a := store.Account{
		ID:                  token.MustNewAccountID(),
		Name:                "test account",
		Created:             time.Now(),
		SalesforceAccountID: sfid,
	}
	err = s.store.AddAccount(ctx, a, "AAAAAA")
	c.Assert(err, qt.Equals, nil)

	a2, err := s.store.AccountForSalesforceID(ctx, sfid)
	c.Assert(err, qt.Equals, nil)
	c.Assert(a2, deepEquals, a)
}

func (s *StoreSuite) TestAddAccount(c *qt.C) {
	ctx := s.ctx

	a := store.Account{
		ID:      token.MustNewAccountID(),
		Name:    "test account",
		Created: time.Now(),
	}
	err := s.store.AddAccount(ctx, a, "AAAAAA")
	c.Assert(err, qt.Equals, nil)

	// Attempting to create an account a second time fails with
	// store.ErrAccountExists
	err = s.store.AddAccount(ctx, a, "AAAAAA")
	c.Assert(errgo.Cause(err), qt.Equals, store.ErrAccountExists)

	// Creating a different account works though.
	a2 := store.Account{
		ID:      token.MustNewAccountID(),
		Name:    "test account 2",
		Created: time.Now(),
	}
	err = s.store.AddAccount(ctx, a2, "AAAAAA")
	c.Assert(err, qt.Equals, nil)
}

func (s *StoreSuite) TestAddAccountDuplicateSalesforceAccountID(c *qt.C) {
	ctx := s.ctx

	a := store.Account{
		ID:                  token.MustNewAccountID(),
		Name:                "test account",
		Created:             time.Now(),
		SalesforceAccountID: "sfacc1",
	}
	err := s.store.AddAccount(ctx, a, "AAAAAA")
	c.Assert(err, qt.Equals, nil)

	// Attempting to create another account with the same salesforce ID fails with
	// store.ErrDuplicateSalesforceAccount.
	a2 := store.Account{
		ID:                  token.MustNewAccountID(),
		Name:                "test account 2",
		Created:             time.Now(),
		SalesforceAccountID: a.SalesforceAccountID,
	}
	err = s.store.AddAccount(ctx, a2, "AAAAAA")
	c.Assert(errgo.Cause(err), qt.Equals, store.ErrDuplicateSalesforceAccount)
}

func (s *StoreSuite) TestAddAccountOriginID(c *qt.C) {
	ctx := s.ctx

	a := store.Account{
		ID:       token.MustNewAccountID(),
		Name:     "test account",
		Created:  time.Now(),
		Origin:   store.AwsOrigin,
		OriginID: "oacc1",
	}
	err := s.store.AddAccount(ctx, a, "")
	c.Assert(err, qt.Equals, nil)

	a2, err := s.store.Account(ctx, a.ID)
	c.Assert(err, qt.Equals, nil)

	c.Assert(a2, deepEquals, a)
}

func (s *StoreSuite) TestAddAccountDuplicateOriginID(c *qt.C) {
	ctx := s.ctx

	a := store.Account{
		ID:       token.MustNewAccountID(),
		Name:     "test account",
		Created:  time.Now(),
		Origin:   store.AwsOrigin,
		OriginID: "oacc1",
	}
	err := s.store.AddAccount(ctx, a, "")
	c.Assert(err, qt.Equals, nil)

	// Attempting to create another account with the same origin ID fails with
	// store.ErrDuplicateOriginAccount.
	a2 := store.Account{
		ID:       token.MustNewAccountID(),
		Name:     "test account 2",
		Created:  time.Now(),
		Origin:   store.AwsOrigin,
		OriginID: "oacc1",
	}
	err = s.store.AddAccount(ctx, a2, "")
	c.Assert(errgo.Cause(err), qt.Equals, store.ErrDuplicateOriginAccount)

	// Creating another account with the same ID but different origin works though.
	a3 := store.Account{
		ID:       token.MustNewAccountID(),
		Name:     "test account 3",
		Created:  time.Now(),
		Origin:   store.AzureOrigin,
		OriginID: "oacc1",
	}
	err = s.store.AddAccount(ctx, a3, "")
	c.Assert(err, qt.Equals, nil)
}

func (s *StoreSuite) TestSetAccountAccess(c *qt.C) {
	ctx := s.ctx

	u := store.User{
		ID:          "AAAAAA",
		Username:    "test-user",
		DisplayName: "Test User",
		Email:       "test@example.com",
		Verified:    true,
	}
	err := s.store.EnsureUser(ctx, u)
	c.Assert(err, qt.Equals, nil)

	a := store.Account{
		ID:      token.MustNewAccountID(),
		Name:    "test account",
		Created: time.Now(),
	}

	err = s.store.AddAccount(ctx, a, "AAAAAA")
	c.Assert(err, qt.Equals, nil)

	accounts, err := s.store.UserAccounts(ctx, u.ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(accounts, deepEquals, []store.AccountWithAccess{{
		Account: a,
		Access:  store.AdminAccountAccess,
	}})

	u2 := store.User{
		ID:          "BBBBBB",
		Username:    "test-user",
		DisplayName: "Test User",
		Email:       "test2@example.com",
		Verified:    true,
	}
	err = s.store.EnsureUser(ctx, u2)
	c.Assert(err, qt.Equals, nil)

	access, err := s.store.AccountAccess(ctx, a.ID, u2.ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(access, qt.Equals, store.NoAccountAccess)

	err = s.store.SetAccountAccess(ctx, a.ID, u2.ID, store.DelegatedAccountAccess)
	c.Assert(err, qt.Equals, nil)

	access, err = s.store.AccountAccess(ctx, a.ID, u2.ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(access, qt.Equals, store.DelegatedAccountAccess)

	accounts, err = s.store.UserAccounts(ctx, u2.ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(accounts, deepEquals, []store.AccountWithAccess{{
		Account: a,
		Access:  store.DelegatedAccountAccess,
	}})
}

func (s *StoreSuite) TestAccountContracts(c *qt.C) {
	ctx := s.ctx

	u := store.User{
		ID:          "AAAAAA",
		Username:    "test-user",
		DisplayName: "Test User",
		Email:       "test@example.com",
		Verified:    true,
	}
	err := s.store.EnsureUser(ctx, u)
	c.Assert(err, qt.Equals, nil)

	a := store.Account{
		ID:      token.MustNewAccountID(),
		Name:    "test account",
		Created: time.Now(),
	}

	err = s.store.AddAccount(ctx, a, "AAAAAA")
	c.Assert(err, qt.Equals, nil)

	s.addTestProduct(ctx, c, s.store, "p1")
	ct := store.Contract{
		Name:       "test-contract",
		AccountID:  a.ID,
		ProductIDs: []string{"p1"},
		Created:    time.Now(),
	}
	ct.ID, err = s.store.AddContract(ctx, ct)
	c.Assert(err, qt.Equals, nil)

	contracts, err := s.store.AccountContracts(ctx, a.ID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(contracts, deepEquals, []store.FullContract{{
		Account:  a,
		Contract: ct,
	}})

	s.addTestProduct(ctx, c, s.store, "p2")
	ct2 := store.Contract{
		Name:        "test-contract-2",
		AccountID:   a.ID,
		ProductIDs:  []string{"p2"},
		Created:     time.Now(),
		EffectiveTo: newTime(time.Now().Add(time.Hour)),
	}
	ct2.ID, err = s.store.AddContract(ctx, ct2)
	c.Assert(err, qt.Equals, nil)

	contracts, err = s.store.AccountContracts(ctx, a.ID)
	c.Assert(err, qt.Equals, nil)

	expected := []store.FullContract{{
		Account:  a,
		Contract: ct,
	}, {
		Account:  a,
		Contract: ct2,
	}}

	sort.Slice(contracts, func(i, j int) bool {
		return bytes.Compare(contracts[i].Contract.ID[:], contracts[j].Contract.ID[:]) < 0
	})
	sort.Slice(expected, func(i, j int) bool {
		return bytes.Compare(expected[i].Contract.ID[:], expected[j].Contract.ID[:]) < 0
	})
	c.Assert(contracts, deepEquals, expected)
}

func (s *StoreSuite) TestServiceAccount(c *qt.C) {
	ctx := s.ctx

	_, err := s.store.ServiceAccount(ctx, "test-account")
	c.Assert(errgo.Cause(err), qt.Equals, store.ErrServiceAccountNotFound)

	err = s.store.EnsureServiceAccount(ctx, store.ServiceAccount{
		Name:         "test-account",
		PasswordSalt: []byte("salt"),
		PasswordHash: []byte("hash"),
	})
	c.Assert(err, qt.Equals, nil)

	sa, err := s.store.ServiceAccount(ctx, "test-account")
	c.Assert(err, qt.Equals, nil)
	c.Assert(sa, qt.DeepEquals, store.ServiceAccount{
		Name:         "test-account",
		PasswordSalt: []byte("salt"),
		PasswordHash: []byte("hash"),
	})
}

func (s *StoreSuite) TestUpdateContractProducts(c *qt.C) {
	ctx := s.ctx

	accID := token.MustNewAccountID()
	err := s.store.AddAccount(ctx, store.Account{
		ID:   accID,
		Name: "test account",
	}, "userid")
	c.Assert(err, qt.Equals, nil)
	s.addTestProduct(ctx, c, s.store, "productID")
	s.addTestProduct(ctx, c, s.store, "productID2")
	contractID, err := s.store.AddContract(ctx, store.Contract{
		Name:       "contract-1",
		AccountID:  accID,
		ProductIDs: []string{"productID"},
	})
	c.Assert(err, qt.Equals, nil)
	err = s.store.UpdateContractProducts(ctx, contractID, []string{"productID", "productID2"})
	c.Assert(err, qt.Equals, nil)
	contract, err := s.store.Contract(ctx, contractID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(contract, qt.DeepEquals, store.FullContract{
		Account: store.Account{
			ID:   accID,
			Name: "test account",
		},
		Contract: store.Contract{
			ID:         contractID,
			Name:       "contract-1",
			AccountID:  accID,
			ProductIDs: []string{"productID", "productID2"},
		},
	})
}

func (s *StoreSuite) TestUpdateContractProductsWithConfiguredContract(c *qt.C) {
	ctx := s.ctx

	accID := token.MustNewAccountID()
	err := s.store.AddAccount(ctx, store.Account{
		ID:   accID,
		Name: "test account",
	}, "userid")
	c.Assert(err, qt.Equals, nil)
	s.addTestProduct(ctx, c, s.store, "productID")
	s.addTestProduct(ctx, c, s.store, "productID2")
	contractID, err := s.store.AddContract(ctx, store.Contract{
		Name:       "contract-1",
		AccountID:  accID,
		ProductIDs: []string{"productID"},
	})
	c.Assert(err, qt.Equals, nil)
	err = s.store.ConfigureContract(ctx, contractID, []store.ProductResource{{
		ResourceType: "fips",
		Data:         `{"AptURL": "https://0.1.2.3/onsite-fips"}`,
	}, {
		ResourceType: "esm",
		Data:         `{"AptURL": "https://0.1.2.3/onsite-esm"}`,
	}})
	c.Assert(err, qt.Equals, nil)
	err = s.store.UpdateContractProducts(ctx, contractID, []string{"productID", "productID2"})
	c.Assert(err, qt.Equals, nil)
	contract, err := s.store.Contract(ctx, contractID)
	c.Assert(err, qt.Equals, nil)
	c.Assert(contract, qt.DeepEquals, store.FullContract{
		Account: store.Account{
			ID:   accID,
			Name: "test account",
		},
		Contract: store.Contract{
			ID:         contractID,
			Name:       "contract-1",
			AccountID:  accID,
			ProductIDs: []string{"productID", "productID2"},
		},
		Config: []store.ProductResource{{
			ResourceType: "fips",
			Data:         `{"AptURL": "https://0.1.2.3/onsite-fips"}`,
		}, {
			ResourceType: "esm",
			Data:         `{"AptURL": "https://0.1.2.3/onsite-esm"}`,
		}},
	})
}

// We don't require the database to store times at nanosecond resolution,
// so compare them as millisecond resolution only.
var deepEquals = qt.CmpEquals(equateApproxTime(time.Millisecond))

func equateApproxTime(margin time.Duration) cmp.Option {
	return cmp.Comparer(func(t0, t1 time.Time) bool {
		if t0.IsZero() && t1.IsZero() {
			return true
		}
		if t0.IsZero() || t1.IsZero() {
			return false
		}
		diff := t0.Sub(t1)
		if diff < 0 {
			diff = -diff
		}
		return diff <= margin
	})
}

// addTestProduct adds an arbitrary product with the given ID to the given store.
func (s *StoreSuite) addTestProduct(ctx context.Context, c *qt.C, st store.Store, productID string) {
	err := s.store.AddProduct(ctx, store.Product{
		ID:   productID,
		Name: "some product name",
		Resources: []store.ProductResource{{
			ResourceType: "fips",
			Data: `{
				"series": ["xenial"],
				"architectures": ["amd64"]
			}`,
		}},
	})
	c.Assert(err, qt.Equals, nil)
}

func newTime(t time.Time) *time.Time {
	return &t
}
