// Copyright 2019 Canonical Ltd.

// Package storeutil implements operations on top of the
// store.Store interface. It's higher level than the
// store package itself, but lower level than the API
// package.
package storeutil

import (
	"context"
	"sort"
	"strings"
	"time"

	"github.com/CanonicalLtd/salesforce"
	sfparams "github.com/CanonicalLtd/salesforce/params"
	"gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/internal/awsauth"
	"github.com/CanonicalLtd/ua-contracts/internal/azureauth"
	"github.com/CanonicalLtd/ua-contracts/internal/monitoring"
	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/token"
	"github.com/CanonicalLtd/ua-contracts/params"
)

var ErrSalesforceAccountNotFound = errgo.New("salesforce account not found")

// SalesforceClient represents an instance of Salesforce.
type SalesforceClient interface {
	// GetUser gets the user information for the account with the
	// given email address.
	GetUser(ctx context.Context, email string) (*sfparams.User, error)

	// GetAssetsForAccount returns all the assets associated with
	// the account with the given accountID.
	GetAssetsForAccount(ctx context.Context, salesforceID string) ([]salesforce.Asset, error)
}

type SalesforceDisabledClient struct{}

func (SalesforceDisabledClient) GetUser(ctx context.Context, email string) (*sfparams.User, error) {
	return nil, sfparams.ErrNotFound
}

func (SalesforceDisabledClient) GetAssetsForAccount(ctx context.Context, accountID string) ([]salesforce.Asset, error) {
	return nil, nil
}

// DefaultProductID holds the ID of the product that's associated with
// an account by default by EnsureDefaultAccount.
const DefaultProductID = "free"

// GetAccounts gets the accounts for the authenticated user. If there are
// no stored accounts for this user then a default account will be
// created.
func GetAccounts(ctx context.Context, client SalesforceClient, s store.Store, user store.User) ([]store.AccountWithAccess, error) {
	accounts, err := s.UserAccounts(ctx, user.ID)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	if len(accounts) > 0 {
		return accounts, nil
	}

	sfuser, err := client.GetUser(ctx, user.Email)
	switch {
	case errgo.Cause(err) == sfparams.ErrNotFound:
		if err := ensureDefaultAccount(ctx, s, user); err != nil {
			// No entry for the user in salesforce, create a "free" account.
			return nil, errgo.Notef(err, "cannot create default account")
		}
	case err == nil:
		if err := ensureDefaultAccount(ctx, s, user); err != nil {
			return nil, errgo.Notef(err, "cannot create default account")
		}
		if err := ensureSalesforceAccount(ctx, s, sfuser, user); err != nil {
			return nil, errgo.Notef(err, "cannot create default account")
		}
	default:
		monitoring.SalesforceErrorCount.Inc()
		return nil, errgo.Mask(err)
	}
	accs, err := s.UserAccounts(ctx, user.ID)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	sort.SliceStable(accs, func(i, j int) bool {
		return accs[i].Name < accs[j].Name
	})
	return accs, nil
}

// ensureDefaultAccount creates a default account for the given user and
// makes the user the admin of the account. The account ID will be
// predictable based on the user ID so there can only be one default
// account for a given user.
func ensureDefaultAccount(ctx context.Context, s store.Store, user store.User) error {
	account := store.Account{
		ID:      token.AccountIDForUserID(user.ID),
		Name:    user.Email,
		Created: time.Now().UTC().Round(time.Second),
	}
	err := s.AddAccount(ctx, account, user.ID)
	if err != nil && errgo.Cause(err) != store.ErrAccountExists {
		return errgo.Mask(err)
	}

	return nil
}

// ensureSalesforceAccount creates (if necessary) an account for the
// salesforce account specified for the given salesforce user. The given
// user is then made an admin of that account.
func ensureSalesforceAccount(ctx context.Context, s store.Store, sfuser *sfparams.User, user store.User) error {
	accountID, err := token.NewAccountID()
	if err != nil {
		return errgo.Mask(err)
	}
	account := store.Account{
		ID:                  accountID,
		Name:                sfuser.Account.Name,
		Created:             time.Now().UTC().Round(time.Second),
		SalesforceAccountID: sfuser.Account.ID,
	}
	err = s.AddAccount(ctx, account, user.ID)
	if errgo.Cause(err) == store.ErrDuplicateSalesforceAccount {
		account, err = s.AccountForSalesforceID(ctx, sfuser.Account.ID)
		if err != nil {
			return errgo.Mask(err)
		}
	}
	if err != nil {
		return errgo.Mask(err)
	}
	return errgo.Mask(s.SetAccountAccess(ctx, account.ID, user.ID, store.AdminAccountAccess))
}

// GetContractsForAccount rerieves the contracts for given account ID. If
// there are no stored contracts for the account then a contract is
// generated. If the account has a SalesforceAccountID then salesforce
// will be queried to determine the required contract, otherwise a
// default "free" contract will be created.
func GetContractsForAccount(ctx context.Context, client SalesforceClient, s store.Store, id token.AccountID) ([]store.FullContract, error) {
	contracts, err := s.AccountContracts(ctx, id)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	if len(contracts) > 0 {
		return contracts, nil
	}
	// We need to create a default contract for this account.
	a, err := s.Account(ctx, id)
	if err != nil {
		return nil, errgo.Mask(err, errgo.Is(store.ErrAccountNotFound))
	}
	if a.SalesforceAccountID != "" {
		if err := ensureSalesforceContract(ctx, client, s, a); err != nil {
			return nil, errgo.Notef(err, "cannot ensure contract")
		}
	} else {
		if err := ensureDefaultContract(ctx, s, a); err != nil {
			return nil, errgo.Notef(err, "cannot ensure contract")
		}
	}
	return s.AccountContracts(ctx, id)
}

// ensureDefaultContract creates the default "free" contract.
func ensureDefaultContract(ctx context.Context, s store.Store, a store.Account) error {
	_, err := s.AddContract(ctx, store.Contract{
		Name:          a.Name,
		AccountID:     a.ID,
		ProductIDs:    []string{DefaultProductID},
		Created:       time.Now(),
		EffectiveFrom: newTime(time.Now()),
		Origin:        store.FreeOrigin,
	})
	if err != nil && errgo.Cause(err) != store.ErrContractExists {
		return errgo.Mask(err)
	}
	return nil
}

func ensureSalesforceContract(ctx context.Context, client SalesforceClient, s store.Store, a store.Account) error {
	assets, err := client.GetAssetsForAccount(ctx, a.SalesforceAccountID)
	if err != nil && errgo.Cause(err) != params.ErrNotFound {
		monitoring.SalesforceErrorCount.Inc()
		return errgo.Mask(err)
	}
	productIDs, endDate, supportLevel := getProductIDs(assets)
	if len(productIDs) == 0 {
		return nil
	}
	name := "Ubuntu Advantage"
	if supportLevel != "" {
		name += " - " + supportLevel
	}

	_, err = s.AddContract(ctx, store.Contract{
		Name:          name,
		AccountID:     a.ID,
		ProductIDs:    productIDs,
		Created:       time.Now(),
		EffectiveTo:   newTime(endDate),
		EffectiveFrom: newTime(time.Now()),
		Origin:        store.SalesforceOrigin,
	})
	if err != nil && errgo.Cause(err) != store.ErrContractExists {
		return errgo.Mask(err)
	}
	return nil
}

// UpdateContractsForSalesforceAccount updates the contract associated with the given salesforce user.
// Will return a ErrSalesforceAccountNotFound if the salesforce account is not found.
func UpdateContractsForSalesforceAccount(ctx context.Context, client SalesforceClient, s store.Store, sfID string) ([]store.FullContract, error) {
	assets, err := client.GetAssetsForAccount(ctx, sfID)
	if err != nil {
		if errgo.Cause(err) == sfparams.ErrNotFound {
			return nil, errgo.WithCausef(nil, ErrSalesforceAccountNotFound, "")
		}
		monitoring.SalesforceErrorCount.Inc()
		return nil, errgo.Mask(err)
	}
	productIDs, endDate, supportLevel := getProductIDs(assets)

	account, err := s.AccountForSalesforceID(ctx, sfID)
	if err != nil {
		if errgo.Cause(err) == store.ErrAccountNotFound {
			/// Nothing to do if the account doesn't exist.
			return nil, nil
		}
		return nil, errgo.Mask(err)
	}
	existingContracts, err := s.AccountContracts(ctx, account.ID)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	var c *store.FullContract
	for _, contract := range existingContracts {
		if contract.Contract.Origin == store.SalesforceOrigin {
			c = &contract
		}
	}
	if c == nil {
		if len(productIDs) != 0 {
			name := "Ubuntu Advantage"
			if supportLevel != "" {
				name += "- " + supportLevel
			}
			if _, err := s.AddContract(ctx, store.Contract{
				Name:          account.Name,
				AccountID:     account.ID,
				ProductIDs:    productIDs,
				Created:       time.Now(),
				EffectiveTo:   newTime(endDate),
				EffectiveFrom: newTime(time.Now()),
				Origin:        store.SalesforceOrigin,
			}); err != nil {
				return nil, errgo.Mask(err)
			}
		}
	} else {
		if stringsEqual(c.Contract.ProductIDs, productIDs) {
			return existingContracts, nil
		}
		if err := s.UpdateContractProducts(ctx, c.Contract.ID, productIDs); err != nil {
			return nil, errgo.Mask(err)
		}
	}
	contracts, err := s.AccountContracts(ctx, account.ID)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return contracts, nil
}

// UpdateContractsByEmail updates the contract associated with the given salesforce user.
// Will return a ErrSalesforceAccountNotFound if the salesforce account is not found.
func UpdateContractsByEmail(ctx context.Context, client SalesforceClient, s store.Store, email string) ([]store.FullContract, error) {
	user, err := s.FindUserByMail(ctx, email)
	if err != nil {
		return nil, errgo.Mask(err, errgo.Is(store.ErrUserNotFound))
	}
	accs, err := GetAccounts(ctx, client, s, user)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	for _, acc := range accs {
		if acc.SalesforceAccountID != "" {
			contracts, err := UpdateContractsForSalesforceAccount(ctx, client, s, acc.SalesforceAccountID)
			if err != nil {
				return nil, errgo.Mask(err)
			}
			return contracts, nil
		}
	}
	return nil, errgo.WithCausef(nil, ErrSalesforceAccountNotFound, "")
}

func stringsEqual(a []string, b []string) bool {
	if len(a) != len(b) {
		return false
	}
	for i, v := range a {
		if v != b[i] {
			return false
		}
	}
	return true
}

func newTime(t time.Time) *time.Time {
	if t.IsZero() {
		return nil
	}
	return &t
}

func getProductIDs(assets []salesforce.Asset) ([]string, time.Time, string) {
	var products []string
	var endDate time.Time
	var supportLevel string
	for _, asset := range assets {
		if time.Now().After(asset.UsageEndDate.T) || asset.Status != "Active" {
			continue
		}
		// endDate will be the minimum end date of any valid assets found.
		if endDate.IsZero() || asset.UsageEndDate.T.Before(endDate) {
			endDate = asset.UsageEndDate.T
		}
		if asset.Product.ESM.B {
			products = append(products, "esm")
		}
		if asset.Product.Livepatch.B {
			// Find a way to get the kernel support for ibm or acquia.
			products = append(products, "livepatch")
		}
		if asset.Product.FIPS() {
			products = append(products, "fips")
		}
		if level := asset.Product.SupportLevel(); level != "" {
			products = append(products, "support-"+level)
			supportLevel = strings.Title(level)
		}
	}
	return products, endDate, supportLevel
}

// GetAWSAccount retrieves an account for the given AWS instance information
// document, creating an account if necessary.
func GetAWSAccount(ctx context.Context, s store.Store, doc awsauth.InstanceInformationDocument) (store.Account, error) {
	aid := token.AccountIDForAWSAccountID(doc.AccountId)
	acc, err := getAccount(ctx, s, aid, store.Account{
		ID:       aid,
		Name:     doc.AccountId,
		Created:  time.Now(),
		Origin:   store.AwsOrigin,
		OriginID: doc.AccountId,
	})
	return acc, errgo.Mask(err)
}

// GetAzureAccount retrieves an account for the given Azure attested data,
// creating an account if necessary.
func GetAzureAccount(ctx context.Context, s store.Store, compute *params.AzureComputeInstanceMetadata) (store.Account, error) {
	aid := token.AccountIDForAzureSubscriptionID(compute.SubscriptionId)
	acc, err := getAccount(ctx, s, aid, store.Account{
		ID:       aid,
		Name:     compute.SubscriptionId,
		Created:  time.Now(),
		Origin:   store.AzureOrigin,
		OriginID: compute.SubscriptionId,
	})
	return acc, errgo.Mask(err)
}

func getAccount(ctx context.Context, s store.Store, id token.AccountID, a store.Account) (store.Account, error) {
	acc, err := s.Account(ctx, id)
	if errgo.Cause(err) != store.ErrAccountNotFound {
		return acc, errgo.Mask(err)
	}

	if err := s.AddAccount(ctx, a, ""); err != nil {
		if errgo.Cause(err) != store.ErrAccountExists {
			return store.Account{}, errgo.Mask(err)
		}
	}
	acc, err = s.Account(ctx, id)
	return acc, errgo.Mask(err)
}

// GetAWSContract retrieves a contract for the given AWS instance information
// document, creating an account and contract if necessary.
func GetAWSContract(ctx context.Context, s store.Store, doc awsauth.InstanceInformationDocument) (store.FullContract, error) {
	acc, err := GetAWSAccount(ctx, s, doc)
	if err != nil {
		return store.FullContract{}, errgo.Mask(err)
	}

	fcs, err := s.AccountContracts(ctx, acc.ID)
	if err != nil {
		return store.FullContract{}, errgo.Mask(err)
	}
	for _, fc := range fcs {
		if appropriateAWSContract(fc, doc) {
			return fc, nil
		}
	}
	c := store.Contract{
		Name:          doc.AccountId,
		AccountID:     acc.ID,
		ProductIDs:    []string{"aws"},
		Created:       time.Now(),
		CreatedBy:     doc.AccountId,
		EffectiveFrom: newTime(time.Now()),
		Origin:        store.AwsOrigin,
	}
	_, err = s.AddContract(ctx, c)
	if err != nil && errgo.Cause(err) != store.ErrContractExists {
		return store.FullContract{}, errgo.Mask(err)
	}
	fcs, err = s.AccountContracts(ctx, acc.ID)
	if err != nil {
		return store.FullContract{}, errgo.Mask(err)
	}
	for _, fc := range fcs {
		if appropriateAWSContract(fc, doc) {
			return fc, nil
		}
	}
	// If we get here something has gone wrong.
	return store.FullContract{}, errgo.New("internal error getting contract")
}

func appropriateAWSContract(fc store.FullContract, doc awsauth.InstanceInformationDocument) bool {
	// TODO use instance information to determine which contract to
	// use.

	for _, p := range fc.Contract.ProductIDs {
		if p == "aws" {
			return true
		}
	}
	return false
}

// GetAzureContract retrieves a contract for the given Azure compute instance
// metadata and attested data document, creating an account and contract if
// necessary.
func GetAzureContract(ctx context.Context, s store.Store, compute *params.AzureComputeInstanceMetadata, signed azureauth.AttestedData) (store.FullContract, error) {
	acc, err := GetAzureAccount(ctx, s, compute)
	if err != nil {
		return store.FullContract{}, errgo.Mask(err)
	}

	fcs, err := s.AccountContracts(ctx, acc.ID)
	if err != nil {
		return store.FullContract{}, errgo.Mask(err)
	}
	for _, fc := range fcs {
		if appropriateAzureContract(fc) {
			return fc, nil
		}
	}
	c := store.Contract{
		Name:          compute.SubscriptionId,
		AccountID:     acc.ID,
		ProductIDs:    []string{"azure"},
		Created:       time.Now(),
		CreatedBy:     compute.SubscriptionId,
		EffectiveFrom: newTime(time.Now()),
		Origin:        store.AzureOrigin,
	}
	_, err = s.AddContract(ctx, c)
	if err != nil && errgo.Cause(err) != store.ErrContractExists {
		return store.FullContract{}, errgo.Mask(err)
	}
	fcs, err = s.AccountContracts(ctx, acc.ID)
	if err != nil {
		return store.FullContract{}, errgo.Mask(err)
	}
	for _, fc := range fcs {
		if appropriateAzureContract(fc) {
			return fc, nil
		}
	}
	// If we get here something has gone wrong.
	return store.FullContract{}, errgo.New("internal error getting contract")
}

func appropriateAzureContract(fc store.FullContract) bool {
	// TODO use metadata or attested data to determine which contract to use?
	for _, p := range fc.Contract.ProductIDs {
		if p == "azure" {
			return true
		}
	}
	return false
}
