package token_test

import (
	"bytes"
	"encoding/base64"
	"encoding/json"
	"testing"

	"github.com/btcsuite/btcutil/base58"
	qt "github.com/frankban/quicktest"

	"github.com/CanonicalLtd/ua-contracts/internal/token"
)

func TestContractIDMarshaling(t *testing.T) {
	c := qt.New(t)

	t1 := token.MustNewContractID()
	var t2 token.ContractID

	data, err := json.Marshal(t1)
	c.Assert(err, qt.Equals, nil)
	err = json.Unmarshal(data, &t2)
	c.Assert(err, qt.Equals, nil)
	c.Assert(t2, qt.DeepEquals, t1)
}

func TestAccountIDMarshaling(t *testing.T) {
	c := qt.New(t)

	t1 := token.MustNewAccountID()
	var t2 token.AccountID

	data, err := json.Marshal(t1)
	c.Assert(err, qt.Equals, nil)
	err = json.Unmarshal(data, &t2)
	c.Assert(err, qt.Equals, nil)
	c.Assert(t2, qt.DeepEquals, t1)
}

func TestContractTokenRoundTrip(t *testing.T) {
	c := qt.New(t)

	ct := token.MustNewContractToken()
	s := ct.String()
	ct2, err := token.ParseContractToken(s)
	c.Assert(err, qt.Equals, nil)
	c.Assert(ct2, qt.Equals, ct)
}

func TestParseContractTokenError(t *testing.T) {
	c := qt.New(t)
	_, err := token.ParseContractToken("")
	c.Assert(err, qt.ErrorMatches, `invalid token`)

	_, err = token.ParseContractToken(token.MustNewAccountID().String())
	c.Assert(err, qt.ErrorMatches, `invalid token`)

	_, err = token.ParseContractToken("C+" + base58.CheckEncode([]byte{}, 0))
	c.Assert(err, qt.ErrorMatches, `invalid token: invalid format: version and/or checksum bytes missing`)

	_, err = token.ParseContractToken("C" + base58.CheckEncode([]byte{}, 1))
	c.Assert(err, qt.ErrorMatches, `invalid token`)
}

func TestContractIDRoundTrip(t *testing.T) {
	c := qt.New(t)

	cid := token.MustNewContractID()
	s := cid.String()
	cid2, err := token.ParseContractID(s)
	c.Assert(err, qt.Equals, nil)
	c.Assert(cid2, qt.Equals, cid)
}

func TestParseContractIDError(t *testing.T) {
	c := qt.New(t)
	_, err := token.ParseContractID("")
	c.Assert(err, qt.ErrorMatches, `invalid id`)

	_, err = token.ParseContractID(token.MustNewAccountID().String())
	c.Assert(err, qt.ErrorMatches, `invalid id`)

	_, err = token.ParseContractID(base64.RawURLEncoding.EncodeToString([]byte{}))
	c.Assert(err, qt.ErrorMatches, `invalid id`)
}

func TestContractIDIsZero(t *testing.T) {
	c := qt.New(t)

	var cid token.ContractID
	c.Assert(cid.IsZero(), qt.Equals, true)
	cid = token.MustNewContractID()
	c.Assert(cid.IsZero(), qt.Equals, false)
}

func TestMachineTokenBodyRoundTrip(t *testing.T) {
	c := qt.New(t)

	body := token.MachineTokenBody{
		MachineID:  "00001",
		ContractID: token.MustNewContractID(),
	}
	b, err := body.MarshalBinary()
	c.Assert(err, qt.Equals, nil)

	var body2 token.MachineTokenBody
	err = body2.UnmarshalBinary(b)
	c.Assert(err, qt.Equals, nil)
	c.Assert(body2, qt.DeepEquals, body)
}

func TestMachineTokenBodyUnmarshalErrors(t *testing.T) {
	c := qt.New(t)

	var body token.MachineTokenBody
	err := body.UnmarshalBinary([]byte{})
	c.Assert(err, qt.ErrorMatches, `invalid machine token`)

	err = body.UnmarshalBinary(bytes.Repeat([]byte{1}, 1+1+len(token.ContractID{})))
	c.Assert(err, qt.ErrorMatches, `invalid machine token`)
}

func TestMachineTokenBodyIsZero(t *testing.T) {
	c := qt.New(t)

	var body token.MachineTokenBody
	c.Assert(body.IsZero(), qt.Equals, true)
	body.MachineID = "1"
	c.Assert(body.IsZero(), qt.Equals, false)
}

func TestMachineTokenRoundTrip(t *testing.T) {
	c := qt.New(t)

	body := token.MachineTokenBody{
		MachineID:  "00001",
		ContractID: token.MustNewContractID(),
	}
	codec := token.NewMachineTokenCodec([]byte("supersecret"))
	mt, err := codec.Encode(body)
	c.Assert(err, qt.Equals, nil)
	s := mt.String()
	mt2, err := token.ParseMachineToken(s)
	c.Assert(err, qt.Equals, nil)
	body2, err := codec.Decode(mt2)
	c.Assert(err, qt.Equals, nil)
	c.Assert(body2, qt.DeepEquals, body)
}

func TestParseMachineTokenError(t *testing.T) {
	c := qt.New(t)
	_, err := token.ParseMachineToken("")
	c.Assert(err, qt.ErrorMatches, `invalid token`)

	_, err = token.ParseMachineToken(token.MustNewAccountID().String())
	c.Assert(err, qt.ErrorMatches, `invalid token`)

	_, err = token.ParseContractToken(base64.RawURLEncoding.EncodeToString([]byte{}))
	c.Assert(err, qt.ErrorMatches, `invalid token`)
}

func TestAccountIDRoundTrip(t *testing.T) {
	c := qt.New(t)

	aid := token.MustNewAccountID()
	s := aid.String()
	aid2, err := token.ParseAccountID(s)
	c.Assert(err, qt.Equals, nil)
	c.Assert(aid2, qt.Equals, aid)
}

func TestAccountIDForUserID(t *testing.T) {
	c := qt.New(t)

	aid := token.AccountIDForUserID("AAAAAA")
	aid2 := token.AccountIDForUserID("AAAAAA")
	c.Assert(aid2, qt.Equals, aid)
	c.Assert(aid.String(), qt.Equals, "aAYQ_ZZc4w8vp4fnYo6HavRb6N4N_g_HrChY-b7r_DKo")

	aid3 := token.AccountIDForUserID("AAAAAB")
	c.Assert(aid3, qt.Not(qt.Equals), aid)
}

func TestParseAccountIDError(t *testing.T) {
	c := qt.New(t)
	_, err := token.ParseAccountID("")
	c.Assert(err, qt.ErrorMatches, `invalid id`)

	_, err = token.ParseAccountID(token.MustNewContractID().String())
	c.Assert(err, qt.ErrorMatches, `invalid id`)

	_, err = token.ParseAccountID(base64.RawURLEncoding.EncodeToString([]byte{}))
	c.Assert(err, qt.ErrorMatches, `invalid id`)
}

func TestAccountIDForAWSAccount(t *testing.T) {
	c := qt.New(t)

	aid := token.AccountIDForAWSAccountID("123456789012")
	aid2 := token.AccountIDForAWSAccountID("123456789012")
	c.Assert(aid2, qt.Equals, aid)
	c.Assert(aid.String(), qt.Equals, "aAha_2hQ1PW7UoDj-1xIDx40kXg-iBBw8X62-noHGkPM")

	aid3 := token.AccountIDForAWSAccountID("123456789013")
	c.Assert(aid3, qt.Not(qt.Equals), aid)
}
