package v1

import (
	"context"
	"encoding/json"
	"reflect"
	"sort"
	"strings"
	"time"

	errgo "gopkg.in/errgo.v1"

	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/params"
)

func storeProduct(p params.Product) store.Product {
	return store.Product{
		ID:        p.Id,
		Name:      p.Name,
		Resources: storeProductResources(p.Resources),
	}
}

func storeProductResources(rs []params.ProductResource) []store.ProductResource {
	rs1 := make([]store.ProductResource, len(rs))
	for i, r := range rs {
		rs1[i] = storeProductResource(r)
	}
	return rs1
}

func storeProductResource(r params.ProductResource) store.ProductResource {
	data, err := json.Marshal(r.Data)
	if err != nil {
		// Should never happen because r has been
		// unmarshaled from JSON.
		panic(err)
	}
	return store.ProductResource{
		ResourceType: r.ResourceType,
		Data:         string(data),
	}
}

func paramsProduct(p store.Product) params.Product {
	return params.Product{
		Id:        p.ID,
		Name:      p.Name,
		Resources: paramsProductResources(p.Resources),
	}
}

func paramsProductResources(rs []store.ProductResource) []params.ProductResource {
	rs1 := make([]params.ProductResource, len(rs))
	for i, r := range rs {
		rs1[i] = paramsProductResource(r)
	}
	return rs1
}

func paramsProductResource(r store.ProductResource) params.ProductResource {
	type badJSON struct {
		InvalidJSON string `json:"invalid-json"`
	}
	var data interface{}
	if err := json.Unmarshal([]byte(r.Data), &data); err != nil {
		// This should never happen unless we've got a corrupt
		// database, but return the bad JSON anyway
		// if this happens rather than making the whole
		// API request fail with an internal error.
		data = badJSON{
			InvalidJSON: r.Data,
		}
	}
	return params.ProductResource{
		ResourceType: r.ResourceType,
		Data:         data,
	}
}

var resourceEntitlementInfo = map[string]struct {
	paramsType      reflect.Type
	makeEntitlement func(p interface{}, rtype string) (params.Entitlement, error)
}{
	"fips": {
		paramsType:      reflect.TypeOf(aptParams{}),
		makeEntitlement: makeAPTEntitlement,
	},
	"fips-updates": {
		paramsType:      reflect.TypeOf(aptParams{}),
		makeEntitlement: makeAPTEntitlement,
	},
	"esm": {
		paramsType:      reflect.TypeOf(aptParams{}),
		makeEntitlement: makeAPTEntitlement,
	},
	"cc-eal": {
		paramsType:      reflect.TypeOf(aptParams{}),
		makeEntitlement: makeAPTEntitlement,
	},
	"livepatch": {
		paramsType:      reflect.TypeOf(livepatchParams{}),
		makeEntitlement: makeLivepatchEntitlement,
	},
	"support": {
		paramsType:      reflect.TypeOf(supportParams{}),
		makeEntitlement: makeSupportEntitlement,
	},
	"uca": {
		paramsType:      reflect.TypeOf(aptParams{}),
		makeEntitlement: makeAPTEntitlement,
	},
}

// aptParams holds product parameters for an APT-based resource.
type aptParams struct {
	EnableByDefault    bool                          `json:"enableByDefault"`
	Architectures      []string                      `json:"architectures"`
	AdditionalPackages []string                      `json:"additionalPackages"`
	AptKey             string                        `json:"aptKey"`
	AptURL             string                        `json:"aptURL"`
	Suites             []string                      `json:"suites"`
	Series             map[string]aptOverwriteParams `json:"series"`
}

type aptOverwriteParams struct {
	EnableByDefault    *bool    `json:"enableByDefault"`
	Architectures      []string `json:"architectures"`
	AdditionalPackages []string `json:"additionalPackages"`
	AptKey             *string  `json:"aptKey"`
	AptURL             *string  `json:"aptURL"`
	Suites             []string `json:"suites"`
}

// supportParams holds parameters for a support resource.
type supportParams struct {
	SupportLevel string `json:"supportLevel"`
}

// livepatchParams holds parameters for a livepatch resource.
type livepatchParams struct {
	Tier             string                              `json:"tier"`
	EnableByDefault  bool                                `json:"enableByDefault"`
	Architectures    []string                            `json:"architectures"`
	KernelFlavors    []string                            `json:"kernelFlavors"`
	MinKernelVersion string                              `json:"minKernelVersion"`
	CaCerts          string                              `json:"caCerts"`
	RemoteServer     string                              `json:"remoteServer"`
	Series           map[string]livepatchOverwriteParams `json:"series"`
}

type livepatchOverwriteParams struct {
	Tier             *string  `json:"tier"`
	EnableByDefault  *bool    `json:"enableByDefault"`
	Architectures    []string `json:"architectures"`
	KernelFlavors    []string `json:"kernelFlavors"`
	MinKernelVersion *string  `json:"minKernelVersion"`
	CaCerts          *string  `json:"caCerts"`
	RemoteServer     *string  `json:"remoteServer"`
}

func accountInfo(a store.Account) params.AccountInfo {
	return params.AccountInfo{
		Id:        a.ID.String(),
		Name:      a.Name,
		CreatedAt: a.Created.Truncate(time.Second),
	}
}

// contractEntitlements returns entitlement data for the given contract,
// keyed by resource type.
// Note that it does not return entries for non-entitled entitlements.
func (h *handler) contractEntitlements(ctx context.Context, c store.FullContract) (map[string]params.Entitlement, error) {
	rs, err := h.productResources(ctx, c.Contract.ProductIDs)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	entitlements := make(map[string]params.Entitlement)
	for rtype, r := range rs {
		var rConfig *store.ProductResource
		if r.CanOverride {
			for i := range c.Config {
				r1 := &c.Config[i]
				if r1.ResourceType == rtype {
					rConfig = r1
					break
				}
			}
		}
		e, err := makeEntitlement(r, rConfig)
		if err != nil {
			return nil, errgo.Mask(err)
		}
		entitlements[r.ResourceType] = e
	}
	return entitlements, nil
}

func (h *handler) productResources(ctx context.Context, productIDs []string) (map[string]store.ProductResource, error) {
	rs := make(map[string]store.ProductResource)
	for _, productID := range productIDs {
		p, err := h.p.Store.Product(ctx, productID)
		if err != nil {
			return nil, errgo.Newf("cannot get product %s", productID)
		}
		for _, r := range p.Resources {
			if _, ok := rs[r.ResourceType]; ok {
				continue
			}
			rs[r.ResourceType] = r
		}
	}
	return rs, nil
}

func (h *handler) productForResource(ctx context.Context, productIDs []string, resourceType string) (string, error) {
	for _, productID := range productIDs {
		p, err := h.p.Store.Product(ctx, productID)
		if err != nil {
			return "", errgo.Newf("cannot get product %s", productID)
		}
		for _, r := range p.Resources {
			if r.ResourceType == resourceType {
				return productID, nil
			}
		}
	}
	return "", errgo.Newf("cannot determine product for resource %s", resourceType)
}

// contractInfo returns information on the given contract suitable for
// returning to the API client. If resourceType
func (h *handler) contractInfo(ctx context.Context, c store.FullContract) (params.ContractInfo, error) {
	entitlements, err := h.contractEntitlements(ctx, c)
	if err != nil {
		return params.ContractInfo{}, errgo.Mask(err)
	}
	entitlementSlice := make([]params.Entitlement, 0, len(store.ResourceTypes))
	// Use the ResourceTypes slice so we always get the entitlement slice
	// in a predictable order.
	for _, rt := range store.ResourceTypes {
		if e, ok := entitlements[rt]; ok {
			entitlementSlice = append(entitlementSlice, e)
		}
	}
	c1 := params.ContractInfo{
		Id:                   c.Contract.ID.String(),
		Name:                 c.Contract.Name,
		Origin:               strings.ToLower(string(c.Contract.Origin)),
		CreatedAt:            c.Contract.Created,
		CreatedBy:            c.Contract.CreatedBy,
		EffectiveFrom:        c.Contract.EffectiveFrom,
		EffectiveTo:          c.Contract.EffectiveTo,
		ResourceEntitlements: entitlementSlice,
	}

	return c1, nil
}

// makeEntitlement returns entitlement data derived from the given product resource
// parameters.
func makeEntitlement(r store.ProductResource, rConfig *store.ProductResource) (params.Entitlement, error) {
	info, ok := resourceEntitlementInfo[r.ResourceType]
	if !ok {
		return nil, errgo.Newf("unknown resource type %q", r.ResourceType)
	}
	p := reflect.New(info.paramsType).Interface()
	if err := json.Unmarshal([]byte(r.Data), p); err != nil {
		return nil, errgo.Notef(err, "cannot unmarshal resource data %q into %T", r.Data, p)
	}
	if rConfig != nil {
		// We've got some per-contract configuration data. It overrides
		// anything in the main parameters.
		if err := json.Unmarshal([]byte(rConfig.Data), p); err != nil {
			return nil, errgo.Notef(err, "cannot unmarshal resource config data %q into %T", r.Data, p)
		}
	}

	e, err := info.makeEntitlement(p, r.ResourceType)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return e, nil
}

func makeAPTEntitlement(p0 interface{}, rt string) (params.Entitlement, error) {
	p := p0.(*aptParams)
	ent := params.EntitlementAptRepository{
		Type:     rt,
		Entitled: true,
		Affordances: params.AffordancesAptRepository{
			Architectures: p.Architectures,
		},
		Directives: params.DirectivesAptRepository{
			AdditionalPackages: p.AdditionalPackages,
			AptURL:             p.AptURL,
			AptKey:             p.AptKey,
			Suites:             p.Suites,
		},
		Obligations: params.Obligations{
			EnableByDefault: p.EnableByDefault,
		},
	}
	if len(p.Series) > 0 {
		series := make(map[string]map[string]interface{}, len(p.Series))
		for k, v := range p.Series {
			ent.Affordances.Series = append(ent.Affordances.Series, k)
			affordances := make(map[string]interface{})
			if len(v.Architectures) > 0 {
				affordances["architectures"] = v.Architectures
			}
			directives := make(map[string]interface{})
			if len(v.AdditionalPackages) > 0 {
				directives["additionalPackages"] = v.AdditionalPackages
			}
			if v.AptURL != nil {
				directives["aptURL"] = *v.AptURL
			}
			if v.AptKey != nil {
				directives["aptKey"] = *v.AptKey
			}
			if len(v.Suites) > 0 {
				directives["suites"] = v.Suites
			}
			obligations := make(map[string]interface{})
			if v.EnableByDefault != nil {
				obligations["enableByDefault"] = v.EnableByDefault
			}
			series[k] = make(map[string]interface{})
			if len(affordances) > 0 {
				series[k]["affordances"] = affordances
			}
			if len(directives) > 0 {
				series[k]["directives"] = directives
			}
			if len(obligations) > 0 {
				series[k]["obligations"] = obligations
			}
		}
		ent.Series = series
	}
	sort.Strings(ent.Affordances.Series)
	return ent, nil
}

func makeLivepatchEntitlement(p0 interface{}, rt string) (params.Entitlement, error) {
	p := p0.(*livepatchParams)
	ent := params.EntitlementLivepatch{
		Type:     rt,
		Entitled: true,
		Affordances: params.AffordancesLivepatch{
			Architectures:    p.Architectures,
			KernelFlavors:    p.KernelFlavors,
			MinKernelVersion: p.MinKernelVersion,
			Tier:             p.Tier,
		},
		Directives: params.DirectivesLivepatch{
			CaCerts:      p.CaCerts,
			RemoteServer: p.RemoteServer,
		},
		Obligations: params.Obligations{
			EnableByDefault: p.EnableByDefault,
		},
	}
	if len(p.Series) > 0 {
		series := make(map[string]map[string]interface{}, len(p.Series))
		for k, v := range p.Series {
			ent.Affordances.Series = append(ent.Affordances.Series, k)
			affordances := make(map[string]interface{})
			if len(v.Architectures) > 0 {
				affordances["architectures"] = v.Architectures
			}
			if len(v.KernelFlavors) > 0 {
				affordances["kernelFlavors"] = v.KernelFlavors
			}
			if v.MinKernelVersion != nil {
				affordances["minKernelVersion"] = *v.MinKernelVersion
			}
			if v.Tier != nil {
				affordances["tier"] = *v.Tier
			}
			directives := make(map[string]interface{})
			if v.CaCerts != nil {
				directives["caCerts"] = *v.CaCerts
			}
			if v.RemoteServer != nil {
				directives["remoteServer"] = *v.RemoteServer
			}
			obligations := make(map[string]interface{})
			if v.EnableByDefault != nil {
				obligations["enableByDefault"] = v.EnableByDefault
			}
			series[k] = make(map[string]interface{})
			if len(affordances) > 0 {
				series[k]["affordances"] = affordances
			}
			if len(directives) > 0 {
				series[k]["directives"] = directives
			}
			if len(obligations) > 0 {
				series[k]["obligations"] = obligations
			}
		}
		ent.Series = series
	}
	sort.Strings(ent.Affordances.Series)
	return ent, nil
}

func makeSupportEntitlement(p0 interface{}, rt string) (params.Entitlement, error) {
	p := p0.(*supportParams)
	return params.EntitlementSupport{
		Type:     rt,
		Entitled: true,
		Affordances: params.AffordancesSupport{
			SupportLevel: p.SupportLevel,
		},
	}, nil
}

func getAffordances(e params.Entitlement) interface{} {
	if v := reflect.ValueOf(e).FieldByName("Affordances"); v.IsValid() {
		return v.Interface()
	}
	return nil
}
