package v1

import (
	"context"
	"net/http"

	"github.com/juju/zaputil"
	"github.com/juju/zaputil/zapctx"
	"go.uber.org/zap"
	errgo "gopkg.in/errgo.v1"
	httprequest "gopkg.in/httprequest.v1"

	"github.com/CanonicalLtd/ua-contracts/internal/monitoring"
	"github.com/CanonicalLtd/ua-contracts/params"
)

// errorMapper implements an error mapper for an httprequest server.
func errorMapper(ctx context.Context, err error) (int, interface{}) {
	errorBody := errorResponseBody(ctx, err)
	status := http.StatusInternalServerError
	switch params.ErrorCode(errorBody.Code) {
	case params.ErrBadRequest:
		status = http.StatusBadRequest
	case params.ErrAlreadyExists,
		params.ErrResourceForbidden:
		status = http.StatusForbidden
	case params.ErrUnauthorized:
		status = http.StatusUnauthorized
	case params.ErrNotFound:
		status = http.StatusNotFound
	case params.ErrForbidden:
		status = http.StatusForbidden
	}
	if status >= http.StatusInternalServerError {
		zapctx.Error(ctx, err.Error(), zap.Int("status", status), zaputil.Error(err))
	} else {
		zapctx.Info(ctx, err.Error(), zap.Int("status", status))
	}
	return status, errorBody
}

// errorResponse returns an appropriate error response for the provided error.
func errorResponseBody(ctx context.Context, err error) *params.Error {
	errResp := &params.Error{
		Message: err.Error(),
		TraceId: monitoring.TraceIDFromContext(ctx),
	}
	cause := errgo.Cause(err)
	if code, ok := cause.(params.ErrorCode); ok {
		errResp.Code = code
	} else if errgo.Cause(err) == httprequest.ErrUnmarshal {
		errResp.Code = params.ErrBadRequest
	}
	return errResp
}

func badRequestf(underlying error, f string, a ...interface{}) error {
	err := errgo.WithCausef(underlying, params.ErrBadRequest, f, a...)
	err.(*errgo.Err).SetLocation(1)
	return err
}

func unauthorizedErrorf(err error, f string, a ...interface{}) error {
	err = errgo.WithCausef(err, params.ErrUnauthorized, f, a...)
	err.(*errgo.Err).SetLocation(1)
	return err
}

func forbiddenErrorf(err error, f string, a ...interface{}) error {
	err = errgo.WithCausef(err, params.ErrForbidden, f, a...)
	err.(*errgo.Err).SetLocation(1)
	return err
}
