package params_test

import (
	"encoding/json"
	"reflect"
	"testing"
	"time"

	qt "github.com/frankban/quicktest"

	"github.com/CanonicalLtd/ua-contracts/params"
)

var expected = time.Date(2016, 12, 24, 14, 0, 0, 0, time.UTC)

var marshalTests = []struct {
	testName string
	data     string
	val      interface{}
}{{
	testName: "GetResourceMachineAccessResponse",
	data: `
{
	"resourceToken": "resource-token",
	"entitlement": {
		"type": "esm",
		"enabled": true,
		"directives": {
			"aptURL": "get it here",
			"aptKey": "trust me"
		},
		"series": {
			"xenial": {},
			"bionic": {},
			"disco": {}
		}		
	}
}`[1:],
	val: &params.GetResourceMachineAccessResponse{
		ResourceToken: "resource-token",
		Entitlement: params.EntitlementAptRepository{
			Type: "esm",
			Directives: params.DirectivesAptRepository{
				AptKey: "trust me",
				AptURL: "get it here",
			},
			Series: map[string]interface{}{
				"bionic": map[string]interface{}{},
				"disco":  map[string]interface{}{},
				"xenial": map[string]interface{}{},
			},
		},
	},
}, {
	testName: "ContractInfo",
	data: `
{
	"createdAt": "2016-12-24T14:00:00Z",
	"createdBy": "me",
	"effectiveFrom": "2016-12-24T14:00:00Z",
	"effectiveTo": "2016-12-24T14:00:00Z",
	"id": "3bd80eb3-e764-40f9-a40e-156385167817",
	"name": "UA Advanced site-license 1yr",
	"resourceEntitlements": [{
			"affordances": {
				"kernelFlavors": [
					"generic",
					"aws",
					"gcp",
					"azure",
					"ibm"
				]
			},
			"directives": {
				"remoteServer": "https://livepatch.canonical.com"
			},
			"entitled": true,
			"type": "livepatch"
		}, {
			"directives": {
				"aptKey": "-----BEGIN PGP PUBLIC KEY BLOCK-----\n...\n-----END PGP PUBLIC KEY BLOCK-----",
				"aptURL": "deb http://esm.archive.ubuntu.com {{series}} security"
			},
			"entitled": true,
			"series": {
				"xenial": {},
				"bionic": {}
			},
			"type": "esm"
		}, {
			"directives": {
				"aptKey": "-----BEGIN PGP PUBLIC KEY BLOCK-----\n...\n-----END PGP PUBLIC KEY BLOCK-----",
				"aptURL": "deb http://fips.archive.ubuntu.com {{series}} security"
			},
			"entitled": true,
			"type": "fips"
		}, {
			"affordances": {
				"supportLevel": "community"
			},
			"obligations": {
				"enableByDefault": true
			},
			"type": "support"
		}
	]
}
	`,
	val: &params.ContractInfo{
		Id:            "3bd80eb3-e764-40f9-a40e-156385167817",
		Name:          "UA Advanced site-license 1yr",
		CreatedAt:     expected,
		CreatedBy:     "me",
		EffectiveFrom: &expected,
		EffectiveTo:   &expected,
		ResourceEntitlements: []params.Entitlement{
			params.EntitlementLivepatch{
				Type:     "livepatch",
				Entitled: true,
				Affordances: params.AffordancesLivepatch{
					KernelFlavors: []string{
						"generic",
						"aws",
						"gcp",
						"azure",
						"ibm",
					},
				},
				Directives: params.DirectivesLivepatch{
					RemoteServer: "https://livepatch.canonical.com",
				},
			},
			params.EntitlementAptRepository{
				Type:     "esm",
				Entitled: true,
				Directives: params.DirectivesAptRepository{
					AptKey: "-----BEGIN PGP PUBLIC KEY BLOCK-----\n...\n-----END PGP PUBLIC KEY BLOCK-----",
					AptURL: "deb http://esm.archive.ubuntu.com {{series}} security",
				},
				Series: map[string]interface{}{
					"xenial": map[string]interface{}{},
					"bionic": map[string]interface{}{},
				},
			},
			params.EntitlementAptRepository{
				Type:     "fips",
				Entitled: true,
				Directives: params.DirectivesAptRepository{
					AptKey: "-----BEGIN PGP PUBLIC KEY BLOCK-----\n...\n-----END PGP PUBLIC KEY BLOCK-----",
					AptURL: "deb http://fips.archive.ubuntu.com {{series}} security",
				},
			},
			params.EntitlementSupport{
				Type: "support",
				Affordances: params.AffordancesSupport{
					SupportLevel: "community",
				},
				Obligations: params.Obligations{
					EnableByDefault: true,
				},
			},
		},
	},
}}

func TestMarshalUnmarshal(t *testing.T) {
	c := qt.New(t)
	for _, test := range marshalTests {
		c.Run("unmarshal_"+test.testName, func(c *qt.C) {
			v := reflect.New(reflect.TypeOf(test.val).Elem()).Interface()
			err := json.Unmarshal([]byte(test.data), v)
			c.Assert(err, qt.Equals, nil)
			c.Assert(v, qt.DeepEquals, test.val)

			// Check that it round trips OK.
			data, err := json.Marshal(v)
			c.Assert(err, qt.Equals, nil)

			v = reflect.New(reflect.TypeOf(test.val).Elem()).Interface()
			err = json.Unmarshal([]byte(data), v)
			c.Assert(err, qt.Equals, nil)
			c.Assert(v, qt.DeepEquals, test.val)
		})
	}
}

func TestUnmarasalUnknownEntitlement(t *testing.T) {
	c := qt.New(t)

	data := `{
		"createdAt": "2016-12-24T14:00:00Z",
		"createdBy": "me",
		"effectiveFrom": "2016-12-24T14:00:00Z",
		"effectiveTo": "2016-12-24T14:00:00Z",
		"id": "3bd80eb3-e764-40f9-a40e-156385167817",
		"name": "UA Advanced site-license 1yr",
		"resourceEntitlements": [{	
			"type": "unknown-type",
			"entitled": true,
			"affordances": {
				"unkown": "unknown"
			}
		}]
	}`

	var ci params.ContractInfo
	err := json.Unmarshal([]byte(data), &ci)
	c.Assert(err, qt.Equals, nil)
	c.Assert(ci.ResourceEntitlements, qt.HasLen, 1)
	c.Assert(ci.ResourceEntitlements[0], qt.DeepEquals, params.EntitlementUnknown{
		Type: "unknown-type",
		Entitlement: json.RawMessage(`{	
			"type": "unknown-type",
			"entitled": true,
			"affordances": {
				"unkown": "unknown"
			}
		}`),
	})
}
