// Copyright 2019 Canonical Ltd.

// Package contracts provides the main server for the ua-contracts API.
package contracts

import (
	"context"
	"crypto/rand"
	"crypto/rsa"
	"crypto/x509"
	"encoding/pem"
	"net/http"

	"github.com/CanonicalLtd/salesforce"
	aclstore "github.com/juju/aclstore/v2"
	"github.com/juju/simplekv/memsimplekv"
	"github.com/juju/zaputil/zapctx"
	"github.com/julienschmidt/httprouter"
	"go.uber.org/zap"
	errgo "gopkg.in/errgo.v1"
	"gopkg.in/macaroon-bakery.v2/bakery"

	"github.com/CanonicalLtd/ua-contracts/internal/awsauth"
	"github.com/CanonicalLtd/ua-contracts/internal/azureauth"
	"github.com/CanonicalLtd/ua-contracts/internal/memstore"
	"github.com/CanonicalLtd/ua-contracts/internal/monitoring"
	"github.com/CanonicalLtd/ua-contracts/internal/saauth"
	"github.com/CanonicalLtd/ua-contracts/internal/ssoauth"
	"github.com/CanonicalLtd/ua-contracts/internal/store"
	"github.com/CanonicalLtd/ua-contracts/internal/storeutil"
	v1 "github.com/CanonicalLtd/ua-contracts/internal/v1"
)

type Params struct {
	// ACLStore holds the store used for the ACLs for admin commands.
	// If it is nil an new in-memory ACL store will be used.
	ACLStore aclstore.ACLStore

	// AWSCertificates contains a set of AWS certificates to use to
	// verify AWS credentials. If there are no certificates then AWS
	// registrations will not be possible.
	AWSCertificates []*x509.Certificate

	// AzureCertificates is a pool of intermediate certificates with which
	// Azure attested data PKCS#7 certificates are verified. If not configured,
	// or the intermediate certificates are out of date with Microsoft Azure,
	// Azure registrations will not be possible.
	AzureCertificates *x509.CertPool

	// BakeryRootKeyStore holds the store to use for root keys. If it
	// is nil a bakery.NewMemRootKeyStore will be used to create an
	// in-memory store.
	BakeryRootKeyStore bakery.RootKeyStore

	// Logger contains a zap.Logger to attach to all requests. If
	// this is nil then the default logger will be used.
	Logger *zap.Logger

	// SSOLocation holds the location of the SSO service. If this is
	// not specified then login.ubuntu.com will be used.
	SSOLocation string

	// SSOPublicKey holds the public key of the SSO service specified
	// in SSOLocation. If this is not specified then the default
	// public key for login.ubuntu.com will be used.
	SSOPublicKey *rsa.PublicKey

	// Store contains the store the server will use for contracts
	// data. If this is not specified then a new in-memory store will
	// be used.
	Store store.Store

	// SalesforceConfig holds the configuration used to access
	// salesforce. If this is not specified then access to salesforce
	// will be disabled.
	SalesforceConfig *salesforce.Config

	// TokenSecret contains a shared secret used when creating new
	// tokens. If this is not specified a new random token will be
	// created.
	TokenSecret []byte
}

// NewServer creates a new http.Handler that serves the ua-contracts API.
func NewServer(ctx context.Context, p Params) (http.Handler, error) {
	if p.Logger != nil {
		ctx = zapctx.WithLogger(ctx, p.Logger)
	}
	if p.ACLStore == nil {
		p.ACLStore = aclstore.NewACLStore(memsimplekv.NewStore())
	}
	if p.BakeryRootKeyStore == nil {
		p.BakeryRootKeyStore = bakery.NewMemRootKeyStore()
	}
	if p.SSOLocation == "" {
		p.SSOLocation = "login.ubuntu.com"
	}
	if p.SSOPublicKey == nil {
		p.SSOPublicKey = ubuntuSSOPublicKey
	}
	if p.Store == nil {
		p.Store = memstore.New()
	}
	if p.TokenSecret == nil {
		zapctx.Warn(ctx, "TOKEN_SECRET not set, generating randomly")
		p.TokenSecret = make([]byte, 32)
		if _, err := rand.Read(p.TokenSecret); err != nil {
			return nil, errgo.Notef(err, "cannot create random token secret")
		}
	}

	var sfclient storeutil.SalesforceClient
	if p.SalesforceConfig == nil {
		zapctx.Warn(ctx, "salesforce has been disabled")
		sfclient = &storeutil.SalesforceDisabledClient{}
	} else {
		var err error
		sfclient, err = salesforce.NewClient(ctx, p.SalesforceConfig)
		if err != nil {
			return nil, errgo.Notef(err, "cannot create salesforce client")
		}
	}

	aclManager, err := aclstore.NewManager(ctx, aclstore.Params{
		Store:             p.ACLStore,
		InitialAdminUsers: []string{"admin"},
	})
	if err != nil {
		return nil, errgo.Notef(err, "cannot create ACL manager")
	}

	var awsDecoder *awsauth.Decoder
	if len(p.AWSCertificates) > 0 {
		awsDecoder = &awsauth.Decoder{
			Certs: p.AWSCertificates,
		}
	}

	var azureDecoder *azureauth.Decoder
	if p.AzureCertificates != nil {
		azureDecoder = &azureauth.Decoder{
			Intermediates: p.AzureCertificates,
		}
	}

	hs, err := v1.Handlers(ctx, v1.Params{
		ACLManager:   aclManager,
		AWSDecoder:   awsDecoder,
		AzureDecoder: azureDecoder,
		Store:        p.Store,
		SSOAuth: ssoauth.New(ssoauth.Params{
			Oven: bakery.NewOven(bakery.OvenParams{
				RootKeyStoreForOps: func([]bakery.Op) bakery.RootKeyStore {
					return p.BakeryRootKeyStore
				},
			}),
			Location:  p.SSOLocation,
			PublicKey: p.SSOPublicKey,
		}),
		TokenSecret:      p.TokenSecret,
		SalesforceClient: sfclient,
	})
	if err != nil {
		return nil, errgo.Notef(err, "cannot create v1 handlers")
	}

	aclHandler := aclManager.NewHandler(
		aclstore.HandlerParams{
			RootPath:     "/acl",
			Authenticate: saauth.AuthenticateACL(p.Store),
		},
	)

	router := httprouter.New()
	monitoring.AddHandlers(router, hs)
	monitoring.AddHTTPHandler(router, "GET", "/acl/*path", aclHandler)
	monitoring.AddHTTPHandler(router, "POST", "/acl/*path", aclHandler)
	monitoring.AddHTTPHandler(router, "PUT", "/acl/*path", aclHandler)
	monitoring.AddMetricsHandler(router, "/metrics")

	// Create pre-router middleware to be performed on all requests
	handler := http.HandlerFunc(func(w http.ResponseWriter, req *http.Request) {
		ctx := req.Context()
		if p.Logger != nil {
			ctx = zapctx.WithLogger(ctx, p.Logger)
		}
		router.ServeHTTP(w, req.WithContext(ctx))
	})
	return handler, nil
}

var ubuntuSSOPublicKey = func() *rsa.PublicKey {
	block, rest := pem.Decode([]byte(ubuntuSSOPublicKeyPEM))
	if len(rest) > 0 {
		panic(errgo.New("cannot decode default public key"))
	}
	pk, err := x509.ParsePKIXPublicKey(block.Bytes)
	if err != nil {
		panic(err)
	}
	pkey, ok := pk.(*rsa.PublicKey)
	if !ok {
		panic(errgo.Newf("unsupported key type %T", pk))
	}
	return pkey
}()

const ubuntuSSOPublicKeyPEM = `-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAtO9bUDClVY5VKD8cLRbo
bw/ruGTTepdvhDc3oewHrnog/i7ymQpB7fCbn3iryNs704p9Yg1UaQM/iFJ75ojB
4gsGHF0yZWj4vf0Wt8vjpY7wdy/7MHerizED132ncoSuc676ulydVzyOPUGKQbPe
/EkYJhfoxyJHrBPvIHj445EK//uScrrKmwXteu0sOytDSLU+EPJgTQQ38sKak9FM
+MAW4R4wPvGZibUsm4JWfGGwf68FPXwZ0a/BIJXL8hBvuGaQlFy9jrafXlJKqY+I
XAQrHqnBTYxhyhhXCQrwBJOxcVTZe/zowGC3W/YO2yIWpqgkgkg4KkmyzayvhpQ2
4wIDAQAB
-----END PUBLIC KEY-----`
