// Copyright 2019 Canonical Ltd.

package contracts_test

import (
	"context"
	"net/http"
	"net/http/httptest"
	"testing"
	"time"

	qt "github.com/frankban/quicktest"
	"github.com/juju/zaputil/zapctx"
	"go.uber.org/zap/zaptest"

	contracts "github.com/CanonicalLtd/ua-contracts"
	"github.com/CanonicalLtd/ua-contracts/contractclient"
	"github.com/CanonicalLtd/ua-contracts/internal/memstore"
	"github.com/CanonicalLtd/ua-contracts/internal/ssoauth"
	"github.com/CanonicalLtd/ua-contracts/internal/ssoauth/ssoauthtest"
	"github.com/CanonicalLtd/ua-contracts/internal/storeutil"
	"github.com/CanonicalLtd/ua-contracts/params"
)

func TestEmptyParams(t *testing.T) {
	c := qt.New(t)
	logger := zaptest.NewLogger(c)
	ctx := zapctx.WithLogger(context.Background(), logger)

	hnd, err := contracts.NewServer(ctx, contracts.Params{
		Logger: logger,
	})
	c.Assert(err, qt.Equals, nil)

	w := httptest.NewRecorder()
	req, err := http.NewRequest("GET", "/debug/info", nil)
	c.Assert(err, qt.Equals, nil)
	hnd.ServeHTTP(w, req)
	resp := w.Result()
	c.Assert(resp.StatusCode, qt.Equals, http.StatusOK)
}

func TestGetContractToken(t *testing.T) {
	c := qt.New(t)
	logger := zaptest.NewLogger(c)
	ctx := zapctx.WithLogger(context.Background(), logger)

	var d ssoauthtest.Discharger
	st := memstore.New()
	err := storeutil.EnsureDefaultProducts(ctx, st)
	c.Assert(err, qt.Equals, nil)

	hnd, err := contracts.NewServer(ctx, contracts.Params{
		Logger:       logger,
		SSOLocation:  d.Location(),
		SSOPublicKey: d.PublicKey(),
		Store:        st,
	})
	c.Assert(err, qt.Equals, nil)

	srv := httptest.NewServer(hnd)
	m, err := contractclient.GetCanonicalSSOMacaroon(ctx, srv.URL)
	c.Assert(err, qt.Equals, nil)

	ms, err := ssoauthtest.Discharge(&d, m, &ssoauth.Account{
		OpenID:      "AAAAAA",
		Username:    "test",
		DisplayName: "Test User",
		Email:       "test@example.com",
		IsVerified:  true,
	}, time.Time{}, time.Time{})
	c.Assert(err, qt.Equals, nil)

	auth, err := contractclient.MacaroonAuthorizer(ms)
	c.Assert(err, qt.Equals, nil)
	client := contractclient.New(contractclient.NewParams{
		BaseURL:    srv.URL,
		Authorizer: auth,
	})

	aresp, err := client.GetAccounts(ctx, &params.GetAccountsRequest{})
	c.Assert(err, qt.Equals, nil)
	c.Assert(aresp.Accounts, qt.HasLen, 1)

	cresp, err := client.GetAccountContracts(ctx, &params.GetAccountContractsRequest{
		Account: aresp.Accounts[0].Id,
	})
	c.Assert(err, qt.Equals, nil)
	c.Assert(cresp.Contracts, qt.HasLen, 1)

	ctresp, err := client.GetContractToken(ctx, &params.GetContractTokenRequest{
		Contract: cresp.Contracts[0].ContractInfo.Id,
	})
	c.Assert(err, qt.Equals, nil)
	c.Assert(ctresp.ContractToken, qt.Not(qt.Equals), "")
}
