// Copyright 2018 Canonical Ltd.

// Package auth implements salesforce specific OAuth 2 login behavior.
package auth

import (
	"context"
	"net/http"
	"net/url"
	"path"
	"strconv"
	"time"

	"github.com/juju/zaputil"
	"github.com/juju/zaputil/zapctx"
	"go.uber.org/zap"
	"golang.org/x/oauth2"
	"gopkg.in/errgo.v1"
	"gopkg.in/httprequest.v1"
	"gopkg.in/retry.v1"
)

const (
	salesforceAPIVersion  = "v41.0"
	resourceURI           = "/services/data"
	DefaultSessionTimeout = 4 * time.Hour
)

// CodeInvalidSession is the error code returned by salesforce if
// the session token has expired or is invalid.
const CodeInvalidSession = "INVALID_SESSION_ID"

// Config contains the configuration required to use username/password
// authentication with salesforce. As opposed to the standard oauth2
// username / password flow the token will be refreshed as necessary.
type Config struct {
	// ClientID holds the client ID of the OAuth application.
	ClientID string

	// ClientSecret holds the client secret of the OAuth application.
	ClientSecret string

	// SessionTimeout holds the amount of time that a token may be
	// used before it must be renewed. This is required as the tokens
	// provided by salesforce do not include the optional expires_in
	// field. If this is zero then DefaultSessionTimeout will
	// be used.
	SessionTimeout time.Duration

	// TokenURL contains the token endpoint to of the salesforce
	// OAuth2 provider.
	TokenURL string

	// Username contains the name of the user that the service will
	// be logging in as.
	Username string

	// Password contains the password for the user that the service
	// will be logging in as.
	Password string

	// SecurityToken contains the security token for the user that
	// the service will be logging in as. See
	// https://help.salesforce.com/articleView?id=remoteaccess_oauth_username_password_flow.htm
	// for details.
	SecurityToken string

	// UnmarshalError is called to unmarshal the error response into an error type.
	UnmarshalError func(*http.Response) error

	// RequestTimeout define the timeout for a call to Salesforce.
	RequestTimeout time.Duration
}

// Token authenticates with salesforce using the username-password OAuth
// flow as described by
// https://help.salesforce.com/articleView?id=remoteaccess_oauth_username_password_flow.htm.
func (c *Config) Token(ctx context.Context) (*oauth2.Token, error) {
	zapctx.Debug(ctx, "retrieving new salesforce OAuth token")
	cfg := oauth2.Config{
		ClientID:     c.ClientID,
		ClientSecret: c.ClientSecret,
		Endpoint: oauth2.Endpoint{
			TokenURL: c.TokenURL,
		},
	}
	// Note that ctx is global, not per-request, because oauth2.Transport doesn't
	// pass through the request context, so we need to add a timeout.
	if c.RequestTimeout != 0 {
		ctx1, cancel := context.WithTimeout(ctx, c.RequestTimeout)
		ctx = ctx1
		defer cancel()
	}
	tok, err := cfg.PasswordCredentialsToken(ctx, c.Username, c.Password+c.SecurityToken)
	if err != nil {
		zapctx.Error(ctx, "retrieving new salesforce OAuth token", zaputil.Error(err))
		return nil, errgo.Mask(err, isRetrieveError)
	}
	if tok.Expiry.IsZero() {
		// Salesforce does not return the optional expires_in
		// field. This causes the oauth2 library to think that
		// the token is always valid and never refresh it. In
		// reality tokens will expire so ensure Expiry is set so
		// that the token will be refreshed.
		timeout := c.SessionTimeout
		if timeout == 0 {
			timeout = DefaultSessionTimeout
		}

		// Despite issued_at being documented as the number of
		// seconds since 1970-01-01T00:00:00Z this gives dates
		// that are many millennia ahead of now, it is actually
		// the number of milliseconds.
		issuedAt := time.Now()
		if issuedAtString, ok := tok.Extra("issued_at").(string); ok {
			ia, err := strconv.ParseInt(issuedAtString, 10, 64)
			if err == nil {
				issuedAt = time.Unix(0, 0).Add(time.Duration(ia) * time.Millisecond)
			}
		}
		tok.Expiry = issuedAt.Add(timeout)
	}
	zapctx.Debug(ctx, "retrieving new salesforce OAuth token", zap.Stringer("expiry", tok.Expiry), zaputil.Error(err))
	return tok, nil
}

// Client requests a new token from salesforce and returns a client using
// that oauth token.
func (c *Config) Client(ctx context.Context) (*Client, error) {
	tok, err := c.Token(ctx)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	client, err := c.client(ctx, tok)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return client, nil
}

func isRetrieveError(err error) bool {
	_, ok := err.(*oauth2.RetrieveError)
	return ok
}

// client returns an HTTP client using the provided token.
// The token will auto-refresh as necessary.
func (c *Config) client(ctx context.Context, t *oauth2.Token) (*Client, error) {
	baseURL := t.Extra("instance_url")
	b, ok := baseURL.(string)
	if !ok || len(b) == 0 {
		return nil, errgo.Newf("cannot get the base url for salesforce from %#v (%T)", baseURL, baseURL)
	}
	u, err := url.Parse(b)
	if err != nil {
		return nil, errgo.Newf("cannot parse the base url %q for salesforce", baseURL)
	}
	u.Path = path.Join(u.Path, resourceURI, salesforceAPIVersion)

	tokenSource := newReuseTokenSource(t,
		func() (*oauth2.Token, error) { return c.Token(ctx) })

	baseClient := &httprequest.Client{
		Doer: &http.Client{
			Transport: &oauth2.Transport{
				Base:   http.DefaultTransport,
				Source: tokenSource,
			},
			Timeout: c.RequestTimeout,
		},
		BaseURL:        u.String(),
		UnmarshalError: c.UnmarshalError,
	}

	retryStrategy := retry.LimitTime(c.RequestTimeout, retry.Exponential{
		Initial:  time.Millisecond,
		Factor:   2,
		MaxDelay: 500 * time.Millisecond,
		Jitter:   true,
	})
	if c.RequestTimeout == 0 {
		// We want at least to retry twice if the token was invalid so it will retry with a new token.
		retryStrategy = retry.LimitCount(2, retry.Regular{
			Min: 2,
		})
	}

	return &Client{
		client:        baseClient,
		tokenSource:   tokenSource,
		retryStrategy: retryStrategy,
	}, nil
}

// Client wraps a httprequest.Client and inspects the responses
// from salesforce calls for errors corresponding to expired tokens.
// In such cases the client forces the token to refresh and retries the request.
type Client struct {
	client        *httprequest.Client
	tokenSource   *reuseTokenSource
	retryStrategy retry.Strategy
}

// Call wraps the underlying httprequest.Client and invalidates the auth token
// if it receives a 401 response.
func (c *Client) Call(ctx context.Context, params, resp interface{}) (err error) {
	r := retry.StartWithCancel(c.retryStrategy, nil, ctx.Done())
	for r.Next() {
		err = c.client.Call(ctx, params, resp)
		if err == nil {
			return nil
		}
		if err1, ok := errgo.Cause(err).(coder); ok {
			// It's an error from the API, so don't retry as
			// we really only want to retry on transient errors,
			// except for auth errors, which we do want to retry
			// for.
			if err1.Code() != CodeInvalidSession {
				return errgo.Mask(err, errgo.Any)
			}
			c.tokenSource.invalidate()
		}
	}
	return errgo.NoteMask(err, "too many retries", errgo.Any)
}

// coder is implemented by params.MultiError. The method returns the error code
// of the salesforce error.
type coder interface {
	Code() string
}
