// Copyright 2018 Canonical Ltd.

package auth

import (
	"sync"

	"golang.org/x/oauth2"
	"gopkg.in/errgo.v1"
)

type tokenFunc func() (*oauth2.Token, error)

// newReuseTokenSource is a copy of oauth2.ReuseTokenSource.
// It uses the same token while it's valid (or is forcibly invalidated).
func newReuseTokenSource(t *oauth2.Token, src tokenFunc) *reuseTokenSource {
	return &reuseTokenSource{
		t:   t,
		new: src,
	}
}

// reuseTokenSource is a TokenSource that holds a single token in memory
// and validates its expiry before each call to retrieve it with
// Token. If it's expired, it will be auto-refreshed using the
// new TokenSource.
type reuseTokenSource struct {
	new tokenFunc // called when t is expired.

	mu          sync.Mutex // guards t
	t           *oauth2.Token
	invalidated bool
}

// invalidate marks the token as invalid.
// This will cause it to be refreshed on next request.
func (s *reuseTokenSource) invalidate() {
	s.mu.Lock()
	s.invalidated = true
	s.mu.Unlock()
}

// Token returns the current token if it's still valid, else will
// refresh the current token (using r.Context for HTTP client
// information) and return the new one.
func (s *reuseTokenSource) Token() (*oauth2.Token, error) {
	s.mu.Lock()
	defer s.mu.Unlock()
	if !s.invalidated && s.t.Valid() {
		return s.t, nil
	}
	t, err := s.new()
	if err != nil {
		return nil, errgo.Mask(err)
	}
	s.invalidated = false
	s.t = t
	return t, nil
}
