// Copyright 2018 Canonical Ltd.

package salesforce

import (
	"context"
	"fmt"
	"net/http"
	"net/url"
	"path"
	"sort"
	"strings"
	"time"

	"github.com/juju/zaputil/zapctx"
	"go.uber.org/zap"
	errgo "gopkg.in/errgo.v1"
	httprequest "gopkg.in/httprequest.v1"

	"github.com/CanonicalLtd/salesforce/auth"
	"github.com/CanonicalLtd/salesforce/params"
)

const (
	// Validation duration of a subscription.
	validityDuration = 3 * 365 * 24 * time.Hour
	defaultMaxUsers  = 5
)

// httpClient is used by the salesforce client to make oauth-authenticated
// http calls.
type httpClient interface {
	// Call wraps the httprequest.Client.Call method.
	Call(ctx context.Context, params, resp interface{}) error
}

// Client to Salesforce.
type Client struct {
	client httpClient
	// productIDs holds a map from support level to the product id for that
	// support level. It is populated when creating the client.
	productIDs map[string]string
	// profileIDs holds a map from support level to the profile id for that
	// support level. It is populated when creating the client.
	profileIDs map[string]string
	// sandboxInstanceName represents the instance name when we are in sandbox mode.
	// This is used as a suffix for all user names so we don't clash with real user names.
	sandboxInstanceName string
	organizationID      string
	signinClient        *signinClient
}

// Config represents the configuration of the salesforce client.
type Config struct {
	ConsumerID     string
	ConsumerSecret string
	User           string
	Password       string
	SecurityToken  string
	URL            string
	// AuthSessionTimeout holds the amount of time that an authorization
	// token may be used before it must be renewed. This is required
	// as the tokens provided by salesforce do not include the
	// optional expires_in field. If this is zero then a default
	// non-zero value will be used.
	AuthSessionTimeout time.Duration
	TenantURL          string
	// https://help.salesforce.com/articleView?id=sso_saml.htm&type=5

	// The following SAML configuration allows a SAML signin client to be set up if SAMLIssuer is not empty.
	SAMLIssuer      string
	SAMLEntityID    string
	SAMLPrivateKey  string
	SAMLCertificate string

	// productNames holds a map from product names to the service levels associated with that product
	// name. This is used to populate the Client.productIDs field.
	ProductNames map[string][]string
	// ProfileNames holds a map from profile names to the service levels associated with that profile
	// name. This is used to populate the Client.profileIDs field.
	ProfileNames map[string][]string

	// RequestTimeout holds the timeout for a call to Salesforce.
	RequestTimeout time.Duration

	authenticate func(ctx context.Context, conf *Config) (httpClient, error)
}

// NewClient creates a new Salesforce client.
func NewClient(ctx context.Context, conf *Config) (*Client, error) {
	var caller httpClient
	var err error
	if conf.authenticate != nil {
		caller, err = conf.authenticate(ctx, conf)
	} else {
		caller, err = authenticateSF(ctx, conf)
	}
	if err != nil {
		return nil, errgo.Mask(err)
	}

	client := &Client{
		client: caller,
	}
	if err := client.init(ctx, conf); err != nil {
		return nil, errgo.Mask(err)
	}
	if conf.SAMLIssuer != "" {
		client.signinClient, err = newSigninClient(conf, client.organizationID)
		if err != nil {
			return nil, errgo.Notef(err, "cannot create a signin client")
		}
	}
	return client, nil
}

// authenticateSF creates an HTTP client that has been authenticated through OAuth2 flow.
func authenticateSF(ctx context.Context, conf *Config) (httpClient, error) {
	tokenURL, err := url.Parse(conf.URL)
	if err != nil {
		return nil, errgo.Notef(err, "cannot parse salesforce url %q", conf.URL)
	}
	tokenURL.Path = path.Join(tokenURL.Path, "/services/oauth2/token")
	cfg := auth.Config{
		ClientID:       conf.ConsumerID,
		ClientSecret:   conf.ConsumerSecret,
		SessionTimeout: conf.AuthSessionTimeout,
		RequestTimeout: conf.RequestTimeout,
		TokenURL:       tokenURL.String(),
		Username:       conf.User,
		Password:       conf.Password,
		SecurityToken:  conf.SecurityToken,
		UnmarshalError: params.UnmarshalError,
	}
	client, err := cfg.Client(ctx)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return client, nil
}

// GetAccount returns the account associated with the accountName.
// Returns an error with an params.ErrNotFound cause if the account does not exist.
//
// Note that this will only work correctly in Azure-GCE context.
func (c *Client) GetAccount(ctx context.Context, accountName string) (*params.Account, error) {
	var accountResp accountResp
	if err := c.client.Call(ctx, &queryRequest{
		Select: &selectQuery{
			Columns: []string{"Id", "Name", "IsDeleted", "Account_Active__c", "Max_Technical_Representatives__c"},
			Table:   "Account",
			Where:   "Name",
			Equals:  accountName,
		},
	}, &accountResp); err != nil {
		return nil, errgo.Mask(err)
	}
	if len(accountResp.errors) > 0 {
		return nil, errgo.Notef(params.MultiError(accountResp.errors), "cannot get account %q", accountName)
	}
	if len(accountResp.Accounts) == 0 {
		return nil, errgo.WithCausef(nil, params.ErrNotFound, "account %q not found", accountName)
	}
	account := accountResp.Accounts[0]
	linkedAsset, err := c.getAssetForAccount(ctx, account.ID)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	account.Enabled = linkedAsset.Status == "Active" && linkedAsset.UsageEndDate.T.After(time.Now())
	account.AssetID = linkedAsset.ID
	account.ContactID = linkedAsset.ContactID

	// Add the support level
	for k, v := range c.productIDs {
		if linkedAsset.Product.ID == v {
			account.SupportLevel = k
			break
		}
	}
	return &account, nil
}

// getAssetForAccount returns the asset corresponding to the account. This will
// only work correctly in the context of the azure-support-interface
// salesforce users.
func (c *Client) getAssetForAccount(ctx context.Context, accountID string) (*Asset, error) {
	assets, err := c.GetAssetsForAccount(ctx, accountID)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	if len(assets) == 0 {
		return nil, errgo.Notef(nil, "no asset associated with account %q", accountID)
	}
	// Pick the first one as there should be only one.
	return &assets[0], nil
}

// CreateAccount creates an account when subscribing to support.
// Be careful to have a unique name for the account as Salesforce accepts the same name for multiple accounts.
func (c *Client) CreateAccount(ctx context.Context, accountName string, supportLevel string) error {
	return c.createAccountAtTime(ctx, accountName, supportLevel, time.Now().UTC())
}

func (c *Client) createAccountAtTime(ctx context.Context, accountName string, supportLevel string, now time.Time) error {
	if _, ok := c.productIDs[supportLevel]; !ok {
		return errgo.Newf("cannot create an account with an unknown support level %q", supportLevel)
	}
	var resp createResp
	if err := c.client.Call(ctx, &createAccountRequest{
		Body: createAccount{
			Name: accountName,
		},
	}, &resp); err != nil {
		return errgo.Mask(err)
	}
	if len(resp.errors) > 0 {
		return errgo.Notef(params.MultiError(resp.errors), "cannot create account %q", accountName)
	}
	if !resp.Success {
		return errgo.Notef(nil, "cannot create account %q", accountName)
	}
	accountID := resp.ID
	return errgo.Mask(c.createAsset(ctx, accountName, accountID, supportLevel, now))
}

// UpdateAccount updates an account when the support level change.
func (c *Client) UpdateAccount(ctx context.Context, account *params.Account, supportLevel string) error {
	return c.updateAccountAtTime(ctx, account, supportLevel, time.Now().UTC())
}

func (c *Client) updateAccountAtTime(ctx context.Context, account *params.Account, supportLevel string, now time.Time) error {
	if len(account.AssetID) == 0 {
		err := c.createAsset(ctx, account.Name, account.ID, supportLevel, now)
		if err != nil {
			return errgo.Mask(err)
		}
	} else {
		err := c.updateAsset(ctx, account.Name, account.AssetID, supportLevel, now)
		if err != nil {
			return errgo.Mask(err)
		}
	}

	// Select/update all users associated with this account to be associated with the right profile.
	users, err := c.getAccountUsers(ctx, account.ID)
	if err != nil {
		return errgo.Mask(err)
	}

	var userErr []error
	for _, user := range users {
		// TODO use a bulk request or do this in concurrently.
		if err := c.updateUserProfile(ctx, &user, supportLevel); err != nil {
			userErr = append(userErr, err)
		}
	}
	if len(userErr) == 1 {
		return errgo.Mask(userErr[0])
	}
	if len(userErr) > 1 {
		zapctx.Info(ctx, "multiple profiles cannot be updated", zap.Errors("errors", userErr))
		return errgo.Notef(userErr[0], "and %d more", len(userErr)-1)
	}
	return nil
}

func (c *Client) getAccountUsers(ctx context.Context, accountID string) ([]params.User, error) {
	var resp userResp
	if err := c.client.Call(ctx, &queryRequest{
		Select: &selectQuery{
			Columns: []string{"Id"},
			Table:   "User",
			Where:   "AccountId",
			Equals:  accountID,
		},
	}, &resp); err != nil {
		return nil, errgo.Mask(err)
	}
	if len(resp.errors) > 0 {
		return nil, errgo.Notef(params.MultiError(resp.errors), "cannot get user associated with account %q", accountID)
	}
	return resp.Users, nil
}

func (c *Client) updateUserProfile(ctx context.Context, user *params.User, supportLevel string) error {
	var resp *http.Response
	if err := c.client.Call(ctx, &updateUserProfileRequest{
		ID: user.ID,
		Body: updateProfile{
			ProfileID: c.profileIDs[supportLevel],
		},
	}, &resp); err != nil {
		return errgo.Notef(err, "cannot update profile on user %s", user.UserName)
	}
	defer resp.Body.Close()
	if err := checkPatchErrors(resp, fmt.Sprintf("cannot update profile on user %s", user.UserName)); err != nil {
		return errgo.Mask(err)
	}
	return nil
}

// createAsset creates an asset associated with an account when subscribing to support.
func (c *Client) createAsset(ctx context.Context, accountName string, accountID string, supportLevel string, now time.Time) error {
	// Create an asset linked to the account with the selected Product2ID.
	var resp createResp
	if err := c.client.Call(ctx, &createAssetRequest{
		Body: createAsset{
			Name:         accountName,
			AccountID:    accountID,
			ProductID:    c.productIDs[supportLevel],
			Status:       "Active",
			UsageEndDate: Time{now.Add(validityDuration)},
			PurchaseDate: Time{now},
			InstallDate:  Time{now},
			Quantity:     1,
		},
	}, &resp); err != nil {
		return errgo.Mask(err)
	}
	if len(resp.errors) > 0 {
		return errgo.Notef(params.MultiError(resp.errors), "cannot create asset linked to account %q", accountName)
	}
	if !resp.Success {
		return errgo.Notef(nil, "cannot create asset linked to account %q", accountName)
	}
	return nil
}

// updateAsset updates an asset associated with an account when subscribing to support.
func (c *Client) updateAsset(ctx context.Context, accountName string, assetID string, supportLevel string, now time.Time) error {
	var resp *http.Response
	if err := c.client.Call(ctx, &updateAssetRequest{
		ID: assetID,
		Body: updateAsset{
			ProductID:    c.productIDs[supportLevel],
			Status:       "Active",
			UsageEndDate: Time{now.Add(validityDuration)},
		},
	}, &resp); err != nil {
		return errgo.Mask(err)
	}
	defer resp.Body.Close()
	if err := checkPatchErrors(resp, fmt.Sprintf("cannot update asset %q for account %q", assetID, accountName)); err != nil {
		return errgo.Mask(err)
	}
	return nil
}

// SetAccountEnabled sets whether the given account is enabled.
// When an account is disabled it is not deleted, but will be marked as expired.
func (c *Client) SetAccountEnabled(ctx context.Context, account *params.Account, enabled bool) error {
	if enabled {
		return errgo.Mask(c.enableAccountAtTime(ctx, account, time.Now().UTC()))
	}
	return errgo.Mask(c.disableAccountAtTime(ctx, account, time.Now().UTC()))
}

func (c *Client) disableAccountAtTime(ctx context.Context, account *params.Account, now time.Time) error {
	var resp *http.Response
	if err := c.client.Call(ctx, &setEnabledRequest{
		ID: account.AssetID,
		Body: setEnabledAsset{
			Status:       "Expired",
			UsageEndDate: Time{now},
		},
	}, &resp); err != nil {
		return errgo.Mask(err)
	}
	defer resp.Body.Close()
	if err := checkPatchErrors(resp, fmt.Sprintf("cannot disable account %q", account.Name)); err != nil {
		return errgo.Mask(err)
	}
	return nil
}

func (c *Client) enableAccountAtTime(ctx context.Context, account *params.Account, now time.Time) error {
	var resp *http.Response
	if err := c.client.Call(ctx, &setEnabledRequest{
		ID: account.AssetID,
		Body: setEnabledAsset{
			Status:       "Active",
			UsageEndDate: Time{now.Add(validityDuration)},
		},
	}, &resp); err != nil {
		return errgo.Mask(err)
	}
	defer resp.Body.Close()
	if err := checkPatchErrors(resp, fmt.Sprintf("cannot enable account %q", account.Name)); err != nil {
		return errgo.Mask(err)
	}
	return nil
}

// GetUser retrieves a user from its name.
// We have a convention to add the sandbox name automatically at the end as the user name needs to be unique in all salesforce
// in the world.
func (c *Client) GetUser(ctx context.Context, mail string) (*params.User, error) {
	userName := c.getUserName(mail)
	var resp userResp
	if err := c.client.Call(ctx, &queryRequest{
		Select: &selectQuery{
			Columns: []string{"Id", "Username", "Email", "Account.Id", "Account.Name", "Account.IsDeleted", "Account.Account_Active__c", "Account.Max_Technical_Representatives__c", "IsActive"},
			Table:   "User",
			Where:   "Username",
			Equals:  userName,
		},
	}, &resp); err != nil {
		return nil, errgo.Mask(err)
	}
	if len(resp.errors) > 0 {
		return nil, errgo.Notef(params.MultiError(resp.errors), "cannot get user %q", userName)
	}
	if len(resp.Users) == 0 {
		return nil, errgo.WithCausef(nil, params.ErrNotFound, "user %s not found", userName)
	}
	return &resp.Users[0], nil
}

// IsContactPresent reports whether a contact exists with the given email address.
// We have a convention to add the sandbox name automatically at the end as the email needs to be unique in all salesforce
// in the world.
func (c *Client) IsContactPresent(ctx context.Context, mail string) (bool, error) {
	userName := c.getUserName(mail)
	var resp idNameResp
	if err := c.client.Call(ctx, &queryRequest{
		Select: &selectQuery{
			Columns: []string{"Id"},
			Table:   "Contact",
			Where:   "Email",
			Equals:  userName,
		},
	}, &resp); err != nil {
		return false, errgo.Mask(err)
	}
	if len(resp.errors) > 0 {
		return false, errgo.Notef(params.MultiError(resp.errors), "cannot get contact %q", userName)
	}
	if len(resp.IDNames) == 0 {
		return false, nil
	}
	return true, nil
}

// CreateUser creates a user with the given name, email address and
// account name. The account should already exist. When in sandbox mode,
// the user name will be suffixed with the sandbox instance name to
// avoid name clashes.
//
// Note that this only works in Azure-GCE context.
func (c *Client) CreateUser(ctx context.Context, firstName string, lastName string, email string, accountName string) error {
	acc, err := c.GetAccount(ctx, accountName)
	if err != nil {
		return errgo.Mask(err)
	}
	if !acc.Enabled {
		return errgo.Notef(nil, "subscription is not enabled for account %s", accountName)
	}
	userName := c.getUserName(email)
	contactID := acc.ContactID
	if len(contactID) == 0 {
		if contactID, err = c.createContact(ctx, userName, firstName, lastName, userName, acc.ID); err != nil {
			return errgo.Mask(err)
		}
		linkedAsset, err := c.getAssetForAccount(ctx, acc.ID)
		if err != nil {
			return errgo.Mask(err)
		}
		if err := c.addContactToAsset(ctx, linkedAsset.ID, contactID); err != nil {
			return errgo.Mask(err)
		}
	}
	userID, err := c.createUser(ctx, userName, firstName, lastName, email, contactID, acc.SupportLevel)
	if err != nil {
		return errgo.Mask(err)
	}
	if err := c.activateUser(ctx, userID); err != nil {
		return errgo.Mask(err)
	}
	return nil
}

func (c *Client) activateUser(ctx context.Context, userID string) error {
	var resp *http.Response
	if err := c.client.Call(ctx, &activateUserRequest{
		ID: userID,
		Body: activateUser{
			IsActive: true,
		},
	}, &resp); err != nil {
		return errgo.Mask(err)
	}
	defer resp.Body.Close()
	if err := checkPatchErrors(resp, fmt.Sprintf("cannot activate user %q ", userID)); err != nil {
		return errgo.Mask(err)
	}
	return nil
}

// createUser creates a new user and activate it returning the its salesforce.
func (c *Client) createUser(ctx context.Context, userName string, firstName string, lastName string, email string, contactID string, supportLevel string) (string, error) {
	var resp createResp
	if err := c.client.Call(ctx, &createUserRequest{
		Body: createUser{
			Username:          userName,
			IsActive:          false,
			ProfileID:         c.profileIDs[supportLevel],
			Email:             email,
			LastName:          lastName,
			FirstName:         firstName,
			Alias:             "none",
			TimeZoneSidKey:    "GMT",
			LocaleSidKey:      "en_US",
			EmailEncodingKey:  "UTF-8",
			LanguageLocaleKey: "en_US",
			ContactID:         contactID,
		}}, &resp); err != nil {
		return "", errgo.Mask(err)
	}
	if len(resp.errors) > 0 {
		return "", errgo.Notef(params.MultiError(resp.errors), "cannot create user %q", userName)
	}
	if !resp.Success {
		return "", errgo.Notef(nil, "cannot create user %q", userName)
	}
	return resp.ID, nil
}

// createContact creates a contact from the first user that will open a support request and
// will become the contact for this account even for other users on this account.
// TODO: Add test.
func (c *Client) createContact(ctx context.Context, userName string, firstName string, lastName string, email string, accountID string) (string, error) {
	var resp createResp
	if err := c.client.Call(ctx, &createContactRequest{
		Body: createContact{
			LastName:  lastName,
			FirstName: firstName,
			Email:     email,
			AccountID: accountID,
		},
	}, &resp); err != nil {
		return "", errgo.Mask(err)
	}
	if len(resp.errors) > 0 {
		return "", errgo.Notef(params.MultiError(resp.errors), "cannot create contact for user %q", userName)
	}
	if !resp.Success {
		return "", errgo.Notef(nil, "cannot create contact for user %q", userName)
	}
	return resp.ID, nil
}

// addContactToAsset adds a contact to the asset. The first user that will open a support request  will become the
// contact for this asset.
func (c *Client) addContactToAsset(ctx context.Context, assetID string, contactID string) error {
	var resp *http.Response
	if err := c.client.Call(ctx, &addContactAssetRequest{
		ID: assetID,
		Body: addContact{
			ContactID: contactID,
		},
	}, &resp); err != nil {
		return errgo.Mask(err)
	}
	defer resp.Body.Close()
	if err := checkPatchErrors(resp, fmt.Sprintf("cannot update asset %q to add contact %q", assetID, contactID)); err != nil {
		return errgo.Mask(err)
	}
	return nil
}

// GetRedirectURL performs the necessary method calls to generate a correct SAML request.
// It returns the single-use url which the user needs to visit to log in.
func (c *Client) GetRedirectURL(ctx context.Context, userID string) (string, error) {
	if c.signinClient == nil {
		return "", errgo.New("saml sign in client not setup")
	}
	redirectURL, err := c.signinClient.GetRedirectURL(ctx, c.getUserName(userID))
	if err != nil {
		return "", errgo.Mask(err)
	}
	return redirectURL, nil
}

// init retrieves the static data from Salesforce like sandbox name and product/profile ids.
func (c *Client) init(ctx context.Context, conf *Config) error {
	if err := c.initOrganization(ctx); err != nil {
		return errgo.Mask(err)
	}
	if err := c.initProfileIDs(ctx, conf.ProfileNames); err != nil {
		return errgo.Mask(err)
	}
	if err := c.initProductIDs(ctx, conf.ProductNames); err != nil {
		return errgo.Mask(err)
	}
	return nil
}

func (c *Client) initProfileIDs(ctx context.Context, profileNames map[string][]string) error {
	if len(profileNames) == 0 {
		return nil
	}
	var profiles []string
	for k := range profileNames {
		profiles = append(profiles, k)
	}
	// sort the profile names for easier testing.
	sort.Strings(profiles)
	var profilesResp idNameResp
	if err := c.client.Call(ctx, &queryRequest{
		Select: &selectQuery{
			Columns: []string{"Id", "Name"},
			Table:   "Profile",
			Where:   "Name",
			In:      profiles,
		},
	}, &profilesResp); err != nil {
		return errgo.Mask(err)
	}
	if len(profilesResp.errors) > 0 {
		return errgo.Notef(params.MultiError(profilesResp.errors), "cannot get profiles")
	}
	if profilesResp.TotalSize == 0 {
		return errgo.WithCausef(nil, params.ErrNotFound, "profiles %s not found", profiles)
	}
	c.profileIDs = map[string]string{}
	for _, p := range profilesResp.IDNames {
		levels := profileNames[p.Name]
		for _, l := range levels {
			c.profileIDs[l] = p.ID
		}
	}
	return nil
}

func (c *Client) initProductIDs(ctx context.Context, productNames map[string][]string) error {
	if len(productNames) == 0 {
		return nil
	}
	var productsResp idNameResp
	var products []string
	for k := range productNames {
		products = append(products, k)
	}
	// sort the product names for easier testing.
	sort.Strings(products)
	if err := c.client.Call(ctx, &queryRequest{
		Select: &selectQuery{
			Columns: []string{"Id", "Name"},
			Table:   "Product2",
			Where:   "Name",
			In:      products,
		},
	}, &productsResp); err != nil {
		return errgo.Mask(err)
	}
	if len(productsResp.errors) > 0 {
		return errgo.Notef(params.MultiError(productsResp.errors), "cannot get products")
	}
	if productsResp.TotalSize == 0 {
		return errgo.WithCausef(nil, params.ErrNotFound, "products %s not found", products)
	}
	c.productIDs = map[string]string{}
	for _, p := range productsResp.IDNames {
		levels := productNames[p.Name]
		for _, l := range levels {
			c.productIDs[l] = p.ID
		}
	}
	return nil
}

// initOrganization initializes the organizationID and sandboxInstanceName
// fields in c by retrieving them from the Salesforce server.
func (c *Client) initOrganization(ctx context.Context) error {
	var resp organizationResp
	if err := c.client.Call(ctx, &queryRequest{
		Select: &selectQuery{
			Columns: []string{"Id", "IsSandbox", "InstanceName"},
			Table:   "Organization",
		},
	}, &resp); err != nil {
		return errgo.Mask(err)
	}
	if len(resp.errors) > 0 {
		return errgo.Notef(params.MultiError(resp.errors), "cannot get organization")
	}
	if len(resp.Organizations) == 0 {
		return errgo.WithCausef(nil, params.ErrNotFound, "organization not found")
	}
	c.organizationID = resp.Organizations[0].ID

	if resp.Organizations[0].IsSandbox {
		c.sandboxInstanceName = strings.ToLower(resp.Organizations[0].InstanceName)
	}
	return nil
}

// UserCounts gets the current number of users for the provided account name and the maximum allowed users for a given
// account.
func (c *Client) UserCounts(ctx context.Context, accountName string) (current int, max int, err error) {
	account, err := c.GetAccount(ctx, accountName)
	if err != nil {
		return 0, 0, errgo.Mask(err)
	}
	users, err := c.getAccountUsers(ctx, account.ID)
	if err != nil {
		return 0, 0, errgo.Mask(err)
	}
	if account.MaxUsers == 0 {
		account.MaxUsers = defaultMaxUsers
	}
	return len(users), int(account.MaxUsers), nil
}

func (c *Client) getUserName(userName string) string {
	if c.sandboxInstanceName != "" {
		userName += "." + c.sandboxInstanceName
	}
	return userName
}

// GetAssetsForUser returns the associated assets for the specified email address.
func (c *Client) GetAssetsForUser(ctx context.Context, email string) ([]Asset, error) {
	user, err := c.GetUser(ctx, email)
	if err != nil {
		return nil, errgo.Mask(err, errgo.Is(params.ErrNotFound))
	}
	return c.GetAssetsForAccount(ctx, user.Account.ID)
}

// GetAssetsForAccount returns all the assets held by the account
// with the given id.
func (c *Client) GetAssetsForAccount(ctx context.Context, accountID string) ([]Asset, error) {
	var assetResp assetResp
	if err := c.client.Call(ctx, &queryRequest{
		Select: &selectQuery{
			Columns: []string{
				"Id",
				"IsDeleted",
				"Status",
				"UsageEndDate",
				"ContactId",
				"Product2.Name",
				"Product2.Id",
				"Product2.Livepatch__c",
				"Product2.ESM__c",
				"Product2.Family",
				"Product2.SLA_Service_Level__c",
			},
			Table:  "Asset",
			Where:  "AccountId",
			Equals: accountID,
		},
	}, &assetResp); err != nil {
		return nil, errgo.Mask(err)
	}
	if len(assetResp.errors) > 0 {
		return nil, errgo.Notef(params.MultiError(assetResp.errors), "cannot get asset linked to account %q", accountID)
	}
	if assetResp.TotalSize == 0 {
		return nil, nil
	}
	return assetResp.Assets, nil
}

func checkPatchErrors(resp *http.Response, message string) error {
	if resp.StatusCode == http.StatusNoContent {
		return nil
	}
	var errs params.MultiError
	err := httprequest.UnmarshalJSONResponse(resp, &errs)
	if err != nil {
		return errgo.Notef(err, message)
	}
	return errgo.Notef(params.MultiError(errs), message)
}
