// Copyright 2018 Canonical Ltd.

package salesforce

import (
	"encoding/json"
	"fmt"
	"strings"
	"time"

	httprequest "gopkg.in/httprequest.v1"

	"github.com/CanonicalLtd/salesforce/params"
)

// Time marshals and unmarshals time values as understood by
// the Salesforce API. It always marshals as UTC ISO8601 and unmarshals
// as either ISO8601 or as yyyy-mm-dd format.
type Time struct {
	T time.Time
}

func (t Time) MarshalText() ([]byte, error) {
	return []byte(t.T.UTC().Format(time.RFC3339)), nil
}

func (t *Time) UnmarshalText(text []byte) (err error) {
	t.T, err = time.Parse(time.RFC3339, string(text))
	if err == nil {
		return nil
	}
	t.T, err = time.Parse("2006-01-02", string(text))
	return err
}

// Bool marshals and unmarshals boolean values as understood by
// the Salesforce API. It always marshals as Yes, No.
type Bool struct {
	B bool
}

func (b *Bool) UnmarshalText(text []byte) (err error) {
	b.B = string(text) == "Yes"
	return nil
}

func (b *Bool) MarshalText() ([]byte, error) {
	if b.B {
		return []byte("Yes"), nil
	}
	return []byte("No"), nil
}

// queryRequest is a generic query request on Salesforce.
type queryRequest struct {
	httprequest.Route `httprequest:"GET /query/"`
	Select            *selectQuery `httprequest:"q,form"`
}

// selectQuery is a simple SOQL(Salesforce Object Query Language) for Salesforce.
// https://developer.salesforce.com/docs/atlas.en-us.soql_sosl.meta/soql_sosl/sforce_api_calls_soql_sosl_intro.htm
type selectQuery struct {
	Columns []string
	Table   string
	Where   string
	Equals  string
	In      []string
}

// MarshalText marshals the query parameter in SOQL format.
func (q *selectQuery) MarshalText() ([]byte, error) {
	if len(q.Where) == 0 {
		return []byte(fmt.Sprintf("SELECT %s FROM %s", strings.Join(q.Columns, ","), q.Table)), nil
	}
	equals := fmt.Sprintf("%s='%s'", q.Where, q.Equals)
	if len(q.In) > 0 {
		equals = fmt.Sprintf("%s IN ('%s')", q.Where, strings.Join(q.In, "','"))
	}
	return []byte(fmt.Sprintf("SELECT %s FROM %s WHERE (%s)", strings.Join(q.Columns, ","), q.Table, equals)), nil
}

// queryBaseResp is alweays include in query results.
type queryBaseResp struct {
	errors    params.MultiError `json:"-"`
	TotalSize int               `json:"totalSize"`
	Done      bool              `json:"done"`
}

// accountResp represents an Account query.
type accountResp struct {
	queryBaseResp
	Accounts []params.Account `json:"records"`
}

func (o *accountResp) UnmarshalJSON(data []byte) error {
	type noUnmarshaller *accountResp
	a := noUnmarshaller(o)
	return unmarshalResponse(data, &a, &o.errors)
}

// idNameResp is a generic select by id and name response.
type idNameResp struct {
	queryBaseResp
	IDNames []idName `json:"records"`
}

// idName represents the generic response from an id/name query.
type idName struct {
	ID   string `json:"ID"`
	Name string `json:"Name"`
}

func (o *idNameResp) UnmarshalJSON(data []byte) error {
	type noUnmarshaller *idNameResp
	a := noUnmarshaller(o)
	return unmarshalResponse(data, &a, &o.errors)
}

// organizationResp is a query to get the organization we are currently in to populate static data.
type organizationResp struct {
	queryBaseResp
	Organizations []organization `json:"records"`
}

// organization represents our own organization to know the instance name and if we are in sandbox or not.
type organization struct {
	ID           string `json:"Id"`
	IsSandbox    bool   `json:"IsSandbox"`
	InstanceName string `json:"InstanceName"`
}

func (o *organizationResp) UnmarshalJSON(data []byte) error {
	type noUnmarshaller *organizationResp
	a := noUnmarshaller(o)
	return unmarshalResponse(data, &a, &o.errors)
}

//userResp is a query to retrieve a user
type userResp struct {
	queryBaseResp
	Users []params.User `json:"records"`
}

func (o *userResp) UnmarshalJSON(data []byte) error {
	type noUnmarshaller *userResp
	a := noUnmarshaller(o)
	return unmarshalResponse(data, &a, &o.errors)
}

// unmarshalResponse check if the response is an error, if not it can then unmarshal it to the proper type.
func unmarshalResponse(data []byte, val interface{}, errs *params.MultiError) error {
	json.Unmarshal(data, errs)
	if len(*errs) > 0 {
		return nil
	}
	return json.Unmarshal(data, val)
}

// assetResp is a response to a query for asset.
type assetResp struct {
	queryBaseResp
	Assets []Asset `json:"records"`
}

func (o *assetResp) UnmarshalJSON(data []byte) error {
	type noUnmarshaller *assetResp
	a := noUnmarshaller(o)
	return unmarshalResponse(data, &a, &o.errors)
}

// Asset represents the fact that we have a commercial item of value(support) with the Account.
type Asset struct {
	ID           string  `json:"Id"`
	Product      Product `json:"Product2"`
	IsDeleted    bool    `json:"IsDeleted"`
	Status       string  `json:"Status"`
	ContactID    string  `json:"ContactId"`
	UsageEndDate Time    `json:"UsageEndDate"`
}

// createAccountRequest is a request to create an account.
type createAccountRequest struct {
	httprequest.Route `httprequest:"POST /sobjects/Account"`
	Body              createAccount `httprequest:",body"`
}

// createAccount represents the parameters to create an account when subscribing to support.
type createAccount struct {
	Name string `json:"Name"`
}

// createAssetRequest is a request to create an asset.
type createAssetRequest struct {
	httprequest.Route `httprequest:"POST /sobjects/Asset"`
	Body              createAsset `httprequest:",body"`
}

// createAsset represents the parameters to create an asset when subscribing to support.
type createAsset struct {
	Name         string  `json:"Name"`
	AccountID    string  `json:"AccountId"`
	ProductID    string  `json:"Product2Id"`
	Status       string  `json:"Status"`
	UsageEndDate Time    `json:"UsageEndDate"`
	PurchaseDate Time    `json:"PurchaseDate"`
	InstallDate  Time    `json:"InstallDate"`
	Quantity     float64 `json:"Quantity"`
}

// updateAssetRequest is a request to update an asset.
type updateAssetRequest struct {
	httprequest.Route `httprequest:"PATCH /sobjects/Asset/:id"`
	ID                string      `httprequest:"id,path"`
	Body              updateAsset `httprequest:",body"`
}

// updateAsset represents the parameters to update an asset when the support level changes.
type updateAsset struct {
	ProductID    string `json:"Product2Id"`
	Status       string `json:"Status"`
	UsageEndDate Time   `json:"UsageEndDate"`
}

// addContactAssetRequest is a request to add a contact to an asset.
type addContactAssetRequest struct {
	httprequest.Route `httprequest:"PATCH /sobjects/Asset/:id"`
	ID                string     `httprequest:"id,path"`
	Body              addContact `httprequest:",body"`
}

// addContact represents the parameters to update an asset with a contact.
type addContact struct {
	ContactID string `json:"ContactId"`
}

// setEnabledRequest is a request to subscribe/unsubscribe from support.
type setEnabledRequest struct {
	httprequest.Route `httprequest:"PATCH /sobjects/Asset/:id"`
	ID                string          `httprequest:"id,path"`
	Body              setEnabledAsset `httprequest:",body"`
}

// setEnabledAsset represents the parameters to enable/disable an asset.
type setEnabledAsset struct {
	Status       string `json:"Status"`
	UsageEndDate Time   `json:"UsageEndDate"`
}

// createUserRequest is a request to create a user when we get a first support request.
type createUserRequest struct {
	httprequest.Route `httprequest:"POST /sobjects/User"`
	Body              createUser `httprequest:",body"`
}

// createUser represents the parameters to create a user.
type createUser struct {
	Username          string `json:"Username"`
	ContactID         string `json:"ContactId"`
	IsActive          bool   `json:"IsActive"`
	ProfileID         string `json:"ProfileId"`
	Email             string `json:"Email"`
	LastName          string `json:"LastName"`
	FirstName         string `json:"FirstName"`
	Alias             string `json:"Alias"`
	TimeZoneSidKey    string `json:"TimeZoneSidKey"`
	LocaleSidKey      string `json:"LocaleSidKey"`
	EmailEncodingKey  string `json:"EmailEncodingKey"`
	LanguageLocaleKey string `json:"LanguageLocaleKey"`
}

// activateUserRequest is a request to activate the user after creation.
type activateUserRequest struct {
	httprequest.Route `httprequest:"PATCH /sobjects/User/:id"`
	ID                string       `httprequest:"id,path"`
	Body              activateUser `httprequest:",body"`
}

// activateUser represents the parameters to activate a user.
type activateUser struct {
	IsActive bool `json:"IsActive"`
}

// UpdateUserProfileRequest updates the profile associated with the user depnding on the support level.
type updateUserProfileRequest struct {
	httprequest.Route `httprequest:"PATCH /sobjects/User/:id"`
	ID                string        `httprequest:"id,path"`
	Body              updateProfile `httprequest:",body"`
}

// updateProfile represents the new profile for the user.
type updateProfile struct {
	ProfileID string `json:"ProfileId"`
}

// createContactrequest represents the first user that will open a support request and will become the contact fror this account.
type createContactRequest struct {
	httprequest.Route `httprequest:"POST /sobjects/Contact"`
	Body              createContact `httprequest:",body"`
}

// createContact represents the parameters to create a contact.
type createContact struct {
	FirstName string `json:"FirstName"`
	LastName  string `json:"LastName"`
	AccountID string `json:"AccountId"`
	Email     string `json:"Email"`
}

// createResp is a generic response for all create request.
type createResp struct {
	errors  params.MultiError `json:"-"`
	ID      string            `json:"id"`
	Success bool              `json:"success"`
	// TODO: Are those errors/warning useful if we already have errors coming instead of the Json body??
	// TODO even the success seems weird.
	//`errors" : [ ],
	//"warnings" : [ ]
}

func (o *createResp) UnmarshalJSON(data []byte) error {
	type noUnmarshaller *createResp
	a := noUnmarshaller(o)
	return unmarshalResponse(data, &a, &o.errors)
}

// Product represents a product.
type Product struct {
	// ID holds the id of the product.
	ID string `json:"Id"`

	// Name holds the name of the product.
	Name string `json:"Name"`

	// Livepatch holds whether the product includes a livepatch
	// subscription.
	Livepatch Bool `json:"Livepatch__c"`

	// Livepatch holds whether the product includes an ESM
	// subscription.
	ESM Bool `json:"ESM__c"`

	// Family holds the product family.
	Family string `json:"Family"`

	// ServiceLevel holds the service level of the product.
	ServiceLevel string `json:"SLA_Service_Level__c"`
}

// FIPS reports whether the product includes a FIPS subscription.
func (p Product) FIPS() bool {
	return p.Name == "Canonical FIPS Service"
}

var supportLevels = []string{"standard", "essential", "advanced"}

// SupportLevel returns the support level associated with
// the product. If the product does not imply support,
// then it returns the empty string.
func (p Product) SupportLevel() string {
	if p.Family != "Support" {
		return ""
	}
	if p.ServiceLevel != "" && p.ServiceLevel != "Not-Applicable" {
		level := strings.ToLower(p.ServiceLevel)
		if level == "bitstream" {
			return "essential"
		}
		return level
	}
	name := strings.ToLower(p.Name)
	for _, level := range supportLevels {
		if strings.Contains(name, level) {
			return level
		}
	}
	return ""
}
