// Copyright 2018 Canonical Ltd.

package params

import (
	"fmt"
	"net/http"

	errgo "gopkg.in/errgo.v1"
	httprequest "gopkg.in/httprequest.v1"
)

// These constants represent the possible error codes
// returned by the API server.
var (
	ErrBadRequest   = errgo.New("bad request")
	ErrUnauthorized = errgo.New("unauthorized")
	ErrNotFound     = errgo.New("not found")
)

func UnmarshalError(resp *http.Response) error {
	var errs MultiError
	err := httprequest.UnmarshalJSONResponse(resp, &errs)
	if err != nil {
		return errgo.Mask(err)
	}
	return errs
}

// ErrorResponse represents the errors for all requests to Salesforce.
type ErrorResponse struct {
	Code    string `json:"errorCode"`
	Message string `json:"message"`
}

// Account represents an organization/person we have to support.
type Account struct {
	Name      string `json:"Name"`
	ID        string `json:"Id"`
	IsDeleted bool   `json:"IsDeleted"`
	// Custom field from canonical to indicate the account as active.
	IsActive     bool    `json:"Account_Active__c"`
	MaxUsers     float64 `json:"Max_Technical_Representatives__c"`
	Enabled      bool    `json:"-"`
	SupportLevel string  `json:"-"`
	AssetID      string  `json:"-"`
	ContactID    string  `json:"-"`
}

// User is one of the user of the Account
type User struct {
	ID       string  `json:"ID"`
	Email    string  `json:"Email"`
	UserName string  `json:"UserName"`
	Account  Account `json:"Account"`
	IsActive bool    `json:"IsActive"`
}

// MultiError represents the error coming back from Salesforce but mostly we will get only one unless we are
// batching requests.
type MultiError []ErrorResponse

// Error implements the error interface.
func (m MultiError) Error() string {
	if len(m) == 0 {
		return "no errors"
	}
	if len(m) > 1 {
		return fmt.Sprintf("error %s (and %d more)", m[0].Message, len(m)-1)
	}
	return m[0].Message
}

// Code returns the error code of the first error in the response.
func (m MultiError) Code() string {
	if len(m) == 0 {
		return "no errors"
	}
	return m[0].Code
}
