// Copyright 2018 Canonical Ltd.

package salesforce

import (
	"bytes"
	"context"
	"crypto/rand"
	"crypto/rsa"
	"crypto/x509"
	"encoding/base64"
	"encoding/hex"
	"encoding/pem"
	"fmt"
	"io"
	"io/ioutil"
	"net/http"
	"net/url"
	"strings"
	"text/template"
	"time"

	"github.com/beevik/etree"
	"github.com/russellhaering/goxmldsig"
	"gopkg.in/errgo.v1"
)

// signinClient represents a client that can generate a login success response to get a redirect URL to salesforce for
// a user to be directly signin to the salesforce portal as this user.
type signinClient struct {
	organisationId string
	destination    string
	issuer         string
	audience       string
	signinContext  *dsig.SigningContext
}

// newSigninClient creates a signinClient.
func newSigninClient(conf *Config, organizationId string) (*signinClient, error) {
	if len(organizationId) < 3 {
		return nil, errgo.Notef(nil, "organization id %q must be at least of 3 characters", organizationId)
	}
	keyBlock, _ := pem.Decode([]byte(conf.SAMLPrivateKey))
	if keyBlock == nil {
		return nil, errgo.Notef(nil, "no key found")
	}
	if keyBlock.Type != "PRIVATE KEY" {
		return nil, errgo.Newf("unknown certificate type %q", keyBlock.Type)
	}
	key, err := x509.ParsePKCS8PrivateKey(keyBlock.Bytes)
	if err != nil {
		return nil, errgo.Notef(err, "cannot parse key")
	}
	certBlock, _ := pem.Decode([]byte(conf.SAMLCertificate))
	if certBlock == nil {
		return nil, errgo.Notef(err, "no certificate found")
	}
	if certBlock.Type != "CERTIFICATE" {
		return nil, errgo.Newf("unknown certificate type %q", certBlock.Type)
	}
	privateKey, ok := key.(*rsa.PrivateKey)
	if !ok {
		return nil, errgo.Newf("unknown private type %T", key)
	}
	// Get the organization id without the 3 digit at the end which represents the checksum of the organization id.
	// https://developer.salesforce.com/forums/?id=906F00000008yhKIAQ
	organizationId = organizationId[:len(organizationId)-3]

	return &signinClient{
		organisationId: organizationId,
		destination:    fmt.Sprintf("%s/selfservice/login?so=%s", conf.TenantURL, organizationId),
		issuer:         conf.SAMLIssuer,
		audience:       conf.SAMLEntityID,
		signinContext: dsig.NewDefaultSigningContext(&memoryX509KeyStore{
			privateKey: privateKey,
			cert:       certBlock.Bytes,
		}),
	}, nil
}

// GetRedirectURL performs the necessary method calls to generate a correct SAML request.
// It returns the single-use url which the user needs to visit to log in.
func (c *signinClient) GetRedirectURL(ctx context.Context, userId string) (string, error) {
	document, err := c.getSignedDocument(userId)
	if err != nil {
		return "", errgo.Mask(err)
	}
	httpReq, err := c.createSAMLRequest(ctx, document)
	if err != nil {
		return "", errgo.Notef(err, "cannot build SAML request")
	}
	loc, err := c.requestRedirect(httpReq)
	if err != nil {
		return "", errgo.Notef(err, "cannot get redirect URL")
	}

	return loc, nil
}

func (c *signinClient) getSignedDocument(userId string) ([]byte, error) {
	rendered, err := c.renderTemplate(userId)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	signed, err := signSAMLDocument(c.signinContext, rendered)
	if err != nil {
		return nil, errgo.Notef(err, "cannot sign SAML document")
	}
	return signed, nil
}

func (c *signinClient) createSAMLRequest(ctx context.Context, document []byte) (*http.Request, error) {
	data := url.Values{
		"SAMLResponse": {base64.StdEncoding.EncodeToString(document)},
	}
	httpReq, err := http.NewRequest("POST", c.destination, strings.NewReader(data.Encode()))
	if err != nil {
		return nil, errgo.Notef(err, "cannot create SAML request")
	}
	httpReq.Header.Set("Content-Type", "application/x-www-form-urlencoded")
	return httpReq.WithContext(ctx), nil
}

// requestRedirect makes the supplied request and extracts the redirect URL.
func (c *signinClient) requestRedirect(req *http.Request) (string, error) {
	client := &http.Client{
		CheckRedirect: func(*http.Request, []*http.Request) error {
			return http.ErrUseLastResponse
		},
	}
	resp, err := client.Do(req)
	if err != nil {
		return "", errgo.Mask(err)
	}
	defer io.Copy(ioutil.Discard, resp.Body)
	defer resp.Body.Close()

	if resp.StatusCode/100 != 3 {
		return "", errgo.Newf("cannot get a redirect url, got a status %v", resp.Status)
	}
	loc, err := resp.Location()
	if err != nil {
		return "", errgo.Notef(err, "cannot parse response location")
	}
	if strings.HasSuffix(loc.Path, "/SamlError") {
		return "", errgo.Newf("SAML request failed")
	}
	return loc.String(), nil
}

// renderTemplate renders the SAML Template before signin.
func (c *signinClient) renderTemplate(userId string) ([]byte, error) {
	now := time.Now()
	values := &values{
		ResponseId:  genHexID(),
		AssertionId: genHexID(),
		Instant:     now.UTC().Format(time.RFC3339),
		TimeLimit:   now.UTC().Add(time.Minute).Format(time.RFC3339),
		UserId:      userId,
		Destination: c.destination,
		Issuer:      c.issuer,
		Audience:    c.audience,
	}
	var buf bytes.Buffer
	if err := samlTemplate.Execute(&buf, &values); err != nil {
		return nil, err
	}
	return buf.Bytes(), nil
}

// signSAMLDocument sign the SAML response to a login to act as an identity provider.
func signSAMLDocument(signinContext *dsig.SigningContext, samlData []byte) ([]byte, error) {
	doc := etree.NewDocument()
	if err := doc.ReadFromBytes(samlData); err != nil {
		return nil, errgo.Mask(err)
	}
	doc.Indent(etree.NoIndent)
	elem := doc.FindElement("//samlp:Response")
	if elem == nil {
		return nil, errgo.Notef(nil, "no response element found in saml document")
	}
	// Sign the element
	signed, err := signinContext.SignEnveloped(elem)
	if err != nil {
		return nil, errgo.Mask(err)
	}
	if signed == elem {
		return nil, errgo.Notef(nil, "saml document is not signed")
	}
	parent := elem.Parent()
	parent.InsertChild(elem, signed)
	elem.Parent().RemoveChild(elem)
	signedTemplate, err := doc.WriteToBytes()
	if err != nil {
		return nil, errgo.Mask(err)
	}
	return signedTemplate, nil
}

type values struct {
	ResponseId  string
	AssertionId string
	Instant     string
	TimeLimit   string
	UserId      string
	Destination string
	Issuer      string
	Audience    string
}

// genHexID generates a random hex string
func genHexID() string {
	buf := make([]byte, 12)
	_, err := rand.Read(buf)
	if err != nil {
		panic("cannot read random bytes: " + err.Error())
	}
	return hex.EncodeToString(buf)
}

var samlTemplate = template.Must(template.New("").Parse(`
<?xml version="1.0" encoding="UTF-8"?>
<samlp:Response xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol" Destination="{{.Destination}}" ID="_{{.ResponseId}}" IssueInstant="{{.Instant}}" Version="2.0">
  <saml:Issuer xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion">{{.Issuer}}</saml:Issuer>
  <samlp:Status>
    <samlp:StatusCode Value="urn:oasis:names:tc:SAML:2.0:status:Success"/>
  </samlp:Status>
  <saml:Assertion xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion" ID="_{{.AssertionId}}" IssueInstant="{{.Instant}}" Version="2.0">
    <saml:Issuer>{{.Issuer}}</saml:Issuer>
    <saml:Subject>
      <saml:NameID Format="urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified">{{.UserId}}</saml:NameID>
      <saml:SubjectConfirmation Method="urn:oasis:names:tc:SAML:2.0:cm:bearer">
        <saml:SubjectConfirmationData NotOnOrAfter="{{.TimeLimit}}" Recipient="{{.Destination}}"/>
      </saml:SubjectConfirmation>
    </saml:Subject>
    <saml:Conditions NotBefore="{{.Instant}}" NotOnOrAfter="{{.TimeLimit}}">
      <saml:AudienceRestriction>
        <saml:Audience>{{.Audience}}</saml:Audience>
      </saml:AudienceRestriction>
    </saml:Conditions>
    <saml:AuthnStatement AuthnInstant="{{.Instant}}">
      <saml:AuthnContext>
        <saml:AuthnContextClassRef>urn:oasis:names:tc:SAML:2.0:ac:classes:unspecified</saml:AuthnContextClassRef>
      </saml:AuthnContext>
    </saml:AuthnStatement>
  </saml:Assertion>
</samlp:Response>
`))

type memoryX509KeyStore struct {
	privateKey *rsa.PrivateKey
	cert       []byte
}

func (ks *memoryX509KeyStore) GetKeyPair() (*rsa.PrivateKey, []byte, error) {
	return ks.privateKey, ks.cert, nil
}
